// ------------------------------------------------------------------
// module       : frm_hdf_io.h
// purpose      : implement HDF4, HDF54 basic set of I/O wrappers
// author       : joe glassy, Lupine Logic Inc.
// revision     : v1.9-2014-09-30T17:01, added frm_hdf5_GetDatasetId() and
//                frm_hdf5_CloseDataset() as these were missing earlier.
// history      : v1.8-2014-08-08T17:08:00-07:00, refined frm_hdf5_GetDatasetInfo(), 
//                and frm_hdf5_DatasetExists()
// history      : v1.7-2014-06-24T13:06:00, added sprintf() style attribute
//                writer function
// history     : v1.6-2014-06-13T11:51:00, added several inquiry functions
//                and improved robustness of attribute reading,writing.
//                added: frm_hdf5_GetObjectCount(), frm_hdf5_GetAttributeCount()
//                and others.
// history     : v1.5-2014-02-23T21:00,added missing frm_OpenfileAppend()
// history     : v1.4-2014-02-21T00:37:00-07:00,jmg
// history     : v1.3-2014-02-11T14:27:00-07:00,jmg
// history      : v1.2-2014-02-10T16:48:00-07:00,jmg, added more fx
// history      : v1.1-2014-01-21T11:49:00-07:00
// history      : 2014-01-05T21:06:00-07:00 
// history      : 2013-12-03T12:49:00-07:00 refactor pass 1, for D4
// version      : see VERSTAG macro, v1.1-2014-01-21T11:49:00-07:00
// dependencies : HDF4 v4.2.6.x lib, HDF5 v1.8.x lib 
// limitations  : n/a
// ------------------------------------------------------------------

#ifndef FRMWRK_HDF_IO_DEF
#define FRMWRK_HDF_IO_DEF  1

#define FRM_HDF_IO_VERSTAG  "v1.8-2014-08-08T17:08:00-07:00"

#include "frm_api.h"

#include "hdf5.h"
#include "hdf5_hl.h"

// verify: constants are needed as well
// #include "l4c_gridconst.h"

// ------------------------------------------------------------------
// Sub-API specific manifest constants
#define FRM_MAX_HDF4_DATASET_NAMELEN   256

// max n. of supported dimensions per object-layer
#define FRM_MAX_HDF_MAXDIMS            FRM_API_MAXDIMS

#define FRM_MAX_HDF5_DATASET_NAMELEN   256

#define FRM_HDF_DFLT_COMPRESSION_LEV      7


// Stores separate dcpl's and dspaces for each supported grid
typedef struct tagHdf5GridProps {
    int     easeGridId  ; // IsEG_Global_1km,IsEG_Global_3km,IsEG_Global_9km
    hid_t   dcpl_1km  ; 
    hid_t   dcpl_3km  ; 
    hid_t   dcpl_9km  ; 
    hid_t   dspace_1km ; 
    hid_t   dspace_3km ; 
    hid_t   dspace_9km ; 
} Hdf5GridProps ;


// ------------------------------------------------------------------
// ADTs -- for HDF4 I/p processing
typedef struct tagHdf4Obj {
 int32 sd_id     ;
 int32 sds_id    ;
 int32 attrib_id ;
 char  h4DatasetName[FRM_MAX_HDF4_DATASET_NAMELEN] ;
} HDF4_ObjAdt ;


// ------------------------------------------------------------------
// ADTs -- for HDF4 I/p processing
typedef struct tagHdf5Obj {
 hid_t   h5File_id   ;
 hid_t   h5Dspace_id ;
 hid_t   h5Dset_id   ;
 hid_t   h5Dcpl_id   ; // plist, dataset-creation property list (dcpl) etc
 int     h5Rank      ;
 hsize_t hDims[FRM_MAX_HDF_MAXDIMS]; // dimensions list
 hsize_t h5Chunk_size[2] ;  // shared chunks 
 char    h5DatasetName[FRM_MAX_HDF5_DATASET_NAMELEN];
} HDF5_ObjAdt ;


// ------------------------------------------------------------------
// FUNCTION PROTOTYPES (components common to both HDF variants)
char *frm_hdf_GetVersTag(void);

// ------------------------------------------------------------------
// FUNCTION PROTOTYPES (HDF4 specific components)
int32  frm_hdf4_Openfile(const char *hdf4Path);
int32  frm_hdf4_Closefile(int32 sd_id);
int32  frm_hdf4_Createfile(const char *hdf4Path);

int32 frm_hdf4_CloseAttrib(void);
int32 frm_hdf4_CreateAttrib(void);
int32 frm_hdf4_CreateDataset(void);
int32 frm_hdf4_ReadAttrib(void);
int32 frm_hdf4_WriteAttrib(void);
int32 frm_hdf4_ReadData(const int32 sd_id,const int32 sds_id, void *data_p);
int32 frm_hdf4_WriteData(const int32 sd_id,const int32 sds_id, void *data_p);

// ------------------------------------------------------------------
// FUNCTION PROTOTYPES (HDF5 specific components)
// ------------------------------------------------------------------

// <JMG> 06/17/2014, added orig.
int frm_hdf5_ErrorStackEnable(H5E_auto_t errFxHandlerPtr,void *clientData_p);
// returns 0 to indicated sys error stack is now off
int frm_hdf5_ErrorStackDisable(void);

hid_t frm_hdf5_TIV_datatype_to_H5T(const int TIV_dType);
// future: not yet enabled..
int   frm_hdf5_H5T_Datatype_to_TIV(const hid_t h5_dType);

// <JMG> 09/30/2014, added as this was missing earlier, ‘frm_hdf5_GetDatasetId’
hid_t frm_hdf5_GetDatasetId(const hid_t file_id,const char *datasetName_p) ;
int   frm_hdf5_CloseDataset(const hid_t dataset_id);
// <JMG> 6/16/2014, renamed from DoesDatasetExist to DatasetExists for
// consistency with others
int  frm_hdf5_DatasetExists(hid_t file_id,const char *datasetName_p);

// <JMG> 06/13/2014, added orig
int32 frm_hdf5_ClassifyHdf5Datatype(hid_t loc_id);

// <JMG> 6/13/2014 added orig
int frm_hdf5_GetObjectInfo(hid_t loc_id,char *myName_p,int *nAttrib,int *nObj);


size_t frm_hdf5_GetSizeOfOne(hid_t dsTypeId);
hid_t  frm_hdf5_GetDatasetInfo(hid_t file_id,const char *datasetName_p,
                hid_t *hDtype, hid_t *dcpl,hid_t *dspace,
                int *hRank,hsize_t *dims);
// <JMG> 6/12/2014, added orig
size_t frm_hdf5_GetAttributeCount(hid_t loc_id);
// <JMG> 6/13/2014, added orig
char *frm_hdf5_GetComment(hid_t object_id,char *commentStr_p,size_t bufsize);
// <JMG> 6/13/2014, added orig
void frm_hdf5_SetComment(hid_t object_id,const char *commentStr_p);

void   *frm_hdf5_DiscoverAndRead(hid_t file_id,char *datasetName_p,
                int *dRank,hsize_t *dims);              

size_t frm_hdf5_GetBasicProperties(hid_t file_id,char *datasetName_p,
                int *rtnRank, hsize_t *rtnDims);
                
// <JMG> 06/25/2014, if used, must call prior to CreateFile call
void   frm_hdf5_SetUserblocksize(size_t newUbSize);
size_t frm_hdf5_GetUserblocksize(const hid_t file_id);
// <JMG> 06/25/2014, warning: these (2) are not yet fully tested
void frm_hdf5_SetUserblock(const char *hdf5Path,const size_t ubDataSiz,char *userBlockdata_p);
void frm_hdf5_GetUserblock(const char *hdf5Path,const size_t ubDataSiz,char *userBlockdata_p);

hid_t  frm_hdf5_Createfile(char *hdf5Path,char *callerLbl_p);
hid_t  frm_hdf5_Openfile(const char *hdf5Path,char *callerLbl_p);
hid_t  frm_hdf5_OpenfileAppend(char *hdf5Path,char *callerLbl_p);
herr_t frm_hdf5_Closefile(const hid_t file_id,char *callerLbl_p);
// <JMG> 2014-06-24T18:03:00, added orig.
herr_t frm_hdf5_Flushfile(const hid_t file_id);

int frm_hdf5_ReadData(const hid_t thisFile_id,const hid_t dsTypeId,char *datasetname_p,
                void *dataBuffer_p,char *callerLbl_p);

// <JMG> 2014-07-28, re-ord arg call list to match ReadData better
int frm_hdf5_WriteData(const hid_t file_id,const hid_t dsTypeId, char *datasetname_p,
                void *dataBuffer_p,const int rank,hsize_t *dims,char *callerLbl_p);
                        
// use this, as CreateDataset is not yet complete and may be deprecated.
int frm_hdf5_Writedataset(const hid_t file_id,const int rank,hsize_t *dims,
    const hid_t dsTypeId, char *datasetname_p,
    void *dataBuffer_p,char *callerLbl_p) ;

// <JMG> 2014-06-13T19:07, added orig.
// returns boolean TRUE if attribute exists, otherwise FALSE.
int frm_hdf5_AttributeExists(hid_t loc_id,const char *attrName_p);

// <JMG> 2014-06-16, orig stub, not yet fully tested.
int frm_hdf5_GroupExists(hid_t loc_id,const char *groupName_p);

hid_t frm_hdf5_GetAttribInfo(const hid_t owner_id,int *nAttribs, 
        const char *attrName_p, hid_t *attrDspace_p,hsize_t *attrSiz_p,
        hid_t *attrDatatype_p);
            
void *frm_hdf5_ReadAttribute(const hid_t owner_id,hid_t *rtn_AttrDatatype, 
                const char *attrName,void *attrValue_p) ;

char *frm_hdf5_ReadAttributeString(const hid_t owner_id,
                const char *attrName,char *attrValue_p) ;

hid_t frm_hdf5_CreateStringAttribute(const hid_t owner_id,
                const char *attrLabel,  const char *attrValue_p);
                
// <JMG> 2014-06-13T18:30, added orig
// specify attributes datatype using TIV style -- IsFloat32,IsString etc
int frm_hdf5_WriteAttributeAsType(hid_t loc_id,const char *grpQualifier,
    const char *attrName_p,const int frmApiDatatype,const int nElem,
    void *attrData_p);

// JMG> 2014-06-24T13:05, added orig. custom sprintf() style attribute
int frm_hdf5_WriteAttributeCustomString(hid_t loc_id,
    const char *grpQualifier, const char *attrName_p,char *fmtSpecStr_p, ...);

// <JMG> 2014-06-13T18:31, added orig
hid_t  frm_hdf5_CreateGroup(hid_t loc_id,const char *newGroupName_p);
hid_t  frm_hdf5_OpenGroup(hid_t loc_id, const char *groupName_p);
void   frm_hdf5_CloseGroup(hid_t groupId);


#ifdef HDF5_NOT_FINISHED_YET
hid_t frm_hdf5_CreatePrep(const int compLevel,const int hRank, 
        hsize_t *dims, hsize_t *chunkSizes,  hid_t *resultDspace_p, const float64 fillval);
        
herr_t frm_hdf5_FetchDcplAndDspaceByGrid(const int dsGridRes,
    Hdf5GridProps *hGridProp_p, hid_t *dcpl_p,  hid_t *dspace_p) ;

// This calls frm_hdf5_CreatePrep() ONCE per supported EASGRID res {1km,3km,9km}
// and stores all results in hGridProp_p object
herr_t frm_hdf5_GridInit(const int dsGridRes,Hdf5GridProps *hGridProp_p);

hid_t frm_hdf5_CreateDataset(Hdf5GridProps *hGridProp_p,char *groupIdStr,char *granulePath_p,
                char *datasetName, const int dsType_id,const int dsGridRes,
                hid_t dcpl, hid_t dspace, hid_t file_id) ;

// <JMG> added 2014-09-05T13:34, as per Joe A suggestion
char *frm_hdf5_GetLibraryVersion(char *hdf5VersionStr);
                
#endif

#endif
