// ---------------------------------------------------------------
// module       : frm_tiv_api.h
// purpose      : Implement "framework" component headers for the 
//                universal-union-of-typeful-atoms: a 'type-independent variable' 
//                API adapted from the NASA EOS UM MODIS (MUM) API.
// project      : SMAP L4C science development.
// dependencies : HDF4 typedefs e.g. 'float32'|'int32' and DFNT_UINT64, DFNT_FLOAT64, etc
// revision    : v2.4-2014-03-08T12:44:00, changed TI_ATOM* TIV_ATOM* to 
//               separate namespace from Lucas uuta.h/uuta.c TI_ATOM
// history     : v2.3-2014-02-15T19:41:00,added idempotence test on TIV_ATOM_DEF also
// history     : v2.3-2014-02-15T19:41:00,added idempotence test on TIV_ATOM_DEF also
// history     : v2.2-2014-02-10T11:57:00, added TIV_ATOM_PTR_DEF and TIV_INT64_DEF tests
// history     : v2.0-2014-01-21T18:04:00-07:00,jmg
// history      : 2014-01-21T11:17:00, minor doc additions only.
// history      : 2014-01-08T09:01:00-- introduces UUTAObj for transition to
//                thread-safe use, and re-factor as TIV_xxx for use in 
//                latest: memObjCalloc(),memCalloc(), xLogMsg(),xLocalLogMsg()
// history      : 2014-01-06T03:10:00-07:00,madison,jmg
// history      : 2013-11-08T08:00, re-factored and extended MUM API UUTAs
// author       : joe glassy, Lupine Logic Inc.
// notes        : 
// ---------------------------------------------------------------


#ifndef FRM_TIV_API_DEF 
#define FRM_TIV_API_DEF 1

// define version tag
#define FRM_TIV_VERSTAG "v2.4-2014-03-08T12:44:00"

// define identity signature
#define HAVE_FRM_TIV   1

// include all common, core symbols for frm_api
#include "frm_core.h"

// max length of a UUTA internal label
#define MAX_UUTA_LBL_LEN  64

#ifndef TIV_ATOM_DEF
#define TIV_ATOM_DEF 1
// TBD: migrate namespace to TIV_ATOM
typedef union {
 int8    aInt8   ;
 uint8   aUint8  ;
 int16   aInt16  ;
 uint16  aUint16 ;
 int32   aInt32  ;
 uint32  aUint32 ;
 int64   aInt64  ;
 uint64  aUint64 ;
 float32 aFlt32  ;
 float64 aFlt64  ; 
} TIV_ATOM ;
#endif

// Type-independent union (translation-layer) of pointers to atomic types
// TBD: migrate namespace to TIV_ATOM_PTR

#ifndef TIV_ATOM_PTR_DEF
#define TIV_ATOM_PTR_DEF 1

typedef union {
 int8    *aInt8_p   ;
 uint8   *aUint8_p  ;
 int16   *aInt16_p  ;
 uint16  *aUint16_p ;
 int32   *aInt32_p  ;
 uint32  *aUint32_p ;
 int64   *aInt64_p  ;
 uint64  *aUint64_p ;
 float32 *aFlt32_p  ;
 float64 *aFlt64_p  ; 
 void    *aVoid_p   ;
} TIV_ATOM_PTR ;
#endif


// IMPLEMENTATION NOTES:
//  --by "atSlot" we refer to an array index, and NOT to a byte offset.
//  --use: sizOfOne = DFKNTsize(DFNT_FLOAT32) to determine n. bytes per single scalar object
//         of a given supported datatype

// --------------------------------------------------------------------
// wrapper mini-container package for one UUTA instance, storing basic metadata
typedef struct tagTivUUTAObj {
 TIV_ATOM_PTR uuta;
 int32  uDatatype ;// NCSA style datatype code for this UUTA, e.g. DFNT_INT32
 size_t uCount   ; // n. elements allocated,owned by this UUTAObj <JMG> added 7/8/2013
 char   uLabel[MAX_UUTA_LBL_LEN]; // optional text label for this high level UUTAObj
} UUTAObj ;


// --------------------------------------------------------------------
// Higher level TIV FUNCTIONS
char  *tiv_GetVersTag(void);

int32  tiv_SizeofTivType(const int tivType);

// Interpret contents of varBuf through context of a TIV_ATOM
//  to yield a strongly typed numeric value of tivDatatype.
void *frm_tivTypedValue(void *varBuf,const int tivDatatype);

int    tiv_Hdf4TypeToTivType(const int32 hdfDatatype);
int32  tiv_TivTypeToHdf4Type(const int tivType);
char   *tiv_NameofTivType(const int tivType,int *sizOfOne);

// added 1/28/2014..
int tiv_NameToTivType(const char *typeLabel,int32 *hdf4DatatypeId);


// UUTAObj higher level services
UUTAObj *tiv_CreateObj(const int32 dataType,const size_t nElements,char *varName);
void    tiv_ReleaseObj(UUTAObj *uObj_p);

void    *tiv_GetObjRawPtr(UUTAObj *uObj_p);

void    tiv_SetRawIntoObj(UUTAObj *uObj_p,void *srcData_p,const int32 dataType,const size_t srcSlot,const size_t nElements);

void    *tiv_setIntoUUTAObjData(UUTAObj *uObj_p,void *srcData_p,const size_t srcSlot,
			const int32 dataType,const size_t nElements);

void   *tiv_SetObjToObj(UUTAObj *destUObj_p, UUTAObj *srcUObj_p,
			const size_t destSlot,const size_t atSlot,const size_t nElements);

void    tiv_setIntoUUTAObj(UUTAObj *uObj_p,void *srcData_p,const int32 dataType,
			const size_t srcSlot,const size_t nElements);

void    tiv_setUUTAObjDatatype(UUTAObj *uObj_p,const int32 newDatatype);
int32   tiv_getUUTAObjDatatype(UUTAObj *uObj_p);

void    tiv_setUUTAObjLabel(UUTAObj *uObj_p,const char *newLabel);
char    *tiv_getUUTAObjLabel(UUTAObj *uObj_p);

void    tiv_setUUTAObjCount(UUTAObj *uObj_p,const size_t newCount);
size_t  tiv_getUUTAObjCount(UUTAObj *uObj_p);


// --------------------------------------------------------------------
// Simple (low-level) TIV FUNCTIONS
void *tiv_AllocRaw(const int32 dataType,const size_t nElements,char *varName);
void  tiv_ReleaseRaw(const TIV_ATOM_PTR *uuta_p,char *varName);

// Copy raw data from SOURCE array of datatype, to destination array of same datatype, 
//      and rtn ptr to new dest as well. Prior, was tiv_copyUUTA
void *tiv_CopyRawFromSlot(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot,const size_t nElements);
// void *l4c_copyUUTA(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot,const size_t nElements);



// PRIOR: void *l4c_copyUUTA_to_UUTA(void *vDest_p,void *vSource_p,const int32 dataType,
//                        const size_t srcSlot,const size_t dstSlot,size_t nElements);
// copy nElements of SOURCE UUTA, starting at SrcOffset, into DESTINATION UUTA, at DestOffset.
void *tiv_CopyRawToRaw(void *vDest_p,void *vSource_p,const int32 dataType,
			const size_t srcSlot,const size_t destSlot, size_t nElements);


// Retrieve ONE element from  UUTA SOURCE, starting at offset, storing in a destination (scalar)
//          where each share a common datatype. 
void *tiv_GetOneFromRaw(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot);

// Assign ONE element from vSource_p of dataType, placed into vDest_p at position 'atSlot'
void tiv_SetOneFromSourceSlot(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot);

// Retrieve ONE element from vSource_p of dataType, placed into vDest_p, at position 'atSlot'
void *tiv_GetOneFromArray(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot);


#endif



