// ---------------------------------------------------------------
// module  : nsb_api_prot.h
// purpose : implement function prototypes for the thread-safe concurrent
//           string-block (NSB) in-memory DB implementation of the 
//           name-value-pair (NVP) processing api
// revised : 2015-09-28T16:43, added nsbCreateDatabase() for custom user sized NSBs
// revised : 2014-09-03T15:01, added nsbDoesKeyExist() boolean..
// revised : 2014-02-12T22:52:00,madison,jmg. added nsbSchemaPath[] capability
// history : 2014-01-03T13:42:00, initial re-factoring from nvp_strblk generation.
// author  : joe glassy, Lupine Logic Inc.
// notes   : n/a
// ---------------------------------------------------------------

#ifndef NSB_API_PROT_DEF
#define NSB_API_PROT_DEF 1

// ---------------------------------------------------------------
// import local headers

#include "frm_api.h"

#include "nsb_api.h"

// ---------------------------------------------------------------
// FUNCTION PROTOTYPES
// ---------------------------------------------------------------

// ------------------------------------------------------------
// module: nsb_dbsys.c
int nsbParseSchema(NsbObj *nsb_p,const char *keywordToken, char *valueToken);
int nsbParseInstance(NsbObj *nsb_p,const char *keywordToken,char *valueToken);
int nsbParseImport(NsbObj *nsb_p,const char *keywordToken,char *valueToken);
int nsbParseExport(NsbObj *nsb_p,const char *keywordToken,char *valueToken);

// <JMG> 2015-09-30, modified semantics
NsbObj *nsbImportDbFromTextfile(const char *nsbTextPath_p,const char *keyValDelims_p,int *nRecordsFound);

// <JMG> 2015-09-30, nsbExportDbToText name changed to nsbExportDbToTextfile() and
// semantics modified so that if nsbOutputTextPath_p is NULL or empty string, this
// causes output to be routed to stdout.
int    nsbExportDbToTextfile(NsbObj *nsb_p,char *nsbOutputTextPath_p,const int delimChr);


// Remember: user defined 'first' and 'last' indices here are Base0 not Base1
// and in order for delimChr to be used it MUST be >= 10 e.g. linefeed, where
// 13=carriage return, or a literal symbol like "," may also be specified.
char **nsbExportDbToStringtable(NsbObj *nsb_p,const int userFirstIdx,const int userLastIdx,const int delimChr);

size_t nsbExportDb(NsbObj *nsb_p,char *nsbOutputPath_p);

int    nsbGetStringblockRecsize(NsbObj *nsb_p);
void   nsbSetStringblockRecsize(NsbObj *nsb_p,const size_t newRecsize);

void   nsbSetStringblockRecordCount(NsbObj *nsb_p,const size_t newNvpcount);
size_t nsbGetStringblockRecordCount(NsbObj *nsb_p);

// <JMG> 2015-09-24T12:35, added for v2.x of NSB
int nsbGetMaxKeyWidth(NsbObj *nsb_p);
int nsbGetMaxValueWidth(NsbObj *nsb_p);

// Set|Get special delimiters for separating KEYWORD <delim> VALUE
// PROPERTY = {VALUE}, up to 15 different symbols may be specified.
char *nsbGetDelimiters(NsbObj *nsb_p,char *userDelimiterBuffer_p);
int  nsbSetDelimiters(NsbObj *nsb_p,const char *newCustomDelimiterSymbols);

// Set|Get special delimiters for separating multi-value items in a single
// string, such as properties of a dataset, up to 15 chars, e.g. 
// PROPERTY = {VALUE}|"myVar,float32,minVal,maxVal,units|longName|shortName"
int  nsbSetDelimiterMultiValue(NsbObj *nsb_p,const char *compoundItemDelimters_p);
char *nsbGetDelimiterMultiValue(NsbObj *nsb_p,char *userBufferCompoundDelimiters_p);

// ------------------------------------------------------------
// module: nsb_api.c

// mandatory 'constructor' call to create fresh but empty parent
// Note that when complete nsb_p->nsbLabel contains a fully qualified
// name that can be passed to memObjFree() to release its memory
NsbObj *nsbCreateInstance(const char *newNsbLabel);

// <JMG> 2015-09-28T16:43:00
// New,optional direct method for users to create their own custom
//  sized NSB (empty) database that does NOT depend on parsing a RIP file.
//  To be used with caution, observing current key and value width limits
//  that are easily retrieved.
int nsbCreateDatabase(const size_t nvpMaxRecordCount,
 const int keyMaxWidth,const int valMaxWidth,const char *newNsbDatabaseLabel);

// 'destructor' call.. or rely on memObjGarbageCollect()
void nsbReleaseInstance(NsbObj *nsb_p);
        
// service fx to fetch fully qualified names, of: instance, keytable, and 'core' stringblock.
char  *nsbGetInstanceName(NsbObj *nsb_p,char *keyTblName,char *coreName);

int    nsbSearchForSupportedKey(NsbObj *nsb_p,const char *targetKey) ;
char **nsbSetSupportedKeyDictionary(NsbObj *nsb_p,const char *keycatalogFilepath);

int    nsbInitStringblock(NsbObj *nsb_p,const size_t nNvps,
            const int keyMaxwidth,const int valMaxwidth);
            
int    nsbFreeStringblock(NsbObj *nsb_p);

// optional, allows user to declare an ancillary or schema RIP file to also
// size the database to fit.
int nsbDeclareSchemaPathToParse(NsbObj *nsb_p,const char *schemaRipPath_p) ;

// ------------------------------------------------------------
// module: nsb_accessors.c

// <JMG> 2014-07-31 added.
char *nsbGetVersion(char *versBuffer_p);

// <JMG> 2014-09-03, added , was missing earlier. returns 1 if keyword exists, otherwise 0.
int  nsbDoesKeyExist(NsbObj *nsb_p,const char *targetKey_p);

// <JMG> 2014-07-31 added.
int  nsbGetKeywordCount(NsbObj *nsb_p,const char *targetKey_p,const int matchPolicy);


// low level 'store' of a key,value (by index), and combined key,value into stringblock
int   nsbStoreKey(NsbObj *nsb_p,const int ithElement,const char *keyword);
int   nsbStoreValue(NsbObj *nsb_p,const int ithElement,const char *valueStr);
int   nsbStoreKeyAndValue(NsbObj *nsb_p,const int ithElement,const char *keyword,const char *valueStr);
// low level 'retrieves' of a key,value (by index), and combined key,value from stringblock
char *nsbRetrieveKey(NsbObj *nsb_p,const int ithElement,char *keyword);
char *nsbRetrieveValue(NsbObj *nsb_p,const int ithElement,char *valueStr);
int   nsbRetrieveKeyAndValue(NsbObj *nsb_p,const int ithElement,char *keyword,char *valueStr);

int   nsbStoreProperties(NsbObj *nsb_p,const int ithElement,const char *property_p);
char *nsbRetrieveProperties(NsbObj *nsb_p,const int ithElement,char *property_p);

char *nsbGetKeywordByIndex(NsbObj *nsb_p,const int ithNvp,char *keyBuffer);
char *nsbGetValueByIndex(NsbObj *nsb_p,const int ithNvp,char *valueBuffer);
int  nsbSetValueByIndex(NsbObj *nsb_p,const int ithNvp,char *valueBuffer);

char *nsbGetRecordByIndex(NsbObj *nsb_p,const int ithNvp,char *recordBuffer_p);

// TBD --> not coded or implemented yet (verify)
int  nsbSetKeywordByIndex(NsbObj *nsb_p,const int ithNvp,char *valueBuffer);

// PRIMARY ACCESSORS to GET or SET a given name-value-pair
char *nsbGetValueByKey(NsbObj *nsb_p,const char *targetKey_p,char *valueBuffer);
int  nsbSetValueByKey(NsbObj *nsb_p,const char *targetKey_p,char *valueBuffer);
// this erases or 'clears' the NVP value assoc with 'targetKey_p'
int  nsbClearValueByKey(NsbObj *nsb_p,const char *targetKey_p);

// <JMG> 2014-09-03T15:38, added, to support extended (compound) NVP 'value'
// strings consisting of a collection of comma-delimited property values
char    *nsbGetPropertyValueByKey(NsbObj *nsb_p,const char *targetKey_p,const int propOrderIdx,char *valueBuffer);
float64 nsbGetPropertyNumericByKey(NsbObj *nsb_p,const char *targetKey_p,const int propOrderIdx,float64 *resultFloat64_p);

// New accessors to directly interpret NVP string value as a "generic numeric" float64 value
float64 nsbGetNumericByKey(NsbObj *nsb_p,const char *targetKey_p,float64 *valFlt64_p);
float64 nsbGetNumericByIndex(NsbObj *nsb_p,const int ithNvp,float64 *valFlt64_p);

// <JMG> 2015-09-30, extended syntax to recognize preface symbol "!" as a
//   macro, meaning its 'value' string is interpreted as a file path, that 
//   when opened and read into the stringtable, itself contains an arbitrary 
//   number of 'value' items, one per line.
char **nsbGetMultiValueByKey(NsbObj *nsb_p,const char *targetKey_p,int *nElem);
// <JMG> 7/23/2014, extended syntax to match on partial keys as well
char **nsbGetMultiValueByPartialKey(NsbObj *nsb_p,const char *targetKey_p,int *nElem);

// verify, test yet to do
int  nsbAppendIntoCatalog(NsbObj *nsb_p,const char *keywordTok,const char *valuTok,const int ithNvp);

// scans a primary RIP submitted, and calls nsbScanSchemaRip() if optional schema path
// is ALSO made available.
int  nsbScanRip(NsbObj *nsb_p,FILE *ripf,int *nPrimaryNvps);

// <JMG> 2/16/2014 added user's delimStr_p option
int  nsbParseIntoStringblock(NsbObj *nsb_p,char *delimStr_p,FILE *ripf);

// This erases or 'clears' the NVP at the base0 index 'ithNvp'
int  nsbClearValueByIndex(NsbObj *nsb_p,const int ithNvp);

// <JMG> added 2/16/2014
void nsbSetSupportedKeyPolicy(NsbObj *nsb_p,const int newKeyPolicy);
int  nsbGetSupportedKeyPolicy(NsbObj *nsb_p);

#endif

