// --------------------------------------------------------------------
// module   : pkg_mgr.h
// purpose  : implement ADT's for PkgMgr mini API
// author   : joe glassy, Lupine Logic Inc.
// revision : 2020-07-08T19:17, changed int to int64_t for better hid_t compatibility in SetObjOwner,GetObjOwner,SetId,GetId
// revision : 2020-06-24, changed int to int64 for hid_t compatibility in SetObjOwner,GetObjOwner,SetId,GetId
// revision : 2.2-2017-02-27T16:17:00-07:00, doc updates only
// revision : 2014-08-21 added 'gain' pkObjGain for HDF4 style pack,unpacking
//            and pkObjOffset 
// revision : 2.0-2014-08-14T17:04, added to PkgObj hid_t pkObjH5d{type,dspace,dcpl,dsetId}
// revision : 1.9-2014-07-23T18:20:00, added to PkgObj valid_min,valid_max
//            along with assoc "get"|"set" functions.
// revision : 1.8-2014-02-26T19:08:00, changed pkgMgr_CreateObj() to 
//            accept objOwnerId not *PkgTheme which wasn't used and added
//            pkgMgr_CreateTheme(),pkgMgr_ReleaseTheme(), to support
//            singleton, anonymous un-bound Themes.
// revision : 1.7-2014-06-10T14:02, added to PkgObj: shortName,longName, genericPath,
//            and added several fx for robustness (pkgMgr_ReleaseAllObjects() etc)
// revision : 1.6-2014-03-05T16:08, changed PkgObj nElem and related counts to size_t
//            and added pkObjSelectFilter to facilitate discontiguous subset selection
//            of PkgObj from a common theme, added pkThemeSelectFilter, implementation TBD
// revision : 1.5-2014-02-19T13:00, added pkThemeMetadata[],pkObjMetadata[]
// revision : 1.4-2014-02-19T12:58:00,jmg, added pkgThemeReport() prototype
// revision : 1.3-2014-02-18T12:58:00,jmg, added initial PkgMg<-->memObj coupling
// revision : 1.2-2014-02-18T16:12:00-07:00, cross-integrated PkgMgr+MemObj by
//            adding to PkgTheme the cross-indexing member, (MemObj *pkThemeMemRegistry_p)
// and in PkgObj, added PKGMGR_DNA_STRLEN, and  pkObjDna[PKGMGR_DNA_STRLEN] to match up
//          : with memObj's vDnaLabel[]
// revision : 1.1-2014-02-01T10:56:00:00,jmg, continued initial devel
// history  : 1.0-2014-01-15T14:24:00,jmg (original)
//
// version  : 2.1
// notes    : Package Manager (PkgMgr) is a mini API implementing a session level,
//               2-level "in-memory registry" for an application, for selectively
//              exposing data and metadata within a large application. It is does
//              currently persist across sessions.

// notes    : v1.2, added to PkgTheme: PkgObj *pkTheme_PkgObjAry_p, to allow for
//            optional cross-integration of PkgMgr with MemObj registries, either 
//            on a straight 1:1 basis, or, allowing a given Theme to index into
//            any subset of contiguous or dis-contiguous MemObj slots. This also
//            allows PkgMgr mechanism, at the Theme organizational level,
//            to have complete access to memObj via either the direct memobj
//            accessors (Get|Set), or via special PkgMgr access calls. Further
//            by mapping a sub-collection of entities from a MemObj registry
//            into a contiguous collection of PkgObj entities, this yields
//            unparalleled flexibility for PkgMgr to serve as a scoping tool.
// --------------------------------------------------------------------

#ifndef PKG_MGR_DEF 
#define PKG_MGR_DEF 1

#define PKGMGR_VERS_TAG "2.2-2017-02-27T16:17:00-07:00"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <stddef.h>
#include <stdint.h>
#include <stdarg.h>   // for va_start(), va_end() etc
#include <assert.h>
#include <limits.h>
#include <math.h>
#include <sys/types.h>  // for getpid()
#include <unistd.h>     // for gethostname() and getpid()

#ifdef HAVE_UUID
#include <uuid/uuid.h>  // for uuid services, uuid_generate() etc..
#endif

// import HDF4 and HDFEOS required constants
#include "hdf.h"
#include "mfhdf.h"
// TBD: verify this HDFEOS dependency is now really needed
// #include "HdfEosDef.h"

// import HDF5 
// #include "hdf5.h"
// #include "hdf5_hl.h"

#include "frm_api.h"


// #define PKGMGR_DIAG 1 

// compile-time COUNTS 
// PkgMgr by default will create this many PkgThemes unless overridden
#define PKGMGR_MAX_THEMES    16
// PkgMgr by default will create this many PkgObj per PkgTheme, unless overridden

// <JMG> 2016-01-06, increased this hard limit to 128 until a dynamic
// (single or double linked list) or other heap based scheme is implemented
#define PKGMGR_MAX_OBJ      128

#define PKGMGR_DNA_STRLEN    32

// max length of a PkgObj or PkgTheme mid-sized label
#define PKGMGR_MAX_LBL_LEN    128

// max length of a PkgObj or PkgTheme name or label
#define PKGMGR_MAX_NAMELEN    512
// max length of a UUID string (sized extra intentionally)
#define PKGMGR_MAX_UUIDLEN     64
// max length of a units string descriptor
#define PKGMGR_MAX_UNITSTRLEN  32

// Careful, this is less than many HDF4/HDF5 hard limits, but in practice
// we rarely encounter arrays of greater rank (dimensionality) than 4, though
// this can easily be increased at build time
#define PKGMGR_MAX_DIMS         4

// PkgObj (or PkgThemes) possess a descriptor that communicates their
//  function, role, or use-context. Here are a few examples. These
//  are defined but not yet implemented in accessor member functions
enum PkgMgrUseContexts {
 IsThemeScalarVar ,   IsThemeArrayVar,   IsThemeMetadata, IsThemeUnits, 
 IsThemeFillvalue ,   IsThemeMinValid,   IsThemeMaxValid,
 IsThemeInputTheme,   IsThemeMeterology, IsThemeEcosystem,
 IsIsThemeException,  IsThemeSessionLog, IsThemeCustom
} ;


// when a PkgMgr's PkgTheme is "used", its state is set to IsOccupied..
enum PkgMgrStates {
  IsPkgUnoccupied = -1 , IsPkgOccupied = 0
} ;

// namespace: pkObjXXXXXXX
// The primary ADT is the multiple-instance PkgObj object, which stores
//   information pertinent to a single variable, string, entry, NVP, etc.
//   The properties shown here represent an initial set emphasizing HDF5
//   and CF use-cases. The plan in the future is to bind to a user defined
//   domain-specific, property structure at runtime via void *pkObjProperties_p
//   that matches a well-known set of {Get|Set| accessor functions servicing
//   the domain-specific use case.
//
//   As a "well-known-named" registry component, PkgObj emphasis is NOT
//   to create or act as the canonical STORE of a data variable, but is
//   intended to point to prior instances of a variable, array, or
//   function residing in memory elsewhere. As such, PkgMgr facilities are
//   never used to 'free' a referAgain, this “build-time” limit may be increased, but after doing so, the API must be rebuilt using Make.  ent pointed to by pkPbjContent_p. Freeing
//   memory for actual variables must be done elsewhere.
typedef struct tagPkgObj {
  // <JMG> 2020-06-24, changed pkObjId datatype of int to int64 for hid_t compatibility 
  int64_t     pkObjId   ;        // unique id, if interfacing to NSB, its the base0 index of the NVP. -1 =IsUnoccupied
  // <JMG> 2020-07-08 changed pkObjOwner datatype of int to int64_t for hid_t compatibility 
  int64_t pkObjOwner;        // ID of PkgTheme who owns me. If -1, no one owns me, so I am IsUnoccupied
  int     pkObjCategoryCode; // classifier, from predefined enum set {function, role, purpose, or use-context}
  int     pkObjDatatype;     // TIV , mem_obj style datatype descriptor: IsString,IsFloat32,IsUint32, etc.
  int     pkObjResolution;   // int code for dataset resolution, or may be expressed as a unit 1=1km, 3=3km ,9=9km
  int     pkObjRank ;        // rank if referent is an array, or pkObjRank=1 if it is a IsString|IsFilespec,etc
  int     pkObjMemSlotIdx;   // <JMG> added, 2014-02-18, to support cross-integration into memObj registry
  hid_t   pkObjH5dtype ;     // <JMG> added 2014-08-14, for more direct L4_C support
  hid_t   pkObjH5dspace;     // <JMG> added 2014-08-14, for more direct L4_C support
  hid_t   pkObjH5dcpl  ;     // <JMG> added 2014-08-14, for more direct L4_C support
  hid_t   pkObjH5dsetId;     // <JMG> added 2014-08-14, for more direct L4_C support
  uint8   pkObjSelectFilter; // boolean, allows for discontiguous subset selection within a theme.
  size_t  pkObjDims[PKGMGR_MAX_DIMS];  // dims[0..3], if referent is an array...
  size_t  pkObjSubtaskStartIdx; // index of 1st item to process (for threaded use)
  size_t  pkObjSubtaskEndIdx  ; // index of last item to process (for threaded use)
  size_t  pkObjNelem  ;         // total count of referent's elements across all dims, or, stringlength  
  double  pkObjFillval        ; // generic _Fillvalue, interp type as pkObjDatatype. Use as missing_value too.
  double  pkObjValidMin ;       // <JMG> 7/23/2014, added: valid_min
  double  pkObjValidMax ;       // <JMG> 7/23/2014, added: valid_max
  double  pkObjGain   ;  // <JMG> 2014-08-21 added 'gain' for HDF4 style pack,unpacking
  double  pkObjOffset ;  // <JMG> 2014-08-21 added 'offset for HDF4 style pack,unpacking
  char    pkObjUnitStr[PKGMGR_MAX_UNITSTRLEN];    // units string
  char    pkObjShortname[PKGMGR_MAX_LBL_LEN]; // <JMG> 6/10/2014, added, limit 128 chars 
  char    pkObjLongname[PKGMGR_MAX_LBL_LEN] ; // <JMG> 6/10/2014, added, limit 128 chars
  char    pkObjGenericPath[FILENAME_MAX];     // <JMG> 6/10/2014, added, limit FILENAME_MAX chars
  char    pkObjKey[PKGMGR_MAX_NAMELEN]; // key, text label identifys PkgObj, limit 512 chars
  char    pkObjDna[PKGMGR_DNA_STRLEN];  // <JMG> added 2/18/2014, for cross-integration of memObj regis entries
  char    pkObjMetadata[FILENAME_MAX];  // <JMG> added 2/25/2014,generic metadata container for PkgObj
  void    *pkObjContent_p;              // ptr to primary referent obj as defined and allocated elsewhere
} PkgObj ;



// multiple-instance PkgTheme object itself stores a collection of K PkgObj instances
// a THEME is defined here as a unifying classifier such that the collection of PkgObj's owned
// by this THEME all share a common underlying purpose, role, identity, or function.
// --the PkgThemeCategory describes my function,role, purpose or use-context
// Notes: 2/18/2014, added: pkgThemeMemRegistry_p, pointer to allow for tight cross integration
// between PkgMgr and MemObj
// Notes: Class/Instance namespace distinction uses camel-case: ADT PkgTheme Instance: pkgTheme
// Notes: member namespace: pkThemeXXXXXXX
typedef struct tagPkgTheme {
  int    pkThemeId  ;       // unique ID of this PkgTheme instance
  int    pkThemeOwner;      // id of PkgMgr who owns me
  char   pkThemeLabel[PKGMGR_MAX_NAMELEN]; // Text label identifying this THEME
  char   pkThemeUUID[PKGMGR_MAX_UUIDLEN]  ; // UUID assigned to identify this THEME
  char   pkThemeDna[PKGMGR_DNA_STRLEN];     // self-referential DNA signature of myself
  char   pkThemeMemRegDna[PKGMGR_DNA_STRLEN]; // signature of WHICH exact memobj registry I work within
  char   pkThemeHostPidThrdAddrId[PKGMGR_MAX_NAMELEN];// host-pid-thread-addr 4 part, dot-qualified identifier
  char   pkThemeCategory[PKGMGR_MAX_NAMELEN]; // categorical descriptor, for K "peer" themes that share a common bond
  char   pkThemeMetadata[FILENAME_MAX]; // added 2/25/2014,generic metadata for Theme
  int    pkThemeObjCount ;        // count of max available PkgObj instances, of which only a subset may be in use
  uint8  pkThemeSelectFilter;   // allows for theme level discontiguous subset selection
  MemObj *pkThemeMemRegistry_p; // points directly a memObjRegistry[MAX_MEMOBJ_CANDIDATES] ;
  PkgObj *pkTheme_PkgObjAry_p;  // pointer to an array of PkgObj instances, one per scalar, array, NVP, string,etc
} PkgTheme ;


// namespace: pkMgrXXXXXX
// single overarching "overseer" controls all lower level registry abstractions
typedef struct tagPkgMgr {
 int  pkMgrId ; // unique ID of this PkgMgr instance
 char pkMgrLabel[PKGMGR_MAX_NAMELEN] ; // text label identifying or describing this PkgMgr
 char pkMgrUUID[PKGMGR_MAX_UUIDLEN]  ; // UUID uniquely identifying this PkgMgr instance
 char pkMgrDna[PKGMGR_DNA_STRLEN]    ; // self-referential DNA signature of myself
 char pkMgrThemeCommonCategory[PKGMGR_MAX_NAMELEN]; // optional,as appropriate, a shared-theme-category descriptor
 int  pkMgrThemeCount  ;  // count of max available PkgTheme's defined, some used, some empty
 PkgTheme *pkMgrThemeAry_p ; // ptr to array of pkgMgr_nMaxThemes, of which only pkgMgr_nThemesUsed are in use.
} PkgMgr ; 

#endif
