// ------------------------------------------------------------------
// module       : frm_hdf_io.c
// purpose      : Implement "framework" level HDF4, HDF54 basic set of I/O wrappers
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2015-08-25T16:00:00, adjusted doVerboseMessaging=0, back to 0 for now
// revised      : 2014-09-30T10:17:00, frm_hdf5_DatasetExists() refined,
//                changed to use H5Dopen(),H5Dclose()
//                atomic open|close instead of using H5LTfind_dataset()
//                while fails to find group-nested datasets.
// history      : TBD 2014-09-02, in frm_hdf5_WriteAttributeAsType() must
//                re-confirm if H5LTset_attribute_xxx() family works
// history      : v1.7.6-2014-08-08T17:06:00, refined frm_hdf5_DatasetExists() and
//                frm_hdf5_GetDatasetInfo(), which occassionally reported a dataset
//                missing when it was present in a HDF5 (SOC) file.
// history      : 2014-07-294, in frm_hdf5_Openfile(), 
//                removed call to H5Pcreate() until further work is done
// history      : 2014-07-17T18:29:001, fixed frm_hdf5_Openfile() bug in
//                facpl ref.
// history      : 2014-06-16T12:31, added existence test in fx like
//                
// history      : 2014-06-13T11:49:00, added serveral inquiry functions:
//                frm_hdf5_GetAttributeCount(0, frm_hdf5_GetObjectCount()
//                frm_hdf5_GetComment(),SetComment() etc.
// history      : 2014-03-09T15:26, added 'doVerboseMessaging' to allow suppression
//                of too much statu info..
// history      : 2014-02-21T20:36, checked..:
// history      : 2014-02-10T16:32:00-07:00, added more, to work off TBDs 
// history      : 2014-01-20T09:54:00-07:00, added more, to work off TBDs 
// history      : 2014-01-05T21:06:00-07:00, added more, to work off TBDs
// history      : 2013-12-03T12:49:00-07:00 refactor pass 1, for D4
// version      : see VERS macro
// dependencies : HDF4 v4.2.6.x lib, HDF5 v1.8.x lib 
// notes        : If libuuid.a available, use via HAVE_UUID.
//              : If HDF4 unavailable, unset HAVE_HDF4 for a HDF5 only build
//              : If HDF5 unavailable, unset HAVE_HDF5 for a HDF4 only build
// ------------------------------------------------------------------


// includes frm_api.h as well
#include "frm_hdf_io.h"

// default: set detailed status messaging to off=0
static int doVerboseMessaging = 0 ;

// <JMG> 2014-06-25, temporary long-lived storage of user-block size
static size_t masterUserBlockSize  = 0 ;
// used for retrieving user block size of most recently opened HDF5
static size_t existingUserblocksize = 0;


// ------------------------------------------------------------------
// function     : frm_hdf_GetVersTag()
// purpose      : retrieve  version tag for caller
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-01-21T11:51:00-07:00 
// history      : 2013-12-03T12:49:00-07:00 refactor pass 1, for D4
// notes        : The lack of a "hdf4" or "hdf5" specific qualifier is intentional
//                since this package applies to both variants.
char *frm_hdf_GetVersTag(void)
{
    return((char *)FRM_HDF_IO_VERSTAG);
    
} // end:: frm_hdf_GetVersTag()


// ------------------------------------------------------------------
// function     : frm_hdf5_SetUserblocksize(newSizeInBytes)
// purpose      : Assign a (global) userblock size, must be called
//                prior to creating a new HDF5 file.
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-06-25T17:41:00-07:00 
// history      : n/a
// notes        : If cleared using frm_hdf_SetUserblocksize(0) the
//                default userblock size is then used.
// notes        : Warning, not yet thread safe since a static
//                is referenced; TBD Must enclose in mutex locks.
void frm_hdf5_SetUserblocksize(size_t newUbSize)
{
    masterUserBlockSize = newUbSize;
    
} // end:: frm_hdf_GetVersTag()


// ------------------------------------------------------------------
// function     : frm_hdf5_SetUserblock()
// purpose      : Assign userblock data to an HDF5 file. 
// inputs       : hid_t file_id,
//                size_t ubDataSiz --size in bytes of data to write,
//                  includes any null term if a string etc.
//              void *userBlockdata_p -- ptr to data itself
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-06-25T18:46:00, orig
// history      : n/a
// notes        : Not yet TESTED! This is trivial when HDF5 
//                file is closed, but trickier to do when open, on-the-fly
void frm_hdf5_SetUserblock(const char *hdf5Path,const size_t ubDataSiz,char *userBlockdata_p)
{
 char fxName[256] = {0} ;
 unsigned char padfill = '\0' ;
 size_t ubExistingSize = 0 ;
 // size_t padByteCount = 0 ;
 
 int    i =0 ;
 FILE *inf = NULL;
 
        strcpy(fxName,(char *)__func__);
        // TBD: first open as an HDF5 file and check size of userblock to 
        // assure there is (ubDataSiz) bytes of room available.
        
        xLogMsg("@W In(%s) USE-with-CAUTION, NOT YET TESTED!\n~",fxName);
        
        // file_id = frm_hdf5_Openfile(hdf5Path,fxName);
        // ubExistingSize = frm_hdf5_GetUserblocksize(file_id);
        // frm_hdf5_Closefile(file_id,fxName);
        ubExistingSize  =0 ;  // for now, until tested, force at 0
        
        // open HDF5 as a raw binary file in read+write (update) mode
        if( ubExistingSize > 0 && 
            ubExistingSize >= ubDataSiz && (inf=fopen(hdf5Path,"rb+"))!=NULL)
        {
                fseek(inf,0L,SEEK_SET);
                fwrite((char *)userBlockdata_p,sizeof(char),ubDataSiz,inf);
                // If user's userblock data does not fill up the user block
                // pad it out with nulls for now
                if(ubExistingSize > ubDataSiz)
                {   // pad out to end of userblock
                    for(i=0; i < (int)(ubExistingSize - ubDataSiz); i++)
                        fwrite(&padfill,sizeof(char),1,inf);
                }
                fclose(inf);
        } // end:: if valid ub size and if HDF5 file could be opened ok
    
} // end:: frm_hdf5_SetUserblock()


// ------------------------------------------------------------------
// function : frm_hdf5_GetUserblock()
// purpose  : retrieve contents of an HDF5 file's user block, which is 
//            totally opaque to the HDF5 API library proper.
// inputs   : const char *hdf5Path,const size_t ubDataSiz,char *userBlockdata_p
// revised  : 2014-06-25T18:46:00, orig
// returns  : void
// notes    : Recommend user does NOT perform this operation on an open 
//            HDF5 file.
// notes    : <JMG> 2014-06-25, Warning: Minimally implemented but not 
//             yet tested.
void frm_hdf5_GetUserblock(const char *hdf5Path,const size_t ubDataSiz,char *userBlockdata_p)
{
 char fxName[256] = {0};
 
    strcpy(fxName,(char *)__func__);
    
    frm_ioReadbinary(hdf5Path,IsUint8,ubDataSiz,userBlockdata_p);
    
    xLogMsg("@W In(%s) Minimally implemented, not yet fully tested!\n~",fxName);
    
} // end:: frm_hdf5_GetUserblock()


// ------------------------------------------------------------------
// function     : s = frm_hdf5_GetUserblocksize(fileCreationPropListID)
// inputs       : hid_t file_id -- HDF5 file handle file_id
//                for a presumed open file.
// purpose      : Retrieve a (global) userblock size
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-06-25T17:41:00-07:00 
// history      : n/a
// notes        : when called with an id =0, this returns the
//                last value stored in (static) existingUserblocksize
//                instead, which may be 0.
// notes        : Warning, not yet thread safe since a static
//                is referenced; TBD Must enclose in mutex locks.
size_t frm_hdf5_GetUserblocksize(const hid_t file_id)
{
 herr_t ret = -1;
 hsize_t ubSize = 0;
 hid_t   fcpl_id = 0;
 
    if(file_id >0)
    {
            // fetch cpl for this (open) file_id
            fcpl_id = H5Fget_create_plist(file_id);
            // fetch userblock size for this file_id
            ret = H5Pget_userblock(fcpl_id, &ubSize);
            // if successful, updates 1 static var, where value
            // set may be 0 when no userblock exists.
            if( ret > 0)
                existingUserblocksize = (size_t)ubSize;
    } else {
            ubSize = (hsize_t)existingUserblocksize;
    }
    return ubSize;
} // end:: frm_hdf5_GetUserblocksize()


// ------------------------------------------------------------------
// HDF5 I/O components
// ------------------------------------------------------------------

hid_t frm_hdf5_TIV_datatype_to_H5T(const int TIV_dType)
{
 hid_t rtnDatatype = -1;
 
        switch(TIV_dType)
        {
        case  IsInt8    : rtnDatatype = H5T_NATIVE_INT8  ; break;
        case  IsUint8   : rtnDatatype = H5T_NATIVE_UINT8 ; break;       
        case  IsInt16   : rtnDatatype = H5T_NATIVE_INT16 ; break;       
        case  IsUint16  : rtnDatatype = H5T_NATIVE_UINT16; break;       
        case  IsInt32   : rtnDatatype = H5T_NATIVE_INT32 ; break;       
        case  IsUint32  : rtnDatatype = H5T_NATIVE_UINT32; break;       
        case  IsInt64   : rtnDatatype = H5T_NATIVE_INT64 ; break;       
        case  IsUint64  : rtnDatatype = H5T_NATIVE_UINT64; break;       
        case  IsFloat32 : rtnDatatype = H5T_NATIVE_FLOAT ; break;       
        case  IsFloat64 : rtnDatatype = H5T_NATIVE_UINT64; break;       
        default :
            xLogMsg("@E frm_hdf5_H5T_Datatype_to_TIV: TIV type(%d) not yet supported~\n",
                    TIV_dType);
            break;
        } // end : switch on TIV datatype identifier, return H5T equivilent
        
 return rtnDatatype;
    
} // end:: frm_hdf5_TIV_datatype_to_H5T()



// notes: requires C99 level . Uses TIV enums...
int frm_hdf5_H5T_Datatype_to_TIV(const hid_t h5_Dtype)
{
 int rtnDatatype = IsNoType;

#ifdef NOT_DONE_H5T_DOES_NOT_REDUCE_TO_INT_CONSTANTS  
        switch(h5_Dtype)
        {
        case H5T_NATIVE_INT8  : rtnDatatype = IsInt8   ; break;
        case H5T_NATIVE_UINT8 : rtnDatatype = IsUint8  ; break;     
        case H5T_NATIVE_INT16 : rtnDatatype = IsInt16  ; break;     
        case H5T_NATIVE_UINT16: rtnDatatype = IsUint16 ; break;     
        case H5T_NATIVE_INT32 : rtnDatatype = IsInt32  ; break;     
        case H5T_NATIVE_UINT32: rtnDatatype = IsUint32 ; break;     
        case H5T_NATIVE_INT64 : rtnDatatype = IsInt64  ; break;     
        case H5T_NATIVE_UINT64: rtnDatatype = IsUint64 ; break;     
        case H5T_NATIVE_FLOAT : rtnDatatype = IsFloat32; break;     
        case H5T_NATIVE_UINT64: rtnDatatype = IsFloat64; break;     
        default :
            xLogMsg("@E frm_hdf5_H5T_Datatype_to_TIV: HDF5 type(%d) not yet supported!~\n",
                    (int)h5_Dtype);
            break;
        } // end : switch on HDF5 datatype identifier
#endif  

        if(h5_Dtype == H5T_NATIVE_INT8) {
            rtnDatatype = IsInt8 ;
        } else if (h5_Dtype == H5T_NATIVE_UINT8)  {
            rtnDatatype = IsUint8 ;         
        } else if (h5_Dtype == H5T_NATIVE_INT16)  {
            rtnDatatype = IsInt16 ;         
        } else if (h5_Dtype == H5T_NATIVE_UINT16) {
            rtnDatatype = IsUint16 ;            
        } else if (h5_Dtype == H5T_NATIVE_INT32)  {
            rtnDatatype = IsInt32 ;         
        } else if (h5_Dtype == H5T_NATIVE_UINT32) {
            rtnDatatype = IsUint32 ;
        } else if (h5_Dtype == H5T_NATIVE_INT64)  {
            rtnDatatype = IsInt64 ;         
        } else if (h5_Dtype == H5T_NATIVE_UINT64) {
            rtnDatatype = IsUint64 ;            
        } else if (h5_Dtype == H5T_NATIVE_FLOAT)  {
            rtnDatatype = IsFloat32 ;           
        } else if (h5_Dtype == H5T_NATIVE_DOUBLE) {
            rtnDatatype = IsFloat64;
        }
        
  return rtnDatatype;   
} // end::frm_hdf5_H5T_Datatype_to_TIV()


// ------------------------------------------------------------------
// function     : frm_hdf5_DatasetExists()
// purpose      : Boolean, returns 1 if dataset exists, otherwise 0
// inputs       : hid_t file_id,const char *datasetName_p
// returns      : int, 1 if dataset exists, 0 if it does not
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-09-30T10:17:00, changed to use H5Dopen(),H5Dclose()
//                atomic open|close instead of using H5LTfind_dataset()
//                while fails to find group-nested datasets.
// history      : 2014-01-21T11:51:00-07:00 
// history      : 2013-12-03T12:49:00-07:00 refactor pass 1, for D4
//
// notes        : old logic:    ret = H5LTfind_dataset(file_id, datasetName_p);
//              : but call interface left identical to orig
int frm_hdf5_DatasetExists(hid_t file_id,const char *datasetName_p)
{
 char fxName[256] = {0};
 int doesExist = 0 ; // boolean, True=1 when exists, False=0 when absent
 herr_t ret = 0 ;
 hid_t  dset_id = -1 ;
 
    strcpy(fxName,__func__);
    
    // Returns NEGATIVE id if failed, otherwise returns a non-negative valid ID
    //  although in practice, I've never seen a 0 returned as a valid ID even
    //  though it is technically a positive number.
    if((dset_id= H5Dopen(file_id,datasetName_p,H5P_DEFAULT)) > 0)
    {
        // Next, to double-check, make sure the close is also successful!
        // H5Dclose() "Returns a non-negative value if successful; otherwise returns a negative value"
        if((ret = H5Dclose(dset_id)) < 0)
        {
            xLogMsg("@E In(%s) ERROR failed to close (%s) verification-dataset(%d) in file(%d)~\n",
                fxName, datasetName_p, dset_id, file_id);
        } else {
            doesExist = 1 ;     
        }
    } else {
        doesExist = 0 ; // failed outright.
        xLogMsg("@E In(%s) ERROR failed to identify or locate dataset(%d) in file(%d)~\n",
                fxName, dset_id, file_id);
    } // end:: block assigning test-dataset id

 return doesExist ;
 
} // end:: frm_Hdf5_DatasetExists()


// ------------------------------------------------------------------
// function : frm_hdf5_GetSizeOfOne()
// returns  : size_t , n. of bytes for one element of dsTypeId
// purpose  : 
size_t frm_hdf5_GetSizeOfOne(hid_t dsTypeId)
{
 size_t sizOfOne = 0;
 
 sizOfOne = H5Tget_size(dsTypeId);

 return sizOfOne;   
} // end:: frm_hdf5_GetSizeOfOne()
 

// ------------------------------------------------------------------
// function     : frm_hdf5_GetAttributeCount()
// purpose      : Retrieve n. of attributes owned by file,group,dataset object
//                identified by loc_id. 
// inputs       : hid_t loc_id -- id of object (file,group,dataset,etc) 
// returns      : size_t numberOfAttributes owned by object identified via loc_id
// revision     : 2014-06-12T17:12:00, added orig.
// notes        : Retrieves dataset ID,datatype,dcpl,dspace,rank and dims
size_t frm_hdf5_GetAttributeCount(hid_t loc_id)
{
    herr_t ret = -1;
    H5O_info_t object_info ;
    
    ret = H5Oget_info(loc_id, &object_info);    
    
    return (size_t)object_info.num_attrs ;

} // end:: frm_hdf5_GetAttributeCount() 
 

// ------------------------------------------------------------------
// function     : frm_hdf5_GetObjectCount()
// purpose      : Retrieve n. of first-class-citizen objects owned
//                at any of these levels:  by file,group,dataset object
//                identified by loc_id. 
// inputs       : hid_t loc_id -- id of object (file,group,dataset,etc) 
// outputs      : int *nAttrib --n. of attributes owned if any
//              ; size_t numberOfAttributes
// returns      : size_t number of objects (groups|datasets) owned 
//                by the object identified via input ID, loc_id
// revision     : 2014-06-13TT10:40:00, added orig.
// notes        : n/a
size_t frm_hdf5_GetObjectCount(hid_t loc_id,int *nAttrib)
{
    herr_t ret = -1;
    hsize_t nObjects = 0; // n. "first-class-citizen" objects   
    H5O_info_t object_info ;
    
    // first, fetch count of any attributes owned directly by loc_id
    ret = H5Oget_info(loc_id, &object_info);    

    // also assigns the count of attributes this loc_id owns, if any    
    *nAttrib = (int)object_info.num_attrs ;

    ret = H5Gget_num_objs(loc_id, &nObjects);
    
    return (size_t)nObjects ;

} // end:: frm_hdf5_GetObjectCount()
 

// ------------------------------------------------------------------
// function     : frm_hdf5_GetObjectInfo()
// purpose      : Retrieve basic object metadata descriptor info.
//                loc_id. 
// inputs       : hid_t loc_id -- id of object (file,group,dataset,attribute) 
// outputs      : int *nAttrib --n. of attributes owned if any
//              ; size_t numberOfAttributes
// returns      : int , equiv FRM_API enum identifier: IsHdf5File,
//                 IsHdf5Group,IsHdf5Dataset, IsHdf5Attribute, etc.
// revision     : 2014-06-13TT13:34:00, added orig.
// notes        : HDF5 designers made this a lot harder than it needed
//                to be, as same problem with not returning datatypes
//                as simple integer id values..sigh.
// notes        : choices: IsHdf5Group, IsHdf5Dataset, IsHdf5NamedDatatype, 
//              IsHdf5Attribute, IsHdf5UnknownType
//
int frm_hdf5_GetObjectInfo(hid_t loc_id,char *myName_p,int *nAttrib,int *nObj)
{
    herr_t retInfo = -1;
    herr_t retNobj = -1;
    
    hsize_t nObjects = 0; // n. "first-class-citizen" objects   
    H5O_info_t object_info ;
    // H5O_type_t objClass ;
    H5I_type_t objIdentityType ; // used by H5I_get_type(file_id)
    // H5I_type_t childsType ; // used by H5I_get_type(file_id)
    int rtnObjectType = -1 ;
    char fxName[256] = {0} ;

    *nAttrib = 0 ; // initialize
    *nObj    = 0 ; 
    
    strcpy(fxName,(char *)__func__);
    
    // TBD: check if we can first verify if this loc_id even exists
    
    // fetch CORE HDF5 type as:
    // {H5I_FILE, H5I_GROUP,H5I_DATATYPE,H5I_DATASPACE,
    //   H5I_DATASET,H5I_ATTR,H5_IBADID}
    objIdentityType = H5Iget_type(loc_id);
    
    // get name of loc_id, myName should be at least size 256
    // WARNING: the 'myName_p' returned here may NOT be 
    // what you expect, so check it. When loc_id is 'file_id'
    // the name returned is NOT the filenames name, but usually
    // is something like "/",the first or root group.
    H5Iget_name(loc_id,myName_p,256-1);
    
    switch(objIdentityType)
    {
    case H5I_FILE      : rtnObjectType = IsHdf5File ; break;
    case H5I_GROUP     : rtnObjectType = IsHdf5Group; break;
    case H5I_DATATYPE  : rtnObjectType = IsHdf5NamedDatatype; break;
    case H5I_DATASPACE : rtnObjectType = IsHdf5Dataspace; break;
    case H5I_DATASET   : rtnObjectType = IsHdf5Dataset  ; break;
    case H5I_ATTR      : rtnObjectType = IsHdf5Attribute; break;
    case H5I_BADID     : rtnObjectType = IsHdf5InvalidObj;break;
    default : break;
    } // end:: switch on core identity
    
    
    // see: H5I interface..
    objIdentityType = H5Iget_type(loc_id);

    if(objIdentityType != H5I_BADID)
    {
        // first, fetch count of any attributes owned directly by loc_id
        retInfo = H5Oget_info(loc_id, &object_info);
        retNobj = H5Gget_num_objs(loc_id, &nObjects);
    } else {
        xLogMsg("@E %s detected a Invalid-Hdf5-ID(%ld)~\n",
            fxName, (long)loc_id);
    }
    
    if(object_info.type == H5O_TYPE_GROUP) 
    {
        rtnObjectType = IsHdf5Group;
        // also assigns the count of attributes this loc_id owns, if any    
        *nAttrib = (int)object_info.num_attrs ;
        *nObj = (int)nObjects ;         
    } else if(object_info.type == H5O_TYPE_DATASET ) {
        rtnObjectType = IsHdf5Dataset ;
        *nObj = (int)nObjects ;     
        // also assigns the count of attributes this loc_id owns, if any    
        *nAttrib = (int)object_info.num_attrs ;
    } else if(object_info.type == H5O_TYPE_NAMED_DATATYPE ) {
        rtnObjectType = IsHdf5NamedDatatype ;
        *nObj = (int)nObjects ;     
        // also assigns the count of attributes this loc_id owns, if any    
        *nAttrib = (int)object_info.num_attrs ;

    } else if(object_info.type == H5O_TYPE_UNKNOWN && rtnObjectType < 0 )  {
            rtnObjectType = IsHdf5UnknownType ;
    }
    
    // NOTE: original logic here ,to spin through 'objects' under loc_id
    // did not work out, so code removed  until further understood.
    
    // returns the FRM_API object type identifier
    return rtnObjectType ;

} // end:: frm_hdf5_GetObjectInfo()


// ------------------------------------------------------------------
// function     : frm_hdf5_GetComment()
// purpose      : Retrieve text of an object level comment owned by
//                the  loc_id, stored in loc_id's private header space
// inputs       : hid_t loc_id -- id of object (file,group,dataset,etc) 
//                char *commentStr_p --ptr to string holding comment text
//                limited to bufsize chars.
// returns      : char *, ptr to comment string, may often be empty!
// revision     : 2014-06-13TT10:40:00, added orig.
// notes        : n/a
char *frm_hdf5_GetComment(hid_t object_id,char *commentStr_p,size_t bufsize)
{
 ssize_t rsltSiz = 0;
 char fxName[256] = {0} ;
 
  strcpy(fxName,(char *)__func__);
    
  if((rsltSiz =  H5Oget_comment(object_id,commentStr_p,bufsize)) < 0)
  {
      xLogMsg("@E %s Error retrieving object comment(%s) at ID %ld~\n",
        fxName, commentStr_p, object_id);
  }
  
  
  // as a convenience also returns pointer to the resulting string
  return commentStr_p;
  
} // end:: frm_hdf5_GetComment()


// ------------------------------------------------------------------
// function     : frm_hdf5_SetComment()
// purpose      : Assign comment text mapped to the object at loc_id
//                where the comment is stored in object's private
//                header space.
// inputs       : hid_t loc_id -- id of object (file,group,dataset,etc) 
//                char *commentStr_p --ptr to string holding comment text
//                limited to bufsize chars.
// returns      : char *, ptr to comment string, may often be empty!
// revision     : 2014-06-13TT10:40:00, added orig.
// notes        : n/a
void frm_hdf5_SetComment(hid_t object_id,const char *commentStr_p)
{
 ssize_t rsltSiz = 0;
 char fxName[256] = {0} ;
 
  strcpy(fxName,(char *)__func__);
  
  if((rsltSiz = H5Oset_comment(object_id, commentStr_p))< 0)
  {
      xLogMsg("@E %s Error assigning object comment(%s) to ID %ld~\n",
        fxName, commentStr_p, object_id);
  }
  
} // end:: frm_hdf5_SetComment()


// ------------------------------------------------------------------
// function     : frm_hdf5_ClassifyHdf5Datatype()
// purpose      : Given an HDF5 object id (loc_id), discover its 
//                HDF45 datatype and using this, express it as a HDF4 
//                and FRM_API datatype enum,integer style identifier.
// inputs       : hid_t loc_id -- id of object (file,group,dataset,etc) 
// returns      : int32, FRM_API datatype identifier like IsFloat32,IsUint8 etc
//                noting that the FRM_API datatypes are now sync with 
//                prior HDF4 datatype identifiers.
// revision     : 2014-06-13TT13:02:00, added orig.
// notes        : sign indentifiers are: H5T_SGN_NONE (0), Unsigned integer type.
//                or H5T_SGN_2 (1) Two's complement signed integer 
// notes         ; datatype classes are: {H5T_INTEGER,H5T_FLOAT,
//                 H5T_STRING,H5T_BITFIELD, H5T_OPAQUE,H5T_COMPOUND,
//                 H5T_REFERENCE,H5T_ENUM,H5T_VLEN,H5T_ARRAY}
int32 frm_hdf5_ClassifyHdf5Datatype(hid_t loc_id)
{
 hid_t   hDatatype = -1 ;
 size_t  datatypeSizOfOne = 0;
 H5T_sign_t whichSign ;
 H5T_class_t datatypeClass ;
 int32 hdf4Datatype = -1 ;  
 
        // next, get its HDF5 datatype identifier ,works now only
        // for datasets, not yet for attributes
        hDatatype = H5Dget_type(loc_id) ;
        // fetch size (in bytes) of one instance of this datatype
        datatypeSizOfOne = H5Tget_size(hDatatype);
        // fetch class of datatype
        datatypeClass    = H5Tget_class(hDatatype);
        // if integer type, retrieve its "signedness"
        // otherwise, check if a float..
        if(datatypeClass == H5T_INTEGER)
        {
            whichSign = H5Tget_sign(hDatatype);
            if(whichSign == H5T_SGN_NONE)
            {
                switch(datatypeSizOfOne)
                {
                    case 1: hdf4Datatype = IsInt8 ; break;
                    case 2: hdf4Datatype = IsInt16; break;
                    case 4: hdf4Datatype = IsInt32; break;
                    case 8: hdf4Datatype = IsInt64; break;
                    default : break;
                } // end:: switch on an integers size   
            } else {
                switch(datatypeSizOfOne)
                {
                    case 1: hdf4Datatype = IsUint8  ; break;
                    case 2: hdf4Datatype = IsUint16 ; break;
                    case 4: hdf4Datatype = IsUint32 ; break;
                    case 8: hdf4Datatype = IsUint64 ; break;
                    default : break;
                } // end:: switch on an integers size   
            } // end:: if an unsigned integer

        } else if(datatypeClass == H5T_FLOAT) {
            
            if(     datatypeSizOfOne == 4)
                hdf4Datatype = IsFloat32;
            else if(datatypeSizOfOne == 8)
                hdf4Datatype = IsFloat64;
                
        } else if(datatypeClass == H5T_STRING) {
                hdf4Datatype = IsString;
        }
        
        return hdf4Datatype ;

} // end:: frm_hdf5_ClassifyDatatype()

 
// ------------------------------------------------------------------
// function     : frm_hdf5_GetDatasetInfo()
// purpose      : Fetch common dataset metadata info
//
// inputs       :   hid_t file_id -- id of file or group owning this dataset
//                  const char *datasetName_p -- fully qualified dataset name,
//                  hid_t *hDtype --returned H5T datatype identifier
//                  hid_t *dcpl -- returned plist
//                  hid_t *dspace -- returned dataspace id
//                  int *hRank   -- returned rank
//                  hsize_t *dims -- returned dims[]
//
// returns      : hid_t --dataset's unique ID , or -1 if failed or 
//                does not exist
// revised      : 2014-06-16, added test if dataset exists prior to
//                attempting fetch of its properties.
// revised      : 2014-06-13T18:30, added call to new function, 
//                    frm_hdf5_ClassifyHdf5Datatype(datasetId)
//
// history      : n/a
// notes        : Retrieves dataset ID,datatype,dcpl,dspace,rank and dims
// notes        : Should add more properties returned, such as:
//                return {gain,offset}, fillvalue too, and return a HDF4 style
//                datatype identifier like DFNT_FLOAT32 etc...
hid_t frm_hdf5_GetDatasetInfo(hid_t file_id,const char *datasetName_p,
    hid_t *hDtype, hid_t *dcpl,hid_t *dspace, int *hRank,hsize_t *dims)
{
 herr_t ret      = -1 ;
 hid_t datasetId = -1 ;
 hid_t plist     = -1 ;
 hid_t hDatatype = -1 ;
 hid_t hDataspace = -1;
 hsize_t nElem =0; // n. elements across all dimensions
 int rank = -1;
 char fxName[256]= {0} ;
 char fillBuffer[16] = {0}; // fetched, not yet returned.
 int32 hdf4Datatype = 0 ;
 
 // initialized , to fix valgrind "uninitialized var was  branched on"
 // error
 H5D_fill_value_t fvstatus = 0;
 
 
    strcpy(fxName,(char *)__func__);
    
    // <JMG> 2014-09-30, changed to use test that recognizes nested datasets
    if( frm_hdf5_DatasetExists(file_id,datasetName_p) != 1)
    {
        // conservatively: now triggers a fatal error
        xLogMsg("@F In(%s) dataset(%s) does NOT appear to exist at id(%d)\n",
            fxName, datasetName_p, file_id);
    }
    
    // first, get the dataset's ID, using default access plist for now
    // Returns a dataset identifier if successful; otherwise returns a negative value
    if((datasetId = H5Dopen(file_id,datasetName_p,H5P_DEFAULT)) >= 0)
    {
        // next, get its datatype
        hDatatype = H5Dget_type(datasetId) ;
        *hDtype = hDatatype;
        hdf4Datatype = frm_hdf5_ClassifyHdf5Datatype(datasetId);
        
        // next, get its dataset creation property list (dcpl)
        plist = H5Dget_create_plist( datasetId ) ;
        *dcpl = plist ;
        
        // next get its dataspace
        hDataspace = H5Dget_space(datasetId ) ;
        *dspace = hDataspace;
        
        // fetch rank and dimensions..
        rank = H5Sget_simple_extent_dims(hDataspace,dims,dims);
        *hRank = rank ;
        nElem= H5Sget_simple_extent_npoints(hDataspace);
        
        // fetch existing fill value, first checking if one is present
        // then retrieving it, if it is present
        H5Pfill_value_defined(datasetId, &fvstatus);        
        
        // check fvstatus== H5D_FILL_VALUE_UNDEFINED, then NOT defined.
        // herr_t H5Pfill_value_defined(hid_t plist_id, H5D_fill_value_t *status )
        if(fvstatus != H5D_FILL_VALUE_UNDEFINED || fvstatus == H5D_FILL_VALUE_USER_DEFINED)
            ret =  H5Pget_fill_value(plist, hDatatype,fillBuffer) ;
            
    } else {
        xLogMsg("@E In(%s) FAILED query on dataset(%s) owned by %d ~\n",
            fxName,datasetName_p, file_id);
    }

 return datasetId;
} // end::frm_hdf5_GetDatasetInfo()


// ------------------------------------------------------------------
// function     : frm_hdf5_Createfile()
// purpose      : create a new HDF5 file using  H5F_ACC_TRUNC semantics.
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-01-05T21:06:00-07:00 
// history      : 2013-12-03T12:49:00-07:00 refactor pass 1, for D4
// dependencies : hdf5, HDF5 
// notes        : Reminder on "Mode" options: H5F_ACC_TRUNC, H5F_ACC_RDWR
// notes        : Caller should always make sure file_id returned is not -1.
// notes        :  herr_t H5Pset_userblock (hid_t plist, hsize_t size) 
// notes        :  herr_t H5Pget_userblock (hid_t plist, hsize_t *size)
hid_t frm_hdf5_Createfile(char *hdf5Path,char *callerLbl_p)
{
 hid_t file_id = -1  ;
 hid_t fcpl_id = -1  ;  // FILE creation property list ID ..
 hsize_t ubSize= 1024;  // default user block size is 1024, can be larger
 herr_t ret = -1 ;
 
    // create a file creation property list. Choices:{H5P_FILE_CREATE|H5P_FILE_ACCESS}
    fcpl_id = H5Pcreate(H5P_FILE_CREATE);
    
    // Assign a non-zero user-block size to facilitate special tag 
    // inclusions. Must call frm_hdf5_SetUserblocksize(n) prior to this 
    // call in order to enable a userblock.
    if(masterUserBlockSize >0)
    {
        ubSize = masterUserBlockSize ;
        ret = H5Pset_userblock(fcpl_id,ubSize);
    }
 
    // create the empty shell of the new file
    // H5Fcreate(char *path,mode,hid_t fcpl_id,fapl_id)
    // hid_t H5Fcreate( const char *name, unsigned flags, hid_t fcpl_id, hid_t fapl_id )
    if((file_id = H5Fcreate(hdf5Path, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT))<0)
    {
        xLogMsg("@F frm_hdf5_Createfile:: (%s) FAILED creating new HDF5 [%s]~\n",
                    callerLbl_p, hdf5Path);
    } 
    
    // Close out the new file's creation property list 
    // TBD: verify if we can skip this until the file is closed
    H5Pclose(fcpl_id);

 return file_id;
 
} // end:: frm_hdf5_Createfile()


// ------------------------------------------------------------------
// function     : frm_hdf5_OpenfileAppend()
// purpose      : open in Append (Read+Write) mode HDF5 file
// inputs       : char *hdf5Path,char *callerLbl_p
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-01-05T21:06:00-07:00 
// history      : 2013-12-03T12:49:00-07:00 refactor pass 1, for D4
// dependencies : HDF5 , uses standard file drivers for now
// notes        : Mode options: H5F_ACC_TRUNC, H5F_ACC_RDWR,
hid_t frm_hdf5_OpenfileAppend(char *hdf5Path,char *callerLbl_p)
{
 hid_t file_id = -1 ;
 
    // open a pre-existing file
    if((file_id = H5Fopen(hdf5Path, H5F_ACC_RDWR, H5P_DEFAULT)) < 0)
    {
        xLogMsg("@F frm_hdf5_OpenfileAppend: (%s) FAILED creating new HDF5 [%s]~\n",
                    callerLbl_p, hdf5Path);
    } else {
        xLogMsg("@A In(%s) Caller:(%s) OPEN-APPEND HDF5 (%s) ID %d~\n",
                    (char *)__func__, callerLbl_p,  hdf5Path, file_id);
    }

 return file_id;
 
} // end:: frm_hdf5_OpenfileAppend()


// ------------------------------------------------------------------
// function     : frm_hdf5_Openfile()
// purpose      : open an existing HDF5 file
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-06-25T16:21:00, added frm__ioDoesFileExist() check
//                prior to trying to open, and fetch user-block size
//                returned for now only via static masterUserBlockSize
// revised      : 2014-01-05T21:06:00-07:00 
// history      : 2013-12-03T12:49:00-07:00 refactor pass 1, for D4
// version      : see VERS macro
// dependencies : HDF5 libraries, v1.8 or  higher
// notes        : Options: H5F_ACC_RDONLY,H5F_ACC_RDWR, H5F_ACC_TRUNC etc
//              : for now suppress advisory traffic to console...
// notes        : The HDF5 file driver,SEC2 or H5FD_SEC2 is used by default
//
//                <JMG> 7/29/2014, Removed call to H5Pcreate() until further
//               debugging is performed.
//
hid_t frm_hdf5_Openfile(const char *hdf5Path,char *callerLbl_p)
{
 hid_t file_id  = -1;
 hid_t fcpl_id  = -1;
 hid_t fapl_id  = -1;
 herr_t ret     = -1;
 hsize_t ubSize = 0 ;
 // TBD: not yet implemented
 // H5O_info_t     hFilInfo ; // low level (object) properties 
 // H5O_hdr_info_t hHdrInfo ; // low level (object) header properties
  
 // TBD: test if file exists first and don't bother opening it if it doesn't
 if(frm_ioDoesFileExist(hdf5Path))
 {

#ifdef NOT_READY_YET
    // fetch creation and access property list ID's
    fapl_id = H5Pcreate(H5P_FILE_ACCESS) ;
    // <JMG> 7/17/2014 fixed bug: changed next from H5P_FILE_ACCESS to H5P_FILE_CREATE
    fcpl_id = H5Pcreate(H5P_FILE_CREATE) ;
    file_id = H5Fopen(hdf5Path, H5F_ACC_RDONLY, fapl_id) --> does NOT work as expected yet
#endif  

    // Open file 
    if((file_id = H5Fopen(hdf5Path, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0)
    {
        xLogMsg("@F frm_hdf5_Openfile: (%s) Cannot find or open HDF5 file (%s) status: %d ~\n",
                callerLbl_p, hdf5Path, file_id);
    } else {
        // TBD: Fetch other metadata. Not tested (2014-06-25) yet.
        // herr_t H5Oget_info( hid_t object_id, H5O_info_t *object_info )
        
        if( doVerboseMessaging)
            xLogMsg("@A frm_hdf5_Openfile: (%s) HDF5 file %s opened. ID assigned: %d ~\n",
                    callerLbl_p, hdf5Path, file_id);
        ret =0;
    }
    
 } else {
        xLogMsg("@E frm_hdf5_Openfile: (%s) Cannot open non-existingHDF5 file (%s)~\n",
                callerLbl_p, hdf5Path);
        // returns -1 if file cannot be opened due to not existing
 }
 
 return file_id ;
 
} // end:: frm_hdf5_Openfile()


// ------------------------------------------------------------------
// function     : frm_hdf5_Closefile()
// purpose      : close an open HDF5 file.
// history      : n/a
// dependencies : HDF5 lib, FRM API lib
// notes        : n/a
herr_t frm_hdf5_Closefile(const hid_t file_id,char *callerLbl_p)
{
 herr_t ret = -1 ;

 // pre-emptively FLUSH at least THIS file...
 H5Fflush(file_id,H5F_SCOPE_LOCAL);
  
 if((ret = H5Fclose(file_id)) < 0)
 {
    xLogMsg("@A frm_hdf5_Closefile: caller(%s) closed id(%d)~\n",
        callerLbl_p, file_id);   
 }

 if( doVerboseMessaging)
        xLogMsg("@A frm_hdf5_Closefile: caller(%s) closed id(%d)~\n",
        callerLbl_p, file_id);
 
 return ret ;
} // end::frm_hdf5_Closefile()


// ---------------------------------------------------------------------
// function  : frm_hdf5_CloseDataset()
// purpose   : Retrieve HDF5 dataset ID if available, leaving it open
// inputs    : const hid_t dataset_id -- active, open ID of dataset to close
// returns   : int , ret verbatim from H5Dclose()...
// revised   : 2014-09-30T16:55:00, added, as this was missing earlier
// notes     : Caller must also be responsible for closing a datasets
//             dcpl, dspace, and other machinery etc.
int frm_hdf5_CloseDataset(const hid_t dataset_id)
{
 herr_t ret = -1;
 char fxName[256] = {0} ;
 
    strcpy(fxName,__func__);
    
    if((ret= H5Dclose(dataset_id)) < 0)
    {
        xLogMsg("@E In(%s) ERROR closing dataset (%s) ret (%d)\n",
            fxName, dataset_id, ret);
    }

 return (int)ret ;  
 
} // end:: frm_hdf5_CloseDataset()


// ---------------------------------------------------------------------
// function  : frm_hdf5_GetDatasetId()
// purpose   : Retrieve HDF5 dataset ID if available, leaving it open
// inputs    : const hid_t file_id,
//             const char *datasetName_p
// returns   : hid_t dataset_id -- valid ID if available, or neg value if not
// revised   :
// notes     : Caller MUST be responsible for CLOSING dataset after use!!
//             via a call to frm_hdf5_CloseDatasetId(dset_id)
hid_t frm_hdf5_GetDatasetId(const hid_t file_id,const char *datasetName_p)
{
    
 hid_t dset_id = -1 ;
 char fxName[256] = {0};
 
    strcpy(fxName,__func__);
 
    // first, open the dataset and fetch its unique id
    if((dset_id = H5Dopen(file_id,datasetName_p,H5P_DEFAULT)) < 0)
    {
        xLogMsg("@E In(%s) ERROR retrieving (%s) dataset id (%d) in file_id %d\n",
            fxName, datasetName_p, dset_id, file_id);
    }

 // returns valid dataset id if available OR a negative number if not
 return dset_id ;
    
} // end:: frm_hdf5_GetDatasetId()


// ---------------------------------------------------------------------
// function     :frm_hdf5_ReadData() 
// purpose      : read a HDF5 dataset using simple all-at-once semantics
// history      : n/a
// inputs       : const hid_t thisFile_id,
//              : const hid_t dsTypeId, HDF5 H5T type identifier like
//                H5T_NATIVE_UINT8,H5T_IEEE_FLOAT32 etc
//              ; char *datasetname_p
//              : void *dataBuffer_p
// returns      : int , (cast from herr_t) 'ret' status
// revised      : 2014-09-30T16:25, revised to use H5Dopen() and H5Dread()
//                instead of "lite" fx
// dependencies : HDF5 library
// notes        : A flaw in this fx is that the dset_id is never returned,
//                where it would be very useful to have it passed back
//                since it was accessed!
// notes        : It is critical that the dataBuffer_p passed in is 
//                sized for the right datatype and and no. of elements
// notes        : dsTypeId should be one of below, or other H5T family of type identifiers...
//              :  H5T_NATIVE_UINT8, H5T_NATIVE_INT16,H5T_NATIVE_UINT8, 
//                 H5T_NATIVE_FLOAT, H5T_NATIVE_DOUBLE or  H5T_IEEE_F32LE etc.
//
// dspace_id = H5Dget_space(dataset_id);
//  dtype_id = H5Dget_type(dataset_id);
//  plist_id = H5Dget_create_plist(dataset_id);
//
// Here below is the long form syntax for H5Dread()
//   herr_t H5Dread( hid_t dataset_id, hid_t mem_type_id, hid_t mem_space_id, hid_t file_space_id, hid_t xfer_plist_id, void * buf )
//
// Example: status = H5Dread(dataset_id, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, dset_data_p);
//
int frm_hdf5_ReadData(const hid_t thisFile_id,const hid_t dsTypeId,char *datasetname_p,
                void *dataBuffer_p,char *callerLbl_p)
{

 herr_t ret = 0;
 herr_t ret2 = 0;
 hid_t  dset_id = -1;
 char fxName[256] = {0};  
  
    strcpy(fxName,__func__);

    // EVALUATE new logic
    // first, open the dataset and fetch its unique id
    if((dset_id = H5Dopen(thisFile_id,datasetname_p,H5P_DEFAULT)) >= 0)
    {
        // next, read it and inspect return status
        // Returns a non-negative value if successful; otherwise returns a negative value.
        if((ret = H5Dread(dset_id, dsTypeId,H5S_ALL, H5S_ALL, H5P_DEFAULT,dataBuffer_p)) < 0)
        {
            xLogMsg("@E In(%s) FAILED ret(%d) H5Dread() dataset(%s)ID(%d) owned by file_id %d ~\n",
                fxName,ret, datasetname_p, dset_id, thisFile_id);
        }
        // Regardless of status, close the dataset id
        ret2 = H5Dclose(dset_id);
    } else {
        xLogMsg("@A frm_hdf5_ReadData: [%s] READ dataset(%s) OK from file_id %d\n",
            callerLbl_p, datasetname_p, thisFile_id);
    }
    
#ifdef OLD_ORIG_LOGIC
    if((ret = H5LTfind_dataset(thisFile_id, datasetname_p)) >= 0)
    {
        // use 'lite' interface, read entire dataset into buffer of HDF5 datatype, in one call
        ret = H5LTread_dataset(thisFile_id, datasetname_p, dsTypeId,(uint8 *)dataBuffer_p);

        if(ret < 0)
        {
            xLogMsg("@F frm_hdf5_ReadData: [%s] READ dataset(%s) FAILED ret %d from file_id %d ~\n",
                    callerLbl_p, datasetname_p, ret, thisFile_id);
        } else {
            xLogMsg("@A frm_hdf5_ReadData: [%s] READ dataset(%s) OK from file_id %d\n",
                    callerLbl_p, datasetname_p, thisFile_id);
        } // end:: block if data was read ok
        
    } else {
        xLogMsg("@E frm_hdf5_ReadData: ret(%d) [%s] dataset(%s) NOT available from file_id %d ~\n",
                ret, datasetname_p, thisFile_id);       
    }
#endif

  // return status from the H5Dread() itself
 return (int)ret ;

}  // end:: frm_hdf5_ReadData()


// ------------------------------------------------------------------
// function     :frm_hdf5_WriteData() 
// purpose      : write a HDF5 dataset using simple all-at-once semantics
// revised      : 2014-09-29T15:32:00, just cleaned up.
// history      : 2014-07-28T19:16, re-org arg call params to better match
//                frm_hdf5_ReadData()'s.
// history      : n/a
// inputs       :   const hid_t file_id,
//                  const int rank,hsize_t *dims,
//                  const hid_t dsTypeId,
//                  char *datasetname_p,
//                  void *dataBuffer_p,
//                  char *callerLbl_p
// dependencies : 
// notes        : Must open file in H5F_ACC_RDWR or in H5F_ACC_TRUNC for this to work
// notes        : It is critical that the dataBuffer_p passed in has data of the correct
//                datatype in it, matching dsTypeId
// notes        : dsTypeId should be one of below, or other H5T family of type identifiers...
//              :  H5T_NATIVE_UINT8, H5T_NATIVE_INT16,H5T_NATIVE_UINT8, 
//                 H5T_NATIVE_UINT8,H5T_NATIVE_UINT8, H5T_NATIVE_FLOAT,
//                 H5T_NATIVE_UINT8, H5T_NATIVE_UINT8, H5T_NATIVE_UINT8, or  H5T_IEEE_F32LE etc.
int frm_hdf5_WriteData(const hid_t file_id,const hid_t dsTypeId, char *datasetname_p,
                void *dataBuffer_p,const int rank,hsize_t *dims,char *callerLbl_p)
{
 int status = -1 ;
 herr_t ret = 0;
 char fxName[256]= {0} ;
 
 
    strcpy(fxName,__func__);
    
    // using 'lite' interface, WRITE entire dataset into buffer of HDF5 datatype, in one call
    ret = H5LTmake_dataset(file_id, datasetname_p, rank, dims, dsTypeId, (uint8 *)dataBuffer_p);

    // error check...
    if(ret < 0)
    {
        xLogMsg("@F In(%s): [%s] WRITE dataset(%s) FAILED ret %d from file_id %d ~\n",
                fxName,callerLbl_p, datasetname_p, ret, file_id);
    } else {
        xLogMsg("@A In(%s): [%s] WRITE dataset(%s) OK from file_id %d\n",
                fxName,callerLbl_p, datasetname_p, file_id);
    } // end:: block if data was written ok
 
 return ret ;
 
}  // end:: frm_hdf5_WriteData()


// ------------------------------------------------------------------
// function  : frm_hdf5_CreateStringAttribute()
// inputs    : const hid_t owner_id,const char *attrLabel,const char *attrValue_p
// purpose   : Create a string attribute (a string attribute, owned by owner_id)
// revision  : 2015-08-05T11:38, added test on new FRM_API_VERBOSE switch
//             to trim excessive chatter
// notes     : Once an attribute if closed, like any object, its 'attr_id'
//             is essentially worthless since it is returned to the free
//             pool of attribute ID's assignable to the next one requesting
//             and id.
hid_t frm_hdf5_CreateStringAttribute(const hid_t owner_id,
                const char *attrLabel,  const char *attrValue_p)
{
 herr_t ret =0;
 hid_t attr_dspace = H5Screate(H5S_SCALAR);
 hid_t attr_type   = -1;
 hid_t attr_id =-1;
 
    attr_type   = H5Tcopy(H5T_C_S1);
    // Set the SIZE of the attribute as the length of the single string
    H5Tset_size(attr_type, strlen(attrValue_p)) ;
    // Create the container for the string attribute
    attr_id = H5Acreate(owner_id, attrLabel, attr_type, attr_dspace, H5P_DEFAULT, H5P_DEFAULT);
    // Write the string attribute's value, held in string "unitStrValue" 
    ret = H5Awrite(attr_id, attr_type, attrValue_p); 
    // Close out the attributes dataspace
    H5Sclose(attr_dspace);  
    // close out the attribute object's private mechanism
    H5Aclose(attr_id);

// <JMG> 2015-08-05T11:36 added qualifier to mute excessive chatter...    
#ifdef FRM_API_VERBOSE    
    xLogMsg("@A In(%s) wrote string-attribute (%s) ret: %d ~\n",
        (char *)__func__, attrLabel, (int)ret);
#endif        
    
 // return attributes unique handle, in case its useful outside..
 return attr_id ;
 
} // end::frm_hdf5_CreateStringAttribute()


// ------------------------------------------------------------------
// function : int frm_hdf5_AttributeExists(hid_t loc_id,const char *attrName_p)
// purpose  : Evaluate if attribute exists as named and owned, returning
//            TRUE|FALSE.
// inputs   : hid_t loc_id,const char *attrName_p
// returns  : TRUE if attribute 'attrName_p' exists as owned by loc_id,
//            otherwise returns FALSE.
// revision : 2014-06-13T19:07:00-07, orig
// author   : joe glassy
int frm_hdf5_AttributeExists(hid_t loc_id,const char *attrName_p)
{
 htri_t doesThisExist = FALSE ;
 
    doesThisExist = H5Aexists(loc_id,attrName_p);
    
    return (int)doesThisExist ;
    
} // end:: frm_hdf5_AttributeExists()


// ------------------------------------------------------------------
// function : int frm_hdf5_GroupExists(hid_t loc_id,const char *groupName_p)
// purpose  : Evaluate if a group exists as named and owned, returning
//            TRUE|FALSE.
// inputs   : hid_t loc_id,const char *groupName_p
// returns  : TRUE if GROUP'groupName_p' exists as owned by loc_id,
//            otherwise returns FALSE.
// revision : 2014-06-16T12:38:00-07, orig
// author   : joe glassy
// notes    : There is no H5Gexists() function per se, so this logic
//            needs to be implemented in a more granular way.
int frm_hdf5_GroupExists(hid_t loc_id,const char *groupName_p)
{
 htri_t doesThisExist = FALSE ;
 herr_t ret= -1 ;
 H5G_info_t grpInfo ;
 
 char fxName[256] = {0};
 
    strcpy(fxName,(char *)__func__);
    
    // doesThisExist = H5Aexists(loc_id,attrName_p);
    // ret = H5Gget_info_by_name( hid_t loc_id, const char *group_name, H5G_info_t *group_info, hid_t lapl_id )
    ret = H5Gget_info_by_name(loc_id,groupName_p, &grpInfo,H5P_DEFAULT);
    // returns a non-negative no. if OK, or a negative number (like -1) if failed.
    doesThisExist = (ret >= 0) ? 1 : 0 ;
    // xLogMsg(@E In(%s) frm_hdf5_GroupExists is NOT YET IMPLEMENTED!~\n");
    
    return (int)doesThisExist ;
    
} // end:: frm_hdf5_GroupExists()


// ------------------------------------------------------------------
// function  : frm_hdf5_getAttribInfo()
// purpose   : Retrieve metadata (datatype,dataspace,sized) assoc
//             with an attribute named "attrName" owned by owner_id
// inputs    : const hid_t owner_id,const char *attrLabel,char *attrValue_p
// returns   : hid_t, unique ID of this attribute
// outputs   : int   *nAttributes --total n. owned by object
//           : hid_t *attrDspace_p -- dataspace of attribute
//             hsize_t size in bytes of attributes requirement.
// notes     : Using this requires some care+awareness since the
//             ID's may be transient. TBD: verify if it is preferable
//             to explictly CLOSE OUT the dataspace as part of this call.
hid_t frm_hdf5_GetAttribInfo(const hid_t owner_id,int *nAttribs, 
        const char *attrName_p, hid_t *attrDspace_p,hsize_t *attrSiz_p,
        hid_t *attrDatatype_p)
{
 hid_t   attr_id = -1;
 herr_t ret = -1;
 H5O_info_t object_info ;
 
    // fetch total n. of attributes this object owns
    // using preferred method (not deprecated one)
    ret = H5Oget_info(owner_id, &object_info);  
    
    *nAttribs =  object_info.num_attrs;
   
   //  fetch THIS attribues unique ID, and characteristics
   attr_id = H5Aopen(owner_id,attrName_p,H5P_DEFAULT);
   
   *attrDatatype_p = H5Aget_type(attr_id) ;
   *attrDspace_p   = H5Aget_space(attr_id) ;
   *attrSiz_p      = H5Aget_storage_size(attr_id) ;

 // returns attr_id 
 return attr_id  ;
    
} // end::frm_hdf5_GetAttribInfo()


// ------------------------------------------------------------------
// function  : frm_hdf5_ReadAttributeString()
// inputs    : const hid_t owner_id,
//             const char *attrName_p,
//             char *attrValue_p
// purpose   : Read one string attribute, owned by 'owner_id'
// returns   : char *attrValue_p - ptr to data itself
// outputs   : hid_t *attrValue_p -- ptr to data itself
// example   : frm_hdf5_ReadAttributeString(const hid_t owner_id,
//              const char *attrName_p,char *attrValue_p)
// notes     : TBD: more testing is essential (as of 6/12/2014)
char *frm_hdf5_ReadAttributeString(const hid_t owner_id,
                const char *attrName_p,char *attrValue_p)
{
 herr_t ret = -1;
 hid_t attr_id = -1;
 hid_t attrDatatype = -1 ;

 // fetch attribute, by joint identifiers {owner_id+name}
 if((attr_id = H5Aopen(owner_id, attrName_p, H5P_DEFAULT)) >= 0)
 {
     attrDatatype = H5Aget_type(attr_id);
     ret  = H5Aread(attr_id, attrDatatype, attrValue_p) ;
 } else {
    xLogMsg("@F ERROR (%s) on attrib read of (%s) ~\n",(char *)__func__,
        attrName_p);
 }
 // close attribute
 H5Aclose(attr_id);
 
 // return string read...
 return attrValue_p;
    
} // end: frm_hdf5_ReadAttributeString


// ------------------------------------------------------------------
// function  : frm_hdf5_ReadAttribute()
// inputs    : const hid_t owner_id,
//             const char *ownerLoc_p,
//             hid_t *rtn_AttrDatatype,
//             const char *attrName_p,
//             void *attrValue_p
// purpose   : read an attribute of a supported datatype.
// revised   : 2014-06-13T18:47, removed unused parameter, ownerLoc_p.
// returns   : void *attrValue_p - ptr to attribute data read
// outputs   : *attrValue_p -- ptr to attribute data and 
//             hid_t *rtn_AttrDatatype -- returns H5T datatype of attribute
// example   : frm_hdf5_ReadAttribute(file_id,"/",&aType,"soc_calendar",soc_cal_p);
//
// notes     : // notes     : TBD: more testing is essential (as of 6/12/2014)
//
//  The "lite" method is also suitable for non-string types:
//  H5LTget_attribute(file_id,"/", "soc_calendar",H5T_NATIVE_UINT8,soc_calendar_p)
//
//  herr_t H5LTget_attribute( hid_t loc_id, const char *obj_name,
//       const char *attr_name,  hid_t mem_type_id, void *data )
// herr_t H5LTget_attribute_string( hid_t loc_id, const char *obj_name,
//       const char *attr_name,  char *data )
// herr_t H5LTget_attribute( hid_t loc_id, const char *obj_name, 
//       const char *attr_name,  char *data )
//
// to fetch n. of attributes an object owns, use:
//   int H5Aget_num_attrs( hid_t loc_id ) 
//    hid_t H5Aget_type(hid_t attr_id) 
//    hsize_t H5Aget_storage_size(hid_t attr_id) 
//    hid_t H5Aget_space(hid_t attr_id) 
//
//  The manual way:
//   hid_t H5Aopen( hid_t obj_id, const char *attr_name, hid_t aapl_id )
//   attr_id = H5Aopen(file_id,"my_attrib", H const char *attr_name, H5P_DEFAULT);
//   herr_t H5Aread(hid_t attr_id, hid_t mem_type_id, void *buf ) ;
//   ret =  H5Aread(attr_id, H5T_NATIVE_UINT8, uBytBuffer) ;
//   ret =  H5Aclose(attr_id) ;
//  Deprecated method: (used ownerLoc_p param)
// ret = H5LTget_attribute(owner,"/", "soc_calendar",H5T_NATIVE_UINT8, soc_calendar_p)  
// ret = H5LTget_attribute(owner,ownerLoc_p,attrName, attrDatatype,attrValue_p) ;
//
void *frm_hdf5_ReadAttribute(const hid_t owner_id,hid_t *rtn_AttrDatatype, 
        const char *attrName_p,void *attrValue_p)
{
 herr_t  ret = -1;
 hid_t   attr_id = -1;
 hid_t   attrDatatype = -1 ;
 hsize_t nBytesRequired = 0;
 
 if((attr_id = H5Aopen(owner_id, attrName_p, H5P_DEFAULT)) >= 0)
 {
     attrDatatype =  H5Aget_type(attr_id);
     *rtn_AttrDatatype = attrDatatype ;
     // the destination buffer should be at least this size...
     nBytesRequired= H5Aget_storage_size(attr_id);
     // read the attribute 
     ret  = H5Aread(attr_id, attrDatatype, attrValue_p) ;
 } else {
    xLogMsg("@F ERROR (%s) on attrib read of (%s) ~\n",
        (char *)__func__,attrName_p);
 }
 // close attribute
 H5Aclose(attr_id);

 return attrValue_p ;
    
} // end:: frm_hdf5_ReadAttribute()

// ------------------------------------------------------------------
// function : frm_hdf5_WriteAttributeAsType()
// purpose  : Write a typed attribute to loc_id. This may be a scalar
//            when nElem=1, or a 1D array when nElem >1.
// inputs   :hid_t loc_id -- id of owner, like file_id or dataset_id
//           const char *grpQualifier -- group qualifier like ".","SOC"
//           note that this should NOT typically be prefaced with a slash
//           like h5py, as the slash is implied.
//           const char *attrName_p,
//           const int frmApiDatatype,
//           const int nElem,
//           void *attrData_p
// returns  : int, status
// revision : 2014-06-13T11:24:00-07, orig
// author   : joe glassy
// notes    : TBD: consider modifying to return the new unique attr_id!!
//            Using H5LTset_attribute_xxx() is a drawback since the attr_id
//            is not returned, forcing us to have to get it another way...
// notes    : TBD 2014-09-02, re-confirm if H5LTset_attribute_xxx() family works
//            with-or-without a pre-existing attribute. e.g. If it already
//            exists, does it overwrite it? (it should), and,  if it 
//            does not already exist, does it create it then write it?
int frm_hdf5_WriteAttributeAsType(hid_t loc_id,const char *grpQualifier,
    const char *attrName_p,const int frmApiDatatype,const int nElem,
    void *attrData_p)
{
 char fxName[256]= {0} ;
 herr_t ret = -1;
 
  strcpy(fxName,__func__);
  
  switch(frmApiDatatype)
  {
    case IsString :ret = H5LTset_attribute_string(loc_id,grpQualifier,
                    attrName_p,(char *)attrData_p);
                    break;
    case IsInt8   : ret = H5LTset_attribute_char(loc_id,grpQualifier,
                        attrName_p, (char *)attrData_p, nElem);
                    break;
    case IsUint8  : ret = H5LTset_attribute_uchar(loc_id,grpQualifier,
                        attrName_p, (unsigned char *)attrData_p, nElem);
                    break;
    case IsInt16  : ret =  H5LTset_attribute_short(loc_id,grpQualifier,
                            attrName_p,(int16 *)attrData_p, nElem);
                    break;
    case IsUint16 : ret =  H5LTset_attribute_ushort(loc_id,grpQualifier,
                            attrName_p,(uint16 *)attrData_p, nElem);
                    break;
    case IsInt32  : ret = H5LTset_attribute_int(loc_id,grpQualifier,
                    attrName_p, (int32 *)attrData_p, nElem);
                    break;
    // here a 'ulong' H5T_NATIVE_ULON evaluates to a uint64 not a uint32!
    case IsUint32 : ret = H5LTset_attribute_uint(loc_id,grpQualifier,
                    attrName_p, (uint32 *)attrData_p, nElem);
                    break;
    case IsInt64  : ret = H5LTset_attribute_long(loc_id,
                        grpQualifier,attrName_p,(int64 *)attrData_p,nElem);
                    break;
    case IsUint64  : ret = H5LTset_attribute_ulong(loc_id,
                        grpQualifier,attrName_p,(uint64 *)attrData_p,nElem);
                    break;
                    
    case IsFloat32: ret = H5LTset_attribute_float(loc_id,
                        grpQualifier,attrName_p,(float32 *)attrData_p,nElem);
                    break;  
    case IsFloat64: ret = H5LTset_attribute_double(loc_id,
                        grpQualifier,attrName_p,(float64 *)attrData_p,nElem);
                    break;
    default :   xLogMsg("@E In(%s) unsupported FRM_API (TIV) datatype(%d)\n",
                        fxName, frmApiDatatype);
                    break;
  } // end:: switch on datatype

 // return status
 return ret ;
 
} // end:: frm_hdf5_WriteAttributeAsType()

#ifdef DISABLED_AS_DEVEL_PENDING
// -----------------------------------------------------------------------------
// Function : frm_hdf5_WriteNvpAsAttribute()
// Purpose  : using a supported NVP keyword, create a string attribute
//            using the RIP keyword as the attribute name by default, OR,
//            if the user supplies a 'attrForceName', use that as the attribute name
// Inputs   : NsbObj *nsb_p,
//              const hid_t loc_id -- the file_id,group_id, or dataset_id
//              that owns this attribute.
//              hid_t attrDatatype -- H5T_NATIVE_FLOAT32 or whatever...
//              const char *keyword -- RIP keyword
//              char *attrForceName --optional, if supplied, used as the
//              HDF5 attribute name verbatim
// Author   : joe glassy
// Returns  : hid_t attr_id if successful, or -1 if not
// Notes    : The HDF5 attribute written using the keyword as the
//            attributes name and the NVP's 'value' as the attribute value

hid_t frm_hdf5_WriteNvpAsAttribute(NsbObj *nsbObj_p, hid_t loc_id, 
            const hid_t attrDatatype,
            const char *keyword,
            const char *attrForceName)
{
 char attrValu[FILENAME_MAX] = {0} ;
 char nvpBuf[FILENAME_MAX]   = {0} ;
 char fxName[256] = {0} ;
 int keyResult = -1 ;
 hid_t attr_id = -1 ;
 
    strcpy(fxName,__func__);
    
    // fetch value of this supported keyword
    strcpy(attrValu, nsbGetValueByKey(nsbObj_p,keyword,nvpBuf));
        
    // Force use of 'attrForceName' instead of 'keyword' as the
    // HDF5 attribute name or label.
    if(attrForceName != NULL && strlen(attrForceName)>= 1)  
    {
        attr_id = l4c_createHdf5Attribute(file_id,attrForceName, attrValu);

    } else {
        // use the 'keyword' verbatim as the HDF5 attribute name
        attr_id = l4c_createHdf5Attribute(file_id,keyword, attrValu);
    }
    // xLogMsg("@A I(%s) wrote key(%s) value(%s) ~\n",fxName,keyword,attrValu);
        
    
 return attr_id;
} // end::frm_hdf5_WriteNvpAsAttribute()
#endif

// ------------------------------------------------------------------
// function : frm_hdf5_WriteAttributeCustomString()
// purpose  : Write a sprintf() style custom string attribute to loc_id.
// inputs   :hid_t loc_id -- id of owner, like file_id or dataset_id
//           const char *grpQualifier -- group qualifier like ".","SOC"
//           note that this should NOT typically be prefaced with a slash
//           like h5py, as the slash is implied.
//           const char *attrName_p,
//           char *fmtSpecString -- sprintf() style custom formatted string
// returns  : int, status
// revision : 2014-06-14T12:58:00-07, orig
// author   : joe glassy
// notes    : The attribute 'attrName_p' is overwritten if pre-existing.
int frm_hdf5_WriteAttributeCustomString(hid_t loc_id,
    const char *grpQualifier, const char *attrName_p,char *fmtSpecStr_p, ...)
{
    char fxName[256] = {0} ;
    char *userMessage= NULL; // POSIX...
    va_list argptr ;   // ptr to list of var args
    int     fmt_count =0;   
    herr_t ret = -1;
 
    strcpy(fxName,__func__);
    
    // allow up to a 4Kb string to be written out
    userMessage = (char *)calloc(4096,sizeof(char));
    // format it according to format spec string user submitted
    fmt_count= vsprintf(userMessage,fmtSpecStr_p, argptr);
    
    // write the formatted string out as an HDF5 attribute string
    // to the attribute named 'attrName_p'.
    ret = H5LTset_attribute_string(loc_id,grpQualifier,
                    attrName_p,(char *)userMessage);

    // close out the var args facility
    va_end(argptr);
    // free the local buffer storing the custom string
    free(userMessage);
    
 // return status
 return ret ;
 
} // end:: frm_hdf5_WriteAttributeCustomString()


// ------------------------------------------------------------------
// function: frm_hdf5_GetBasicProperties()
// purpose : retrieve basic properties of a dataset named
//          datasetName_p
// inputs  : hid_t file_id,char *datasetName_p,
//          int *rtnRank, hsize_t *rtnDims
// returns : Total elements across all dimensions
// notes   : Add a call to get the dataset's dcpl and dspace too if
//           possible.
size_t frm_hdf5_GetBasicProperties(hid_t file_id,char *datasetName_p,
                int *rtnRank, hsize_t *rtnDims)
{
 herr_t ret = -1;
 size_t s,sizOfOne = 0 ;
 // size_t localDims[8] = {0};
 hsize_t nTotal =0;
 size_t nTotalElem =0;
 // H5T_class_t *typClass ;
 
 if((ret=H5LTfind_dataset (file_id,datasetName_p))< 0)
 {
   xLogMsg("@E frm_hdf5_GetBasicProperties:  dataset(%s) is NOT present in file_id %d ~\n",
        datasetName_p,file_id);
    return ret ;
 } else {
   ret = 1;
 }
 
 // Fetch (minimalist properties) --> the dimensions of the dataset and size of it 
 // note that if typClass passed in, it needs to be 'closed' later, but this is tricky
 // ret  = H5LTget_dataset_info(file_id,datasetName_p,rtnDdims, typClass, &sizOne);
 
 ret  = H5LTget_dataset_info(file_id,datasetName_p, rtnDims, NULL, &sizOfOne);
 
 // fprintf(stderr,"@E H5LTget_dataset_info status %d for dataset %s sizOne %zu\n",ret,datasetName,sizOne);
 
 // Next, fetch this dataset's properties:: rank, then calc dims
 ret = H5LTget_dataset_ndims(file_id,datasetName_p, rtnRank);   
 
 // calc total n. of elements across all dimensions
 for(nTotal=1,s=0; s < (size_t)(*rtnRank); s++)
     nTotal *= rtnDims[s];
    
 nTotalElem = (size_t)nTotal;
 
 // return total n. elements across all dimensions
 return nTotalElem ;    
 
} // end::frm_hdf5_GetBasicProperties()


// -----------------------------------------------------------------------------
// function : frm_hdf5_DiscoverAndRead()
// inputs   : hid_t file_id,char *datasetName,int *dRank,hsize_t *dims
// Purpose  : For a variable 'datasetName_p' discover its basic properties, create
//            an appropriate buffer, then read the data and return it in a buffer
// notes    : this needs additional refinement to make it type generic..
//
void *frm_hdf5_DiscoverAndReadFloat32(hid_t file_id,char *datasetName_p,
            int *dRank,hsize_t *dims)
{
 herr_t ret    = -1;
 // size_t d;
 size_t nElem =0;
 // size_t sizOne = 0;
 void *dsData_p= NULL;
  
 if((ret=H5LTfind_dataset(file_id,datasetName_p))< 0)
 {
   xLogMsg("@A frm_hdf5_DiscoverAndRead: dataset %s NOT present in file_id %d ~\n",
    datasetName_p,file_id);
   exit(-1);
 } else {
   // fprintf(stdout,"@A frm_hdf5_DiscoverAndRead: dataset %s with dataset-id %d\n",datasetName,ret);
   ret = 1;
 }

    // fetch dataset's rank, dims, and total n. elem 
  nElem = frm_hdf5_GetBasicProperties(file_id,
                datasetName_p,dRank, dims);
  
 // here we assume we know the datatype directly
 dsData_p =(float32 *)memCalloc(nElem,sizeof(float32));
 
    if( dsData_p != NULL)
    {
        // READ the entire dataset into the buffer in one call
        if((ret=H5LTread_dataset_float(file_id,datasetName_p,(float32 *)dsData_p))<0)
                xLogMsg("@F dataset %s READ FAILED!~\n",datasetName_p);
        else 
                ret = 1;
        
    }
 // NOTE: using H5LTxxx family fx we don't retrieve the dataset_ID explicitly, so don't close it here
 
 // return pointer to data read
 return dsData_p;
} // end:: frm_hdf5_DiscoverAndRead()


// ------------------------------------------------------------------
// function : frm_hdf5_CreateGroup()
// purpose  : create a new group. This fails if the group already exists
// inputs   : hid_t loc_id,const char *newGroupName_p
// revision : 2014-06-17T17:33:00, add test if group already exists so
//            to prevent over-writing error on existing groups.
// revision : 2014-06-13T17:33:00, added orig
// author   : joe glassy
// returns  : new groups id
// notes    : The newly created HDF5 group is intentionally LEFT OPEN
//            after creation, so if it really needs to be closed, use
//            frm_hdf5_CloseGroup() to do this explicitly
// notes    : When specifying a nested group name here, we typically
//            do not lead with a (root-group) slash qualifier,
//            but we DO include a slash separator, like:
//            "BaseGroup/NestGroup1" or "METADATA/MoreMetaGroup"
hid_t frm_hdf5_CreateGroup(hid_t loc_id,const char *newGroupName_p)
{
 hid_t newGroup_id = -1 ;
 char fxName[256] = {0} ;
  
    strcpy(fxName,(char *)__func__);
    
    // Check if group does not already exist before trying to 
    // create it, since it is an error to try to re-create existing groups
    if(!frm_hdf5_GroupExists(loc_id, newGroupName_p))
    {
        newGroup_id = H5Gcreate(loc_id, newGroupName_p, H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
        xLogMsg("@A In(%s) created group(%s) at id (%d) owned by id(%d)~\n",
            fxName, newGroupName_p, (int)newGroup_id, (int)loc_id);
            
    } else {
        xLogMsg("@E In(%s) Cannot Create a pre-Existing group(%s) owned by id (%ld)~\n",
                fxName, newGroupName_p,loc_id);
    }

    return newGroup_id ;
    
} // end:: frm_hdf5_CreateGroup()


// ------------------------------------------------------------------
// function : frm_hdf5_OpenGroup()
// inputs   : hid_t loc_id, const char *groupName_p
// purpose  : open an existing group. 
// revision : 2014-06-13T18:25:00-07:00, orig
// returns  : new groups id, or -1 if failed.
// notes    : When specifying a group, don't lead with a slash
//            like in h5py, but do include a slash separator if
//            it is a nested group, relative to loc_id position.
hid_t frm_hdf5_OpenGroup(hid_t loc_id, const char *groupName_p)
{
 hid_t groupId = -1;
 char fxName[256] = {0} ;
  
    strcpy(fxName,(char *)__func__);
    
    if( frm_hdf5_GroupExists(loc_id, groupName_p) )
    {
        groupId =  H5Gopen(loc_id, groupName_p, H5P_DEFAULT) ;
        xLogMsg("@A In(%s) opened group(%s) at id (%ld)~\n",
                fxName, groupName_p, groupId);
    } else {
        xLogMsg("@E In(%s) Cannot open a Non-Existent group(%s) at id (%ld)~\n",
                fxName, groupName_p, groupId);
    }

 return groupId ;
 
} // end::frm_hdf5_OpenGroup()


// ------------------------------------------------------------------
// function : frm_hdf5_CloseGroup()
// inputs   : hid_t groupId -- id of group to close
// purpose  : close an existing group. 
// returns  : void
// revision : 2014-06-13T18:25:00-07:00, orig
void frm_hdf5_CloseGroup(hid_t groupId)
{
 herr_t ret = -1;
 
 ret =  H5Gclose(groupId ) ;

  xLogMsg("@A In(%s) closed group id (%ld) with ret %d~\n",
        __func__, groupId, (int)ret);
 
} // end::frm_hdf5_CloseGroup()

// ------------------------------------------------------------------
// function : frm_hdf5_ErrorStackEnable()
// inputs   : H5E_auto2_t *fxPtr -- ptr to fx assigned as error handler
//            which defaults to H5Eprint2() if passed in here as NULL
//          : void *clientData_p -- ptr to client data to pass in to
//            error handler. Not yet implemented.
// purpose  : Enable (turn on) the default HDF5 error stack mechanism
// returns  : int 1, allow storing state in a sesCtl variable
// example  : frm_hdf5_ErrorStackEnable(NULL,NULL)  ..idiom here uses defaults
// revision : 2014-06-17T10:40:00-07, orig
// notes    : reminder: herr_t H5Eprint2( hid_t estack_id, FILE * stream)
// notes    : Mode with user errFx passed in has NOT yet been tested.
int frm_hdf5_ErrorStackEnable(H5E_auto_t errFxHandlerPtr,void *clientData_p)
{
    
  hid_t error_stack = H5E_DEFAULT; // indicates WHICH stack to reference
  
    if(errFxHandlerPtr == NULL && clientData_p == NULL)
        H5Eset_auto(error_stack, (H5E_auto_t)H5Eprint,stderr); // uses H5Eprint2() of course...
    else
        H5Eset_auto(error_stack, errFxHandlerPtr,clientData_p);
        
    // returns 1 to indicate that default error stack is now ON
    return 1;
        
} // end:: frm_hdf5_ErrorStackEnable()


// ------------------------------------------------------------------
// function : frm_hdf5_ErrorStackDisable()
// inputs   : void
// purpose  : Disables (turns off) the default HDF5 error stack mechanism
// returns  : int, 0, indicating HDF5 error stack mechanism is now OFF.
// example  : frm_hdf5_ErrorStackDisable()  ..uses defaults
// revision : 2014-06-17T10:40:00-07, orig
int frm_hdf5_ErrorStackDisable(void)
{
hid_t error_stack = H5E_DEFAULT; // indicates WHICH stack to reference
    
    H5Eset_auto(error_stack, NULL, NULL);
    
    // returns 0 to indicate default error stack is now OFF
    return 0 ;
            
} // end:: frm_hdf5_ErrorStackDisable()


// ------------------------------------------------------------------
// function : frm_hdf5_GetLibraryVersion()
// purpose  : Retrieve (major,minor,releaseId) for linked in HDF5 library
char *frm_hdf5_GetLibraryVersion(char *hdf5VersionStr)
{
 // Retrieve HDF5 version no: major,minor,and release ID
 unsigned hdf5MajorVersionId = 0;
 unsigned hdf5MinorVersionId = 0;
 unsigned hdf5ReleaseId = 0;
 herr_t ret = -1 ;

    // fetch HDF5 version ID's (major,minor,releaseID)
    ret = H5get_libversion(&hdf5MajorVersionId,&hdf5MinorVersionId,&hdf5ReleaseId);
    // format all these into a string to return
    sprintf(hdf5VersionStr,"HDF5 version %u.%u.%u",
        hdf5MajorVersionId, hdf5MinorVersionId, hdf5ReleaseId);
        
 // return ptr to library version string
 return hdf5VersionStr ;

} // end:: frm_hdf5_GetLibraryVersion()

