// ---------------------------------------------------------------
// Purpose : frm_log_msg.c provides a stand-alone, self-contained,
//           lite, var-args style log write and exception handler
//           that does NOT require any external symbols or session mechanisms
// Dependencies : frm_log_msg.c/.h
// Revision: 2015-05-29T17:55, no changes or additions, added stub for FRM_LOG_FUTURE only
// history : 2014-02-09T15:48, made xLogMsg() safer by reverting to stdout if currentLogf not yet open
// history : 2014-01-05T23:11:00, introduced new LogMgr layer, to begin transition to thread-safe operation
//           although currently, this is still a single-stream system.
// history : 2013-11-04T16:35:00
// history : 2013-11-01, added new "TILDE" semantics, where is a '~' symbol is present on the
//           format specifier line to nvp_sll_logMsg(), the output is "tee'd" to also route to stdout.
// Author  : joe glassy, Lupine Logic Inc.
// Notes   : --include a tilde in any xLogMsg() expression to "T" traffic to console
// Notes   : Future feature request (Lucas): add a xLogSetting(enumOp) function to allow user
//           to {enable|disable} various logging behaviors, e.g. temporarily suspend output
//           to console, or file, or other etc.

// TBD: consider reducing or eliminate dependency on frm_mem_obj from frm_log_msg if possible!

#include "frm_mem_obj.h"
#include "frm_log_msg.h"

// master (private, shared) instance of log file handle -- NOT THREAD SAFE
static FILE *masterDefaultLogf = NULL;
static FILE *currentLogf = NULL; // ptr to active log stream


#ifdef FRM_LOG_FUTURE
// xLogSetting() allows user to manually enable|disable a given supported
// behavior from anywhere in the application, e.g. selectively turn on or off
// output to console, a text file, an HTML,JSON,YAML file, etc
// enums: logEnableConsole,logDisableConsole,logEnableText,logDisableText,
//        logEnableHtml,logDisableHTML etc.
// Notes: Requires access to a thread-aware central state repository, to allow
// various state or behavior settings to be "sticky" as needed
void xLogSetting(const int opCode)
{
   // <JMG> 5/29/2015  Development of xLogSetting() is TBD  
} // end::xLogSetting()
#endif

// ---------------------------------------------------------------
// function: xLogGetVersTag()
// purpose : Retrieve log and exception handler API version tag
// return  : char * version tag string, like: v2.0-2014-01-21T11:27:00-07:00
//           as defined in frm_log_msg.h
// revised : 2014-01-21T11:30:00-07:00,jmg
// notes   : n/a
char *xLogGetVersTag(void)
{

    return( (char *)FRM_LOG_VERSTAG );
    
} // end::xLogGetVersTag()

// ---------------------------------------------------------------
// function: xLogGetFilehandle()
// purpose : return handle to log file, only valid after opened.
// revised : 2015-10-14T17:09, new form, migrate ALL usage to this
//           and soon deprecate xGetLogfileHandle()
// notes   : This is NOT yet thread safe but can be made thread safe by
//           bracketing in a pthread mutex lock
FILE *xLogGetFilehandle(void)
{
     return currentLogf ;
} // end::xLogGetFilehandle()


// ---------------------------------------------------------------
// function: xGetLogfileHandle()
// purpose : return handle to log file, only valid after opened.
// notes   : DEPRECATE THIS FUNCTION VERY SOON,to use new xLogGetfileHandle()
// revised : 
// notes   : This is not yet thread safe but can be made thread safe by
//           bracketing in a pthread mutex lock
// 
FILE *xGetLogfileHandle(void)
{
     return currentLogf ;
} // end::xGetLogfileHandle()


// ---------------------------------------------------------------
// function: xSetCurrentLogHandle()
// purpose : return handle to log file, only valid after opened.
// notes   : not yet thread safe..
void xSetCurrentLogHandle(FILE *usersLogf)
{
        currentLogf = usersLogf ;
} // end:: xSetCurrentLogHandle()


// ---------------------------------------------------------------
// function: xLogOpenfile()
// inputs  : logPath -- string containing full path to text log file to open
//           Note that no FILE *logf is now passed in, as this form
//           requires use of default internal 'masterLogf'
// returns : returns time_t , starting seconds mark for session.
// revised : 2015-10-14T17:04:00
// purpose : Opens the single, internal default session log file handle
// notes   : To close log file,must now use: xLogClose(xLogGetFilehandle()) 
//
time_t xLogOpenfile(char *logPath)
{
 time_t strtSec = time(0L);

    // <JMG> 2015-10-14T16:49, requires use of internal default
    // logf file handle, 'masterDefaultLogf'
            
    if((masterDefaultLogf=fopen(logPath,"wt"))==NULL)
    {
        fprintf(stderr,"@F openLogfile::FAILED opening logfile(%s)\n",
            logPath);
        exit(-1);
    }
    
    // Set ptr to current,active log stream handle
    currentLogf = masterDefaultLogf; 

 // return the starting seconds count
 return(strtSec);

} // end::xOpenLogfile()


// ---------------------------------------------------------------
// function: xOpenLogfile()
// inputs  : logPath -- string containing full path to text log file to open
// returns : returns time_t , starting seconds mark for session.
// revised : 2013-11-04 (back-dated)
// purpose : open the session log file. If logf==NULL, then the 
//           single,shared default handle is used. If logf is valid,
//           it is used, and the file is opened for write, and the
//           currentLogf handle is set to this log handle.
// notes   : This function SHOULD currently be called using the
//           (non-obvious) usage of xOpenLogfile(path,NULL) and should
//           NOT be called with a user submitted 'logf' handle as
//           written. The plan is to change the call interface to
//           be more explicit, as in: time_t xOpenLogfile(char *logPath) 
//           forcing the use of internal 'masterDefaultLogf' handle,
//           which works fine in serial mode but is NOT thread safe.
//
//           Alternatively, another strategy will be to alter the call
//           interface to: time_t xOpenLogfile(char *logPath,FILE **logf)
//           such that we pass in the file handle, FILE *myLogf, like this:
//              xOpenLogfile(logPath,&myLogf);
//           This usage can also be confusing, so a preferable method is
//           just use the traditional idiom:
//              myLogf = xOpenLogfile(logPath);
//           which allows us to retain the termination side call:
//              xLogClose(myLogf);

time_t xOpenLogfile(char *logPath,FILE *logf)
{
 time_t strtSec = time(0L);
 FILE *workLogf = NULL;

    // this is sketchy logic, will be improved as per notes above
    if(logf == NULL)
            workLogf = masterDefaultLogf ;
    else
            workLogf = logf ;
            
    if((workLogf=fopen(logPath,"wt"))==NULL)
    {
        fprintf(stderr,"@F openLogfile::FAILED opening logfile(%s)\n",
            logPath);
        exit(-1);
    }
    // set ptr to current,active log stream handle
    currentLogf = workLogf ; 

 // return the starting seconds count
 return(strtSec);

} // end::xOpenLogfile()


// ---------------------------------------------------------------
// function: xCloseLog()
// purpose : this is an alias to xLogClose() kept only for backward compat
// notes   : This alias xCloseLog() is SOON to be deprecated, keeping only xLogClose()
void xCloseLog(FILE *logf)
{
    xLogClose(logf);
} // end::xCloseLog()


// ---------------------------------------------------------------
// function: LogClose()
// purpose : close out the private, single-stream log file handle
// inputs  : FILE *logf -- optional user defined file handle or if
//           this is NULL, uses shared default
// notes   : 
void xLogClose(FILE *logf)
{
    if(logf == NULL)
    {
        // <JMG> 2015-10-12, added test on null for more robustness
        if( currentLogf != NULL)
        {
            fflush(currentLogf);
            fclose(currentLogf);
        }
        
    } else {
        // <JMG> 2015-10-12, get segfault here on certain cases even
        // when logf is NOT null. Verify if after a fclose() handle is
        // explicitly set to null or not
        fflush(logf);
        fclose(logf);
    }
    
} // end:: xLogClose() 


// ----------------------------------------------------------------------------
// function: xLogMsg()
// inputs  : msg_p -- vsprintf style formatted string
// purpose : writes a formatted text message to the currently
//           active log file steam, pointed to by currentLogf.
// limitations : this function does NOT yet work with LogMgr
//              enabled concurrent log streams
void xLogMsg(char *fmtMsg_p,...)
{
  char    userMessage[FILENAME_MAX] = {0}; // POSIX...
  va_list argptr ;   // ptr to list of var args
  int     fmt_count =0;
  FILE *localf = NULL;
  
    va_start(argptr,fmtMsg_p);

    fmt_count= vsprintf(userMessage,fmtMsg_p,argptr);

    va_end(argptr);

    // bind to a file stream presumed open, but if not
    // re-route to safe stdout
    if( currentLogf != NULL)
        localf = currentLogf ;
    else
        localf = stdout ;

    // <JMG> 2016-03-03T18:35, re-arrange calls to put to
    // console first, so we can then strip out the annoying tilde mark
    // prior to output to file.
    
    // use presence of tilde to trigger echo to console
    // such that if meta-symbol '~' is omitted we suppress
    // output to console to avoid clutter
    if(strchr(userMessage,'~') != NULL)
    {
        fflush(stdout);
        // <JMG> 2016-03-03, if present, remove the tilde
        frm_strChrRemove(userMessage,'~');
        fprintf(stdout,"%s",userMessage);
    }
        
    // output message to safely open stream 
    fprintf(localf,"%s",userMessage);
    fflush(localf);

    // TBD: if FATAL, refine to try to first call the 
    // garbage collector. For now, just aborts. TBD:
    // re-test via exit(-1) to trigger the 'atexit' 
    // registered handler...
    if(strstr(userMessage,"@F") != NULL)
    {
        abort();
    }

  // completed...
} // end::xLogMsg()


// ----------------------------------------------------------------------------
// function: xLocalLogMsg(filehandle_f,char *fmtMsg_p)
// inputs  : FILE *thisLogf -- locally provided file stream handle
//           fmtMsg_p, ... -- vsprintf style formatted string with args
// purpose : writes a formatted text message to the currently
//           active log file steam, pointed to by currentLogf, which
//           eventually could be one of K active log streams or channels
// limitations : this function does NOT yet tested to work with LogMgr
//              enabled concurrent log streams
void xLocalLogMsg(FILE *thisLogf,char *fmtMsg_p,...)
{
  char    userMessage[FILENAME_MAX] = {0}; // POSIX...
  va_list argptr ;   // ptr to list of var args
  int     fmt_count =0;

    va_start(argptr,fmtMsg_p);

    fmt_count= vsprintf(userMessage,fmtMsg_p,argptr);

    va_end(argptr);

    fprintf(thisLogf,"%s",userMessage);
    fflush(thisLogf);

    // if FATAL, call the garbage collector and stop NOW
    // This needs more work to be robust...
    /*
    if(strstr(userMessage,"@F")!=NULL)
    {
        sesClose(nvp_sll_GetSesCtl());
        exit(-1);
    }
    */

    // use presence of tilde to trigger echo to console
    if(strchr(userMessage,'~')!= NULL)
    {
        fflush(stdout);
        fprintf(stdout,"%s",userMessage);
    }

} // end::xLocalLogMsg()


// ---------------------------------------------------------------
// LOG MANAGER services
// used to implement multiple, concurrent streams or channels of log
// traffic (e.g. one per thread at worst) to allow thread safe ops
// with a later consolidation pass as necessary back into one result log


// ----------------------------------------------------------------------------
// function: logMgrCreate()
// inputs  : const int nLogObjects-- count of log streams to create (dummy placeholder)
// purpose : Create one instance of a LogMgr, where each encapsults one log stream
//           to facilitate thread-safe use of the session logs, with one log per thread
LogMgr *logMgrCreate(char *thisLogMgrName)
{
    LogMgr *logMgr_p = NULL;

    
    if(memObjInitCondition() == 0)
    {
        fprintf(stderr,"@F logMgrCreate: memObjInit NOT yet called! cannot continue!");
        exit(-1);
    }
    
    // <JMG> 2/9/2014, switched to use simpler memCalloc()
    logMgr_p= memCalloc(1,sizeof(LogMgr));
    
    // intialize..
    strcpy(logMgr_p->logMgrName,thisLogMgrName);
    logMgr_p->logMgrMasterOwner = -1 ; // not defined yet.use process ID or thread id
    logMgr_p->nLogObjects = 0 ; // none yet...
    logMgr_p->logObj_p = NULL; // none yet.. may be 1 or more...
    // return pointer to the parent container of K concurrent log streams
    return logMgr_p ;
} // end::logMgrCreate()


// ----------------------------------------------------------------------------
// function: logMgrInit()
// inputs  : LogMgr *logMgr_p -- ptr to LogMgr 
//           const int nLogObj -- count of LogObj to create
//           char **logPathTable_p -- table of nLogObj paths (not yet implemented)
// purpose : create nLogObj instances of LogObj, provide a temporary filepath, and 
//           open each log file. This is evolving towards use in a thread pool,
//           such that each thread will be given its own log stream.
// notes   : char **logPathTable_p is passed in but has not yet been fully tested. This will
//           allow the caller to supply an explicit list of nLogCount paths, used to
//           open each log session file, for thread-safe operations
int logMgrInit(LogMgr *logMgr_p,const int nLogCount,char **logPathTable_p)
{
 LogObj *lObj_p = NULL;
 int a=0;   
 char lognamePath[256]={0};

    assert(nLogCount >= 1) ;
    logMgr_p->logObj_p = memObjCalloc("name",sizeof(LogObj),nLogCount,IsCustomUserType,"","");
    
    for(a=0; a < nLogCount; a++)
    {
        // for now, implement a auto-naming scheme. Can use rename() later
        // to rename these...
        if( logPathTable_p != NULL) 
            strcpy(lognamePath,logPathTable_p[a]);
        else
            sprintf(lognamePath, "LOGMGR_%-6.6d.log",a+1);
        lObj_p = &logMgr_p->logObj_p[a] ;
        // ALLOCATE as a registered object, using memObjCalloc()...
        lObj_p =(LogObj *)memObjCalloc(lognamePath, sizeof(LogObj),1,IsCustomUserType,"","");
        
        if((lObj_p->logObjf = fopen(lognamePath,"wt"))==NULL)
        {
            fprintf(stderr,"logMgrInit: cannot open log file(%s)\n",lognamePath);
            exit(-1);
        } else {
            // Store path internally, adjust states, assign ID, etc.
            strcpy(lObj_p->logObjPath,lognamePath);
            lObj_p->logObjLinesWritten = 0;
            lObj_p->logObjState = IsAvailable;
            lObj_p->logObjOwner = a+1; // only provisionally assigned...
        } // end:: block if log opened OK..
    } // end:: a loop through log streams to create and open
    
    logMgr_p->nLogObjects = nLogCount;
    
  return a ;
    
} // end:: logMgrInit()


// ----------------------------------------------------------------------------
// function: logMgrRelease()
// inputs  : LogMgr *logMgr_p -- ptr to log manager object.
// purpose : Free up one instance of a LogMgr, where each encapsults one log stream
// inputs  : LogMgr *logMgr_p -- ptr to one log manager container to free memory of
// notes   : If child log object files are open, these are closed.
// notes   : works on...nLogObjects-- count of log streams to release (dummy placeholder)
void logMgrRelease(LogMgr *logMgr_p)
{
 LogObj *thisLogObj_p = NULL;
 
 int k =0;
    for(k=0; k < logMgr_p->nLogObjects; k++)
    {
        thisLogObj_p = &logMgr_p->logObj_p[k] ;
        // first, make sure to close the file inside if open!
        if(thisLogObj_p->logObjState == IsAvailable && thisLogObj_p->logObjf != NULL)
        {
            thisLogObj_p->logObjState = IsNotAvailable;// explicitly MARK off limits
            fclose(thisLogObj_p->logObjf);
        }
        // now, release memory for the LogObj itself. Switched to use
        // simpler memFree() 2/9/2014...    
        memFree(thisLogObj_p);
    }   // end:: k loop through log stream objects to release
} // end::logMgrRelease()


// ----------------------------------------------------------------------------
// function: logMgrMessage()
// inputs  : LogObj *logObj_p,char *fmtMsg_p,...
// purpose : emit a formatted log message to a specific LogObj -- logObj_p.
// inputs  : LogMgr *logMgr_p -- ptr to one log manager container to free memory of
// notes   : If child log object files are open, these are closed.
//         : TBD- establish a method to pre-register a pointer-to-function, such as done via
//           atexit(fxPtr), to facilitate trapping fatal errors via "@F"
void   logMgrMessage(LogObj *logObj_p,char *fmtMsg_p,...)
{
  LogObj *lObj_p = NULL; // ptr to one LogObj owned by LogMgr parent
  char    userMessage[FILENAME_MAX] = {0}; // POSIX...
  va_list argptr ;   // ptr to list of var args
  int     fmt_count =0;

    va_start(argptr,fmtMsg_p);

    fmt_count = vsprintf(userMessage,fmtMsg_p,argptr);

    va_end(argptr);

    // Make sure its ready for use!
    assert(logObj_p->logObjState == IsAvailable);
    
    fprintf(logObj_p->logObjf,"%s",userMessage);
    fflush(logObj_p->logObjf);
    
    // Increment count of lines this stream has emitted
    lObj_p->logObjLinesWritten++ ;

    // If FATAL, call the garbage collector and stop NOW
    // This needs more work to be robust...
    if(strstr(userMessage,"@F")!=NULL)
    {
        abort();
    }

    // use presence of tilde to trigger echo to console
    if(strchr(userMessage,'~')!= NULL)
    {
        fflush(stdout);
        fprintf(stdout,"%s",userMessage);
    }   
    
} // end::logMgrMessage()


// ----------------------------------------------------------------------------
// function: logObjCreate()
// inputs  : LogMgr *logMgr_p,char *logObjName, char *logPath
// purpose : create a single instance of a LogObj, where each encapsulates one log stream
//           typically used in a serial process application if one, or 
//           multiple if in a multi-threaded app
// inputs  : LogMgr *logMgr_p -- ptr to one log manager container to free memory of
//           char logObjName -- name for memObjCalloc
//           char logPath -- files path to open write traffic to
// notes   : the log file is subsequently opened after creation, using logPath
LogObj *logObjCreate(LogMgr *logMgr_p,char *logObjName, char *logPath)
{
 LogObj *lObj_p = NULL;

    if(memObjInitCondition() == 0)
    {
        fprintf(stderr,"@F logObjCreate: memObjInit NOT yet called! cannot continue!");
        exit(-1);
    }
 
    // <JMG> 2014-02-09, switched to use simpler memCalloc()
    logMgr_p->logObj_p = (LogObj *)memCalloc(1,sizeof(LogObj));
 
    lObj_p = logMgr_p->logObj_p ; // set a convenience pointer
    // Post the name internally for self-referential use by memObjFree() later
    strcpy(lObj_p->logObjName,logObjName);
 
    logMgr_p->nLogObjects = 1; // be sure to indicate we are in SINGLE log mode!

    // OPEN just a single log file
    if((lObj_p->logObjf =fopen(logPath,"wt"))==NULL)
    {
        fprintf(stderr,"logObjCreate: cannot open log file(%s)\n",logPath);
        exit(-1);
    } else {
        lObj_p->logObjState = IsAvailable ;
    }
  
 // return LogObj instance
 return lObj_p ;
    
} // end::logObjCreate()


// ----------------------------------------------------------------------------
// function  : logObjRelease()
// purpose   : Release (close and free) one LogObj, using either a base0 index or name
// inputs    : LogMgr *logMgr_p,char *logName,const int logIdx
// notes     : --to search on name "logName" set logIdx = -1
//           : --to search on base0 index logIdx, set >=0 and set logName=""
void  logObjRelease(LogMgr *logMgr_p,char *logName,const int logIdx)
{
 LogObj *lObj_p = NULL;
 int k= 0;
 int nAvail = logMgr_p->nLogObjects;
 char logname[256] ={0} ;
 
     for(k=0; k < nAvail; k++)
     {
            if(logIdx >= 0 && logIdx == k)
            {
                lObj_p = &logMgr_p->logObj_p[k];
                strcpy(logname,lObj_p->logObjName);
                break;
            } else {
                lObj_p = &logMgr_p->logObj_p[k];
                if( strlen(logName) >=1 && logName != NULL)
                    if(strcmp(logName,lObj_p->logObjName)==0)
                        break;
            }
     } // end:: k 
     
     // if (registered) log object correctly identified, free it
     // <JMG> 2014-02-09, switched to use simpler memFree()
     if(strlen(logname)>= 1)
        memFree(lObj_p);
     
} // end::logObjRelease()


// ----------------------------------------------------------------------------
// function  : getLogObj() 
// purpose   : fetch one log object (stream) via name or base0 index
// inputs    : LogMgr *logMgr_p,char *logName,const int logIdx
// notes     : to search on name "logName" set logIdx = -1
//           : to search on base0 index logIdx, set >=0 and set logName=""
LogObj *getLogObj(LogMgr *logMgr_p,char *logName,const int logIdx)
{
 LogObj *lObj_p = NULL;
 int k= 0;
 int nAvail = logMgr_p->nLogObjects;
 
     for(k=0; k < nAvail; k++)
     {
            if(logIdx >= 0 && logIdx == k)
            {
                lObj_p = &logMgr_p->logObj_p[k];
                break;
            } else {
                lObj_p = &logMgr_p->logObj_p[k];
                if( strlen(logName) >=1 && logName != NULL)
                    if(strcmp(logName,lObj_p->logObjName)==0)
                        break;
            }
     } // end:: k 
     
     // return pointer to selected LogObj, or NULL if not found
     return lObj_p ;
} // end:: getLogObj()



#ifdef FRM_LOG_UNIT_TEST
int main(int argc,char *argv[])
{
 FILE *testlogf = NULL;
 char testLogpath[FILENAME_MAX]= {0};
 char dtBufr[32] = {0} ;
 time_t strtSec = 0;
 
 // Test single-stream log facility. 
 
 // Fetch the default log file handle
 testlogf = xGetLogfileHandle() ;
 
 // Exercise original single-stream log facility, 
 //  using USE_DEFAULT_LOG (which is NULL) semantics
 strcpy(testLogpath,"frm_log_msg.log");
 strtSec = xOpenLogfile(testLogpath,USE_DEFAULT_LOG);
 sprintf(dtBufr,"%s",ctime(&strtSec));
 dtBufr[strlen(dtBufr)-1] = '\0' ; // clip newline
 
 xLogMsg("@A Test of ADVISORY class message from frm_log_msg facility on %s~\n",dtBufr);
 xLogMsg("@W Test of WARNING  class message echoing to console with tilde ~\n");
 xLogMsg("@E Test of ERROR    class message echoing to console with tilde ~\n");
 
 strtSec = time(0);
 // xLogMsg("@F Test of FATAL    class message at %s\n",ctime(&strtSec));
 xCloseLog(USE_DEFAULT_LOG);
 
 // FUTURE: test the multiple-stream (thread-safe) LogMgr based
 // session logging facility. TBD...
  
 return 0;
} // end::main()
#endif

