
// ---------------------------------------------------------------
// module       : frm_mem_obj.c
//
// purpose      : The "memObj" API component of the UM Framework API implements 
//                 a session oriented, registry based, journaled 
//                 heap memory management scheme such that full reporting of 
//                 all memory objects allocated is available. The number of
//                 variables and how much memory each uses is included in 
//                 reports, along with estimates of the "high-water-mark" used
//                 within the session. In the memObj scheme, memory objects may
//                 be "named" (if allocated and free'd using memObjCalloc(),memObjFree()),
//                 or they may be "anonymous" (if allocated and free'd 
//                 using {memCalloc(),memFree()}. The default memObj facility uses a
//                 single (static,fixed size) memory object registry (or dictionary),
//                 whose size is fixed at build-time.  This default registry scheme is not
//                 guaranteed to be thread safe.  An extended memObj facility called
//                 the MemMgr interface, also in development, offers the programmer a
//                 completely independent set of concurrent registers, so each thread
//                 or process can be provided with its own memory registry/dictionary.
// 
//  Capabilities:  Primary capabilities offered by memObj facility are to:
//  -- allocate, register a user defined set of (heap) memory variables, using 
//     either memObjCalloc() (for named objects), or memCalloc() for anonymous objects.
//  -- provide object-oriented "name" based access methods (memObjGet(objname)) to allow
//     a function anywhere in the application to fetch a pointer to an objects memory by name
//     or slot index.
//  -- track a given heap memory object through its entire life-cycle, once registered.
//  -- track the amount of memory used, which caller requested it, and n. bytes requested
//  -- provide an end-of-session garbage collector that frees any registered objects not otherwise
//     free'd.
//  -- maintain a session wise "high-water-mark" of memory consumed within a session life-cycle.
//  -- offer name oriented, matching calloc + free facilities where each variable
//     is referenced with a text label (typically the same as its variable name),
//     stored in a session table of memory objects, such that a
//     report of all registered memory objects is available at any time (memObjDump(),
//     memObjReport() etc.
//  -- provide support for single-handle vectors as well as string-table (compound) 2-level objects
//  -- provide an event-driven method to write out the binary contents of a memory object
//     to disk, to eventually cooperate within a 'warm-restart' capability if a long-lived 
//   session is interrrupted or dies.
//

// author       : joe glassy, Lupine Logic Inc.
// revision     : 2016-01-07, in memObjSetName() added check to dis-allow (name) strings that
//                are too long , e.g. exceed MAX_MEMOBJ_VNAME_LEN. More of these checks
//                should also be added throughout API!
// revision     : 2014-08-21T16:44, made error handling more robust in memObjPrivateMsg()
// revised      : 2014-02-23T10:50, added memObjGetName(),memObjSetName(),
//                memObjGetNelem,memObjSetMetadata() etc
// history      : 2014-01-21T11:22:00-07:00
// history      : 2014-01-21T11:22:00, added memObjGetVersTag()
// revised      : 2014-01-05T21:06:00-07:00, now frees IsStringTable correctly
// history      : 2013-07-12T12:49:00-07:00 for D4
// builds       : use -lefence to test, if available.
// test         : use valgrid frm_mem_mgr.ex after building in unit test mode
// version      : see VERS macro
// dependencies : frm_mem.h, esp defining MAX_MEMOBJ_CANDIDATES .
// limitations  : the n. of registered memory variables is MAX_MEMOBJ_CANDIDATES set at compile time.
//
// example : typical workflow and usage:
//   memObjInit();
//   memObjCalloc(varTextLabel,SizeInBytesOfOne,nElements,
//      enumDataTypeIdentifier,"optionalFunctionName","optionalCallerLabel");

//             myVar_p = memObjCalloc("myVarsLabel",sizOfOne,nElem,
//                enumDatatypeId,"optionalFxName","optionalCallerLbl") ;
//                 ...now use myVar_p throughout application...or, get it locally via
//                 ...myLocal_ptr_p =  memObjGet("myVarsLabel");
//                 ...use it...
//             memObjFree("myVarsLabel",NULL); ..lets manager find the ptr by name
//             memObjFreeAll()       ...may be used within-session..
//             memGarbageCollect();  used only at end-of-session, ..let's manager 
//                to finalize and free any objects not yet free
//
//  Protocols:
//  --when a variable is registered to memObj using memObjCalloc(),
//  via a text label, this label must subsequently ALWAYS be matched EXACTLY when it is
//  later released by memObjFree(). Failing to match the name exactly will result
//  in temporarily orphaned memory, BUT, even this memory is released at the 
//  end-of-session when a call is made to memObjGarbageCollect(), which doesn't need names.
//
//  Release Notes  2012-11-20T16:29:00-07:00,Boson
//  This code implements a stand-alone memObj facility on Linux and Windows 7
//  On Linux, any 2.6.x kernel will do.
//  On Windows 7, it is typically built using Visual Studio Express 2008,2010,2012,2013

// Debugging tips
// To route valgrind output to a file, use:
// valgrind -v ./FrmApiTemplate.ex >& FrmApiTemplate-valgrindReport.txt  


// ---headers-----------------------------------------------------

#include "frm_mem_obj.h"


// ---module-level static symbols --------------------------------

// Define a fixed size array of pointers to void, such that each memory object
//  binds to one slot, and always requests memory via calloc to initalize the variable.
//  Note that these memory slots are re-usable within a given execution life-cycle,
//  so MAX_MEMOBJ_CANDIDATES just refers to the n. of <concurrent> memory objects active, and NOT to
//  the total no. of memory objects actually created and used within a session.

// allows check if memObjInit() properly called first
static int    memObjFacilityInitialized = 0 ; 
// nMemObjRegistered counts the number of variables actively registered in the facility
static size_t nMemObjRegistered  =0;
// nMemObjBytesTracked counts the number of bytes under management
static size_t nMemObjBytesTracked = 0 ;
static size_t nMemHighWaterMark   = 0 ;
static size_t highestSlotReferenced = 0 ;

// memObjRegistry is an array of (parent level) memory container objects
static MemObj memObjRegistry[MAX_MEMOBJ_CANDIDATES] ;

static FILE *memObjCurrLogf = NULL ;

// <JMG> 2014-07-30, added, for deeper diagnostic debugging
static int memVerboseMsg = 0 ;

// ---------------------------------------------------------------
// function: memObjAdjustVerbose(()
// purpose : Allows external caller to adjust the boolean setting
//           of 'memVerboseMsg'. A 1 setting VERBOSE ON and
//           0= DISABLES VERBOSE messaging.
// returns : void
// revision: 2014-07-30T03:06:00-07:00
// author  : joe glassy
// notes   : n/a
void memObjAdjustVerbose(const int newSetting)
{
    if(newSetting < 0 || newSetting > 1)
    {
        fprintf(stderr,"@F In(%s) INVALID memObj setting (%d) for memVerboseMsg!\n",
                (char *)__func__, newSetting);
        abort();
    }
    memVerboseMsg = newSetting;

} // end::memObjAdjustVerbose()

// ---------------------------------------------------------------
// function: memObjRegisterLog()
// purpose : caller must now pass in their own FILE *userf for mem_obj
//           to use as its destination steam for log traffic.
// returns : void
// revision: 2014-01-21T11:25:0-07:00
// author  : joe glassy
// notes   : TBD,make sure this also allows caller to pass in 'stdout'
//           or 'stderr' etc.
void memObjRegisterLog(FILE *whichLogf)
{
    memObjCurrLogf = whichLogf ;
} // end::memObjRegisterLog() 


// ---------------------------------------------------------------
// function: memObjPrivateMsg()
// purpose : emit one sprintf() style formatted log message
//           to a private log instance, where the dest FILE *logf 
//           must have been pre-registered using memObjRegisterLog()
// returns : void
// revision: 2014-08-21T16:44, made error handling more robust
// revision: 2014-02-12T21:39-07:00, avoided dup messages to stdout.
// author  : joe glassy
// notes   : TBD,make sure this also allows caller to pass in 'stdout'
//           or 'stderr' etc.
// notes   : TBD, reorganize order where assert() is, and replace
//           assert() with a softer error trap mechanism!
void memObjPrivateMsg(char *fmtMsg_p,...)
{
  char    userMessage[FILENAME_MAX] = {0}; // POSIX...
  char fxName[256] = {0};
  va_list argptr ;   // ptr to list of var args
  int     fmt_count =0;

    
    strcpy(fxName,__func__);
    // format the message and emit it
    va_start(argptr,fmtMsg_p);
    fmt_count= vsprintf(userMessage,fmtMsg_p,argptr);
    va_end(argptr);

    // minimally verify that we have a valid, open stream to write to
    // assert(memObjCurrLogf != NULL);
    if(memObjCurrLogf == NULL) 
    {
        fprintf(stderr,"%s",userMessage);
        fprintf(stderr,"@F In(%s) UNEXPECTED null log stream!\n",fxName);   
        fflush(stderr);
        abort();
    } else {
        fprintf(memObjCurrLogf,"%s",userMessage);
        fflush(memObjCurrLogf);
    }
    

    if(strstr(userMessage,"@F") != NULL)
        abort();
        
    // use presence of tilde to trigger echo to console
    if(strchr(userMessage,'~') != NULL)
    {
        fflush(stdout);
        // avoid dupicate messages by checking that the
        // current logf stream hasn't already been set to stdout
        if( memObjCurrLogf != stdout)
            fprintf(stdout,"%s",userMessage);
    }
} // end::memObjPrivateMsg()


// ---------------------------------------------------------------
// function: memObjGetVersTag()
// purpose : retrieve MEM_OBJ_API_VERSTAG to allow caller to confirm version
// returns : char *, string containing full version tag
// revision: 2014-01-21T11:25:0-07:00
// author  : joe glassy
// notes   : rarely used or needed, may be deprecated.
char *memObjGetVersTag(void)
{
    return((char *)MEM_OBJ_API_VERSTAG);
} // end:: memObjGetVersTag()

// ---------------------------------------------------------------
// Function: memObjExport()
// Purpose : export visibility of master memory object dictionary to outside caller,
//           primarily used by name-value-pair support functions.
// Returns : ptr to master instance of memory object dictionary
// Notes   : rarely used or needed, may be deprecated.
MemObj *memObjExport(void)
{
 return memObjRegistry ;
} // end::memObjExport()


// ---------------------------------------------------------------
// function : memObjGetSlot()
// inputs   : void *referent_p --ptr to a prior registered variable
// returns  : base0 index of slot mapped to this variable if found, or -1 if none was found.
// purpose  : return base0 slot index of a prior occupied slot
int memObjGetSlot(void *referent_p)
{
 int iSlot     = 0 ;
 int foundSlot = -1;
 char addrLbl[MAX_MEM_DNA_LEN] = {0} ;
 char *dnaName_p = NULL;
 
    // capture a tissue sample of my DNA 
    sprintf(addrLbl, "%p", (uint8 *)referent_p);
    
    if(strstr(addrLbl,"nil") != NULL)
    {
        fprintf(stderr,"@W warning: memObjGetSlot: referent DNA (%s) may be 'nil'\n",addrLbl);
        return -1 ;
    }
    
    // Perform linear scan of fixed length registry table for slot matching this DNA
    for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
        // pre-load the object's internal DNA level identifier
        dnaName_p = memObjRegistry[iSlot].vDnaLabel ;
        // test for a DNA match, exit when its found
        if(strcmp(addrLbl, dnaName_p)== 0)
        {
            foundSlot = iSlot;
            break;
        }
  } // end:: iSlot scanning loop
  
  // return base0 slot as {0..n-1} or -1 if not found at all
  return foundSlot ;
} // end::memObjGetSlot()


// ---------------------------------------------------------------
// function : memObjGetNextSlot()
// inputs   : void
// returns  : base0 index of next available slot, or -1 if none was found.
// purpose  : return base0 slot index of next unoccupied slot or -1 if none found
// notes    : 
int memObjGetNextSlot(void)
{
 int iSlot     = 0 ;
 int foundSlot = -1;
 
    // Perform Linear scan of the fixed length registry table to locate
    // the next available slot. This is done since there may be interior 
    // "holes" or gaps in the registry, as vacated by released slots.
    for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
            if( memObjRegistry[iSlot].vNelem == 0     &&
                memObjRegistry[iSlot].vContents_p == NULL )
                    
            {
                    foundSlot = iSlot ;
                    break;
            }
    } // end:: iSlot scanning loop
  
  // return base0 slot available 
  // if valid, foundSlot is {0..n-1} and points to a slot already
  // occupied, or to the next availalble slot 
  return foundSlot ;
} // end::memObjGetNextSlot()


// --------------------------------------------------------------------
// Purpose: memObjTotalBytesUsed() -- reports current high-water-mark of memory used
size_t memObjTotalBytesUsed(void)
{
 return nMemObjBytesTracked ; 
} // end:: memObjTotalBytesUsed()


// --------------------------------------------------------------------
// function: memObjInit()
// inputs  : FILE *usersLogf
// returns : size_t, total count of fixed-number of memory object slots initialized.
// Purpose : memObjInit() -- mandatory function that must be called prior to
//             using the memObj registry.  This call intializes the session memory 
//             tracking facility e.g. sets all MAX_MEMOBJ_CANDIDATES elements 
//             to a known (resting) state.  it MUST be called exactly once, 
//             and once only, at an early point in the session
//             prior to any calls to memObjCalloc(), memCalloc(), memObjFree() or memFree()
// revised : 2014-02-22T16:13, now sets "stdout" as the default private log dest
//           in case user forgets to call memObjRegisterLog(myLogf)
// Notes   : To adjust the size of the default (static) registry at build time,
//             change the value of MAX_MEMOBJ_CANDIDATES in the mem_obj.h header,
//             then re-build.
size_t memObjInit(FILE *usersLogf)
{
 int  iSlot =0;        // index
 
    // Test if not yet initialized and proceed only if not yet initalized
    assert(memObjFacilityInitialized == 0);

    // As part of initialization, register the user's choice of log
    //  stream right away. This is at least, 'stdout' or 'stderr' etc.
    //  but otherwise is assumed to be an open stream.
    memObjCurrLogf = usersLogf ;

    // fprintf(stdout,"DIAG memObjInit nMemObjRegistered(%u) TotalBytesTracked %u\n",
    // (unsigned int)nMemObjRegistered,(unsigned int)nMemObjBytesTracked);
    nMemObjRegistered     = 0 ;
    nMemObjBytesTracked   = 0 ;
    highestSlotReferenced = 0 ;
    
    for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
        memObjRegistry[iSlot].vContents_p = NULL;
        memObjRegistry[iSlot].vNelem = 0;
        memObjRegistry[iSlot].vBytes = 0;
        // IsNoType,IsBoolean,IsInteger,IsReal,IsString,IsFilespec
        memObjRegistry[iSlot].vDatatype = IsNoType;
        memset(memObjRegistry[iSlot].vName,     '\0',MAX_MEMOBJ_VNAME_LEN);
        memset(memObjRegistry[iSlot].vDnaLabel, '\0',MAX_MEM_DNA_LEN);
    } // end iSlot for loop

  // Alter state of init to now reflect proper initialization
  memObjFacilityInitialized = 1 ;

  // return count of slots initialized    
 return iSlot; 
} // end::memObjInit()


// purpose: return 1 if mem obj system initailized, or 0 if not
int memObjInitCondition(void)
{
    return memObjFacilityInitialized ;
} // end::memObjInitCondition()

// --------------------------------------------------------------------
// memObjCalloc()
//
// Purpose: wrap calloc and keep track of all names allocated and how
//   many bytes each received. Each high level memory object managed herein
//   must be referred to by the same name for both its alloc and free, 
//   and can only be allocated once per a unique name. IsStringTable is now supported.
//
// Inputs:
//   objName_p -- string, text name for variable to be registered to memObj facility
//   sizeOfOne -- size in bytes of scalar instance of datatype. Or if IsStringTable, defines string widths.
//   nElem -- n. of elements of scalar datatype to allocate memory for. If IsStringTable,defines n. of strings.
//   vDtype -- enum identifer for NVP generic class datatypes IsString, IsReal, etc
//   fxLbl -- string identifying calling function, usually fxName == __func__
//   callerStr -- string identifying caller-file, usually uses __FILE__ or equiv here
// Returns:
//   void *ptr -- pointer to 1D serialized memory block (don't use for explicitly multidimensional
//     variable declarations like **myPtr or ***myPtr etc.
// History   : 2013-11-08, added support for 2-level IsStringTable datatype, where sizeOfOne defines the string widths
//             and nElem defines the n. of strings.
// Invocation: 
//   memObjCalloc(varTextLabel,SizeInBytesOfOne,nElements,enumDataTypeIdentifier,"optionalFunctionName","optionalCallerLabel");
//     Note: use enum l4c_daily_adt.h::enum::NvpNumerictypes::IsCustomUserType whenever the datatype is 
//           not a simple scalar defined elsewhere in NvpNumericTypes
// Example:
//   strcpy(moduleFile,__FILE__);
//   pftImg_p= memObjCalloc("PftImage",sizeof(uint8),nPix9km,IsUint8,"myFunc",moduleFile);
//
//  l4cSesCtl_p->landcoverPftImg_p = (uint8 *)memObjCalloc("PftImage",
//          sizeof(uint8), l4cSesCtl_p->nModelPixels9km, IsUint8, "",moduleFile)

void *memObjCalloc(const char *objName_p,const size_t sizeOfOne,const size_t nElem,const int vDtype,const char *fxLbl,const char *callerLbl)
{
 void *new_p = NULL;    // proxy for 1-level allocs, for scalars, or simple arrays
 char **strTbl_p= NULL; // proxy for 2-level allocs, e,g. **ary
 char addrLbl[MAX_MEM_DNA_LEN] = {0} ;
 int  iSlot =0;               // slot index into internal table of memory objects
 
 // int  ithStr =0;  // index for IsStringTable allocation
 int  hwmIndex = MAX_MEMOBJ_CANDIDATES; // high-water-mark base0 index
 int  strTblItemwidth = 0 ;
 int  foundDuplicate = 0 ;
 
 size_t bytesRequested = (sizeOfOne * nElem); // derive n. bytes requested as product of these two
 
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
    // ------------------------------------------------------------------
    strcpy(moduLabel,(char *)__FILE__);
    strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 
    
    // block any progress beyond here if facility is not yet initialized, or insufficient slots left!
    if(memObjFacilityInitialized != 1)
    {
        memObjPrivateMsg("@F memObjCalloc: %s is NOT currently initialized!~","memObjFacility");
    }

    if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
        memObjPrivateMsg("@F memObjCalloc MaxSlotViolation on variable(%s). %zu is >= max allowable objects: %zu!~\n",
            objName_p, nMemObjRegistered, (size_t)MAX_MEMOBJ_CANDIDATES);           

    if(nElem == 0 || sizeOfOne == 0)
        memObjPrivateMsg("@F memObjCalloc: Variable(%s) with FATAL nElem (%zu) or Invalid sizeOfOne %zu!\n",
            objName_p, nElem, sizeOfOne);
    
    // TBD: duplicate-check on prescan works ONLY for non-dot-qualified names at this time...
    for(iSlot =0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
        if(strcmp(objName_p,memObjRegistry[iSlot].vName)==0)
        {
            memObjPrivateMsg("@F memObjCalloc found (%s) already registered at slot %d!~\n",iSlot);
            foundDuplicate = 1 ;
            // STOP IMMEDIATELY!
            exit(-1);
        }
    } // end::iSlot scan 


    // scan for first unused slot
    for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
        if(memObjRegistry[iSlot].vNelem == 0 && strlen(memObjRegistry[iSlot].vName)==0 )
        {
            if( vDtype == IsStringTable) 
            {
                strTblItemwidth = sizeOfOne ;
                strTbl_p= memCreateStringTbl(nElem, strTblItemwidth);
                // capture and store my DNA...
                sprintf(addrLbl,"%p",strTbl_p);
                
                // BIND 2-level returned pointer BACK to parent's 1-level ptr. This
                // is normally quite dangerous, unless always accessed ONLY via a
                // call interface requiring the IsStringTable identifer to ensure it
                // is properly de-referenced later ONLY as a 2-level pointer!
                
                new_p = strTbl_p ; // BIND 2-level ptr back to parent 1-level ptr
                // end:: if vDtype is IsStringTable
            } else {
                // BLOCK for allocating a scalar or 1-level array
                if((new_p= calloc(nElem,sizeOfOne+1))==NULL)
                {
                    memObjPrivateMsg("@F memObjCalloc: FATAL-ERROR allocating(%s) of size(%zu) Nelem %zu\n",
                        objName_p, sizeOfOne,nElem);
                    exit(-1);
                } 
                // capture my DNA
                sprintf(addrLbl,"%p",new_p); // 1/20/2014,add DNA as unique identifier
                
            } // end:: block if a scalar|array, and is NOT a IsStringTable
            
            // TBD: package up all actions assoc. with registering a new object as
            // below, into a separate new function memObjRegisterNewObj(void *new_p,
            //    const char *dnaAddrLbl,const int dataType,const size_t sizOfOne,
            //    const size_t nElem,const size_t bytesRequested)
            // then, simplify all this using memObjGetNextSlot() and memObjRegisterNewObj()
            
            nMemObjRegistered++ ; // increment count of objects successfully registered
            // set tracking pointer to address of this heap memory variable
            memObjRegistry[iSlot].vContents_p  = new_p; // BIND back to master pointer!
            memObjRegistry[iSlot].vNelem     = nElem ;
            memObjRegistry[iSlot].vSizeOfOne = sizeOfOne ; // size of 1 scalar instance
            memObjRegistry[iSlot].vBytes     = bytesRequested;
            // New 1/20/2014, added default unique identifier 
            strcpy(memObjRegistry[iSlot].vDnaLabel, addrLbl);

            // store the code matching the scalar data type this is
            // IsBoolean,IsInteger,IsReal,IsString,IsFilespec, or, IsUint32, IsFloat64, IsCustomUserType, etc
            memObjRegistry[iSlot].vDatatype  = vDtype ;
            // update global facility-wide count of total bytes tracked by memObj
            nMemObjBytesTracked += bytesRequested;

            // update long-lived sentinel, "highest slot" referenced
            if( (size_t)iSlot > highestSlotReferenced)
              highestSlotReferenced = iSlot ;

            // Check if user submitted an explicitly dot-qualified object name
            // TBD: deprecate as soon as possible. Too fragile...
            if(strchr(objName_p,'.') != NULL)
            {
                // Store a user-defined, dot-qualified variable name to registry
                strcpy(memObjRegistry[iSlot].vName,objName_p);
            } else { // end:: if block if a dot-qualifier name WAS found
                // Store an auto-constructed dot-qualified name, prepending 'fxLbl' before the '.'
                // and then appending the objName_p, but only if fxLbl is not empty...
                if(fxLbl != NULL && strlen(fxLbl) >0)
                {   // AUTO-DOT-QUALIFY this memory variable name to guarantee uniqueness on scope
                    strcpy(memObjRegistry[iSlot].vName,fxLbl); // uses fxName usually 
                    strcat(memObjRegistry[iSlot].vName,".");   // add the dot after fxName
                    strcat(memObjRegistry[iSlot].vName,objName_p);// now add the users name
                } else {
                    // allow storing a non-dot-qualified name, presumably NOT a duplicate
                    // to a name that already exists in the slot table.
                    strcpy(memObjRegistry[iSlot].vName,objName_p);
                }
            } // end:: else block if a dot-qualifier syntax was NOT found
#ifdef L4C_DIAG_MEM
            memObjPrivateMsg("memObjCalloc","@A iSlot(%3d) registered <%-32s> allocated %zu bytes by <%s> HWM %zu nReg %zu\n",
                iSlot, memObjRegistry[iSlot].vName, bytesRequested, callerLbl, 
                nMemObjBytesTracked,nMemObjRegistered);         
#endif              
            // if present, stores the callerLbl in the vMetadata entry for this object
            if(strlen(callerLbl)>0)
                strcpy(memObjRegistry[iSlot].vMetadata,callerLbl);

            break;
        } // end:: if slot state is... ready for use

        } // end:: iSlot loop

    // bump up session wise high water mark
    if(nMemObjBytesTracked > nMemHighWaterMark)
        nMemHighWaterMark = nMemObjBytesTracked ;

    hwmIndex = iSlot ; // stores base0 index of slot found  
    // fprintf(stdout,"DIAG memObjCalloc <%s> i-on-Exit(%d) hwmIndex (%zu) nMemObjBytesTracked %zu\n",
    //  objName_p,i, hwmIndex, nMemObjBytesTracked);
      
  // return pointer to new allocated memory    
 return new_p ; 
} // end::memObjCalloc()


// --------------------------------------------------------------------
// function : memObjDuplicateTest()
// returns  : int, 1 if duplicate found, 0= no duplicate found
// purpose  : test callers label as a potentially duplicate name already registered
// returns  : 0=false 1=true, is duplicate
// notes    : a Duplicate (@E ERROR) now treated as recoverable, not FATAL
int memObjDuplicateTest(const char *testObjName_p)
{
 int iSlot          = 0;
 int foundDuplicate = 0 ;
    // TBD: duplicate-check on prescan works ONLY for non-dot-qualified names at this time...
    for(iSlot =0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
        if(strcmp(testObjName_p, memObjRegistry[iSlot].vName)==0)
        {
            memObjPrivateMsg("@F l4c_MemObjCalloc found (%s) already registered at slot %d!~\n",iSlot);
            // exit(-1);
            foundDuplicate = 1 ;
            break;
        }
    } // end::iSlot scan 
 // return 1 if callers name was a duplicate (already registered), otherwise 0
 return foundDuplicate ;

} // end::memObjDuplicate()


// --------------------------------------------------------------------
// function : memObjCallocByName()
// purpose  : using a specific name, register a new memory object and allocate it
// inputs   :const char *objName_p,const size_t sizeOfOne,const size_t nElem,
//                      const int vDtype,const char *fxLbl,const char *callerLbl
// returns  : void 
void *memObjCallocByName(const char *objName_p,const size_t sizeOfOne,const size_t nElem,
                        const int vDtype,const char *fxLbl,const char *callerLbl)
{
 void *new_p = NULL;    // proxy for 1-level allocs, for scalars, or simple arrays
 char **strTbl_p= NULL; // proxy for 2-level allocs, e,g. **ary
 char addrLbl[MAX_MEM_DNA_LEN] = {0} ;
 int  iSlot =0;               // slot index into internal table of memory objects
 // int  ithStr =0;  // index for IsStringTable allocation
 int  hwmIndex = MAX_MEMOBJ_CANDIDATES; // high-water-mark base0 index
 int  strTblItemwidth = 0 ;
 int  foundDuplicate = 0 ;
 size_t bytesRequested = (sizeOfOne * nElem); // derive n. bytes requested as product of these two
 
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
    // ------------------------------------------------------------------
    strcpy(moduLabel,(char *)__FILE__);
    strcpy(fxLabel  ,fxLbl);
    
    // block any progress beyond here if facility is not yet initialized, or insufficient slots left!
    if(memObjFacilityInitialized != 1)
        memObjPrivateMsg("@F memObjCallocByName: %s is NOT currently initialized!~","memObjFacility");

    if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
        memObjPrivateMsg("@F memObjCalloc MaxSlotViolation on variable(%s). %zu is >= max allowable objects: %zu!~\n",
            objName_p, nMemObjRegistered, (size_t)MAX_MEMOBJ_CANDIDATES);           

    if(nElem == 0 || sizeOfOne == 0)
        memObjPrivateMsg("@F memObjCalloc: Variable(%s) with FATAL nElem (%zu) or Invalid sizeOfOne %zu!\n",
            objName_p, nElem, sizeOfOne);

     if((foundDuplicate = memObjDuplicateTest(objName_p)) == 1)
        exit(-1); // FATAL error on duplicates until further analyzed

    // just identify next available slot (so pass in NULL)
    if((iSlot = memObjGetNextSlot() ) > 0)
    {
            if( vDtype == IsStringTable) 
            {
                strTblItemwidth = sizeOfOne ;
                strTbl_p = memCreateStringTbl(nElem,strTblItemwidth) ;
                sprintf(addrLbl,"%p",strTbl_p);
                new_p = strTbl_p ; // BIND it
                // end:: if vDtype is IsStringTable
            } else {
                // Section here allocates a scalar or 1-level array
                if((new_p= calloc(nElem,sizeOfOne+1))==NULL)
                {
                    memObjPrivateMsg("@F memObjCalloc: FATAL-ERROR allocating(%s) of size(%zu) Nelem %zu\n",
                        objName_p, sizeOfOne,nElem);
                    exit(-1);
                } 
                // added 1/20/2014, as default unique identifier
                sprintf(addrLbl,"%p",new_p);
            } // end:: block if a scalar|array, and is NOT a IsStringTable

            nMemObjRegistered++ ; // increment count of objects successfully registered
            // set tracking pointer to address of this heap memory variable
            memObjRegistry[iSlot].vContents_p  = new_p; // BIND back to master pointer!
            memObjRegistry[iSlot].vNelem     = nElem ;
            memObjRegistry[iSlot].vSizeOfOne = sizeOfOne ; // size of 1 scalar instance
            memObjRegistry[iSlot].vBytes     = bytesRequested;
            // New 1/20/2014, added default unique identifier 
            strcpy(memObjRegistry[iSlot].vDnaLabel, addrLbl);
            // store the code matching the scalar data type this is
            // IsBoolean,IsInteger,IsReal,IsString,IsFilespec, or, IsUint32, IsFloat64, IsCustomUserType, etc
            memObjRegistry[iSlot].vDatatype  = vDtype ;
            // update global facility-wide count of total bytes tracked by memObj
            nMemObjBytesTracked += bytesRequested;

            // update highest slot referenced
            if( (size_t)iSlot > highestSlotReferenced)
              highestSlotReferenced = iSlot ;

#ifdef L4C_DIAG_MEM
            memObjPrivateMsg("@A memObjCallocByName: iSlot(%3d) registered <%-32s> allocated %zu bytes by <%s> HWM %zu nReg %zu\n",
                iSlot, memObjRegistry[iSlot].vName, bytesRequested, callerLbl, 
                nMemObjBytesTracked,nMemObjRegistered);         
#endif              
            // if present, stores the callerLbl in the vMetadata entry for this object
            if(strlen(callerLbl)>0)
                strcpy(memObjRegistry[iSlot].vMetadata,callerLbl);

    } else { // a valid unoccupied slot was found
            memObjPrivateMsg("@E memObjCallocByName: NO SLOT available!\n");
    }

    // adjust session wise high water mark
    if(nMemObjBytesTracked > nMemHighWaterMark)
        nMemHighWaterMark = nMemObjBytesTracked ;

    hwmIndex = iSlot ; // stores base0 index of slot found  
  // as a convenience, return pointer to new allocated memory    
 return new_p ; 
} // end::memCallocByName()



// --------------------------------------------------------------------
// function : memCalloc()
// purpose  : using only the calloc() args, register a new scalar 
//            or array memory object, and allocate it using calloc().
//            But memCalloc() does NOT support creation of IsStringTable
//            compound datatypes! See note below.
// inputs   :const size_t nElem,size_t sizeOfOne
// returns  : void * ptr to new memory allocated
// revised  : added memVerboseMsg {0|1} diagnostics
// notes    : Strictly observe call signature order: calloc(nElem,sizeOfOne)
// notes    : WARNING! Do NOT use this for creating a IsStringTable type.
//            Instead, use memCreateStringTbl() if no memObjSetName() is 
//            required, OR, if a name is needed, use this as below:
// memObjCalloc(TblName,MaxWidth,nStrings,IsStringTable,__func__,callerLbl)
//
void *memCalloc(const size_t nElem,size_t sizeOfOne)
{
 void *new_p = NULL;    // proxy for 1-level allocs, for scalars, or simple arrays
 char addrLbl[MAX_MEM_DNA_LEN] = {0} ;
 char fxName[256] ;
 int  iSlot =0;               // slot index into internal table of memory objects
 // int  ithStr =0;  // index for IsStringTable allocation
 int  hwmIndex = MAX_MEMOBJ_CANDIDATES; // high-water-mark base0 index
 size_t bytesRequested = (sizeOfOne * nElem); // derive n. bytes as product of these
 
 
    strcpy(fxName,__func__);
    
    // STOP, if facility is not yet initialized, or insufficient slots left!
    if(memObjFacilityInitialized != 1)
        memObjPrivateMsg("@F memCalloc: %s is NOT currently initialized!~");

    if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
        memObjPrivateMsg("@F memCalloc MaxSlotViolation %zu is >= max allowable objects: %zu!~\n",
            nMemObjRegistered, (size_t)MAX_MEMOBJ_CANDIDATES);          

    if(nElem == 0 || sizeOfOne == 0)
        memObjPrivateMsg("@F memCalloc: FATAL nElem (%zu) or Invalid sizeOfOne %zu!\n",
         nElem, sizeOfOne);

    // Determine base0 slot index of NEXT UNOCCUPIED slot starting at 0
    if((iSlot = memObjGetNextSlot() ) >= 0)
    {
        // BLOCK for allocating a scalar or 1-level array
        // remember call signature: calloc(nMembers,sizeofOne) 
        if((new_p= calloc(nElem, sizeOfOne+1)) == NULL)
        {
            memObjPrivateMsg("@F memCalloc: FATAL-ERROR allocating(%s) of size(%zu) Nelem %zu\n",
                sizeOfOne,nElem);
            exit(-1);
        } 
        // <JMG> added 1/20/2014, store memory address (string) as unique identifier
        sprintf(addrLbl,"%p",(uint8 *)new_p);
        // <JMG> 2014-01-20, added DNA as default unique identifier 
        strcpy(memObjRegistry[iSlot].vDnaLabel, addrLbl);
        
        nMemObjRegistered++ ; // increment count of objects successfully registered
        
        // <JMG> 2014-07-30T02:53:00, added: if verbose, echo out right away
        if(memVerboseMsg == 1)
        {
            fprintf(stderr,"@A In(%s) lastest-dna(%s) slot(%d) n(%zu) siz(%zu) %zu obj-registered\n",
                fxName, addrLbl, iSlot, nElem,sizeOfOne, nMemObjRegistered);
            fflush(stderr);
        }

        // set tracking pointer to address of this heap memory variable
        memObjRegistry[iSlot].vContents_p  = new_p; // BIND back to master pointer!
        memObjRegistry[iSlot].vNelem     = nElem ;
        memObjRegistry[iSlot].vSizeOfOne = sizeOfOne ; // size of 1 scalar instance
        memObjRegistry[iSlot].vBytes     = bytesRequested;
        // zero out regular name in this use-context...until further modified...
        strcpy(memObjRegistry[iSlot].vName,"");
        
        // Datatype is essentially a IsNoType, until overriden with 
        //   a valid TIV datatype like IsFloat32 etc
        memObjRegistry[iSlot].vDatatype  = IsNoType; // unspecified for now;
        
        // udpate the count of all bytes tracked to add in this new assignment
        nMemObjBytesTracked += bytesRequested;

        // update highest slot referenced
        if( (size_t)iSlot > highestSlotReferenced)
            highestSlotReferenced = iSlot ;
      
    } else { // a valid unoccupied slot was found
            memObjPrivateMsg("@E memCalloc: NO SLOT available! highest Slot now already is %d\n",
                    highestSlotReferenced);
    }

    // adjust session wise high water mark
    if(nMemObjBytesTracked > nMemHighWaterMark)
        nMemHighWaterMark = nMemObjBytesTracked ;

    hwmIndex = iSlot ; // stores base0 index of slot found  
    
  // as a convenience, return pointer to new allocated memory    
 return new_p ; 
} // end::memCalloc()

// ---------------------------------------------------------
// function  : memCallocWithName()
// purpose   : Implements a combination of memCalloc() call with immediate
//             memObjSetName()
// revised   : 2015-08-03t12:18
// inputs    : const size_t nElem,
//             size_t sizeOfOne,
//             const char *memObjName_p
// returns   : 
// notes     : This is the corrollary to memObjSetMetadata()
void *memCallocWithName(const size_t nElem,size_t sizeOfOne,const char *memObjName_p)
{
  void *new_p = NULL;    // proxy for 1-level allocs, for scalars, or simple arrays
  char fxName[256] = {0} ;
  
  strcpy(fxName,__func__);
  
  new_p = memCalloc(nElem,sizeOfOne);
  
  memObjSetName(new_p,memObjName_p);

  return new_p ;
  
} // end::memCallocWithName()


// --------------------------------------------------------------------
// function : memObjFree() 
// inputs   : objName -- string name of heap memory variable managed in memObj
// revised  : 2014-08-14T14:11:00, implement relaxed name-matching, to 
//            accept EITHER the user-supplied name, OR the objects DNA label.
//            The prior 'DoFreeAll' directive is deprecated since it was 
//            replaced by memObjFreeAll()
// purpose  : using an objects name to identify it, free up heap memory
//            variable occupying a slot managed by the memObj facility.
//
// Invocation:   memObjFree("nvpIndex",NULL);  ...typical call use-case, 
//                  frees variable named 'nvpIndex',
//   or user can call memObjFree() like:  memObjFree("nvpIndex",nvpIndex); ..used when for clarity 
//   we wish to explicitly pass a local pointer "myPointer", where myPointer internally matches 
//   the contents already referred in the slot table as memObjRegistry[islot].vContents_p
//
// Example: (common usage)  memObjFree(memObjQualify(fxName,"tile.fpar",bufr), arySizt_p);
//   in this example above, the memObjQualify() helper function is used to fully qualify
//   the simple name "tile.fpar", resulting in a dot-qualified name like:
//   "{functionName}.tile.fpar"
//
// Warning: This routine is now configured to free single level pointers 
//   and a specific type of multi-level pointer (IsStringTable).
//   It will NOT work for more elaborately structured multi-level pointer
//   arrangements, employing deeply nested pointer geometry, 
//   like **myPtr, or ***myPtr, or ****myPtr such as those often used
//   for user defined or custom multi-dimensional arrays.
//
// Notes: nMemObjRegistered maintains a session wise global count of the number of
//   discrete memory variables tracked here by the memObj facility.

void memObjFree(char *objName_p,void *mem_p)
{
 int iSlot=0 ;
 int s = 0 ;
 int didMatch = 0 ;
 int doFreeThis = 0;
 int vDtype = -1 ;
 char *vThisName_p = NULL;
 char *vDnaName_p  = NULL; // <JMG> 2014-08-14, implements relaxed matching
 char **strTbl_p = NULL; // mechanism used to free IsStringTable's
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue

  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 

  // block any progress beyond here if facility is not yet initialized
  if(memObjFacilityInitialized != 1)
  {
    memObjPrivateMsg("memObjFree", "@F %s is NOT currently initialized!","memObjFacility");
  }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memObjFree", "@F %zu exceeds maximum allowable registered mem obj %zu!\n",
           nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }
    
  // scan slot-table 
  for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
  {
    // set a convenience pointer
    vThisName_p =  memObjRegistry[iSlot].vName ;
    // retrieve a pointer to object's DNA as known in registry
    vDnaName_p  =  memObjRegistry[iSlot].vDnaLabel;
    
    // ORIGINAL matching logic REQUIRED object's user supplied name
    // if in single-object mode, try to match user name with a slot name..
    // doFreeThis = ( strcmp(objName_p,vThisName_p)==0) ? 1: 0 ;

    // <JMG> 2014-08-14T14:16:00, implements NEW relaxed
    //  matching rules, which allow user to pass in EITHER the
    //  user-supplied name OR the object's DNA, to support a match on 
    //  EITHER of these.
    doFreeThis = 0 ;
    
    if( strcmp(objName_p, vThisName_p)==0 || strcmp(objName_p, vDnaName_p )==0)
                doFreeThis = 1 ;

    // doFreeThis is a boolean that is set true if the name was found in the catalog of actively
    //    defined memory objects.
    if(doFreeThis)
    {
        // set flag True to indicate we did find this objName_p in dictionary of registered objects
        didMatch = 1 ; 
        // Note: we only attempt to free a variable if it has not already been freed
        if(memObjRegistry[iSlot].vNelem >0 && memObjRegistry[iSlot].vContents_p != NULL)
        {
                // consult datatype...to see if this is a compound object like IsStringTable
                vDtype = memObjRegistry[iSlot].vDatatype ;
                // verify validity of mem_p if specified
                if(mem_p != NULL)
                    assert(mem_p == memObjRegistry[iSlot].vContents_p);
                    
                if(vDtype != IsStringTable) 
                {
                    // ...simple vector accessible with one handle
                    free( memObjRegistry[iSlot].vContents_p );
                } else {
                    // ...compound (2-level) object, requires 2-level free
                    strTbl_p = memObjRegistry[iSlot].vContents_p;
                    // free the children first...(inner-to-outer..)
                    for(s=0; s < (int)memObjRegistry[iSlot].vNelem; s++)
                        free(strTbl_p[s] );
                    // free the parents
                    free(strTbl_p); 
                }
                nMemObjRegistered-- ;  // decrement long-lived count of occupied slots to FREE UP THIS SLOT             
                // explicitly set the state of vContents_p as NULL to reaffirm for next use
                memObjRegistry[iSlot].vContents_p = NULL;
                // clean up...
                memObjRegistry[iSlot].vNelem  =0 ;
                memObjRegistry[iSlot].vBytes  =0 ;
                memset(memObjRegistry[iSlot].vName,'\0',MAX_MEMOBJ_VNAME_LEN);          
                
                // action satisfied, so we break early..
                break;
                
        } // end:: if met all criteria to free
    } // end:: doFreeThis block...
    
  } //end: iSlot loop

    if(!didMatch)
    {
        memObjPrivateMsg("@E memObjFree: FAILED symbol lookup on name (%s) matching DNA(%s)! ~ \n",
            objName_p, vDnaName_p); 
  }
  
} // end::memObjFree()


// --------------------------------------------------------------------
// function : memFree()
// purpose  : free up an anonymous memory object, identified only via its vDnaLabel
// inputs   : void *mem_p -- ptr to memory previously allocated using memCalloc()
// returns  : void 
void memFree(void *mem_p)
{
 int iSlot=0 ;
 int hitSlot = -1;
 // char **strTbl_p = NULL; // mechanism used to free IsStringTable's
 char addrLbl[MAX_MEM_DNA_LEN] = {0} ;  
 
  // block any progress beyond here if facility is not yet initialized
  if(memObjFacilityInitialized != 1)
  {
    memObjPrivateMsg("memFree", "@F %s is NOT currently initialized!","memObjFacility");
  }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memFree", "@F %zu exceeds maximum allowable registered mem obj %zu!\n",
           nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }

 
  // Attempt to find the slot this object lives at, 
  // using its memory address to identify it
  hitSlot = memObjGetSlot(mem_p);
  sprintf(addrLbl,"%p",(uint8 *)mem_p);  
  // fprintf(stdout,"@A memFree DIAG DNA-addr(%s) at slot %d\n",addrLbl,hitSlot);
  
  if(hitSlot >= 0 && hitSlot < MAX_MEMOBJ_CANDIDATES)
  {
        iSlot = hitSlot ;
        free( memObjRegistry[hitSlot].vContents_p );
        // After an anonymous free, be sure to mark all necessary sentinels so that memFreeAll() or friends
        // do NOT try to re-free this!
        nMemObjRegistered-- ;  // decrement long-lived count of occupied slots to FREE UP THIS SLOT             
        // explicitly set the state of vContents_p as NULL to reaffirm for next use
        memObjRegistry[iSlot].vContents_p = NULL;
        // clean up...
        memObjRegistry[iSlot].vNelem  =0 ;
        memObjRegistry[iSlot].vBytes  =0 ;
        memset(memObjRegistry[iSlot].vName,'\0',MAX_MEMOBJ_VNAME_LEN);

  } else {
    fprintf(stderr,"@W memFree invalid journal slot(%d) for addr(%s) apparently NOT in a condition to free\n",
                    hitSlot, addrLbl);
    exit(-1);
  }
  
} // end::memFree()


// --------------------------------------------------------------------
// Purpose   :  free up ALL remaining heap memory variables occupying ANY slot 
//              managed by the memObj facility.
// Invocation:  memObjFreeAll();
// Returns   : count of objects whose memory was actually released.
// revised   : memObjFreeAll() extended to allow freeing IsStringTable objects
// Notes     : nMemObjRegistered maintains a session wise global count of the number of
//              discrete memory variables tracked here by the memObj facility.
// History   : 2013-10-23T15:32:00, added return, count of n. of objects whose memory was released.
int memObjFreeAll(void)
{
 int s=0, iSlot=0 ;
 int kObjectsReleased = 0 ; // count, returned to caller
 int vDtype = 0 ;
 size_t nElem = 0;
 size_t nBytes= 0;
 char *vDna_p = NULL;
 char *vName_p= NULL;
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
 char **vStrTbl_p = NULL; // ptr to string table for objects defined as such
 
  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 

  // block any progress beyond here if facility is not yet initialized
  if(memObjFacilityInitialized != 1)
  {
    memObjPrivateMsg("memObjFreeAll", "@F %s is NOT currently initialized!","memObjFacility");
  }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memObjFreeAll", "@F %zu exceeds maximum allowable registered mem obj %zu!\n",
           nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }

  // For debug context, first emit a DNA report. Actually, found out this
  //  doesn't really add much more to what we already report in the block below
  // memObjReportDna(memObjCurrLogf, "@A memObjFreeAll-pre-report~");

  // scan slot-table 
  for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
  {
        // Note: we only attempt to free a variable if it has not already been freed
        if(memObjRegistry[iSlot].vNelem > 0 && memObjRegistry[iSlot].vContents_p != NULL)
        {
            vDtype = memObjRegistry[iSlot].vDatatype ;
            vName_p= memObjRegistry[iSlot].vName ;
            vDna_p = memObjRegistry[iSlot].vDnaLabel;
            nElem  = memObjRegistry[iSlot].vNelem;
            nBytes = memObjRegistry[iSlot].vBytes;
            // check if a IsStringTable and handle specially for 2-level release
            if(vDtype != IsStringTable) 
            {
                    // ...simple vector accessible with one handle
                    if( memObjRegistry[iSlot].vContents_p != NULL)
                    {
                        if(memVerboseMsg == 1)
                        {
                            fprintf(stderr,"@A In(%s) Freeing(%s)[%s] nElem %zu bytes: %zu ~ \n",
                                fxLabel, vName_p, vDna_p, nElem, nBytes);
                            fflush(stderr);
                        }
                        free( memObjRegistry[iSlot].vContents_p );
                    }
            } else {
                // here we interpret a (void *) to point to start of (**vStrTbl_p)              
                // ...compound (2-level) object, requires 2-level free
                vStrTbl_p = (char **)memObjRegistry[iSlot].vContents_p;
                // free the children first...(inner-to-outer..)
                for(s=0; s < (int)memObjRegistry[iSlot].vNelem; s++)
                    if(vStrTbl_p[s] !=NULL )
                    {
                        if(memVerboseMsg == 1)
                        {
                            fprintf(stderr,"@A In(%s) Freeing (%s)[%s] stringtbl nElem %zu bytes: %zu ~ \n",
                                fxLabel, vName_p, vDna_p, nElem, nBytes);
                            fflush(stderr);
                        }
                        free( vStrTbl_p[s] );
                    }
                    
                // free the parents, if not NULL...
                if(vStrTbl_p != NULL)
                {
                        if(memVerboseMsg == 1)
                        {
                            fprintf(stderr,"@A In(%s) Freeing (%s)[%s] stringtbl-parent nElem %zu bytes: %zu ~ \n",
                                fxLabel, vName_p, vDna_p, nElem, nBytes);
                            fflush(stderr);
                        }
                        free(vStrTbl_p); 
                } // end:: vStrTbl_p block was not null
            } // end: else block if a stringtable
            
            // REPORT OUT IDENTIFYING INFO on WHICH objects (presumed orphans) were freed
            memObjPrivateMsg("@A memObjFreeAll: released orphan name(%s) dna(%s) Type(%d) nElem(%zu)bytes(%zu)\n",
                vName_p, vDna_p, vDtype, nElem, nBytes);
                
            // can now clear out contents to mark as un-used
            memset(memObjRegistry[iSlot].vName,'\0',MAX_MEMOBJ_VNAME_LEN);
            memObjRegistry[iSlot].vNelem  =0 ;
            memObjRegistry[iSlot].vBytes  =0 ;
            kObjectsReleased++ ; // increment count of n. objects actually released
            // explicitly set the state of vContents_p as NULL to reaffirm for next use
            memObjRegistry[iSlot].vContents_p = NULL;
            // mem obj facility now allows re-cycling of slots...
            nMemObjRegistered-- ;  // decrement long-lived count of occupied slots to FREE UP THIS SLOT
        } // end:: if met all criteria to free
        
  } //end: iSlot loop
  
    // return count of n. of objects whose memory was actually released
    return kObjectsReleased ;
    
} // end::memObjFreeAll()


// --------------------------------------------------------------------
// Function : memObjGarbageCollect()
// Purpose: at end of session, scan table and  perform a
//   forced release of any remaining slots! This eliminates sessions
//   with un-released heap memory managed by memObj from occurring.
void memObjGarbageCollect(void)
{

 int nCollected = 0 ;
 
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
 static int panicSet = 0 ; 
  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 

    // see if we've already been called prior...
    if(panicSet)
        return  ;
  
    // block any progress beyond here if facility is not yet initialized
    assert(memObjFacilityInitialized == 1);
  
    // Unilaterally free all remaining objects and retrieve count of
    // those released any objects released at this point were almost 
    // certainly orphans!!
    nCollected = memObjFreeAll();

    if(nCollected ==0)
    {
        memObjPrivateMsg("@A memObjGarbageCollect: CLEAN-SUCCESS (no orphans detected)~\n");
        // end::nCollected is 0 block
    } else { 
        memObjPrivateMsg("@A memObjGarbageCollect: CLEAN-WITH-ORPHANS released %d ~\n",nCollected);
    }

    memObjPrivateMsg("@A memObjGarbageCollect: collection successful~\n");

 // set flag to indicate we have already been called
 panicSet = 1 ; 
 // fx is now 'void' for compatibility with registration to atexit()
 
} // end:: memObjGarbageCollect()


// --------------------------------------------------------------------
// function : memObjGet()
// purpose  : retrieve pointer to data mapped to MemObj, by "standard" vName
// inputs   : const char *objName_p -- standard vName
// returns  : void *referent_p -- ptr to raw data
// notes    : n/a
void *memObjGet(const char *objName_p)
{
 int  iSlot =0 ;
 // int  foundSlot = -1;
 void *referent_p = NULL;
 char addrLbl[MAX_MEM_DNA_LEN] = {0} ;
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 
 
    // fetch base0 slot mapped to this name. If found, modify ptr to return
    sprintf(addrLbl,"%p",(uint8 *)memObjRegistry[iSlot].vContents_p);
    
    // Perform linear scan on registry to locate slot where name matches
    for(iSlot =0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
        if(strcmp(objName_p, memObjRegistry[iSlot].vName) ==0 )
        {
            // As a double check, make sure its internally assigned 
            //  DNA identify also matches
            if(iSlot == memObjGetSlot(addrLbl) )
            {
                referent_p = memObjRegistry[iSlot].vContents_p ;
                break;
            }
        }
    } // end:: iSlot loop
    
 // Return pointer to raw data mapped to named memory object,
 //  or NULL if not found   
 return referent_p ;    
    
} // end::memObjGet()


// --------------------------------------------------------------------
// function : memObjGetBySlot()
// inputs   : const int ithSlot --base0 index >= 0 and < MAX_MEMOBJ_CANDIDATES
// returns  : ptr to valid memory or NULL if match on this slot not found
// purpose  : retrieve a pointer to a previously registered memory object,
//            referred to by base0 slot index. Be careful with this one!
// notes    : Be careful with this one, what you get may or may not be
//            what you really wanted, unless you independently track
//            the slot indices offline
// notes    : If an obviously bogus slot index is passed in, this will
//            intentionally trigger a fatal abort via the assert call.
void *memObjGetBySlot(const int ithSlot)
{
 void *existing_p =NULL;
 
    // First, perform minimal bounds check
    assert(ithSlot >=0 && ithSlot < MAX_MEMOBJ_CANDIDATES);
    
    // Fetch base0 slot mapped to this name. If found, modify ptr to return
    if(memObjRegistry[ithSlot].vContents_p != NULL)
            existing_p = memObjRegistry[ithSlot].vContents_p ;

 // return pointer to requested memory object, or NULL if not found 
 return existing_p ;    
    
} // end::memObjGetBySlot()


// --------------------------------------------------------------------
// Purpose: retrieve the count of vNelem elements (of a given scalar type)
//   this registered memory object possesses 
size_t memObjGetCount(const char *objName_p)
{
 size_t iSlot =0 ;
 size_t elemAllocated =0 ; 
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 
 
    for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
        if(strcmp(objName_p,memObjRegistry[iSlot].vName)== STRING_MATCHED && memObjRegistry[iSlot].vNelem > 0)
        {
        elemAllocated = memObjRegistry[iSlot].vNelem;
        break;
        } // end:: if slot is selected
    } // end:: iSlot loop
 // return count of n. of scalar elements (not bytes) mapped to this memory object  
 return elemAllocated ;  
} // end::memObjGetCount() 



// --------------------------------------------------------------------
// Purpose: zero out the memory for this memory object, retrieved by name
void memObjZeroOut(const char *objName_p)
{
 size_t iSlot  = 0 ;
 size_t nBytes = 0 ; 
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 

    for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES; iSlot++)
    {
        if(strcmp(objName_p,memObjRegistry[iSlot].vName)== STRING_MATCHED && memObjRegistry[iSlot].vNelem > 0)
        {
            nBytes = memObjRegistry[iSlot].vSizeOfOne * memObjRegistry[iSlot].vNelem;
        memset((unsigned char *)memObjRegistry[iSlot].vNelem, NULL_CH, nBytes);
        break;
        } // end:: if slot is selected
    } // end:: iSlot loop
    
} // end::memObjZeroOut() 


// --------------------------------------------------------------------
// Purpose: count and return the number of memory object slots currently 
int memObjSlotsUsed(void)
{
 int iSlot =0 ;
 int nSlotsUsed = 0 ;
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue

  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 
        
  assert(nMemObjRegistered < MAX_MEMOBJ_CANDIDATES);
  // scan slot-table only as far as we have to...
  for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES ; iSlot++)
    if( strlen(memObjRegistry[iSlot].vName)>0 &&
        memObjRegistry[iSlot].vNelem >0   && 
        memObjRegistry[iSlot].vContents_p != NULL)
            nSlotsUsed++ ;
 // return count of no. of available slots actually in use  
 return nSlotsUsed ;
} // end::memObjSlotsUsed()


// --------------------------------------------------------------------
// Purpose: retrieve various session runtime memObj metrics
size_t memObjGetMetrics(size_t *highestSlot,size_t *totalBytesTracked,size_t *hwmMem)
{
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue

  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 

  // nMemObjBytesTracked counts the number of bytes under management
  *totalBytesTracked = nMemObjBytesTracked ;
  // index of highest numbered slot used
  *highestSlot = highestSlotReferenced ;
  // high water mark  
  *hwmMem = nMemHighWaterMark ;
  
  // return current total of variables registered
  return    nMemObjRegistered ;
} // end::memObjGetMetrics()


// --------------------------------------------------------------------
// function: memObjReportDna()
// Purpose : Using each referent's DNA as a registry check,
//           spin through entire registry and display a summary 
//           of memory symbol table, routed to userRptf if not NULL,
//           or to stdout if user did not specify an open stream.
// Inputs  : FILE *userRptf,const char *reportTitle
// Returns : int, index of last slot accessed.
// Notes   : 7/30/2014, now tests on pos count of vNelem and vBytes
//           
int memObjReportDna(FILE *userRptf,const char *reportTitle)
{
 int iSlot =0;
 
  fprintf(userRptf,"\n@A memObjReportDna report [%s]\n",reportTitle);
 
   for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES ; iSlot++)
  {
    if( memObjRegistry[iSlot].vNelem >0 && 
        memObjRegistry[iSlot].vBytes >0 &&
        strlen(memObjRegistry[iSlot].vDnaLabel)>0)  
        {
            fprintf(userRptf,"MemObj slot(%d) name(%s) dna(%s) type(%d) nElem %zu bytes %zu\n",
                iSlot, 
                memObjRegistry[iSlot].vName, 
                memObjRegistry[iSlot].vDnaLabel,
                memObjRegistry[iSlot].vDatatype, 
                memObjRegistry[iSlot].vNelem,
                memObjRegistry[iSlot].vBytes);
        }
  } // end:: iSlot loop 
  
 // returns base0 index of last slot visited
 return iSlot ;
   
} // end:: memObjReportDna()


// --------------------------------------------------------------------
// function: memObjReport()
// Purpose : Display summary of memory symbol table, routed to 
//           userRptf if not NULL
//           or to stdout if user did not specify an open stream.
// Inputs  : FILE *userRptf,const char *reportTitle
// Returns : int, index of last slot accessed.
// Notes   : n/a
int memObjReport(FILE *userRptf,const char *reportTitle)
{
 size_t iSlot=0;
 FILE *rptf = stderr ;
 char fxName[256] = {0} ;
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxName,__func__);

  // block any progress beyond here if facility is not yet initialized
  assert(memObjFacilityInitialized == 1);
  assert(nMemObjRegistered < MAX_MEMOBJ_CANDIDATES);

  // if rptf is not null, stream results to  it..otherwise,
  // just route to stdout or stderr etc
  if(userRptf == NULL)
    rptf = stdout;
  else
    rptf = userRptf ;

  memObjPrivateMsg("@A Report(%s)~\n",reportTitle);
  memObjPrivateMsg("@A %zu bytes registered across %zu slots\n(HWM %zu) High-slot(%zu) of %u~\n",
    nMemObjBytesTracked, nMemObjRegistered, nMemHighWaterMark, highestSlotReferenced,MAX_MEMOBJ_CANDIDATES);

  fprintf(rptf,"@A memObjReport title(%s)~\n",reportTitle);
  fprintf(rptf,"@A memObjReport %zu bytes registered across %zu slots\n(HWM %zu) High-slot(%zu) of %u~\n",
    nMemObjBytesTracked, nMemObjRegistered, nMemHighWaterMark, highestSlotReferenced,MAX_MEMOBJ_CANDIDATES);
    
  // scan slot-table only as far as we have to...
  for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES ; iSlot++)
  {
    if(memObjRegistry[iSlot].vNelem >0 &&
        strlen(memObjRegistry[iSlot].vName)>0)
    {
        fprintf(rptf,
          "@A memObjReport slot: %4zu name: %32s  datatype %2d nElem( %12zu) nBytes[ %12zu ]~\n",
           iSlot,
           memObjRegistry[iSlot].vName,
           memObjRegistry[iSlot].vDatatype,
           memObjRegistry[iSlot].vNelem,
           memObjRegistry[iSlot].vBytes);
    }
  } // end:: iSlot loop
  
 // returns count of slots reported
 return iSlot; 
} // end::memObjReport()


// --------------------------------------------------------------------
// function : memObjWrite()
// inputs   : const char *objName_p,
//            const char *userPath,
//            const char *allOption
// returns  : int, tbd.
// Purpose: memObjWrite() writes either a single heap memory object
//  to disk, or if the allOption string is non-null ("A"|"All"|"ALL")
//  then it writes
//  EVERY heap memory object defined in the memObj facility to disk files
//  named matching the variable. Warning: This routine will work for any 
//  array declared
//  using a single pointer, so will NOT work for nested pointer arrangements
//  such as those often used for multi-dimensional arrays. 
int memObjWrite(const char *objName_p,const char *userPath,const char *allOption)
{
 FILE *memf = NULL;
 char memFilepath[FILENAME_MAX]= {0} ;
 unsigned char *data_p = NULL;
 size_t nTarg = 0;
 size_t ioStatus =0 ;
 int doWriteAll =0 ;
 int doWriteThis=0 ;
 int nWritten =0 ;
 int iSlot =0;

 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue

  // ------------------------------------------------------------------
  strcpy(moduLabel,(char *)__FILE__);
  strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 

  // block any progress beyond here if facility is not yet initialized
  assert(memObjFacilityInitialized == 1);

  // establish file path to write binary data to
  if(userPath != NULL && strlen(userPath)>0)
  {
    strcpy(memFilepath,userPath);
  }else{
    strcpy(memFilepath,objName_p); 
    strcat(memFilepath,".mem");
  }
  
  // If the WRITE ALL option argument contained any string
  // set flag to write every object. We don't care what the string contents
  // are in this case, so most anything is ok like: {"A"|"a"|"ALL"|"All"|"1"|"junk"}, etc
  if(allOption != NULL && strlen(allOption)>=1)
    doWriteAll = 1 ;
    
    // iterate through current list of candidate memory objects
    for(iSlot=0; iSlot < MAX_MEMOBJ_CANDIDATES ; iSlot++)
    {
        // first eliminate (skip over) any slots with invalid or corrupt info
        if( memObjRegistry[iSlot].vNelem < 1 ||
            strlen(memObjRegistry[iSlot].vName) < 1  ||
            memObjRegistry[iSlot].vContents_p == NULL)
        {
        continue;
        }

        if(doWriteAll) { 
             // unilaterally enable write for every object, but only those with valid contents!
            doWriteThis =1;
            // when using WRITE-ALL mode, override path to use objects name
            strcpy(memFilepath,memObjRegistry[iSlot].vName);
            strcat(memFilepath,".mem");
        }else {
            // allow for normal single-variable criteria to match just one item 
            doWriteThis =(strcmp(memObjRegistry[iSlot].vName,objName_p)==0) ? 1 : 0;
        } 
        // apply selection critiera to current slot     
        if(doWriteThis)
        {
            nTarg = memObjRegistry[iSlot].vBytes ;
            data_p=(unsigned char *)memObjRegistry[iSlot].vContents_p;

            if((memf=fopen(memFilepath,"wb"))== NULL)
            {
                fprintf(stderr,
                    "@F failed to open memory write file %s for output~\n",memFilepath);
                exit(-1);
            }

            // fprintf(stdout,"memObjWrite: writing %zu bytes from var<%s> to file<%s>\n",
            //  memObjRegistry[iSlot].vBytes, objName_p, memFilepath);
            if((ioStatus=fwrite(data_p,sizeof(unsigned char),nTarg,memf))>= nTarg)
            {
                    nWritten++ ; // increment count of variables written to disk
            } else {
                 fprintf(stderr,"@F memObjWrite: error! only %zu of %zu elements written! ~\n",
                    ioStatus,nTarg);
                    exit(-1);
            } // end:: if block where write failed
            fclose(memf);

            // if a single-variable write, always exit after the match event
            if(doWriteAll == 0) 
                break;

        } // end:: if block on match of current heap memory variable name
    } //end:: iSlot loop
    
 // return count of heap variables written to disk files    
 return nWritten;   
} // end::memObjWrite()

// --------------------------------------------------------------------
// function : memCreateStringTable()
// purpose  : Allocate memory assoc. with a "named" 2-level string table
// inputs   : const int nMembers  -- count of substrings in table
//            const int strWidth  -- max width of all substrings in table
//            const char *strTblName_p --ptr to string "name" of entire table
// returns  : char **strTbl_p -- ptr to fully allocated, 2-level table
// author   : joe glassy
// revision : 2016-01-07T10:33, added ability to "name" entire string table
//            when it is created, such that now these stringtables can
//            be released via memFreeStringTbl() like any other memCalloc()
//            managed (journaled) object. Will eventually supercede all 
//            calls to legacy memCreateStringTbl()
// history  : 2014-02-23T09:53:00, documentation added only.
// history  : 2013-10-31
// limitations : string tables created here IS included in the
//             : in the memObj registry and are free'd using memFreeStringTbl()
char **memCreateStringTable(const int nMembers,const int strWidth,const char *strTblName_p)
{
 int i=0;
 char **strTbl_p = NULL;

    // Create outer-shell list of "parent" pointers, now using memCalloc() to allow
    // table itself to be properly registered and journaled etc.
    strTbl_p = (char **)memCalloc(nMembers,sizeof(char *));

    // if strTblName_p valid, assign it. Note length is now checked in
    // memObjSetName() so that if it is too long, error is immediately trapped 
    if(strTblName_p != NULL && strlen(strTblName_p) >= 1)
        memObjSetName(strTbl_p,strTblName_p);
    else
        memObjSetName(strTbl_p,"StringTablePlaceholderName-NotYetAssigned");
    
    // Create list of individual "child" pointers owned by parents
    for(i=0; i < nMembers; i++)
    {       // allocate individual string, adding 1 for null term
            strTbl_p[i] =(char *)memCalloc(strWidth+1,sizeof(char));
    }

 // return pointer to new allocated (named) string table of nMembers, each strWidth wide
 return strTbl_p;

} // end::memCreateStringTable()

// --------------------------------------------------------------------
// function : memReleaseStringTable()
// purpose  : Release memory assoc. with a 2-level string table
// inputs   : char **strTbl_p, const int nMembers
// returns  : void
// author   : joe glassy
// revision : 2016-01-07, added
// limitations : string tables created here are NOT included in the
//             : in the memObj registry. They are intended for internal
//               uses in narrowly defined instances.
// notes    : This function should be used to release memory created using
//            memCreateStringTable() since (memFree() performs ALL 
//            proper clean up of registered, journaled objects
// notes    : When memCreateStringTable() replaces memCreateStringTbl()
//            we plan to rename legacy non-journaled instances to
//            memCreateStringTblLite() and memFreeStringTblLite()
//            these legacy "lite" instances require a lot less memory 
//            and may still be useful in some applications.
void memReleaseStringTable(char **strTbl_p,const int nMembers)
{
 int i=0;

    // first, use full strength memFree() to release list of 
    // registered, individual "child" pointers owned by parents
    for(i=0; i < nMembers; i++)
            memFree(strTbl_p[i] );
    
    // finally, release array of pointers (parent)
    memFree(strTbl_p);
  
} // end::memReleaseStringTable()


// --------------------------------------------------------------------
// function : memCreateStringTbl()
// purpose  : allocate memory assoc. with a 2-level string table
// inputs   : const int nMembers,const int strWidth
// returns  : char **strTbl_p -- ptr to fully allocated, 2-level table
// author   : joe glassy
// revision : 2014-02-23T09:53:00, documentation added only.
// revision : 2013-10-31
// limitations : string tables created here are NOT included in the
//             : in the memObj registry. They are intended for internal
//               uses in narrowly defined instances. This function is
//             superceded by memCreateStringTable() which accepts a name,
//             is registered, and is also released using memFreeStringTbl()
// notes    : TBD, as soon as possible, revise this to use a abbreviated
//            "DNA based" memCalloc(),memFree() style workflow, as many
//          : string tables do NOT need a 'name' and all the rest of 
//            heavy mechanism required by the memObjCalloc() + memObjFree()
//            workflow.
//
char **memCreateStringTbl(const int nMembers,const int strWidth)
{
 int i=0;
 char **strTbl_p = NULL;

        // Create outer-shell list of "parent" pointers
        strTbl_p=(char **)calloc(nMembers,sizeof(char *));
        // Create list of individual "child" pointers owned by parents
        for(i=0; i < nMembers; i++)
        {       // allocate individual string, adding 1 for null term
                strTbl_p[i] =(char *)calloc(strWidth+1,sizeof(char));
        }
        
        // TBD: investigate a way to map this anonymous stringtable to
        // a registered slot in the memObj facility after the fact

 // return pointer to new allocated string table of nMembers, each strWidth wide
 return strTbl_p;
} // end::memCreateStringTbl()


// ---------------------------------------------------------------
// function: memFreeStringTbl()
// purpose : release memory assoc. with a "lite" legacy 2-level string table
// inputs  : char **strTbl_p,const int nMembers
// author  : joe glassy
// returns : void
// historyn: 2013-10-31
// notes    : When memCreateStringTable() replaces memCreateStringTbl()
//            we plan to rename legacy non-journaled instances to
//            memCreateStringTblLite() and memFreeStringTblLite()
//            these legacy "lite" instances require a lot less memory 
//            and may still be useful in some applications.
//
void memFreeStringTbl(char **strTbl_p,const int nMembers)
{
 int i=0;
    // free the children
    for(i=0; i < nMembers; i++)
        if(strTbl_p[i] != NULL) 
               free(strTbl_p[i]);
                
    // free the parents
    if(strTbl_p != NULL)
        free(strTbl_p);
        
} // end::memFreeStringTbl()


// --------------------------------------------------------------------
// memObjQualify()
// Inputs:
//   fxNameStr -- a unique namespace, usually the c99 supplied __func__ function name
//   memObjname-- string identifying the memory variable itself
//   buffer    -- an externally supplied buffer to compose the fully qualified name into
//
// Purpose: fully qualify a memObj name by prepending a namespace identifier
//  which is typically supplied via the c99 symbol, __func__, that names the
//  current function name the memory variable was defined within.
// Invocation example: memObjQualifyName(__func__,"myVarName,buffer);
char *memObjQualify(char *fxNameStr,const char *memObjNameStr,char *buffer)
{
    strcpy(buffer,fxNameStr);
    strcat(buffer,".");
    strcat(buffer,memObjNameStr);
    return buffer;
} // end::memObjQualify()

// ---------------------------------------------------------
// function  : memObjGetNelem()
// purpose   : Retrieves a user defined nElem count for a 
//           : pre-existing MemObj memory object, using native ptr
//           : the search key in the registry
// revised   : 2014-02-23T10:06:00
// inputs    : void *mem_p -- ptr to memory region assigned
// returns   : size_t, vNelem --n. of data elements owned by MemObj
// notes     : There is NO corrollary, memObjSetNelem since this is
//             performed ONLY by a constructor.
size_t memObjGetNelem(void *mem_p)
{
    int iSlot=0 ;
    int hitSlot = -1;
    size_t resultNelem = 0 ;
    char addrLbl[MAX_MEM_DNA_LEN] = {0} ;  
 
  // block any progress beyond here if facility is not yet initialized
  if(memObjFacilityInitialized != 1)
  {
    memObjPrivateMsg("memObjGetNelem",
     "@F memObjFacility is NOT currently initialized!~");
  }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memObjGetNelem",
     "@F %zu exceeds maximum allowable registered mem obj %zu!\n",
     nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }
  // Re-constitute its DNA (unique id) in addrLbl, for labelling only
  sprintf(addrLbl,"%p",(uint8 *)mem_p); 

    // using objects pointer itself, find its base0 index in the registry
    hitSlot = memObjGetSlot(mem_p);
 
    // if object WAS found , fetch and return a pointer to its name
    if(hitSlot >= 0 && hitSlot < MAX_MEMOBJ_CANDIDATES)
    {
        iSlot = hitSlot ;
        resultNelem = memObjRegistry[hitSlot].vNelem;
    } else {
    fprintf(stderr,
    "@E memGetNelem FATAL: Invalid memory journal slot (%d) for existing obj at addr(%s)!\n",
            hitSlot, addrLbl);
    }
 // returns nElem count of items in memObj, or 0 if undefined
 return resultNelem;
 
} // end:: memObjGetNelem()


// ---------------------------------------------------------
// function  : memObjGetName()
// purpose   : Retrieve the user defined label for a memory object,
//             given just a pointer to its memory region.
// revised   : 2014-02-16T08:28:00-07:00
// author    : joe glassy
// inputs    : void *mem_p -- ptr to memory region assigned
// returns   : char *vName_p --ptr to this memory objects text name
// notes     : This is the corrollary to memObjSetName()
// notes     : Keep in mind that not all memObj have names, e.g.
//             anonymous memObj via DNA, etc.
char *memObjGetName(void *mem_p)
{
    int iSlot=0 ;
    int hitSlot = -1;
    char *vLabel_p = NULL; // ptr to memObj's internal label or name

 char addrLbl[MAX_MEM_DNA_LEN] = {0} ;  
 
  // block any progress beyond here if facility is not yet initialized
  if(memObjFacilityInitialized != 1)
  {
    memObjPrivateMsg("memObjGetName", "@F %s is NOT currently initialized!","memObjFacility");
  }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memObjGetName", "@F %zu exceeds maximum allowable registered mem obj %zu!\n",
           nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }

  // Re-constitute its DNA (unique id) in addrLbl, the look for it
  // in the registry
  hitSlot = memObjGetSlot(mem_p);
  sprintf(addrLbl,"%p",(uint8 *)mem_p); 
 
  // if it was found , fetch and return a pointer to its name
  if(hitSlot >= 0 && hitSlot < MAX_MEMOBJ_CANDIDATES)
  {
        iSlot = hitSlot ;
        vLabel_p = memObjRegistry[hitSlot].vName ;
  } else {
    fprintf(stderr,"@E memGetName FATAL: Invalid memory journal slot (%d) for existing obj at addr(%s)!\n",
                    hitSlot, addrLbl);
        // return DNA style of name as a fallback
        vLabel_p = memObjRegistry[hitSlot].vDnaLabel;
  }
  
 return vLabel_p ;
 
} // end:: memObjGetName()


// ---------------------------------------------------------
// function  : memObjSetName()
// purpose   : Assign a user defined label for a memory object,
//             given just a pointer to its memory region.
// revised   : 2014-02-16T08:28:00-07:00
// author    : joe glassy
// inputs    : void *mem_p -- ptr to memory region assigned
//           : char *vNewLabel_p --ptr to (string) new name
// returns   : void
// limits    : Name assigned MUST be < then  MAX_MEMOBJ_VNAME_LEN
//             which is typically 256 characters, so name <= 255 chars.
// notes     : This is the corrollary to memObjGetName()
void memObjSetName(void *mem_p,char *vNewLabel_p)
{
    int iSlot=0 ;
    int hitSlot = -1;
    size_t usersNameLen = 0;
 char fxName[256] = {0} ;
 char addrLbl[MAX_MEM_DNA_LEN] = {0} ;  
 
  strcpy(fxName,__func__);

  if(vNewLabel_p != NULL) 
  {
    usersNameLen = strlen(vNewLabel_p);
  }

  // block any progress beyond here if facility is not yet initialized
  if(memObjFacilityInitialized != 1)
  {
    memObjPrivateMsg(fxName, "@F %s is NOT currently initialized!","memObjFacility");
  }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg(fxName, "@F %zu exceeds maximum allowable registered mem obj %zu!\n",
           nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }

  // <JMG> 2016-01-07, added check on name of users string, to make more robust
  if(usersNameLen ==0 || usersNameLen > MAX_MEMOBJ_VNAME_LEN)
  {
    memObjPrivateMsg(fxName,
     "@E users name string length %zu invalid or exceeds maximum supported! %zu\n",
     usersNameLen, MAX_MEMOBJ_VNAME_LEN);
    return ;
  }

  // Re-constitute its DNA (unique id) in addrLbl, the look for it
  // in the registry
  hitSlot = memObjGetSlot(mem_p);
  // TBD: check if we can now just fetch memObj_p->vDnaLabel  

  sprintf(addrLbl,"%p",(uint8 *)mem_p); 
  // if it was found , fetch and return a pointer to its name
  if(hitSlot >= 0 && hitSlot < MAX_MEMOBJ_CANDIDATES)
  {
        iSlot = hitSlot ;
        // assign the new name to the memObj
        strcpy(memObjRegistry[hitSlot].vName,vNewLabel_p); 
  } else {
    fprintf(stderr,"@E memSetName FATAL: Invalid memory journal slot(%d) for existing obj at addr(%s)!\n",
                    hitSlot, addrLbl);
  }
} // end:: memObjSetName()


// ---------------------------------------------------------
// function  : memObjSetDatatype()
// purpose   : Assign a user defined TIV datatype identifier
//           : for a pre-existing memory object,
// revised   : 2014-02-23T10:06:00
// inputs    : void *mem_p -- ptr to memory region assigned
//           : const int newDatatype -- TIV datatype identifier
// returns   : void
// notes     : This is the corrollary to memObjGetDatatype()
void memObjSetDatatype(void *mem_p,const int newDatatype)
{
    int iSlot=0 ;
    int hitSlot = -1;
    
    char addrLbl[MAX_MEM_DNA_LEN] = {0} ;  

 
    // block any progress beyond here if facility is not yet initialized
    if(memObjFacilityInitialized != 1)
    {
    memObjPrivateMsg("memObjSetDatatype",
     "@F memObjFacility is NOT currently initialized!~");
    }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memObjSetDatatype", "@F %zu exceeds maximum allowable registered mem obj %zu!~\n",
           nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }
  // Re-constitute its DNA (unique id) in addrLbl, for labelling only
  sprintf(addrLbl,"%p",(uint8 *)mem_p); 

    // using objects pointer itself, find its base0 index in the registry
    hitSlot = memObjGetSlot(mem_p);
 
    // if object WAS found , fetch and return a pointer to its name
    if(hitSlot >= 0 && hitSlot < MAX_MEMOBJ_CANDIDATES)
    {
        iSlot = hitSlot ;
        // assign the new TIV datatype identifier to the memObj
        memObjRegistry[hitSlot].vDatatype = newDatatype; 
    } else {
    fprintf(stderr,"@E memSetDatatype FATAL: Invalid mem journal slot(%d) found for existing obj addr(%s)!~\n",
                    hitSlot, addrLbl);
    }
} // end:: memObjSetDatatype()


// ---------------------------------------------------------
// function  : memObjGetDatatype()
// purpose   : Retrieves a user defined TIV datatype identifier
//           : for a pre-existing memory object, using ptr as search key
// revised   : 2014-02-23T10:06:00
// inputs    : void *mem_p -- ptr to memory region assigned
// returns   : int, TIV datatype identifier, is IsFloat32, IsStringTable
// notes     : This is the corrollary to memObjGetDatatype()
int memObjGetDatatype(void *mem_p)
{
    int iSlot=0 ;
    int hitSlot = -1;
    int resultDatatypeId = -1 ;
    char addrLbl[MAX_MEM_DNA_LEN] = {0} ;  
 
  // block any progress beyond here if facility is not yet initialized
  if(memObjFacilityInitialized != 1)
  {
    memObjPrivateMsg("memObjGetDatatype",
     "@F memObjFacility is NOT currently initialized!~");
  }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memObjGetDatatype",
     "@F %zu exceeds maximum allowable registered mem obj %zu!\n",
     nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }
  // Re-constitute its DNA (unique id) in addrLbl, for labelling only
  sprintf(addrLbl,"%p",(uint8 *)mem_p); 

    // using objects pointer itself, find its base0 index in the registry
    hitSlot = memObjGetSlot(mem_p);
 
    // if object WAS found , fetch and return a pointer to its name
    if(hitSlot >= 0 && hitSlot < MAX_MEMOBJ_CANDIDATES)
    {
        iSlot = hitSlot ;
        resultDatatypeId = memObjRegistry[hitSlot].vDatatype;
    } else {
    fprintf(stderr,
    "@E memGetDatatype FATAL: Invalid mem journal slot(%d) for existing obj at addr(%s)!\n",
            hitSlot, addrLbl);
    }

 return resultDatatypeId ;
 
} // end:: memObjGetDatatype()


// ---------------------------------------------------------
// function  : memObjSetMetadata()
// purpose   : Assign a user defined metadata string to a 
//             pre-existing memObj, using its native pointer
//             to find it in registry.
// revised   : 2014-02-23T10:27:00
// inputs    : void *mem_p -- ptr to memory region assigned
//           : const char *newMetadata_p --ptr to new metadata string
// returns   : void
// notes     : This is the corrollary to memObjSetMetadata()
void memObjSetMetadata(void *mem_p,const char *newMetadata_p)
{
    int iSlot=0 ;
    int hitSlot = -1;
    
    char addrLbl[MAX_MEM_DNA_LEN] = {0} ;  
 
    // Block any progress beyond here if facility 
    // is not yet initialized
    if(memObjFacilityInitialized != 1)
    {
    memObjPrivateMsg("memObjSetMetadata",
     "@F memObjFacility is NOT currently initialized!~");
    }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memObjSetMetadata",
     "@F %zu exceeds maximum allowable registered mem obj %zu!~\n",
           nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }
  // Re-constitute its DNA (unique id) in addrLbl, for labelling only
  sprintf(addrLbl,"%p",(uint8 *)mem_p); 

    // using objects pointer itself, find its base0 index in the registry
    hitSlot = memObjGetSlot(mem_p);
 
    // if object WAS found , fetch and return a pointer to its name
    if(hitSlot >= 0 && hitSlot < MAX_MEMOBJ_CANDIDATES)
    {
        iSlot = hitSlot ;
        // pre-emptively clear out ALL of old string, then assign metadata
        memset(memObjRegistry[hitSlot].vMetadata, '\0', MAX_MEMOBJ_VNAME_LEN);
        // FORCE restrict load to prevent buffer overrun
        strncpy(memObjRegistry[hitSlot].vMetadata,
                newMetadata_p,MAX_MEMOBJ_VNAME_LEN-1); 
    } else {
    fprintf(stderr,"@E memSetMetadata FATAL: NO slot found for existing obj addr(%s)!~\n",
                    addrLbl);
    }
} // end:: memObjSetMetadata()


// ---------------------------------------------------------
// function  : memObjGetMetadata()
// purpose   : Retrieves a user defined optional metadata 
//             string for memObj, using objects native
//             ptr as search key
// revised   : 2014-02-23T10:06:00
// inputs    : void *mem_p -- ptr to memory region assigned
// returns   : char *, metadata string
// notes     : This is the corrollary to memObjSetMetadata()
char *memObjGetMetadata(void *mem_p)
{
    int iSlot=0 ;
    int hitSlot = -1;
    
    char *vMeta_p  = NULL;
    char addrLbl[MAX_MEM_DNA_LEN] = {0} ;  
 
  // block any progress beyond here if facility is not yet initialized
  if(memObjFacilityInitialized != 1)
  {
    memObjPrivateMsg("memObjGetMetadata",
     "@F memObjFacility is NOT currently initialized!~");
  }
  
  if(nMemObjRegistered >= MAX_MEMOBJ_CANDIDATES)
  {
    memObjPrivateMsg("memObjGetMetadata",
     "@F %zu exceeds maximum allowable registered mem obj %zu!\n",
     nMemObjRegistered,(size_t)MAX_MEMOBJ_CANDIDATES);
  }
  // Re-constitute its DNA (unique id) in addrLbl, for labelling only
  sprintf(addrLbl,"%p",(uint8 *)mem_p); 

    // using objects pointer itself, find its base0 index in the registry
    hitSlot = memObjGetSlot(mem_p);
 
    // if object WAS found , fetch and return a pointer to its name
    if(hitSlot >= 0 && hitSlot < MAX_MEMOBJ_CANDIDATES)
    {
        iSlot = hitSlot ;
        vMeta_p =  memObjRegistry[hitSlot].vMetadata;
    } else {
    fprintf(stderr,
    "@E memGetMetadata FATAL: NO valid slot for existing obj at addr(%s)!\n",
            addrLbl);
    }
 // returns pointer to pre-existing object's optional metadata string
 return vMeta_p;
} // end:: memObjGetMetadata()



#ifdef MEM_OBJ_STANDALONE

// Purpose: test various (pathological) redundancy use-cases
void memTestRedundancy(void)
{
 size_t i = 0,
        nElements = 4096 * 2;
    
 size_t *arySizt_p = NULL;
 int    *aryInt_p = NULL;
 char bufr[256] = {0} ; // buffer for qualifying memory object names
 char fxName[256] = {0} ;
 char fxLabel[MAX_MODULENAME_LEN]= {0};   // initialized from __func__
 char moduLabel[MAX_MODULENAME_LEN]= {0}; // Module name from __FILE__ to allow as lvalue
 
  // ------------------------------------------------------------------
 strcpy(moduLabel,(char *)__FILE__);
 strcpy(fxLabel  , (char *)__func__); // as defined in C standard c99 
 strcpy(fxName,(char *)__func__); 

  // block any progress beyond here if facility is not yet initialized
  if( memObjFacilityInitialized != 1)
  {
      fprintf(stderr,
      "@F memTestRedundancy: memObjFacility is NOT yet initialized. call memObjInit(logf) prior to use!\n");
      exit(-1);
      return ;
  }

  // test allocation mechanism
  arySizt_p = (size_t *)memObjCalloc("arySizt",sizeof(size_t),nElements,IsSizeT,__func__,"testRundancy");

  memObjReport(NULL,"memTestRedundancy");

  fprintf(stdout,
   "\ntestRundancy:: now populating test variable %s with %zu elements (%d slots-used)\n",
     "arySizt", nElements, memObjSlotsUsed() );

  for(i=0; i < nElements; i++)
  {
    if(i % 2 == 0)
     arySizt_p[i] = ((i+1) * 2) + 14 ;
    else
     arySizt_p[i] = ((i+1343) * 3) + 27 ;
  }
  
 // TEST evaluate what happens if we do 2 back-to-back calloc's with no intervening free 
  fprintf(stdout,
   "\ntestRundancy:: Intentionally re-calloc'ing (arySizt) -- a previously calloc'd variable\n");

 arySizt_p = (size_t *)memObjCalloc("arySizt",sizeof(size_t),nElements,IsSizeT,__func__,"testRundancy");
 for(i=0; i < nElements; i++)
  {
    if(i % 2 == 0)
     arySizt_p[i] = ((i+1) * 2) + 14 ;
    else
     arySizt_p[i] = ((i+1343) * 3) + 27 ;
  }

 fprintf(stdout,
  "\ntestRedundancy:: calling memObjFree to free %s of %zd elements\n","aryInt32",nElements);
  memObjFree(memObjQualify(fxName,"arySizt",bufr), arySizt_p);


  // NOW, allocate a new variable of type int
  aryInt_p = (int *)memObjCalloc("aryInt",sizeof(int),nElements/2,IsInteger,__func__,"testRundancy");
  for(i=0; i < nElements/2; i++)
  {
      aryInt_p[i] = i+1 + 100 ;
  }
  
  // TEST evaluate response when we perform (2) back-to-back 'free's for the same variable
  memObjFree(memObjQualify(fxName,"aryInt",bufr), aryInt_p);
  fprintf(stdout,
   "\ntestRundancy:: Intentionally calling free twice on (arySizt)\n");
  memObjFree(memObjQualify(fxName,"aryInt",bufr), aryInt_p);

  // end of unit test for the memObj heap memory variable management facility  

} // end::memTestRedundancy()


// Purpose: baselineTestMemObjFacility() implements a stand alone test of the 
// the memObjCalloc() wrapper and memObjFree() facility
void memBaselineTestMemObjFacility(void)
{
 size_t i = 0,
       nElements = 4096 * 2;
 // size_t *arySizt_p = NULL;  enable when actually used
 // int    *aryInt_p = NULL;
 double *aryDbl = NULL;
 long   *aryInt32_p = NULL;
 char bufr[256] = {0} ; // buffer for qualifying memory object names 
 char fxName[256] = {0} ;
 
  strcpy(fxName,(char *)__func__); 
   // test allocation mechanism
  aryInt32_p = (long *)memObjCalloc("aryInt32",sizeof(long),nElements,IsInteger,__func__,"testMemObj");
  memObjReport(NULL,"baselineTestMemObjFacility");

#ifdef WIN32
    fprintf(stdout,
   "\ntestMemObjFacility:: now populating test variable %s with %lu elements (%d slots-used)\n",
     "aryInt32", (unsigned long)nElements, memObjSlotsUsed() );
#else
  fprintf(stdout,
   "\ntestMemObjFacility:: now populating test variable %s with %zu elements (%d slots-used)\n",
     "aryInt32",nElements, memObjSlotsUsed() );
#endif

  for(i=0; i < nElements; i++)
  {
    if(i % 2 == 0)
     aryInt32_p[i] = (i+1) * 2 ;
    else
     aryInt32_p[i] = (i+1343) * 3;
  }
  
  // nElements = EG_9KM_ROWS * EG_9KM_COLS * 8 ;  
  nElements = 1624 * 3856 * 8;
  fprintf(stdout,
   "\ntestMemObjFacility:: now populating test variable %s with %zu elements\n","aryDbl",nElements);

  // test array of doubles at 9km EasyGrid times 8 Plant-Function-Type classes
  aryDbl = (double *)memObjCalloc("aryDbl",sizeof(double),nElements,IsReal,__func__,"testMemObj");
        
  fprintf(stdout,
   "\ntestMemObjFacility:: slots used %d\n",memObjSlotsUsed());
        
  // R_RANDOM_PROB()...
  for(i=0; i < nElements; i++)
     aryDbl[i] = (double)rand()/(double)RAND_MAX ; 
  for(i=0; i < nElements; i++)
     aryDbl[i] = (aryDbl[i] * 2.0980984) + ((double)rand()/(double)RAND_MAX) ;

  // test the ability to write binary data out to file
  // memObjWrite(variablesName,"optionUserPath",allOptionString like "All" if want to write all
  // or, the allOption can be specified as either NULL or ""
  memObjWrite("aryInt32","",NULL);

  fprintf(stdout,
   "\ntestMemObjFacility:: calling memObjFree to free %s of %zd elements\n","aryDbl",nElements);
  memObjFree(memObjQualify(fxName,"aryDbl",bufr), aryDbl);
      
    fprintf(stdout,
   "\ntestMemObjFacility:: calling memObjFree to free %s of %zd elements\n","aryInt32",nElements);
  memObjFree(memObjQualify(fxName,"aryInt32",bufr), aryInt32_p);
  
  // end of unit test for the memObj heap memory variable management facility  
  
} // end::memBaselineTestMemObjFacility()


// purpose: test memCalloc(),memFree() the anonymous constructor,destructors
void memTestCallocLite(const int nElem)
{
 int n = 1000;
 int i,k,kTrials =4;
 int assignedSlot = 0 ;
 
 float32 *flt32_p = NULL ;
 float64 *flt64_p = NULL;
 int32   *int32_p = NULL;
 
    // Perform several round-trip trials, to make sure that the
    // full cycle alloc+free's are working correctly.
    for(k=0; k < kTrials; k++)
    {
        
        int32_p = memCalloc(1024,sizeof(int32));
        // populate with some random values
        for(i=0; i < 1024; i++)
            int32_p[i] = FRM_API_RANDOM_INT(10000);
        
        // Set the problem size as a function of which trial we're in
        n = (nElem <= 0)? (k+1 * 1000) : nElem ;
        // allocate anonymous object, flt32_p
        flt32_p = memCalloc(n,sizeof(float32));

        flt64_p = memCalloc(512,sizeof(float64));
        for(i=0; i < 512; i++)
            flt64_p[i] = FRM_API_RANDOM_PROB();
    
        assignedSlot = memObjGetSlot(int32_p) ;
        fprintf(stdout,"int32_p is OK as assigned to slot %d\n",assignedSlot);
        
        assignedSlot = memObjGetSlot(flt64_p) ;
        fprintf(stdout,"flt64_p is OK as assigned to slot %d\n",assignedSlot);

        fprintf(stdout,"memTestCallocLite: Trial(%4d) nElem %d\n",k,n);
        // Assign a dummy value to every cell
        for(i=0; i < n; i++)
            flt32_p[i] = (k+1) * 9999.0 + 1.2345 ;

        // return the base0 slot index of the slot assigned to a variable
        assignedSlot = memObjGetSlot(flt32_p) ;
        fprintf(stdout,"flt32_p is OK as assigned to slot %d\n",assignedSlot);
            
        // Free these out of order, using their DNA to identify them
        memFree(flt32_p);
        memFree(flt64_p);
        memFree(int32_p);
    } // end:: k trials loop
            
} // end:: memTestCallocLite()


#endif
