// ---------------------------------------------------------------
// module       : frm_mmap_api.c
// purpose      : Implement "framework' component instance of the 
//                a mmap facilty for frm use
// project      : SMAP L4C science development.
// dependencies : mem_obj.c, for memObjCalloc(),memObjFree() and log_msg.c, for xLogMsg() etc.
// revision     : v1.0-2014-01-29T14:34:00-07:00,jmg
// history      : n/a
// author       : joe glassy, Lupine Logic Inc.
// build        : gcc -Wall -g -I/usr/include/hdf -o frm_mmap.ex frm_mmap.c -DHAVE_HDF4 -I$FRM_API_INC -L$FRM_API_LIB -L/usr/lib64/hdf -lmfhdf -ldf 
// notes        : caddr_t mmap(caddr_t addr,size_t len,int prot, int flag, int filedescriptor, off_t offset)
//                int munmap(caddr_t addr, size_t len)
//                  prot: PROT_READ,PROT_WRITE,PROT_EXEC,PROT_NONE
//                  map : MAP_FIXED,MAP_SHARED,MAP_PRIVATE
// ---------------------------------------------------------------

// define ADTs and function prototypes
#include "frm_api.h"

// reference various mmap parts and dependencies
#include <sys/types.h>
#include <sys/mman.h>
// references <bits/fcntl.h>....
#include <fcntl.h>

#define FILE_MODE (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)

// purpose: exercise,  write n bytes to a descriptor
size_t exerciseWriten(int fd,const void *vPtr_p,size_t nElem)
{
 size_t n = nElem;
 size_t nleft, nwritten= 0 ;
 const char *ptr_p = NULL;
 
	ptr_p = vPtr_p; // no pointer arith on void allowed, but we can do it on ptr_p
	nleft = n ;
	while(nleft > 0) 
	{
		if((nwritten = write(fd,ptr_p,nleft)) <=0)
			return(nwritten); // error...
		nleft += nwritten ;
		ptr_p += nwritten ;
	} // end:: while
	return n ;
} // end::testWriten()


// purpose: test read to read
size_t exerciseReadn(int fd,const void *vPtr_p,size_t nElem)
{
 size_t n = nElem;
 size_t nleft, nwritten= 0 ;
 size_t nread =0 ;
 const char *ptr_p = NULL;
 
  ptr_p = vPtr_p ;
  nleft = n ;
  while(nleft > 0)
  {
		if((nread=read(fd,ptr_p,nleft)) <0)
			return(nread);
		else if(nread ==0)
			break; // EOF
		nleft -= nread ;
		ptr_p += nread ;
  }
  return(n - nleft); // return for >= 0...
	
} // end::testReadn()


// purpose  : test mmap()
int testMmap(int argc,char *argv[])
{
	int fdin, fdout;
	char *src, *dst ;
	struct stat statbuf;
	char inputPath[FILENAME_MAX]  ={0} ;
	char outputPath[FILENAME_MAX]  ={0} ;
	
	if(argc != 3)
	{
		fprintf(stderr,"usage: %s <fromfile> <tofile>\n",argv[0]);	exit(-1);
	}
	
	strcpy(inputPath,argv[1]);
	strcpy(outputPath,argv[2]);
	fprintf(stdout,"\ntestMmap: copying (%s) to (%s)\n",
			inputPath,outputPath);
	
	if((fdin=open(inputPath, O_RDONLY)) < 0)
	{
		fprintf(stderr,"cannot open %s for reading\n",argv[1]);	exit(-1);
	}
	if((fdout=open(outputPath, O_RDWR | O_CREAT | O_TRUNC, FILE_MODE))<0)
	{
		fprintf(stderr,"cannot create %s for writing\n",argv[1]);exit(-1);
	}
	
	if(fstat(fdin,&statbuf)<0) // need size of input file to mmap
	{
		fprintf(stderr,"fstat failed on fdin \n"); exit(-1);
	}
	// set size of output file, base on input
	if(lseek(fdout,statbuf.st_size-1, SEEK_SET)== -1)
	{
		fprintf(stderr,"lseek error on fdout \n"); exit(-1);
	}
	if(write(fdout,"",1) !=1 )
	{
		fprintf(stderr,"write error on fdout \n"); exit(-1);
	}
	
	// mmap it
	if((src= mmap(0, statbuf.st_size, PROT_READ,MAP_FILE|MAP_SHARED,fdin,0)) == (caddr_t)-1)
	{
		fprintf(stderr,"mmap error on fdin on size %zu \n",statbuf.st_size); exit(-1);		
	}
	
	if((dst=mmap(0,statbuf.st_size, PROT_READ|PROT_WRITE, MAP_FILE|MAP_SHARED,fdout,0)) == (caddr_t)-1)
	{
		fprintf(stderr,"mmap error on output fdout on size %zu \n",statbuf.st_size); exit(-1);		
	}
	
	// both input and output are now wired for memory mapping, so just perform the copy
	memcpy(dst,src, statbuf.st_size);

	fprintf(stdout,"\ntestMmap: copy complete between source and destination!\n");

	return 0;
	
} // end::testMmap()



// purpose  : fill a buffer with nElem random int and share it
int testMmap2(const char *outputPath,size_t nElem)
{
	int fdin, fdout;
	char *src, *dst ; // these could just as well have been TI_ATOM_PTR instances too
	int i=0;
	int ret = -1 ;
	int nSamples = 10;
	int intVal = 0;
	int *intBuffer_p = NULL;
	int *chkBuffer_p = NULL;
	size_t  nOutputBytes = nElem * sizeof(int);
	size_t nRead =0;
	size_t nWritten =0;
	size_t offset =0;
	struct stat statbuf;
	char inputPath[FILENAME_MAX]  = {0} ;

	
	// strcpy(inputPath,argv[1]); // not yet defined...
	// fprintf(stdout,"\ntestMmap: copying (%s) to (%s)\n",inputPath,outputPath);

	fprintf(stdout,"\ntestMmap2: sizeof(int) here is %zu\n",sizeof(int));
	fprintf(stdout,"\ntestMmap2: nElem is %zu\n",nElem);
	
	
	if((fdout=open(outputPath, O_RDWR | O_CREAT | O_TRUNC, FILE_MODE))<0)
	{
		fprintf(stderr,"cannot create %s for writing\n",outputPath);exit(-1);
	}
	
	// set size of NEW output memory/file region, base on input
	if(lseek(fdout, nOutputBytes, SEEK_SET)== -1)
	{
		fprintf(stderr,"lseek error on NEW fdout of %zu elements \n",nElem); exit(-1);
	}
	
	if(write(fdout,"",1) !=1 )
	{
		fprintf(stderr,"write error on fdout \n"); exit(-1);
	}
	
	// create a buffer
	if((intBuffer_p = (int *)calloc(nElem,sizeof(int)))==NULL)
	{
		fprintf(stderr,"calloc error creating buffer of %zu elements \n",nElem);
		 exit(-1);		
	}
	// populate it with test data
	for(i=0; i < nElem; i++)
	{
		intBuffer_p[i] = R_RANDOM_INT(10000);
		if(i < nSamples)
			fprintf(stdout,"OrigDataSample %d = %d\n",i, intBuffer_p[i]);		 
	}

	// Create an alternative, check-receiving buffer
	chkBuffer_p = (int *)calloc(nElem, sizeof(int)) ;

	// mmap (memory map) the new destination area. The connection of interest
	//   here is made between the pointer 'dst' and the stream handle 'fdout'
	if((dst=mmap(0, nOutputBytes, PROT_READ|PROT_WRITE, MAP_FILE|MAP_SHARED,fdout,0)) == (caddr_t)-1)
	{
		fprintf(stderr,"mmap error on output fdout on size %zu \n",nOutputBytes); exit(-1);		
	}

	// use it?
	lseek(fdout,0L,SEEK_SET); // reposition..
	if((nWritten= write(fdout,intBuffer_p,nSamples*sizeof(int) ))<0)
	{
			fprintf(stderr,"write error on mmapped region\n");
			exit(-1);	
	}
	
	// use it?
	lseek(fdout,0L,SEEK_SET); // reposition...
	if((nRead = read(fdout,chkBuffer_p,nSamples*sizeof(int) ))<0)
	{
			fprintf(stderr,"read error on mmapped region\n");
			exit(-1);	
	} else {
			fprintf(stderr,"read %zu\n",nRead);
	}
	
	for(offset=0,i=0; i < nSamples ; i++)
	{
		// fprintf(stdout,"chkBuffer_p Sample %d = %d\n",i, chkBuffer_p[i]);
		memcpy(&intVal,dst+offset,sizeof(int));
		fprintf(stdout,"dst Sampled %d = %d\n",i,intVal );
		offset += sizeof(int);
	}
	
	ret = munmap(dst, nOutputBytes); //  verify...
	fprintf(stdout,"\ntestMmap2: ret %d \n",ret);
	
	// close the stream 
	close(fdout);

	free(intBuffer_p);
	free(chkBuffer_p);

	return 0;
	
} // end::testMmap2()


// invocation: <destinationPath> <sourcePath>
int main(int argc,char *argv[])
{
 int status =0;
 size_t datasetSiz = 10000;
 R_RANDOMIZE();
 
  // status = (int)testMmap(argc, argv ) ;

  testMmap2("testrandom.bin", datasetSiz);
	
 return status;
	
} // end:: main()
