// -----------------------------------------------------
// module       : frm_str_fx.c
// purpose      : Implement as part of L4_C "framework" a set of C string functions 
//                missing from POSIX's string.h
//                to provide token-level parsing, scalar-to-list (to/from) functions.
// revision     : 2014-01-24T13:42:00-07:00,v1.3, jmg, moved fx to fetch file size to frm_util_fx.c
// history      : 2014-01-14T17:18:00,renamed all "lli_" to "frm_" (Framework) for consistency
// history      : 2014-01-07T15:38:00, orig consolidation of MUM fx into one file
// author       : joe glassy
// dependencies : may be built with or without a TIV api dependency, via how HAVE_TIV is set
// notes        : 2014-01-27, TBD: convert all calloc() to use memCalloc() 
// notes        : module tested: frm_strfx.c
//                String Functions included here thus far are:

// char *frm_strGetVersTag(void);
// int   frm_strChrCount(const char *text,const int targetChr) ;
// char *frm_strCpBetween(const char *text,char *resultString,int startTok,int endTok,int *resultLength);
// int   frm_strChrIndex(const char *text,const int targetCh);
// char *frm_strRemoveChr(char *text,const int charToRemove);
// char *frm_strChrRemove(char *text,const int charToRemove);
// int   frm_strCountTokens(char *sourceStr,const char *tokenSepChars);
// char *frm_strGetToken(char *resultStr,char *sourceStr,const int ithToken,const char *tokenSepChars);
// char *frm_srcXtracRoot(char *source,char *dest,const int sep_char);
// char *frm_strXtracExt(char *source,char *dest,const int sep_char);
// char *frm_strPrepend(char *dest_p, char *source_p, const char *prepend_str_p);
// int   frm_strSubstringIndex(char *srcStr_p,char *targStr_p);
// char *frm_strReplaceChr(const char *targetChSet_p,char *srcAndResult_p,const char *newDest_char);
// int   frm_strParseValueList(char *source_p,char *delimStr,const int datatypeCode,void *resultArray_p);
// int   frm_strLoadValueStringFromArray(char *resultStr_p,const char *fmtSpec,const int delimCh,const int nValues,const int tivDatatype,void *resultArray_p);
// char *frm_strCustomFmt(const int op_code,char *origBuffer_p,
//                     char *fmtSpecStr_p,...);
//

#include "frm_api.h"


// -----------------------------------------------------
// function: frm_strGetVersTag()
// purpose : return the string function API version tag
// returns : char * FRM_STR_VERS_TAG as defined in frm_str_fx.h
// inputs  : void
// revision: 2014-01-21T11:47:00,jmg
// author  : joe glassy, Lupine Logic Inc.
char *frm_strGetVersTag(void)
{
  return(FRM_STR_VERS_TAG);
} // end::frm_strGetVersTag()


// -----------------------------------------------------
// function: frm_strToUpper()
// purpose : convert string 'text' to all upper case and return
// inputs  : char *text
// returns : char *text upper cased...
// author  : joe glassy, Lupine Logic Inc.
char *frm_strToUpper(char *text_p)
{
 int a=0;

   for(a=0; a < (int)strlen(text_p); a++)
      text_p[a] = toupper(text_p[a]);
      
 // return upper cased string
 return text_p;
} // end::frm_strToUpper()


// -----------------------------------------------------
// function: frm_strToLower()
// purpose : convert string 'text' to all lower case and return
// inputs  : char *text
// returns : char *text lower cased...
// author  : joe glassy, Lupine Logic Inc.
char *frm_strToLower(char *text_p)
{
 int a=0;

   for(a=0; a < (int)strlen(text_p); a++)
      text_p[a] = tolower( text_p[a] );
      
 // return lower cased string
 return text_p;
} // end::frm_strToLower()


// -----------------------------------------------------
// function: count = frm_strChrCount(text,targetCh) 
// purpose : count the n. of 'targetCh' characters in string 'text'
// inputs  : const char *text,const int targetChr
// returns : int, count of 'targetCh' in text
// author  : joe glassy, Lupine Logic Inc.
int frm_strChrCount(const char *text,const int targetChr)
{
 int a=0;
 int chrCount = 0 ;
 
   for(a=0; a < (int)strlen(text); a++)
      if(text[a] == targetChr) chrCount++ ;
      
// return count of occurrences of targetChr in text
return chrCount;

} // end::frm_strChrCount()


// -----------------------------------------------------
// function: frm_strCpBetween()
// purpose : frm_strCpBetween() is a string function to copy all text between a 
//              starting and ending token character, to 'resultString'
//              While normally 'start' is different from 'end' these may be the same, as for
//              parsing quoted expressions like "this is it!".
// inputs  : const char *text,char *resultString,int startTok,int endTok,int *resultLength
// outputs : ptr to result string AND a int set to the length of the resulting string
// author  : joe glassy, Lupine Logic Inc.
// returns : ptr to result string, as all content from 'text' between 
//           starting character 'startTok' and ending character 'endTok'
char *frm_strCpBetween(const char *text,char *resultString,int startTok,int endTok,int *resultLength)
{
 int a=0;
 int k=0;

    // Initialize at lenght of source string, 
    // to take advantage of all null terminators
    memset(resultString,'\0',strlen(text));

    
    if(strchr(text,startTok) != NULL && strchr(text,endTok)!= NULL )
    {
        // if scanning using quoted string '"' or "'", then make sure
        // that BOTH quote symbols are present
        if(startTok == endTok && (frm_strChrCount(text,startTok) < 2) )
        {
            fprintf(stderr,"@F error! strCpBetween: START and END tokens match but are missing one of these!");
            strcpy(resultString,"");
            return resultString ;
        }
        
        // pre-spin to starting point
        while(text[a] != startTok) 
                a++;
                
        a++ ; // next, step just PAST starting tok now
        while(text[a] != endTok && text)
        {
            resultString[k] = text[a++] ;
            k++;
        } // end:: while text[a]...
        
        // fprintf(stdout,"strCpBetween Delim { %c %c } -->(%s) at-index %d\n",startTok,endTok,resultString,k); 
    } // end: block if validated that both the start,end character do occur at least once

 // set the length of new result string
 *resultLength = k;
 
 // return the resulting string for use as an lvalue
 return resultString;

} // end::frm_strCpBetween()


// -----------------------------------------------------
// function: frm_strChrIndex() 
// inputs  : const char *text,const int targetCh
// returns : the base0 index where 'targetCh' appears in 'text' or -1 if it does NOT occur
// author  : joe glassy, Lupine Logic Inc.
int frm_strChrIndex(const char *text,const int targetCh)
{
  int charAtIndex = -1;
  char *ch_p = NULL;
  
  if((ch_p=strchr(text,targetCh))!= NULL)
  {
     charAtIndex = strlen(text) - strlen(ch_p) ;
     // fprintf(stdout,"strChrIndex: Char<%c> in (%s) at-index %d\n",targetCh,text,charAtIndex);  
  }

  return charAtIndex ;
    
} // end:: frm_strChrIndex()


// -----------------------------------------------------
// function: frm_strChrRemove() 
// inputs  : char *text,const int charToRemove
// returns : string with all 'charToRemove' chars removed from it
// author  : joe glassy, Lupine Logic Inc.
char *frm_strRemoveChr(char *text,const int charToRemove)
{
 int k=0,a=0;
 char *workString_p = NULL;
 
    workString_p= (char *)calloc( strlen(text)+1, sizeof(char));
 
    for(a=0; a < (int)strlen(text); a++)
    {
        if(text[a] != charToRemove)
            workString_p[k++] = text[a];
    }
    strcpy(text,workString_p);
    free(workString_p);
    
 return text;
} // end::frm_strRemoveChr()


// -----------------------------------------------------
// function: frm_strChrRemove() 
// inputs  : char *text,const int charToRemove
// purpose : removes all occurrences of char 'charToRemove' from string
// returns : string with all 'charToRemove' chars removed from it
// author  : joe glassy, Lupine Logic Inc.
char *frm_strChrRemove(char *text,const int charToRemove)
{
 int k=0,a=0;
 char *workString_p = NULL;
 
    workString_p= (char *)calloc( strlen(text)+1, sizeof(char));
 
    for(a=0; a < (int)strlen(text); a++)
    {
        if(text[a] != charToRemove)
            workString_p[k++] = text[a];
    }
    strcpy(text,workString_p);
    free(workString_p);
    
 return text;
} // end::frm_strChrRemove()


// -----------------------------------------------------
// function : frm_strCountTokens()
// inputs   : char *sourceStr,const char *tokenSepChars
// returns  : int, count of delimited tokens present
// author   : joe glassy, Lupine Logic Inc.
int frm_strCountTokens(char *sourceStr,const char *tokenSepChars)
{
 char *ch_p= NULL;
 char *brkb_p= NULL;
 char *workbuffer = NULL;
 int  tokenIndex = -1 ;
 
    // next, shadow the original buffer to protect original
    if((workbuffer=calloc(sizeof(char),strlen(sourceStr)+1))==NULL)
    { 
        fprintf(stderr,"@F frm_strCountToken: fatal error on calloc!\n");
        exit(-1);
    } else {
        strcpy(workbuffer,sourceStr);
        // fetch the ith token...
        for(ch_p =strtok_r(workbuffer,tokenSepChars,&brkb_p);
            ch_p != NULL;
            ch_p = strtok_r(NULL,tokenSepChars,&brkb_p))
        {
                // if we got this far, there is at least one token
                // so adjust...
                if(tokenIndex == -1) 
                   tokenIndex = 0 ;
                   
                tokenIndex++ ; // increment count of tokens encountered
        } // end:: for loop through ch_p tokens
    } // end:: block if tokenizing can proceed
    
    // free local resources
    free(workbuffer);
    
 // return count of tokens found, or -1 if none found
 return tokenIndex;
 
} // end::frm_strCountTokens()


// -----------------------------------------------------
// function : frm_strGetToken()
// inputs   : char *resultStr,char *sourceStr,const int ithToken,const char *tokenSepChars
// returns  : ptr to result string
// author   : joe glassy, Lupine Logic Inc.
char *frm_strGetToken(char *resultStr,char *sourceStr,const int ithToken,const char *tokenSepChars)
{
 char *ch_p= NULL;
 char *brkb_p= NULL;
 char *workbuffer = NULL;
 int  tokenIndex =0 ;
 
    // initialize with backstopped, known result (empty)
    strcpy(resultStr,"");
    // next, shadow the original buffer to protect original
    if((workbuffer=calloc(sizeof(char),strlen(sourceStr)+1))==NULL)
    { 
        fprintf(stderr,"@F frm_strGetToken: fatal error on calloc!\n");
        exit(-1);
    } else {
        strcpy(workbuffer,sourceStr);
        // fetch the ith token...
        for(ch_p =strtok_r(workbuffer,tokenSepChars,&brkb_p);
            ch_p != NULL;
            ch_p = strtok_r(NULL,tokenSepChars,&brkb_p))
        {
            if(tokenIndex == ithToken)
            {
                strcpy(resultStr,ch_p);
                break; // once found, take early out
            }
            tokenIndex++ ; // increment token counter
        } // end:: for loop through ch_p tokens
        
    } // end:: block if tokenizing can proceed
    
    // free local resources
    free(workbuffer);
    
 // return pointer to result token string retrieved, or "" if failed
 return resultStr ;
} // end::frm_strGetNthToken()


// -----------------------------------------------------
// function: frm_strToTypedValue()
// inputs  : void *varBuf,char *valueStr,const int tivDatatype
// returns : void *valuevar --ptr to typed variable's value
// purpose : convert a numeric value in a string to a strongly typed var
void *frm_strToTypedValue(void *varBuf,char *valueStr,const int tivDatatype)
{
 static TIV_ATOM tivAtm = {0} ;
 int32 siz = -1 ;
 
    switch(tivDatatype)
    {
    case IsInt8   :tivAtm.aInt8    = (signed char)valueStr[0]; break;
    case IsUint8  :tivAtm.aUint8   = (unsigned char)valueStr[0]; break;
    case IsInt16  :tivAtm.aInt16   = (short) atoi(valueStr); break;
    case IsUint16 :tivAtm.aUint16  = (uint16)atoi(valueStr); break;
    case IsInt32  :tivAtm.aInt32   = (int32)atol(valueStr); break;
    case IsUint32 :tivAtm.aUint32  = (uint32)atol(valueStr); break;
    case IsInt64  :tivAtm.aInt64   = (int64)  atol(valueStr); break;
    case IsUint64 :tivAtm.aUint64  = (uint64)atol(valueStr); break;
    case IsFloat32:tivAtm.aFlt32 = (float32)atof(valueStr); break;
    case IsFloat64:tivAtm.aFlt64 = (float64)atof(valueStr); break;
    } // end::switch on tivDatatype()
    
    siz = tiv_SizeofTivType(tivDatatype);
    // blast the binary bit pattern into waiting buffer
    memcpy(varBuf,&tivAtm, siz) ;
    // also return pointer to TIV itself
    return &tivAtm ;
} // end:: frmStrtoTypedValue()


// -----------------------------------------------------
// function : frm_strGetTokenAsNumeric()
// inputs   : void *resultValBuffer,--receiving buffer large enough for datatype
//            const int tivDatatype --datatype, IsFloat32, IsUint32, etc.
//            char *sourceStr -- string containing the values representation
//            const int ithToken -- base0 token index in the text line
//            const char *tokenSepChars -- delimiter chars, e.g. "\t= "
// returns  : void *vPtr --ptr to one tivDatatype'd value, or 0 if failed.
// author   : joe glassy, Lupine Logic Inc.
// example  : double x =0.0; 
//            double y, yVal = 0.0;
//            float32 zFlt =0.0;
//            char vBuffer[sizeof(double)] ={0};
//            strcpy(sourceStr,"1,2,3,4.567");
//               frm_strGetTokenAsNumeric(&x,IsFloat64,sourceStr,3,",");
//            After call, x will contain 4.567
//               Or, to use another calling style:
//            memcpy(&y, frm_strGetTokenAsFloat64(&yVal, sourceStr,3,","),
//              sizeof(float64));
//          memcpy(&zFlt, frm_strGetTokenAsNumeric(&vBuffer,IsFloat32,
//              sourcestr,3,","), sizeof(float32));
//            After the call, zFlt will contain 4.567 

// notes    : This is NOT yet thread-safe due to static 'tiv' inside!!
// notes    : TBD, more testing needed.

void *frm_strGetTokenAsNumeric(void *resultValBuffer,const int tivDatatype,
            char *sourceStr,const int ithToken,const char *tokenSepChars)
{
 char *ch_p= NULL;
 char *brkb_p= NULL;
 char *workbuffer = NULL;
 char resultStr[32] = {0} ;
 uint8 vBuffer[32] = {0} ;
 static TIV_ATOM tiv = {0} ;
 int  siz = -1 ;
 int  tokenIndex =0 ;
 
    // resultStr initialized with backstopped, known result (empty)

    // derive size in bytes of ONE scalar instance of this datatype
    siz = tiv_SizeofTivType(tivDatatype);
    
    // next, shadow the original buffer to protect original
    if((workbuffer=calloc(sizeof(char),strlen(sourceStr)+1))==NULL)
    { 
        fprintf(stderr,"@F frm_strGetTokenAsFloat64: fatal error on calloc!~\n");
        exit(-1);
    } else {
        strcpy(workbuffer,sourceStr);
        // fetch the ith token...
        for(ch_p =strtok_r(workbuffer,tokenSepChars,&brkb_p);
            ch_p != NULL;
            ch_p = strtok_r(NULL,tokenSepChars,&brkb_p))
        {
            // values limited to 32 character long in string form
            memset(resultStr,'\0',32);

            if(tokenIndex == ithToken)
            {
                strcpy(resultStr,ch_p);
                // pre-emptively interpret numeric result as a tivDatatype
                memcpy(&tiv,frm_strToTypedValue(vBuffer, resultStr,
                    tivDatatype),siz);
                // TBD verify: ...copy into resultValBuffer?
                memcpy(resultValBuffer, &tiv, siz);
                break; // once found, take early out
            } // end:: block when ithToken matched OK..
            
            tokenIndex++ ; // increment token counter
        } // end:: for loop through ch_p tokens
        
    } // end:: block if tokenizing can proceed
    
    // free local resources
    free(workbuffer);
    
 // return (float64) from result token string retrieved, or 0.0 if failed
 return &tiv ;
} // end::frm_strGetNthTokenAsNumeric()


// -----------------------------------------------------
// function : frm_strGetTokenAsFloat64()
// inputs   : void *resultBuffer,char *sourceStr,const int ithToken,const char *tokenSepChars
// returns  : float64, field's string value interpreted as a float64. No consistency checking is
//            applied yet that assures the string in fact does contain information that we can
//            interpret as a numeric quantity (TBD).
// notes    : A TIV enabled companion function called frm_strGetTokenAsNumeric() is
//            planned frm_strGetTokenAsNumeric(rslt, tivDatatype, srcStr, ithTok, tokSepCh) 
//            that returns a strongly typed numeric result 
//            (e.g. IsInt32,IsFloat32,IsUint8 etc) back in 'rslt'

// example  : double x ; strcpy(sourceStr,"1,2,3,4.567");
//              x = frm_strGetTokenAsFloat64(NULL, sourceStr,3,",");
//            When called, x will contain 4.567, as well 'alternativeResult'.
//            Or, to use another calling style:
//              frm_strGetTokenAsFloat64(&x, sourceStr,3,",");
//            After the call, x will contain 4.567 
//
// author   : joe glassy, Lupine Logic Inc.
//
float64 frm_strGetTokenAsFloat64(void *resultVal,char *sourceStr,const int ithToken,const char *tokenSepChars)
{
 char *ch_p= NULL;
 char *brkb_p= NULL;
 char *workbuffer = NULL;
 char resultStr[64] = {0} ;
 float64 resultValue = -99999.0 ;
 int  tokenIndex =0 ;

 
    // resultStr initialized with backstopped, known result (empty)

    // next, shadow the original buffer to protect original
    if((workbuffer=calloc(sizeof(char),strlen(sourceStr)+1))==NULL)
    { 
        fprintf(stderr,"@F frm_strGetTokenAsFloat64: fatal error on calloc!~\n");
        exit(-1);
    } else {
        strcpy(workbuffer,sourceStr);
        // fetch the ith token...
        for(ch_p =strtok_r(workbuffer,tokenSepChars,&brkb_p);
            ch_p != NULL;
            ch_p = strtok_r(NULL,tokenSepChars,&brkb_p))
        {
            if(tokenIndex == ithToken)
            {
                strcpy(resultStr,ch_p);
                // for now, blindly interpret numeric result as a float64...
                // later we'll remove all whitespace and perform a char-by-char
                // verification to make sure this is indeed a numeric quantity
                resultValue = atof(resultStr);
                //ALSO ...if resultVal is not NULL, then blast
                //   result into a buffer assumed to be correct size
                if(resultVal != NULL)
                    memcpy(resultVal, &resultValue, sizeof(float64));
                break; // once found, take early out
            }
            tokenIndex++ ; // increment token counter
        } // end:: for loop through ch_p tokens
        
    } // end:: block if tokenizing can proceed
    
    // free local resources
    free(workbuffer);
    
 // return (float64) from result token string retrieved, or 0.0 if failed
 return resultValue ;
} // end::frm_strGetNthTokenAsFloat64()




// -----------------------------------------------------
// function: frm_srcXtracRoot()
// purpose : extract root portion of a file path spec string from
//           string 'source' and place the root portion in 'dest'.
//           so if path="/this/is/myfile" result : "/this/is"
// inputs  : char *source,char *dest
//           const int sep_char -- "." usually
// returns : char *, root (preface) portion of filepath spec string
// author  : joe glassy, Lupine Logic Inc.
// notes   : 2014-08-13, corrected fx name from incorrect 'src' to 'str'
//           as in now corrected: frm_strXtracRoot() 
// notes   : 'root' portion is everything up to basename
// example : src="/this/is/a/path/to/myfile.txt
//           result="/this/is/a/path/to"
char *frm_strXtracRoot(char *source,char *dest,const int sep_char)
{
 int sep_symbol = sep_char;

    // if a dummy or invalid separation char passed in, use default
  if(sep_char <= 0)
       sep_symbol = '.';


    //  at least one sep char found, so proceed...
  if(strrchr(source,sep_symbol)!=NULL)
  {
    memset(dest, '\0', strlen(source)+1);
    return strncpy(dest, source,
             (strlen(source)-strlen(strrchr(source,sep_symbol))) );
  } else {

    // no sep. char found, so return pointer to intact string
    return source ;
  }
  
} // end:: frm_strXtracRoot() 


// -----------------------------------------------------
// function: frm_strXtracExt()
// purpose : extracts the "extension"" portion of a file 
//      pathspec from an unambiguous pathspec and returns a pointer
//      to this extracted substring. This function uses the
//      rightmost "." separation symbol as the key navigation mark,
//      though the source string may include multiple "." symbols.
//      This function is useful for building a set of pathnames 
//      from a common root pathname. Example: when the input 
//      full path is "foo.bar", this functions result is "bar".
// inputs  : char *source,char *dest,const int sep_char
// returns : char *, pointer to all text to the right of the rightmost
//      '.' symbol, typically the file's extention.  For example,
//      if the original file is "mod15.am1.h12.v4.hdf" the text returned
//      by this function is "hdf".
// author  : joe glassy, Lupine Logic Inc.
char *frm_strXtracExt(char *source,char *dest,const int sep_char)
{
 int sep_symbol = sep_char;

    // if a dummy or invalid separation char passed in, use default
  if(sep_char <= 0)
       sep_symbol = '.' ;

    // at least one sep char found, so identify rightmost '.' symbol
  if(strrchr(source,sep_symbol)!=NULL)
  {
    strcpy(dest,strrchr(source,(int)sep_symbol)+1 );
    return dest ; 
  } else {
    // no sep. chars were found, so return pointer to intact string
    return source;
  } // end:: block if no sep char found

} // end::frm_strXtracExt() 


// -----------------------------------------------------
// function    : frm_strPrepend()
// inputs      : char *source_p, const char *prepend_p 
// returns     :  string 'source' with a substring "prepend_p" now 
//              inserted in front of original text of "source_p"
// author      : joe glassy, Lupine Logic Inc.
// assumptions : dest_p string MUST be long enough to hold both the
//               original source and prepend_p strings
// notes       : This is functional inverse or opposite of strcat()
char *frm_strPrepend(char *dest_p, char *source_p, const char *prepend_str_p)
{
 /* projected length of new string */
 int src_len =0,
     new_len = 0;
     
 //  note key lengths
 src_len = strlen(source_p)+1; // add one to account for null termination
 new_len = src_len + strlen(prepend_str_p);

 // Explicitly initialize, try to force a trackable runtime 
 //    error if dest not long enough! 
 memset(dest_p, '\0', new_len) ;
 
 strcpy(dest_p, prepend_str_p);
 strcat(dest_p, source_p) ; 

 // As a convenience, returns pointer to destination string
 return dest_p ;

} // end:: frm_strPrepend() 


// -----------------------------------------------------
// function  : frm_strSubstringIndex()
// purpose   : return the base0 index in srcString where substring 'targStr_p' is found
//             or -1 if not found at all.
// inputs    : char *srcStr_p, char *targStr_p
// returns   : int, index where target substring found in Source string or -1 if not found
// author    : joe glassy, Lupine Logic Inc.
// notes     : n/a
//           
int frm_strSubstringIndex(char *srcStr_p,char *targStr_p)
{
 int resultIndex = -1 ;
 char *ptr_p = NULL ;
 
    // first, determine if present at all   
    if((ptr_p=strstr(srcStr_p,targStr_p)) != NULL)
    {
        resultIndex = strlen(srcStr_p) - strlen(ptr_p);
    } else {
        resultIndex = -1;
    }
    
    // return base0 index in source where target substring was found
    // or -1 if not found at all.
    return resultIndex ;
    
} // end:: frm_strSubstrIndex()



// -----------------------------------------------------
// function  : frm_strReplaceChr()
// purpose   : Replace all occurrences of ANY of the characters in source_chars
//             with the single 'dest_char' character in 'buffer' string
// inputs    : char *resultBuffer_p, const char *source_chars,const int dest_char
// returns   : char *, ptr to filtered string, or NULL if empty or failed
// author    : joe glassy, Lupine Logic Inc.
// notes     : n/a
char *frm_strReplaceChr(const char *targetChSet_p,char *srcAndResult_p,const char *newDest_char)
{
 int slen = 0;
 int i = 0 ;
 int j = 0 ;
         
    // block action on null or empty string
   if( srcAndResult_p == NULL )
       return NULL ;
     
   // NOTE: If newDest_char is "" we COULD auto-substitute a call to 
   //   frm_strRemoveChr(srcAndResult_p,targetChSet_p);
   
   if( (slen = strlen(srcAndResult_p)) <=0 )
       return NULL ;
       
    // Replace all occurrences of ANY char inside source string
    //    with the single destination character specified
    for(i=0;  i < slen ; i++)
        for(j=0; j < (int)strlen(targetChSet_p); j++)
        {
            if(srcAndResult_p[i] == targetChSet_p[j])
            {
                if(strlen(newDest_char)>= 1)
                    srcAndResult_p[i] = newDest_char[0];
                else
                    srcAndResult_p[i] = srcAndResult_p[i-1];
            }
        }
        
 return srcAndResult_p;
 
} // end:: frm_strReplaceChr()

// -----------------------------------------------------
// function: frm_StrTrimWhitespace()
// inputs  : str_p -- ptr to source string
// returns : Modified result string where leading,trailing whitespace removed
// invocation: strcpy(destString_p,frm_StrTimeWhitespace(origStr_p));
// purpose : Clip leading and trailing whitespace off a string
// notes   : Caller MUST retrieve modified result via the function return,
//           this source string is passed by value not by reference.
// URL: http://stackoverflow.com/questions/122616/how-do-i-trim-leading-trailing-whitespace-in-a-standard-way
// Note: This function returns a pointer to a substring of the original string.
// If the given string was allocated dynamically, the caller must not overwrite
// that pointer with the returned value, since the original pointer must be
// deallocated using the same allocator with which it was allocated.  The return
// value must NOT be deallocated using free() etc.
// WARNING: proper use is: strcpy(newDest_p,trimwhitespace(source_p));
// since str_p here is NOT modified in place! you MUST retrieve the 
// modified result via the function return!
char *frm_StrTrimWhitespace(char *str_p)
{
  char *end_p = NULL;
  int chLeading = 0 ;
  int chTrailing= 0 ;
  // first, clip off any leading whitespace
  while(isspace((unsigned char)*str_p))
  { 
    str_p++;
    chLeading++; // devel only, confirm how many leading chars clipped
  }
  
  // Return early, when resulting string at this point is all whitespace
  if(*str_p == 0) 
    return str_p;  

  // next, trim off any trailing space. If none exists, that's fine too
  end_p = str_p + strlen(str_p) - 1;
  
  while(end_p > str_p && isspace((unsigned char)*end_p)) 
  {
     end_p--;
     chTrailing++; // devel only, confirm how many trailing chars clipped
  }
  // place a new null terminator at end of modified string
  *(end_p+1) = 0;
  
  // Finally, return the modified result. The caller MUST retrieve the modified
  //   result via this function return, due to call by value..
  return str_p;
} // end_p::frm_StrTrimWhitespace()


// -----------------------------------------------------
// function: frm_StrTrimWhitespaceToBuffer(sourceString,destinationString)
// inputs  : str_p -- ptr to source string
//           dest_p-- ptr to preallocated destination buffer (string)
// returns : Modified result string where leading,trailing whitespace removed
// invocation:  frm_StrTimeWhitespaceToBuffer(origStr_p,destStr_p);
//           : OR, strcpy(destStr_p,frm_StrTrimWhitespaceToBuffer(origStr_p,destStr_p)); 
// purpose : Clip leading and trailing whitespace off a string
// notes   : Caller may retrieve modified via modified dest_p string
//           or via function return.
char *frm_StrTrimWhitespaceToBuffer(char *str_p,char *dest_p)
{
  char *end_p = NULL;
  int chLeading = 0 ;
  int chTrailing= 0 ;
  // create a temporary work area to protect source in this version of fx
  char *orig_p = NULL;
  char *headPtr_p = NULL; // set a ptr to head of buffer

  // create a new temporary, zero'd out buffer to work on
  orig_p    = calloc(sizeof(char),strlen(str_p)+1);
  headPtr_p = orig_p; // capture and retain pointer to head of work buffer
  
  strcpy(orig_p,str_p);
  // fprintf(stdout,"frm_StrTrimWhitespaceToDestTEST initial orig_p [%s]\n",orig_p);
  // fflush(stdout);
  
  // first, clip off any leading whitespace
  while(isspace((unsigned char)*orig_p))
  { 
    orig_p++;
    chLeading++; // devel only, confirm how many leading chars clipped
  }
  
  // Return early, when resulting string at this point is all whitespace
  if(*orig_p == 0) 
  {
    strcpy(dest_p,orig_p);
    return dest_p;  
  }

  // next, trim off any trailing space. If none exists, that's fine too
  end_p = orig_p + strlen(orig_p) - 1;
  
  while(end_p > orig_p && isspace((unsigned char)*end_p)) 
  {
     end_p--;
     chTrailing++; // devel only, confirm how many trailing chars clipped
  }
  // place a new null terminator at end of modified (orig_p) string
  *(end_p+1) = 0;
  
  // Finally, copy to dest_p and return result as a modified destStr
  strcpy(dest_p,orig_p);
  // fprintf(stdout,"frm_StrTrimWhitespaceToDestTEST DEST (%s)\n",dest_p);
  // fflush(stdout);
  
  // point to head, release local, temporary working memory
  free(headPtr_p);
  
  return dest_p;
} // end_p::frm_StrTrimWhitespaceToBuffer()


// -----------------------------------------------------
// function  : frm_strParseValueList()
// purpose   : Scan source string 'source_p', delimited using any char in
//             'delimStr' for a collection of numeric values,and return
//             these in a *void that must be interpreted later by user.
//             Return the count of values as the array size, in nCount.
// inputs    : char *source_p,char *delimStr,int datatypeCode,int *nCount
// returns   : int , count of values parsed or -1 if failed.
// author    : joe glassy, Lupine Logic Inc.
// notes     : the datatypecode must be either IsInt32 or IsFloat64
int frm_strParseValueList(char *source_p,char *delimStr,const int datatypeCode,void *resultArray_p)
{
  char fxName[256] = {0} ;
  char *ch_p = NULL   ; // qualified token pointer
  char *brkb_p = NULL ; // pointer required for strtok_r thread safety
  char *workStr_p = NULL; // temporary buffer, to protect original string
  int srcLen = strlen(source_p); // fetch length of source string
  int nCount =0;
  
  float64 *flt64_p = NULL;
  int32  *int32_p = NULL;
  
    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 

    // simple interface only supports int32 and float64
    if(datatypeCode != IsInt32 && datatypeCode != IsFloat64)
    {
        fprintf(stderr,
        "@F frm_strParseValueList: only int32 and float64 supported!");
        return -1 ;
    }
    
    if(datatypeCode == IsInt32)
        int32_p = (int32 *)resultArray_p ;
    else
        flt64_p = (float64 *)resultArray_p ;

    //  parse the numeric values out and load them
    workStr_p = (char *)calloc(sizeof(char),srcLen+1);
    strcpy(workStr_p,source_p);
    
    for(ch_p = strtok_r(workStr_p, delimStr, &brkb_p);
                ch_p != NULL;
                ch_p = strtok_r(NULL, delimStr, &brkb_p))
    {
            if(datatypeCode == IsInt32)
                int32_p[nCount] = (int32)atoi(ch_p);
            else
                flt64_p[nCount] = (float64)atof(ch_p);
            nCount++ ;
    } // end:: for ch_p loop through tokens
    
    // free up temp buffer
    free(workStr_p);
 // return count of values parsed and loaded
 return nCount;
    
} // end:: frm_strParseValueList()

#ifdef HAVE_TIV
// -----------------------------------------------------
// function  : frm_strLoadValueStringFromArray()
// purpose   : Using a numeric array (datatype identified via TivDatatype),
//             load a delimited string with the initial nValues of the array.
// inputs    : char *resultStr_p,
//             const char *fmtSpec --optional format specifier string
//             const int delimCh,const int nValues,
//             const int tivDatatype,
//             void *resultArray_p
// returns   : char *resultString holding array values, e.g. "1.1,2.2,3.3"
// author    : joe glassy, Lupine Logic Inc.
// notes     : the datatypecode must be either IsInt32 or IsFloat64
int frm_strLoadValueStringFromArray(char *resultStr_p,const char *fmtSpec,const int delimCh,const int nValues,const int tivDatatype,void *resultArray_p)
{
  char fxName[256] = {0} ;
  char *workStr_p = NULL; // temporary buffer, to protect original string
  char numBuf[64] = {0} ;
  
  char workFmt[32] = {0} ;
  const char *dfltFmt[] = {"%d","%d","%d","%u","%d","%u","%ld","%zu","%f","%lf", } ;
  
  int i,sLen = 0 ;
  
  TIV_ATOM_PTR tivObj ;
  
  
    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 
    // Estimate requirelength of result string, giving each value
    // of to 32 ch to be represented
    
    sLen = (nValues * 32) ;
    strcpy(resultStr_p,"");
    workStr_p=(char *)calloc(sLen, sizeof(char));

    // BIND a typed pointer..
    tivObj.aVoid_p = resultArray_p ;
    // INTERPRET each
    for(i=0; i < nValues; i++)
    {
        // Interpret array value as an actionable scalar type
        memset(numBuf,'\0',64);
        
        // set format specifier either as callers (if defined) or as default
        if(fmtSpec != NULL && strlen(fmtSpec)>= 1)
            strcpy(workFmt,fmtSpec);
        else // remember to offset around IsNoType at position 0!
            strcpy(workFmt,dfltFmt[tivDatatype-1] ) ;
            
        switch(tivDatatype)
        {
        case IsInt8   : sprintf(numBuf, workFmt , tivObj.aInt8_p[i]  ) ; break;
        case IsUint8  : sprintf(numBuf, workFmt , tivObj.aUint8_p[i] ) ; break;
        case IsInt16  : sprintf(numBuf, workFmt , tivObj.aInt16_p[i] ) ; break;
        case IsUint16 : sprintf(numBuf, workFmt , tivObj.aUint16_p[i]) ; break;
        case IsInt32  : sprintf(numBuf, workFmt , tivObj.aInt32_p[i] ) ; break;
        case IsUint32 : sprintf(numBuf, workFmt , tivObj.aUint32_p[i]) ; break;
        case IsInt64  : sprintf(numBuf, workFmt , tivObj.aInt64_p[i] ) ; break;
        case IsUint64 : sprintf(numBuf, workFmt , tivObj.aUint64_p[i]) ; break;
        case IsFloat32: sprintf(numBuf, workFmt , tivObj.aFlt32_p[i] ) ; break;
        case IsFloat64: sprintf(numBuf, workFmt , tivObj.aFlt64_p[i] ) ; break;
        default: fprintf(stderr,"Unsupported TIV datatype(%d)",tivDatatype);
        } // end::switch on tivDatatype
        strcat(workStr_p,numBuf);
        sLen = strlen(workStr_p);
        // Add a delimiter char after all but the last element
        if(i < (nValues-1))
            workStr_p[sLen] = delimCh;
    } // end:: i loop
    
    // add final termination char
    workStr_p[sLen+1] = '\0'; 
    
    // transfer final result
    strcpy(resultStr_p, workStr_p);
    
    // free up temp buffer
    free(workStr_p);
 // return count of values loaded
 return i;
    
} // end:: frm_strLoadValueStringFromArray()
#endif


// -----------------------------------------------------
// function  : frm_strCustomFmt()
// purpose   : Either prepend or concatenate a custom sprintf style
//             formatted string, relative to a original source string,
//             returning the result as a ptr and via the call argument.
// inputs    : enum opCode{doConCat,doPrepend}, or 0=ConCat,1=Prepend
//             sourceBuffer_p -- ptr to original source string
//             char *fmtSpecStr 
//             'delimStr' for a collection of numeric values,and return
//             these in a *void that must be interpreted later by user.
//             Return the count of values as the array size, in nCount.
// returns   : int , count of values parsed or -1 if failed.
// author    : joe glassy, Lupine Logic Inc.
// revised   : 2014-02-20T11:45, increased oversize factor from 2 to 6 to
//             make this safer for larger filepath and directory pre-pends
// limitations and assumptions: it is the callers responsibility to make
//             the origBuffer_p is sized large enough to accept the prepend
//             or concatenated result string returned.
// notes     : Useful for a variety of custom string manipulations,
//             or for assembing NSB stringblock 'record' delimited property 
//             entries, etc.
char *frm_strCustomFmt(const int op_code,char *origBuffer_p,
                     char *fmtSpecStr_p,...)
{
 char fxName[256] = {0} ;
 char *workstr_p = NULL;
 char *newform_p = NULL;
 int  origLen = strlen(origBuffer_p);
 va_list argptr;

    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 

    // Invariant: create up-sized, thread-safe instances of (zero'd out) 
    // working buffers
    newform_p = (char *)calloc(origLen * 6 +1, sizeof(char));
    workstr_p = (char *)calloc(origLen * 6 +1, sizeof(char));

    // Build a custom formatted string via vsprintf() using callers fmtSpecStr_p
    va_start(argptr   , fmtSpecStr_p );
    vsprintf(newform_p, fmtSpecStr_p, argptr);
    va_end(argptr);

    // Concatenate OR Prepend, depending on opCode, the formatted string to original
    if(op_code == DoConcatenate)
    {
            strcat(origBuffer_p, newform_p);
    } else if(op_code == DoPrepend) {
            strcpy(workstr_p, newform_p);
            strcat(workstr_p, origBuffer_p);
            strcpy(origBuffer_p, workstr_p);
    }  

    free(workstr_p);
    free(newform_p);
    
    // return pointer to modified string
    return origBuffer_p;
   
} // frm_strCustomFmt()


// ---------------------------------------------------------------
// function  : frm_strCompare()
// inputs    : const void *p1,const void *p2
// returns   : int, 0 on match, standard strcmp() return as {-1|0|1}.
// purpose   : compare (2) null terminated strings for qsort() or bsearch() etc
// source    : The Practice of Programming, by K&R
//             useful and necessary (instead of just strcmp) to managing IsStringTable
//             objects.
int frm_strCompare(const void *p1,const void *p2)
{
  char *v1, *v2 ;
  
   v1 = *(char **)p1 ;
   v2 = *(char **)p2 ;
   
   return strcmp(v1,v2);
} // end:: frm_strCompare()


// ---------------------------------------------------------------------
// function : frm_strExpandStringTbl()
// Purpose  : Using an existing stringtable, expand its width to
//            accomodate a pre-pend operation.
// Inputs   : char **origStrTbl_p
//            const int nOrigStrings -- count
//            const int lenToAdd -- a minimum length to expand stringtable by
// revised  : 2014-07-29T18:55:00
// returns  : char **newStrTbl_p -- a ptr to new string table
// notes    : n/a
char **frm_strExpandStringTbl(const int nOrigStrings,const int expandByCh,char **origStrTbl_p)
{
char **resultStrTbl_p = NULL;
 // to be conservative, double it
 int nNewWidthCh = (strlen(origStrTbl_p[0])+ expandByCh) * 2 ;
 int i = 0 ;
  // create a new stringtable capable of handling orig data plus new
 char **workStrTbl_p = memCreateStringTbl(nOrigStrings,nNewWidthCh);
 
    // copy contents of original string table
    for(i=0; i < nOrigStrings ; i++)
    {
        memset(workStrTbl_p[i],'\0',nNewWidthCh);
        strcpy(workStrTbl_p[i], origStrTbl_p[i]);
    }
    // release original (too narrow for pre-pending)
    memFreeStringTbl(origStrTbl_p,nOrigStrings);
    resultStrTbl_p = workStrTbl_p ;

    // finally, return a pointer to adjusted, populated string table
    return resultStrTbl_p ; 
} // end:: frm_strExpandStringTbl()


// ---------------------------------------------------------------
// function : frm_strAddDatetimeStamp()
// inputs   : const time_t finalSec --user supplied time_t value 
//            const int whichTimezone-- flag indicating which timezone to use
//            UTC(GMT) or local timezone,as: 0=UTC, 1=localtimezone
//            char *prefixPath -- a 'basename' like 'l4c_DAILY' etc
//            char *localTzLabel -- optional local time zone label like "PST","MST","EST" etc
//            where if this is NULL or empty, no localtime zone qualifier label is used.
//            char *finalFilepath -- result with datetime stamp appended
//             
// returns  : char *, ptr to final log path
// revised  : 2014-08-13T11:06:00, added.
// Purpose  : Using an initial prefix pathstring, append a datetime stamped
//            label to the prefix and add a .log extension
// notes    : The finalFilePath must be long enough to accomodate the 
//            datetime text of "YYYY-DD-MMThhmmss"
char *frm_strAddDatetimeStamp(const time_t finalSec,const int whichTimezone,char *prefixPath,char *localTzLabel,char *finalFilepath)
{
 // char timeBuffer[64] = {0} ;
 // buffer for copying log file to final destination
 char fileWorkingPath[FILENAME_MAX] = {0};
 char workingPrefixPath[FILENAME_MAX] = {0} ;
 char tzQualifier[16] = {0} ; // time zone qualifier
 char fxName[256] = {0} ;
 struct tm workingTime  ;
 int curYear,curMonth,curDayOfMonth, curHour,curMin, curSec;


    strcpy(fxName,__func__);
    
    // Call this ONLY AFTER working filepath is closed, datestamp it
    //  and copy it to final log location
    //  Function: struct tm * gmtime_r (const time_t *time, struct tm *resultp)
    // using thread-safe GMT time fx, for final labelling of log file currently
    // NOTE: check if this SHOULD be 'UTC' instead of GMT

    // BUG: orig strXtractRoot() does NOT work as needed here. We need
    // to modify this to extract ALL text EXCEPT the extension.
    
    // Pre-condition the prefix to extract just the root portion of a
    // fully qualified (unambiguous) path.
    frm_strXtracRoot(prefixPath,workingPrefixPath,'/');

    if(whichTimezone == 0)
    {
        gmtime_r(&finalSec, &workingTime);
        strcpy(tzQualifier,"UTC");
    } else {
        localtime_r(&finalSec, &workingTime);
        if(strlen(localTzLabel) >0)
            strcpy(tzQualifier, localTzLabel); // local time zone used...
        else
            strcpy(tzQualifier,"");
    }

    // de-reference the struct tm, noting subtle ways the
    // curYear and curMonth are actually defined as units SINCE a ref mark
    curYear       = workingTime.tm_year + 1900 ;
    curMonth      = workingTime.tm_mon  + 1;
    curDayOfMonth = workingTime.tm_mday;
    curHour       = workingTime.tm_hour ;
    curMin        = workingTime.tm_min  ;
    curSec        = workingTime.tm_sec  ;
    
    if(strlen(tzQualifier)>0)
        sprintf(fileWorkingPath,"%s_%4.4d-%2.2d-%2.2dT%-2.2d%-2.2d%-2.2d_%s.log",
               workingPrefixPath, curYear, curMonth, curDayOfMonth, curHour,curMin,curSec, tzQualifier);
    else
        sprintf(fileWorkingPath,"%s_%4.4d-%2.2d-%2.2dT%-2.2d%-2.2d%-2.2d.log",
               workingPrefixPath, curYear, curMonth, curDayOfMonth, curHour,curMin,curSec);
    
               
    strcpy(finalFilepath, fileWorkingPath);
    
 return finalFilepath;  
} // end:: frm_strAddDatetimeStamp()
