// ---------------------------------------------------------------
// module       : frm_tiv_api.c
// purpose      : Implement "framework' component instance of the 
//                universal-union-of-typeful-atoms: a type-independent variable (TIV) adapted
//                from the NASA EOS UM MODIS (MUM) API.
// project      : SMAP L4C science development.
// dependencies : mem_obj.c, for memObjCalloc(),memObjFree() and log_msg.c, for xLogMsg() etc.
// revision     : v2.0-2014-01-21T18:04:00-07:00,jmg
// history      : 2014-01-21T11:12:00-- documentation clean up and added tiv_GetVersTag() to help caller
//                confirm version used.
// history      : 2014-01-06T00:38:00-- introduces UUTAObj for transition to thread-safe use, 
//                and re-factor as TIV_xxx for use in latest:
//                memObjCalloc(),memCalloc(), xLogMsg(),xLocalLogMsg()
// history      : 2013-11-08T08:32:00
// author       : joe glassy, Lupine Logic Inc.
// build        : UNIT_TEST mode: gcc -Wall -g -I/usr/include/hdf -DUUTA_UNIT_TEST -o tiv_ti_api.ex tiv_ti_api.c -L/usr/lib64/hdf -lmfhdf -ldf
// build        : API mode: gcc -Wall -g -I/usr/include/hdf -c tiv_ti_api.c 
// notes        : In Makefile, specify HAVE_HDF4 if linking with HDF4 headers which
//                provide basic integer and float types, int32, float32, etc.
// ---------------------------------------------------------------


// define ADTs and function prototypes
#include "frm_api.h"


// ---------------------------------------------------------------
// function : tiv_GetVersTag(void)
// purpose  : retrieve TIV version tag, as defined in frm_tiv_api.h
// history  : 2014-01-21, added to help caller confirm version
// inputs   : void
// returns  : char *FRM_TIV_VERSTAG
char *tiv_GetVersTag(void)
{
    return( FRM_TIV_VERSTAG ) ;
} // end::tiv_GetVersTag()


// ---------------------------------------------------------------
// function : tiv_Hdf4TypeToTivType()
// purpose  : convert a HDF4 datatype identifier like DFNT_FLOAT32
//            into its TIV enum equivalent, IsFloat32
// inputs   : const int32 hdfDatatype
// revision : 2014-01-06T01:00:00
// author   : Joe Glassy, Lupine Logic Inc.
// notes    : HDF4 is missing the following datatype descriptors
//             IsString,IsStringTable,IsDirectoryspec, IsFilespec,
//             IsDatetime, IsBoolean, etc.
//
int tiv_Hdf4TypeToTivType(const int32 hdfDatatype)
{
 int tDtype = -1 ; // TIV datatype,like: IsUint8,IsFloat32 etc
 
    switch(hdfDatatype)
    {
        case DFNT_INT8    : tDtype = IsInt8   ; break;
        case DFNT_UINT8   : tDtype = IsUint8  ; break;      
        case DFNT_INT16   : tDtype = IsInt16  ; break;
        case DFNT_UINT16  : tDtype = IsUint16 ; break;
        case DFNT_INT32   : tDtype = IsInt32  ; break;
        case DFNT_UINT32  : tDtype = IsUint32 ; break;
        case DFNT_INT64   : tDtype = IsInt64  ; break;
        case DFNT_UINT64  : tDtype = IsUint64 ; break;
        case DFNT_FLOAT32 : tDtype = IsFloat32; break;
        case DFNT_FLOAT64 : tDtype = IsFloat64; break;
        default : fprintf(stderr,
            "@E tiv_HdfDatatypeToTivType: unrecognized HDF4 datatype! %d\n",hdfDatatype);
    } // end:: switch on HDF4 datatype identifier
    
    // return the TIV type identifier matching the HDF4 datatype
    return tDtype ;
} // end:: tiv_Hdf4TypeToTivType()


// ---------------------------------------------------------------
// function : tiv_TivTypeToHdf4Type
// purpose  : convert a TIV datatype identifier like IsFloat32
//            into its HDF4 equivalent, like DFNT_FLOAT32
// inputs   : const int TIV type identifier
// returns  : int32, HDF4 datatype identifier
// revision : 2014-01-06T01:00:00
// author   : Joe Glassy, Lupine Logic Inc.
// notes    : HDF4 is missing the following datatype descriptors
//             IsString,IsStringTable,IsDirectoryspec, IsFilespec,
//             IsDatetime, IsBoolean, etc.

int32 tiv_TivTypeToHdf4Type(const int tivType)
{
int32 hdfDatatype = -1 ;

    switch(tivType)
    {
        case IsInt8   : hdfDatatype  = DFNT_INT8    ; break;
        case IsUint8  : hdfDatatype  = DFNT_UINT8   ; break;        
        case IsInt16  : hdfDatatype  = DFNT_INT16   ; break;
        case IsUint16 : hdfDatatype  = DFNT_UINT16  ; break;
        case IsInt32  : hdfDatatype  = DFNT_INT32   ; break;
        case IsUint32 : hdfDatatype  = DFNT_UINT32  ; break;
        case IsInt64  : hdfDatatype  = DFNT_INT64   ; break;
        case IsUint64 : hdfDatatype  = DFNT_UINT64  ; break;
        case IsFloat32: hdfDatatype  = DFNT_FLOAT32 ; break;
        case IsFloat64: hdfDatatype  = DFNT_FLOAT64 ; break;
        default : fprintf(stderr,
            "@E tiv_TovTypeToHDF4Type: unrecognized TIV datatype! %d\n",
            tivType);
            break;
    } // end:: switch on HDF4 datatype identifier
    
    return hdfDatatype ;
    
} // end::tiv_TivTypeToHdfType()


// ---------------------------------------------------------------
// function  : tiv_SizeofTivType()
// purpose   : return the size of one numeric scalar TIV datatype
//             identifier (IsInt32) scalar item. This does not support
//             IsString, or IsStringTable or other non-numeric types yet.
// returns   : int32 -- size in bytes of one scalar element of tivType
int32 tiv_SizeofTivType(const int tivType)
{
 size_t siz = 0 ;
 int sizeOfType =0;
 int wasRecognized = 1 ;
 char fxName[256] = {0} ;
 char tivLabel[32] = {0}; 
 
    // <JMG> 2016-02-12T12:22 added
    strcpy(fxName,__func__);
    // strcpy(tivLabel, tiv_NameofTivType(tivType,&sizeOfType));
    
    switch(tivType)
    {
        case IsInt8   : 
        case IsChar   : siz  = sizeof(int8)  ; break;
        case IsUint8  : siz  = sizeof(uint8) ; break;       
        case IsInt16  : siz  = sizeof(int16) ; break;
        case IsUint16 : siz  = sizeof(uint16); break;
        case IsInt32  : siz  = sizeof(int32) ; break;
        case IsUint32 : siz  = sizeof(uint32); break;
        case IsInt64  : siz  = sizeof(int64) ; break;
        case IsUint64 : siz  = sizeof(uint64); break;
        case IsFloat32: siz  = sizeof(float) ; break;
        case IsFloat64: siz  = sizeof(double); break;
        // include traps for these recognized but unsupported TIV types (unknowable)
        case IsBoolean: siz  = sizeof(int)   ; break; // assume, enum "int"
        case IsSizeT  : siz  = sizeof(size_t); break;
        case IsString : siz  = 0; break; // unknowable here 
        case IsStringTable  : siz  = 0; break; // unknowable here 
        case IsDatetime     : siz  = 0; break; // unknowable here 
        case IsFilespec     : siz  = 0; break; // unknowable here 
        case IsDirectoryspec: siz  = 0; break; // unknowable here 
        default : 
             wasRecognized = 0 ;
             fprintf(stderr,
            "@E tiv_SizeofTivType: Sorry, (TIV type %d) was NOT recognized! ~\n",
                tivType);
            break;
    } // end:: switch on HDF4 datatype identifier

  // inform user that this TIV type was recognized but its unit size cannot
  // be determined here
  if(wasRecognized && siz == 0 && tivType >= 54)
  {
    fprintf(stderr,"@W tiv_SizeofTivType: TIV (type %d) recognized but its size cannot be determined here!~\n",
        tivType);

  }
    
  sizeOfType = (int)siz ;
  
  // return size of a single scalar of this type    
  return (int32)sizeOfType;
    
} // end::tiv_SizeOfTivType()


// ---------------------------------------------------------------
// function: tiv_NameofTivType()
// purpose : convert an integer tivType identifier to a text label
// returns : returns text label of TivType, and also sizeOfOne item.
// notes   : The l4c_TIVDatatypeToDatatypeString(tivDatatype,typeBuffer)
//           is a functional alias for this, but THIS function is preferred
//           for use as the l4c_ fx will be deprecated to normalize...
char *tiv_NameofTivType(const int tivType,int *sizOfOne)
{
 static char typeLabel[32] = {0} ;
 
    switch(tivType)
    {
        case IsInt8   : strcpy(typeLabel,"int8"   ) ; break;
        case IsUint8  : strcpy(typeLabel,"uint8"  ) ; break;        
        case IsInt16  : strcpy(typeLabel,"int16"  ) ; break;
        case IsUint16 : strcpy(typeLabel,"uint16" ) ; break;
        case IsInt32  : strcpy(typeLabel,"int32"  ) ; break;
        case IsUint32 : strcpy(typeLabel,"uint32" ) ; break;
        case IsInt64  : strcpy(typeLabel,"int64"  ) ; break;
        case IsUint64 : strcpy(typeLabel,"uint64" ) ; break;
        case IsFloat32: strcpy(typeLabel,"float32") ; break;
        case IsFloat64: strcpy(typeLabel,"float64") ; break;
        case IsBoolean: strcpy(typeLabel,"boolean"); break;
        case IsChar   : strcpy(typeLabel,"char"); break ;
        case IsSizeT  : strcpy(typeLabel,"size_t")  ; break;
        case IsString : strcpy(typeLabel,"string")  ; break;
        case IsStringTable  : strcpy(typeLabel,"stringtable");break;
        case IsDatetime     : strcpy(typeLabel,"datetime")   ;break;        
        case IsFilespec     : strcpy(typeLabel,"filespec")   ;break;
        case IsDirectoryspec: strcpy(typeLabel,"directoryspec");break;
        default : fprintf(stderr,"@E tiv_NameofTivType: unrecognized TIV datatype! %d\n",tivType);
            break;
    } // end:: switch on HDF4 datatype identifier

   if(tivType < 50)
   {
       *sizOfOne = (int)tiv_SizeofTivType(tivType)   ;
   } else {
       // TIV types with an enum value of 50 or higher may be recognized but
       // their sizOfOne cannot be determined since these are mostly compound types
       *sizOfOne = 0 ;
   }
    
  // return pointer to (static, long-lived) label
  return typeLabel;
    
} // end::tiv_NameOfTivType()


// ---------------------------------------------------------------
// function: tiv_NameToTivType()
// purpose : Convert common datatype identifier text label TO 
//           an integer tivType identifier. Also returns the equivalent
//           HDF4 datatype identifier integer like: DFNT_INT32 or DFNT_FLOAT32.
// inputs  : const char *typeLabel, int32 *hdf4DatatypeId
// returns : int, TIV datatype identifier, like IsInt32 or IsFloat32 etc.
// notes   : n/a
int tiv_NameToTivType(const char *typeLabel,int32 *hdf4DatatypeId)
{
 int resultTIV_DatatypeId = -1 ;
 
    if(strcmp(typeLabel,"int8")==0)
    {
        resultTIV_DatatypeId = IsInt8 ;
    } else if(strcmp(typeLabel,"uint8"  )==0) {
        resultTIV_DatatypeId = IsUint8 ;
    } else if(strcmp(typeLabel,"int16"  )==0) {
        resultTIV_DatatypeId = IsInt16 ;
    } else if(strcmp(typeLabel,"uint16" )==0) {
        resultTIV_DatatypeId = IsUint16 ;
    } else if(strcmp(typeLabel,"int32"  )==0) {
        resultTIV_DatatypeId = IsInt32 ;
    } else if(strcmp(typeLabel,"uint32" )==0) {
        resultTIV_DatatypeId = IsUint32 ;
    } else if(strcmp(typeLabel,"int64"  )==0) {
        resultTIV_DatatypeId = IsInt64 ;
    } else if(strcmp(typeLabel,"uint64" )==0) {
        resultTIV_DatatypeId = IsUint64 ;
    } else if(strcmp(typeLabel,"float32")==0) {
        resultTIV_DatatypeId = IsFloat32 ;
    } else if(strcmp(typeLabel,"float64")==0) {
        resultTIV_DatatypeId = IsFloat64 ;
    }

    // Also, as a convenience, return the equivalent HDF5 datatype
    // identifier (integer) identifier
    *hdf4DatatypeId = tiv_TivTypeToHdf4Type(resultTIV_DatatypeId);

  // return enum integer TIV datatype identifier
  return resultTIV_DatatypeId;
    
} // end::tiv_NameToTivType()

// ---------------------------------------------------------------
// function: frm_tivTypedValue()
// purpose : convert a generic buffer into a tivType'd variable, for
//           later interpretation by caller (e.g. cast, extract value
//           from one of the TIV_ATOM's constituent members.
// returns : returns void *ptr, one scalar numeric value interpreted
//           through the context of a TIV_ATOM
void *frm_tivTypedValue(void *varBuf,const int tivDatatype)
{
 static TIV_ATOM tivAtm = {0} ;
 int32 siz = tiv_SizeofTivType(tivDatatype); 
 
    memcpy(&tivAtm, varBuf, siz);
    // also return pointer to TIV itself
    return &tivAtm ;
} // end:: frm_tivTypedValue()

// ---------------------------------------------------------------------
// function: frm_tivStrToTypedValuefrm_strToTypedValue()
// inputs  : void *varBuf,char *valueStr,const int tivDatatype
// returns : void *valuevar --ptr to typed variable's value
// purpose : convert a numeric value in a string to a strongly typed var
// notes   : This is TIV API convenience function, equiv to frm_str
//         : api's 
void *frm_tivStrToTypedValue(void *varBuf,char *valueStr,
            const int tivDatatype)
{
 static TIV_ATOM tivAtm = {0} ;
 int32 siz = -1 ;
 
    switch(tivDatatype)
    {
    case IsInt8   :tivAtm.aInt8    = (signed char)valueStr[0]; break;
    case IsUint8  :tivAtm.aUint8   = (unsigned char)valueStr[0]; break;
    case IsInt16  :tivAtm.aInt16   = (short) atoi(valueStr); break;
    case IsUint16 :tivAtm.aUint16  = (uint16)atoi(valueStr); break;
    case IsInt32  :tivAtm.aInt32   = (int32)atol(valueStr); break;
    case IsUint32 :tivAtm.aUint32  = (uint32)atol(valueStr); break;
    case IsInt64  :tivAtm.aInt64   = (int64)  atol(valueStr); break;
    case IsUint64 :tivAtm.aUint64  = (uint64)atol(valueStr); break;
    case IsFloat32:tivAtm.aFlt32 = (float32)atof(valueStr); break;
    case IsFloat64:tivAtm.aFlt64 = (float64)atof(valueStr); break;
    } // end::switch on tivDatatype()
    
    siz = tiv_SizeofTivType(tivDatatype);
    // blast the binary bit pattern into waiting buffer
    memcpy(varBuf,&tivAtm, siz) ;
    // also return pointer to TIV itself
    return &tivAtm ;
} // end:: frm_tivStrToTypedValue()


// --------------------------------------------------------------------
// function : tiv_AllocRaw()
// purpose  : allocate a named (memObj) style memory object,
//           inside a simple (non-object) UUTA
// inputs   : const int32 dataType,const size_t nElements,char *varName
// returns  : void *ptr -- pointer to allocated memory of nElements of
//            HDF4 integer datatype identifier, like: DFNT_FLOAT32
void *tiv_AllocRaw(const int32 hdf4Datatype,const size_t nElements,char *varName)
{
  TIV_ATOM_PTR *tiv_p = NULL;
  int  thisSize = -1;
  int  tivType = -1 ;
  char varLabel[256] = {0} ;

    // Derive the TIV datatype identifier from the HDF4 identifer ID
    tivType = tiv_Hdf4TypeToTivType(hdf4Datatype);
    
    // Derive the sizeOfOne factor from the TIV type, and, 
    // Auto-derive a prefix name (not currently used yet)
    strcpy(varLabel, tiv_NameofTivType(tivType,&thisSize));

    // ALLOCATE memory for this specific, strongly-typed var
    tiv_p->aUint8_p= (uint8 *)memObjCalloc(varName,thisSize, 
            nElements,tivType,"","");

  // return pointer to new memory within a UUTA
  return tiv_p;
  
} // end::tiv_AllocRaw()


// --------------------------------------------------------------------
// function : tiv_ReleaseRAw() 
// purpose  : release memory allocated for a simple UUTA
void tiv_ReleaseRaw(const TIV_ATOM_PTR *uuta_p,char *varName)
{
  assert(uuta_p->aUint8_p != NULL);
  // this works the same: memObjFree(varName,NULL);
  
  memObjFree(varName,uuta_p->aUint8_p);
  
} // end::tiv_ReleaseRaw()


// --------------------------------------------------------------------
// function : tiv_CopyRawFromSlot()
// purpose  : Copy nElements from SOURCE, beginning at "atSlot",
//            into DEST slot 0, where each shares the same datatype
// revised  : 2014-01-06T00:40:00
// author   : joe glassyRIP_LOCAL/l4c_DAILY_NTSG_2016-02-10.rip
// notes    : n/a
//
void *tiv_CopyRawFromSlot(void *vDest_p,void *vSource_p,const int32 dataType,
                    const size_t atSlot,const size_t nElements)
{
  TIV_ATOM_PTR *src_p = (TIV_ATOM_PTR *)vSource_p; // ...dereference the void* as a real type
  TIV_ATOM_PTR *dst_p = (TIV_ATOM_PTR *)vDest_p;   // ...dereference the void* as a real type
  
  size_t thisOffset =  DFKNTsize(dataType) * atSlot ;  
  size_t nBytes     = DFKNTsize(dataType)  * nElements;
  
  memcpy(dst_p->aUint8_p, src_p->aUint8_p+thisOffset, nBytes);
         
 // return ptr to destination buffer as a syntactic convenience
 return vDest_p ;
} // end::tiv_CopyRawFromSlot()


// --------------------------------------------------------------------
// function : tiv_CopyRawIntoRaw is a powerful but dangerous "any-to-any" UUTA operator.
// purpose  : Copy nElements from SOURCE, beginning at srcSlot, into DEST at destSlot, 
//    where each shares the same datatype. Be CAREFUL using this one, that each
//    UUTA is sized to accomodate all the data requested here! Later we'll add an 
//    extension that internally guards against overcopy via use of a 
//    UUTA's nElemAllocated property.
// inputs      : void *vDest_p -- destination vector
//             : void *vSource_p -- source vector
//             : const int32 dataType -- HDF4 datatype identifier DFNT_FLOAT32 etc.
//             : const size_t srcSlot -- base0 index in source to start at
//             : const size_t destSlot-- base0 index in destination to start,
//             : size_t nElements -- n. of elements to transfer.
// author      : joe glassy
// dependencies: DFKNTsize() function from HDFv4, or, a functional equivalent aliased to this name
// 
// notes       : prior, was "copyUUTA(src_p, dest_p, srcSlot, destSlot, dataType, nelem)"
//
void *tiv_CopyRawToRaw(void *vDest_p,void *vSource_p,const int32 dataType,
            const size_t srcSlot,const size_t destSlot, size_t nElements)
{
  TIV_ATOM_PTR *src_p = (TIV_ATOM_PTR *)vSource_p; // ...essential to dereference the void * as a real type here
  TIV_ATOM_PTR *dst_p = (TIV_ATOM_PTR *)vDest_p;   // ...essential to dereference the void * as a real type here
  
  size_t nBytes    =  DFKNTsize(dataType) * nElements;
  size_t srcOffset =  DFKNTsize(dataType) * srcSlot ;  
  size_t dstOffset =  DFKNTsize(dataType) * destSlot ;  
   
  memcpy(dst_p->aUint8_p+dstOffset, src_p->aUint8_p+srcOffset, nBytes);
         
 // return ptr to populated destination buffer as a syntactic convenience
 return vDest_p ;
} // end::tiv_CopyRawToRaw()



// --------------------------------------------------------------------
// function   : tiv_SetOneIntoRawSlot()
// purpose    : Assign ONE element FROM SOURCE, into DESTINATION UUTA 
//              at dest position "atSlot", where each shares a common dataType.
// Call Example:  float32 myFloat32 = -9999.0;  setOneIntoRawSlot(&destUUTA, &myFloat32, DFNT_FLOAT32, arrayIndex);
void *tiv_SetOneIntoRawSlot(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot)
{
  TIV_ATOM_PTR *src_p = (TIV_ATOM_PTR *)vSource_p;
  TIV_ATOM_PTR *dst_p = (TIV_ATOM_PTR *)vDest_p;
  size_t thisOffset =  DFKNTsize(dataType) * atSlot ;

  memcpy((uint8*)(dst_p->aUint8_p+thisOffset), (uint8*)src_p, DFKNTsize(dataType));
  
  // return ptr to destination buffer as a syntactic convenience
  return vDest_p;
} // end::tiv_setOneIntoRawSlot()


// --------------------------------------------------------------------
// function : tiv_GetOneFromArray()
// purpose  : Retrieve ONE element from  UUTA SOURCE, starting at offset, storing in a destination (scalar),
//            where each share a common datatype.  A call Example:  
//   TIV_ATOM_PTR myUUTA ;  
//   float32 myFloat32 = -9999.0 ;
//   tiv_GetOneFromRaw( &myDestUUTA, &myFloat32, DFNT_FLOAT32, atArrayIndex);
//
void *tiv_GetOneFromArray(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot)
{
  TIV_ATOM_PTR *src_p = (TIV_ATOM_PTR *)vSource_p;

  size_t sizOfOne   = DFKNTsize(dataType);
  size_t thisOffset = sizOfOne * atSlot ; // byte offset into source where starting value lies
  // if variable is "myFloat32", call: tiv_GetOneFromRaw(&myFloat32,&sourceUUTA,arrayIndex);
  memcpy(vDest_p, (uint8*)src_p->aUint8_p+thisOffset,  sizOfOne);
  return(vDest_p);
} // end::tiv_GetOneFromArray()


// --------------------------------------------------------------------
// function : tiv_SetOneFromSourceSlot()
// purpose  : copy-assign ONE value from a source (array) of dataType at srcSlot, into  
//            the base (slot 0) of a preallocated "vDest_p" vector of the same base type

// inputs   : void *vDest_p,void *vSource_p,const int32 dataType,const size_t srcSlot
void tiv_SetOneFromSourceSlot(void *vDest_p,void *vSource_p,const int32 dataType,const size_t srcSlot)
{
 uint8 *ptr_p= NULL;
 size_t sizOfOne   = DFKNTsize(dataType);
 size_t thisOffset =  DFKNTsize(dataType) * srcSlot ;

  ptr_p= (uint8 *)vDest_p;
  memcpy(ptr_p, (uint8 *)vSource_p+thisOffset, sizOfOne);
} // end::tiv_SetOneFromSourceSlot()


// --------------------------------------------------------------------
// HIGH-LEVEL TIV API SERVICE FUNCTIONS
// -----------------------------------RIP_LOCAL/l4c_DAILY_NTSG_2016-02-10.rip---------------------------------

// --------------------------------------------------------------------
// allocate memory for a UUTAObj (parent), for nElements of dataType
UUTAObj *tiv_CreateObj(const int32 dataType,const size_t nElements,char *varName)
{
 UUTAObj *uObj_p     = NULL ;
 TIV_ATOM_PTR  *atm_p = NULL;
 int   tivType       = tiv_Hdf4TypeToTivType(dataType);
 int   siz           = tiv_SizeofTivType(tivType);
 char datasName[256] = {0} ;

    // set default sentinal to up-vote later if succeeeded
    uObj_p->uDatatype = -1;
    uObj_p->uCount    = 0 ;

    // Create parent, empty shell
    uObj_p= (UUTAObj *)memObjCalloc(varName,sizeof(UUTAObj),1,IsCustomUserType,"","");
    // post name inside it
    strcpy(uObj_p->uLabel, varName);
    
    if(uObj_p == NULL)
    {
        fprintf(stderr,"tiv_CreateUUTAObj:: failed on %zu elements of type %d\n",
            nElements, dataType);
            return NULL;
    } else {
        // Set pointer to child to allocate to
        atm_p = &uObj_p->uuta;
        // Give child a derivative name, based on parent, to further qualify
        sprintf(datasName,"%s|DATA",varName);
        
        atm_p->aUint8_p=(uint8*)memObjCalloc(datasName,siz, nElements,tivType,"","");

        if(atm_p->aUint8_p == NULL)
        {
            fprintf(stderr,"tiv_CreateObj:: [%s] failed on %zu elements of type %d\n",
                datasName, nElements, dataType);
                return NULL;
        }
        // up-vote, where a valid datatype now also indicates success
        uObj_p->uDatatype = dataType ;
        uObj_p->uCount    = nElements;
    }
 // return allocated parent object
 return uObj_p;
} // end::tiv_CreateObj()


// --------------------------------------------------------------------
// tiv_ReleaseObj() -- release memory for a UUTA parent object
void tiv_ReleaseObj(UUTAObj *uObj_p)
{
char datasName[256] = {0} ; 

    // First, rrelease from inside,out. First free the children
    sprintf(datasName,"%s|DATA",uObj_p->uLabel);
    memObjFree(datasName,NULL);
    
    // Now free the parent enclosing object
    strcpy(datasName,uObj_p->uLabel);
    memObjFree(datasName, NULL);
    
} // end::tiv_ReleaseObj()


// --------------------------------------------------------------------
// function: tiv_getObjRawPtr() 
// purpose : retrieve void ptr to the start of a UUTAObj's raw data 
//           block in memory
void *tiv_getObjRawPtr(UUTAObj *uObj_p)
{
  return uObj_p->uuta.aVoid_p;
} // end::tiv_getObjRawPtr()


// --------------------------------------------------------------------
// function : tiv_SetRawIntoObj()
// purpose  : copy-assign nElements from srcData_p, of dataType at offset SrcSlot, into a UUTAObj destination.
void tiv_SetRawIntoObj(UUTAObj *uObj_p,void *srcData_p,const int32 dataType,const size_t srcSlot,const size_t nElements)
{
 TIV_ATOM_PTR *tiDest_p = NULL;
 uint8 *ptr_p  = (uint8 *)srcData_p;
 size_t sizOfOne   = DFKNTsize(dataType);
 size_t nBytes   = sizOfOne * nElements;
 size_t srcOffset = sizOfOne * srcSlot;

  // make sure destination object's datatype matches that of the source
  assert(uObj_p->uDatatype == dataType);
  // make sure destination object can accomodate this many elements
  assert(uObj_p->uCount <= nElements);

  tiDest_p = &uObj_p->uuta ; 
  uObj_p->uDatatype = dataType ; // reasign parent Containers datatype to reinforce what we've done
  
  memcpy(tiDest_p->aUint8_p, ptr_p+srcOffset, nBytes);
  
} // end::tiv_setRawIntoObj()



// --------------------------------------------------------------------
// function : tiv_SetObjToObj()
// Purpose  : copy-retrieve nElements from a SOURCE UUTAObj 'srcUObj_p' into the
//          base of DESTINATION UUTAObj 'destUObj_p' at destSlot (usually 0), 
//          starting at source position 'atSlot'
// Inputs   : UUTAObj *destUObj_p -- ptr to receiving object,
//            UUTAObj *srcUObj_p -- ptr to source object,
//            const size_t destSlot -- base0 index in destination to place 1st info at
//            const size_t atSlot   -- base0 index in source vector to start transfer at
//            const size_t nElements-- n. of elements of 'datatype' to copy
//
void *tiv_SetObjToObj(UUTAObj *destUObj_p, UUTAObj *srcUObj_p,const size_t destSlot,const size_t atSlot,const size_t nElements)
{
 uint8 *ptr_p  = (uint8 *)srcUObj_p->uuta.aUint8_p;
 size_t sizOfOne   = DFKNTsize(srcUObj_p->uDatatype);
 size_t nBytes = sizOfOne * nElements;
 size_t srcOffset = sizOfOne * atSlot;

  // make sure source supplying the data has at least this many elements! 
  assert(srcUObj_p->uCount >= nElements);
  // make sure destination can accomodate this many elements
  assert(destUObj_p->uCount <= nElements);
  destUObj_p->uDatatype = srcUObj_p->uDatatype ; // reasign parent Containers datatype to reinforce what we've done
  destUObj_p->uCount    = nElements; // update destinations new n. elements, must be <= its own uCount factor
  
  memcpy(destUObj_p->uuta.aUint8_p+destSlot, ptr_p+srcOffset, nBytes);
  
  // return direct pointer to modified data
  return destUObj_p->uuta.aUint8_p ;
} // end::tiv_SetObjToObj()



// --------------------------------------------------------------------
// tiv_SetRawToObjInside()
// Purpose: From source obj "srcData_p" (e.g. myFlt32Array_p) whose datatype matches uObj_p's, 
//          copy-assign nElements starting at source position "srcSlot"
//          into the designated base0 index (destSlot) in a pre-initialized destination object, UUTAObj.
void *tiv_SetRawToObjInside(UUTAObj *uObj_p,void *srcData_p,const size_t destSlot,const size_t srcSlot, const int32 dataType,const size_t nElements)
{
 uint8 *ptr_p      = (uint8 *)srcData_p;
 size_t sizOfOne   = DFKNTsize(uObj_p->uDatatype);
 size_t nBytes = sizOfOne * nElements;
 size_t srcOffset = sizOfOne * srcSlot;
 size_t dstOffset = sizOfOne * destSlot;

 // make sure scalar or array object srcData_p datatype matches that of receiving uObj_p
 assert(uObj_p->uDatatype == dataType);
 // make sure destination uObj_p has been sufficiently sized to receive at least nElements
 assert(uObj_p->uCount <= nElements);
 
 // perform the block copy-assign here, offsetting into both the source and dest buffer
 memcpy(uObj_p->uuta.aUint8_p+dstOffset, ptr_p + srcOffset, nBytes);
 
 uObj_p->uDatatype = dataType;
 uObj_p->uCount    = nElements;

  return uObj_p->uuta.aUint8_p;
} // end::tiv_setIntoUUTAObjDest()


// --------------------------------------------------------------------
// tiv_setUUTAObjDatatype()
void tiv_setUUTAObjDatatype(UUTAObj *uObj_p,const int32 newDatatype)
{
  uObj_p->uDatatype = newDatatype;
} // end::tiv_setUUTAObjDatatype()

// --------------------------------------------------------------------
// tiv_getUUTAObjDatatype()
int32 tiv_getUUTAObjDatatype(UUTAObj *uObj_p)
{
  return uObj_p->uDatatype ;
} // end::tiv_getUUTAObjDatatype()


// --------------------------------------------------------------------
// tiv_setUUTAObjCount()
void tiv_setUUTAObjCount(UUTAObj *uObj_p,const size_t newCount)
{
  uObj_p->uCount = newCount;
} // end::tiv_setUUTAObjCount()

// --------------------------------------------------------------------
// tiv_getUUTAObjCount()
size_t tiv_getUUTAObjCount(UUTAObj *uObj_p)
{
  return uObj_p->uCount ;
} // end::tiv_getUUTAObjCount()

// --------------------------------------------------------------------
// tiv_setUUTAObjLabel()
void tiv_setUUTAObjLabel(UUTAObj *uObj_p,const char *newLabel)
{
  strcpy( uObj_p->uLabel,newLabel);
} // end::tiv_setUUTAObjLabel()

// --------------------------------------------------------------------
// tiv_getUUTAObjLabel()
char *tiv_getUUTAObjLabel(UUTAObj *uObj_p)
{
  return uObj_p->uLabel ;
} // end::tiv_getUUTAObjLabel()


#ifdef FRM_TIV_UNIT_TEST
// --------------------------------------------------------------------
int main(int argc,char *argv[])
{
 size_t nRows = 2;
 size_t nCols = 4;
 size_t i,ithRow,ithCol;
 size_t nValues = nRows * nCols;
 size_t serialIdx =0;
 size_t nBytes = 0; 
 
 float32 varFlt32 = -9999.0; // example of a scalar float32
 char versTagBuffer[64];
 
 time_t strtSec = time(0L);
 time_t endiSec = time(0L);
 
 TIV_ATOM_PTR tiSrc ;
 TIV_ATOM_PTR tiDest;
 
  // STEP 1: etablish the SOURCE data, and populate it
  // allocate memory for a 2x4 array of DFNT_FLOAT32 values
  FRM_API_RANDOMIZE();
  nBytes = (DFKNTsize(DFNT_FLOAT32) * nValues);
  
  fprintf(stdout,"\n%s started %s\n",argv[0], ctime(&strtSec));

  tivTestIdToNames();

  // Allocate several UUTA's, sized to ultimately accomodate float32's
  // as long as nBytes is correct, actual receiving type can be any member of TIV_ATOM_PTR
    tiSrc.aUint8_p  =(uint8 *)calloc(sizeof(uint8),nBytes) ;
    tiDest.aUint8_p =(uint8 *)calloc(sizeof(uint8),nBytes) ;
  fprintf(stdout,"Buffer addresses (start): Src %p Dest %p\n",tiSrc.aUint8_p,tiDest.aUint8_p);
  
  fprintf(stdout,"\nTEST copyUUTA() - contents of Source (nRows %zu nCols %zu)\n",nRows,nCols);
  for(ithRow=0;ithRow < nRows; ithRow++)
    for(ithCol=0; ithCol < nCols; ithCol++)
    {
      serialIdx = (ithRow *nCols) + ithCol;
      tiSrc.aFlt32_p[serialIdx] = FRM_API_RANDOM_PROB()*10.0;
      fprintf(stdout,"R %3zu C %3zu serialIdx %zu --> %.4f\n",
            ithRow,ithCol, serialIdx,tiSrc.aFlt32_p[serialIdx] );
    } // end::ithCol loop
  // end::ithRow loop
  
  // Execute a UUTA style copy. Offset of '0',with (nRows*nCols) indicates COPY ALL
  // tiv_copyUUTA(&tiDest, &tiSrc, DFNT_FLOAT32, 0, (nRows * nCols));
  // tiv_CopyRawFromSlot(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot,const size_t nElements);
  tiv_CopyRawFromSlot(&tiDest, &tiSrc, DFNT_FLOAT32, 0, (nRows *nCols));

  fprintf(stdout,"\nPOST TEST copyUUTA() - contents of Dest  \n");
  for(ithRow=0;ithRow < nRows; ithRow++)
    for(ithCol=0; ithCol < nCols; ithCol++)
    {
      serialIdx = (ithRow *nCols) + ithCol;
      fprintf(stdout,"R %3zu C %3zu serialIdx %zu --> %.4f\n",
            ithRow,ithCol, serialIdx,tiDest.aFlt32_p[serialIdx] );
    } // end::ithCol loop
  // end::ithRow loop

    // TEST 2: demonstrate a setIntoUUTA() operation
    fprintf(stdout,"\n TEST 2: setIntoUUTA() \n");
    for(ithRow=0;ithRow < nRows; ithRow++)
    {
        for(ithCol=0; ithCol < nCols; ithCol++)
        {
            serialIdx = (ithRow *nCols) + ithCol;
            varFlt32 = FRM_API_RANDOM_PROB() *100.0;
            // DEPRECATED: 
            // tiv_setIntoUUTA(&tiDest, &varFlt32, DFNT_FLOAT32, serialIdx);
            // NEW: note the 'atSlot' refers to the destinary array slot!
            // void *tiv_SetOneIntoRawSlot(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot)
            tiv_SetOneIntoRawSlot(&tiDest, &varFlt32, DFNT_FLOAT32, serialIdx);
            
            // now, dereference the UUTA via its actual datatype
            fprintf(stdout,"tiv_SetOneIntoRawSlot: R %3zu C %3zu  UUTA tiDest(%zu) %.4f\n",
                ithRow,ithCol, serialIdx, tiDest.aFlt32_p[serialIdx] );
        } // end::ithCol loop
    } // end:: ithRow loop
  
    // TEST 3: demonstrate a getFromUUTA() operation
    fprintf(stdout,"\n TEST 3: getfromUUTA() \n");
    for(ithRow=0;ithRow < nRows; ithRow++)
    {
        for(ithCol=0; ithCol < nCols; ithCol++)
        {
            serialIdx = (ithRow *nCols) + ithCol;
            varFlt32 = FRM_API_RANDOM_PROB() *100.0;
            // assign a single value, to 'varFlt32', FROM a source UUTA named tiSrc
            // DEPRECATED: tiv_getFromUUTA(&varFlt32, &tiSrc,  DFNT_FLOAT32, serialIdx);
            // NEW: 
            // void *tiv_GetOneFromArray(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot)
            tiv_GetOneFromArray(&varFlt32, &tiSrc, DFNT_FLOAT32, serialIdx);
            
            // now, dereference the UUTA via its actual datatype
            fprintf(stdout,"getFromUUTA: R %3zu C %3zu UUTA(%zu) varFlt32 = %.4f\n",
                ithRow,ithCol, serialIdx, varFlt32);
        } // end::ithCol loop
    } // end:: ithRow loop
  
  // TEST 4: demonstrate "any-to-any" UUTA to UUTA copy. This examples COPIES EVERYTHING
  // from one UUTA to another.
  fprintf(stdout,"\n TEST 4: copy_UUTA_to_UUTA() now as tiv_CopyRawToRaw \n");
  fprintf(stdout,"\n now named: tiv_CopyRawToRaw\n");
  
  tiv_CopyRawToRaw(&tiDest, &tiSrc, DFNT_FLOAT32, 0,0, nValues);
  
  for(i=0; i < nValues; i++)
  {
            // now, dereference the UUTA via its actual datatype
            fprintf(stdout,"getFromUUTA: R %3zu C %3zu UUTA tiDest(%zu) = %.4f\n",
                ithRow,ithCol, serialIdx, tiDest.aFlt32_p[i]);
  } // end:: i..nValues loop

  fprintf(stdout,"\n TEST 5: ROUND-TRIP copy_UUTA_to_UUTA now as tiv_CopyRawToRaw \n");
  
  // now we copy it back
  // tiv_copyUUTA_to_UUTA(&tiSrc,  &tiDest, DFNT_FLOAT32, 0, 0, nValues);
  // void *tiv_CopyRawToRaw(void *vDest_p,void *vSource_p,const int32 dataType,
  //            const size_t srcSlot,const size_t destSlot, size_t nElements);
  // PRIOR: was tiv_copyUUTA_to_UUTA() now: CopyRawtoRaw
  tiv_CopyRawToRaw(&tiDest, &tiSrc, DFNT_FLOAT32, 0,0, nValues);
  
  for(i=0; i < nValues; i++)
  {
      // tiv_GetOneFromArray(void *vDest_p,void *vSource_p,const int32 dataType,const size_t atSlot)
      
            // now, dereference the UUTA via its actual datatype
            fprintf(stdout,"getFromUUTA: R %3zu C %3zu UUTA tiSrc(%zu) = %.4f\n",
                ithRow,ithCol, serialIdx, tiSrc.aFlt32_p[i]);
  } // end:: i..nValues loop
 
  // free up memory, noting that ANY member will do since they all point to the same block of memory
  free(tiDest.aUint8_p);
  free(tiSrc.aUint8_p) ;
  endiSec = time(0L);
  
  // display FRM API version tag info, from frm_core.c too
  frm_GetApiVersion(versTagBuffer);
  
  fprintf(stdout,"\n");
  fprintf(stdout,"@A FRM API Version-Release Tag %s\n",versTagBuffer );
  fprintf(stdout,"@A%s completed %s\n",argv[0], ctime(&endiSec));

 return 0;
} // end::main()

// end:: block if FRM_TIV_UNIT_TEST
#endif

