// ---------------------------------------------------------------
// module       : nsb_accessors.c --> key,value,property "GET|SET" accessors
// purpose      : Implements replicable (thread-safe) NVP "string-block" get|set accessors
// project      : Lupine Logic Inc, implements generalized name-value-pair stringblock 
//                in-memory database capability.
// dependencies : n/a
// revision     : 2017-04-26T11:00, review behavior of code for "!" macro
//                expansion, e.g. where a "!" prefacing a keyword triggers 
//                expansion of a file that itself contains an arbitrary number
//                of records (usually file paths) to be stored in a stringtable.
//                Note: stack variable keyword[1024] changed to 1024 from 512, but
//                this usage of local stack variables for keyword[] needs
//                a more robust implementation!
// revision     : 2015-09-24T11:23, in nsbGetgMultiValueByKey(), started 
//              : prototyping "!" macro expansion syntax, where a KEYWORD is
//                prefaced with a '!', allowing a single KEY's value to be
//                a path to ONE text file that itself contains multiple items
//                which will often be paths
//                AND, in nsbGetValuebyKey|byIndex(), now autoconverts any
//                quoted-string output of parsers that substitute backticks
//                for spaces, so that we now properly return the verbatim 
//                quoted string that DOES now contain the original embedded
//                spaces (blanks).
// revision     : 2014-09-03T15:01, added nsbDoesKeyExist() boolean..
// history      : 2014-07-23, pending plan to modify strcmp() match on targetKey to allow a
//                "partial key" to match multiple RIP nvp items.
// history      : 2014-02-20T10:28:00, adjust GetValueByKey error to @E if key missing
// history      : 2014-02-09T16:00
// history      : 2014-01-06T10:27:00,jmg --changed to use xxLogMsg() now
// history      : 2014-01-03T13:32:00,jmg --refactored API to provide for 
//                multiple, concurrent, thread-safe of a NVP stringblock instances.
//              : 2013-12-20T12:22:00, pending, eliminated all SLL dependencies
//              : 2013-12-20T02:00:00,milestoned version
//              : 2013-11-19T12:33:00-07:00, original code. 
// author       : joe glassy
// notes        : 
// ---------------------------------------------------------------

#include "nsb_api.h"
#include "nsb_api_prot.h"

// ---------------------------------------------------------------
// function  : nsbSetSupportedKeyPolicy()
// purpose   : assigns a nsbKeyPolicy option, where 0= no supported 
//             keyword table is needed or will be used, and
//             if nsbKeyPolicy==1, a supported keyword table IS required.
// inputs    : NsbObj *nsb_p,const int ithElement,const char *keyword
// returns   : void
// notes     : n/a
void nsbSetSupportedKeyPolicy(NsbObj *nsb_p,const int newKeyPolicy)
{
  nsb_p->nsbKeyPolicy = newKeyPolicy;
} // end::nsbSetSupportedKeyPolicy()


// ---------------------------------------------------------------
// function  : nsbGetSupportedKeyPolicy()
// purpose   : assigns a nsbKeyPolicy option, where 0= no supported 
//             keyword table is needed or will be used, and
//             if nsbKeyPolicy==1, a supported keyword table IS required.
// inputs    : NsbObj *nsb_p,const int ithElement,const char *keyword
// returns   : int, value of nsbKeyPolicy as 0(not required) or 1(is required)
// notes     : n/a
int nsbGetSupportedKeyPolicy(NsbObj *nsb_p)
{

  return nsb_p->nsbKeyPolicy ;
  
} // end::nsbGetSupportedKeyPolicy()


// ---------------------------------------------------------------
// function  : nsbDoesKeyExist()
// purpose   : Tests if 'targetKey_p' keyword exists, returning boolean
// inputs    : const char *targetKey_p,char *valueBuffer
// returns   : int, return boolean: 1=key DOES exist, 0=key DOES NOT exist
// revision  : 2014-09-03T14:57, added, was missing earlier
// notes     : n/a
int nsbDoesKeyExist(NsbObj *nsb_p,const char *targetKey_p)
{
 size_t v =0;
 size_t byteOffsetToKey   = 0 ; 
 size_t byteOffsetToValue = 0 ; 
 int  boolDoesKeyExist  = 0 ; // set to 1 when key exists, 0= key DOES NOT exist
 char keyword[1024] = {0} ;
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
 
    for(v=0; v < nsb_p->nsbActualCount; v++)
    {
        byteOffsetToKey = NSB_OFFSET_TO_KEY(v, nsb_p->nsbRecsize);
        
        // offset --STEP PAST the (properties+keyword) to fetch just the value string
        byteOffsetToValue = NSB_OFFSET_TO_VAL(v, nsb_p->nsbRecsize,
                                                nsb_p->nsbKeyMaxwidth);
        
        // fetch key from in-memory database, at this record, to test
        memcpy(keyword, nsb_p->nsbCoreStringblock_p + byteOffsetToKey, nsb_p->nsbKeyMaxwidth);
        // Force skipping past any @schema(), @instance() statements
        // then Scan for a match between fetched keyword and targetKey_p
        if(strstr(targetKey_p,"@(")== NULL && strcmp(targetKey_p,keyword) == 0 )
        {
            boolDoesKeyExist = 1 ;
            break;
        } // end:: block if target KEY matches key in record...
        
    } // end:: v iteration through list of all NVPs to find keyword

    // return boolean: 1=key DOES exist, 0=key DOES NOT exist
    return boolDoesKeyExist;

} // end::nsbDoesKeyExist()


// ---------------------------------------------------------------
// function  : nsbRetrieveKey()
// purpose   : retrieves a keyword at the ithElement stringblock record
// inputs    : NsbObj *nsb_p,const int ithElement,const char *keyword
// returns   : char *keyword
// notes     : n/a
char *nsbRetrieveKey(NsbObj *nsb_p,const int ithElement,char *keyword)
{
    size_t byteOffsetToKey = 0 ;
    size_t keyLen = strlen(keyword);
    
    memset(keyword,'\0',nsb_p->nsbKeyMaxwidth);
    
    byteOffsetToKey = NSB_OFFSET_TO_KEY(ithElement, nsb_p->nsbRecsize);
    
    // Insert this latest keyword AND valueToken into the master stringblock
    // note: later we'll perform this via a simple function...
    memcpy(keyword, nsb_p->nsbCoreStringblock_p + byteOffsetToKey, keyLen);
    // return ptr
    return keyword;
    
} // end::nsbRetrieveKey()


// ---------------------------------------------------------------
// function  : nsbRetrieveValue()
// purpose   : retrieves a value from the ithElement stringblock record
// inputs    : NsbObj *nsb_p,const int ithElement,const char *valueStr
// returns   : status, length of value string
// notes     : n/a
char *nsbRetrieveValue(NsbObj *nsb_p,const int ithElement,char *valueStr)
{
    size_t byteOffsetToValue = 0 ;
    size_t valLen = strlen(valueStr);
    
    byteOffsetToValue = NSB_OFFSET_TO_VAL(ithElement, nsb_p->nsbRecsize,
                            nsb_p->nsbKeyMaxwidth);
    
    // LOAD the value part of payload
    memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToValue, valueStr,valLen);
    
    // returns value string
    return valueStr;
} // end::nsbRetrieveValue()


// ---------------------------------------------------------------
// function  : nsbStoreKey()
// purpose   : assigns a keyword to the ithElement stringblock record
// inputs    : NsbObj *nsb_p,const int ithElement,const char *keyword
// returns   : status, length of keyword
// revised   : 2015-09-28T17:11:00
// notes     : TBD: need to auto-restrict ALL LOAD operations to the
//             maximum allowable keyword length; currently the user
//             COULD inadvertantly overwrite a keyword slot with a 
//             too-wide keyword...
int nsbStoreKey(NsbObj *nsb_p,const int ithElement,const char *keyword)
{
    size_t byteOffsetToKey = 0 ;
    size_t keyLen = strlen(keyword);
    char *pWork_p = NULL;
    
    byteOffsetToKey = NSB_OFFSET_TO_KEY(ithElement, nsb_p->nsbRecsize);
         
#ifdef READY_TO_IMPLEMENT
    if(keyLen > nsb_p->nsbKeyMaxwidth)
    {
       // add 1 for null termination of temporary working buffer  
       pWork_p = calloc(nsb_p->nsbKeyMaxwidth+1,sizeof(char));
       // AUTO-LIMIT copy to max of nsbKeyMaxwidth chars
       strncpy(pWork_p, keyword, nsb_p->nsbKeyMaxwidth);
       pWork_p[nsb_p->nsbKeyMaxwidth] = '\0' ; // explicitly null terminate
       // Finally, complete the LOAD, using the temporary (clipped) buffer
       memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToKey, pWork_p, keyLen);
       free(pWork_p); // free up the temporary buffer
    } else {
       // LOAD the validated length KEY part of payload
       memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToKey, keyword, keyLen);      
    }
#endif

    // LOAD the keyword part of the payload
    memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToKey, keyword, keyLen);
    
    return keyLen;
    
} // end::nsbStoreKey()


// ---------------------------------------------------------------
// function  : nsbStoreValue()
// purpose   : assigns a value(string) to the ithElement stringblock record
// inputs    : NsbObj *nsb_p,const int ithElement,const char *valueStr
// returns   : status, length of value string
// revised   : 2015-09-28T17:11:00
// notes     : TBD, auto-limit to max value width, using strncpy() or equiv
//             method to refuse overwriting past end of allowed value width
int nsbStoreValue(NsbObj *nsb_p,const int ithElement,const char *valueStr)
{
    size_t byteOffsetToValue = 0 ;
    size_t valLen = strlen(valueStr);
    char *pWork_p = NULL;
    
    byteOffsetToValue = NSB_OFFSET_TO_VAL(ithElement,
                        nsb_p->nsbRecsize,nsb_p->nsbKeyMaxwidth);
    
    // TEST throughly prior to final implementation
#ifdef READY_TO_IMPLEMENT    
    if(valLen > nsb_p->nsbValMaxwidth)
    {
       // add 1 for null termination of temporary working buffer  
       pWork_p = calloc(nsb_p->nsbValMaxwidth+1,sizeof(char));
       // AUTO-LIMIT copy to max of nsbValMaxwidth chars
       strncpy(pWork_p, valueStr,nsb_p->nsbValMaxwidth);
       pWork_p[nsb_p->nsbValMaxwidth] = '\0' ; // explicitly null terminate
       // finally, complete the LOAD, using the temporary (clipped) buffer
       memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToValue, pWork_p,valLen);      
       free(pWork_p); // free up the temporary buffer
    } else {
      // LOAD the value part of payload
      memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToValue, valueStr,valLen);
    }
#endif    
    
    // LOAD the value part of payload
    memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToValue, valueStr,valLen);
    
    // returns lenght of value string stored
    return valLen;
} // end::nsbStoreValue()


// ---------------------------------------------------------------
// function  : nsbStoreKeyAndValue()
// purpose   : assigns a keyword and its associated value 
//             to the ithElement stringblock record
// inputs    : NsbObj *nsb_p,const int ithElement,const char *keyword,const char *valueStr
// returns   : status, length of keyword+value
// revised   : 2015-09-28T17:11:00
// notes     : TBD convert to use nsbStoreKey() and nsbStoreValue() since
//             these will soon properly AUTO-RESTRICT too-width entries...
int nsbStoreKeyAndValue(NsbObj *nsb_p,const int ithElement,const char *keyword,const char *valueStr)
{
    size_t byteOffsetToValue = 0 ;
    size_t byteOffsetToKey   = 0 ;
    size_t keyLen = strlen(keyword);
    size_t valLen = strlen(valueStr);
    char fxName[256] = {0} ;    
    
    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 
    
    byteOffsetToKey = NSB_OFFSET_TO_KEY(ithElement, nsb_p->nsbRecsize);
    
    memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToKey, keyword, keyLen);

    byteOffsetToValue = NSB_OFFSET_TO_VAL(ithElement, nsb_p->nsbRecsize,
                                            nsb_p->nsbKeyMaxwidth);
    
    memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToValue, valueStr,valLen);
    
    return keyLen + valLen ;
} // end::nsbStoreKeyAndValue()


// ---------------------------------------------------------------
// function  : nsbClearValueByKey()
// purpose   : ERASES or clears an NVP value, targetKey_p string, at the 
//             base0 index position where the 'targetKey_p' exists
//             in the separate string dictionary naming all supported keys.
// revised   : 2015-09-30, added documentation
// notes     : This erases (clears) only the VALUE string assoc. with the
//             indicated keyword. It RETAINS the entry and the KEY defining the entry or record.
// notes     : LIMITATION: this method only works for single-valued NVPs,
//             it does NOT work (yet) for multi-valued keys, since each
//             key appears only ONCE in the supported key dictionary!
int nsbClearValueByKey(NsbObj *nsb_p,const char *targetKey_p)
{
 size_t v =0;
 size_t byteOffsetToValue = 0 ; 
 
 char fxName[256] = {0} ;
 int atIndex = 0 ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 

    for(v=0; v < nsb_p->nsbActualCount; v++)
    {
        // fetch key from SUPPORTED KEYWORD dictionary, and if present, 
        // CLEAR the value-string at this record but RETAINS the keyword
        if((atIndex = nsbSearchForSupportedKey(nsb_p,targetKey_p)) != -1)
        {
            // be sure to step PAST key, to start of VALUE string itself
            byteOffsetToValue = NSB_OFFSET_TO_VAL(atIndex, nsb_p->nsbRecsize,
                                            nsb_p->nsbKeyMaxwidth);
                        
            memset(nsb_p->nsbCoreStringblock_p + byteOffsetToValue,
                    '\0', nsb_p->nsbValMaxwidth);
            break;
        } // end:: block v if keyword IS SUPPORTED
    } // end:: v iteration through list of all NVPs to find keyword
    
    if(v >= nsb_p->nsbActualCount)
    {
        xLogMsg("@F In(%s): failed on key(%s)! ~\n",fxName,targetKey_p);
    }
   
   return 0;
} // end::nsbClearValueByKey()


// ---------------------------------------------------------------
// function  : nsbClearValueByIndex()
// purpose   : clears an NVP value, at base0 index ithNvp, while retaining
//             the current keyword.
// inputs    : const int ithNvp,char *valueBuffer
// returns   : int, 
// notes     : This erases (clears) only the VALUE string assoc. with the
//             indicated keyword, so it retains the keyword itself.
// notes     : Assuming nTotalNvp is a count of ALL (e.g. both unique
//              and non-unique keyword NVPs, this function works for both
//              single-valued NVPs and multi-valued NVPs.
int nsbClearValueByIndex(NsbObj *nsb_p,const int ithNvp)
{
 int status = -1 ;
 
 size_t byteOffsetToKey   = 0 ; 
 size_t byteOffsetToValue = 0 ; 
 char fxName[256] = {0} ;

    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 
 
    // offset to fetch just the keyword...
    byteOffsetToKey = NSB_OFFSET_TO_KEY(ithNvp, nsb_p->nsbRecsize);
 
   if(ithNvp < (int)nsb_p->nsbActualCount)
   {
        // offset --remember to also STEP PAST the keyword to fetch JUST the value string!
        byteOffsetToValue = NSB_OFFSET_TO_VAL(ithNvp, nsb_p->nsbRecsize,
                            nsb_p->nsbKeyMaxwidth);

        // preemptively CLEAR the value string in the database
        memset(nsb_p->nsbCoreStringblock_p + byteOffsetToValue,
            '\0', nsb_p->nsbValMaxwidth);
   } else {
       xLogMsg("@F In(%s): FATAL-ERROR invalid NVP index %d\n",fxName,ithNvp);
   }
   
  // return status, as value's string length or -1 if failed
  return status;
    
} // end::nsbClearValueByIndex()


// ---------------------------------------------------------------
// function  : nsbGetValueByKey()
// purpose   : retrieves an NVP value, searched by targetKey_p string
// inputs    : const char *targetKey_p,char *valueBuffer
// returns   : char *,string 'value' mapped to 1st occurrence of 'targetKey_p' string
// revision  : 2015-09-24T12:20, autoconvert any quoted-string output of
//             parser that substitutes backticks for spaces, so that we return
//             the verbatim quoted string that DOES now contain the original
//             embedded spaces (blanks).
// revision  : 2014-02-20T10:28:00, adjusted error level to @E if key is missing
// notes     : LIMITATION! this method only currently works for single-valued
//             NVPs; it does NOT work for multi-valued collections of 
//             NVPs whose elements all share a common keyword.
char *nsbGetValueByKey(NsbObj *nsb_p,const char *targetKey_p,char *valueBuffer)
{
 size_t v =0;
 size_t byteOffsetToKey   = 0 ; 
 size_t byteOffsetToValue = 0 ; 
 char keyword[1024] = {0} ;
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    // clear out the yet-to-be-returned value string first
    memset(valueBuffer,'\0', nsb_p->nsbValMaxwidth);
 
    for(v=0; v < nsb_p->nsbActualCount; v++)
    {
        byteOffsetToKey = NSB_OFFSET_TO_KEY(v, nsb_p->nsbRecsize);
        
        // offset --STEP PAST the (properties+keyword) to fetch just the value string
        byteOffsetToValue = NSB_OFFSET_TO_VAL(v, nsb_p->nsbRecsize,
                                                nsb_p->nsbKeyMaxwidth);
        
        // fetch key from in-memory database, at this record, to test
        memcpy(keyword, nsb_p->nsbCoreStringblock_p + byteOffsetToKey, nsb_p->nsbKeyMaxwidth);
        
        // Force skipping past any @schema(), @instance() statements
        // then Scan for a match between fetched keyword and targetKey_p
        if(strstr(targetKey_p,"@(")== NULL && strcmp(targetKey_p,keyword) ==0 )
        {
            // retrieve value string at the record matching the targetKey_p string
            memcpy(valueBuffer, nsb_p->nsbCoreStringblock_p + byteOffsetToValue, nsb_p->nsbValMaxwidth);
            // take the early out
            break;
        } // end:: block if target KEY matches key in record...
        
    } // end:: v iteration through list of all NVPs to find keyword
    
    // <JMG> 2015-09-24 auto convert parser output where a quoted string
    //  was encountered and backticks were substituted for spaces, SO, here
    //  we convert all backticks BACK to spaces prior to returning the
    //  value string.
    if(strchr(valueBuffer,'`')!=NULL)
    {
        frm_strReplaceChr("`",valueBuffer," ") ;
    }
    
    if(v >= nsb_p->nsbActualCount)
    {
        xLogMsg("@E In(%s): failed on key(%s)! ~\n",fxName, targetKey_p);
    }
   
   // return ptr to value string, or empty string if failed
   return valueBuffer;
    
} // end::nsbGetValueByKey()

// ---------------------------------------------------------------
// function  : nsbGetPropertyValueByKey()
// purpose   : Using keyword and a base0 index into a comma-delimited 
//             string that is a NVP's "value", this retrieves the 
//             sub-string value that is the ith item associated with the
//            'targetKey_p' KEYWORD string from an NSB database.
// inputs    : NsbObj *nsb_p -- ptr to the NSB database container.
//             const char *targetKey_p -- ptr to the KEYWORD
//             const int propOrderIdx  -- integer base0 index of the item in the
//             a comma-delimited list of variable properties. Typically
//             an enum member is used as this index. If propOrderIdx 
//             references an item out-of-range, an empty string is returned
//             and a error message is emitted.
//             char *valueBuffer -- ptr to the returned 'value' string
// returns   : char *, ptr to the return value string.
// notes     : Currently, comma-delimiting MUST separate the sub-values
//             of a supported compound-value string with no embedded blanks, e.g.
//             KEYWORD = Name,subvalueA,subvalueB,subValueC,etc
//              
// notes     : Example of a compound 'value' string, always starting with
//             the name: 
//    L4C_OUTPUT_DATASET_NEE_MEAN = NEE/nee_mean,"units",-9999.0,0,10000.0,My-Long-Name
//
char *nsbGetPropertyValueByKey(NsbObj *nsb_p,const char *targetKey_p,const int propOrderIdx,char *resultBuffer)
{
 char fxName[256] = {0} ;
 char privateCompoundValue[FILENAME_MAX] = {0} ; // input data is stored here...
 int  tokenCount = 0 ; // count of comma-delimited sub-properties in compound NVP string
 
    strcpy(fxName,__func__);
    strcpy(resultBuffer,"");
    
    if( nsbDoesKeyExist(nsb_p,targetKey_p) == 1)
    {
        // Using 'targetKey_p' store 'VALUE' result in privateCompoundValue string
        nsbGetValueByKey(nsb_p,targetKey_p,privateCompoundValue);
        // Fetch the count of "sub-value" delimited tokens within the
        //  privateCompoundValue string to verify this is an actionable request
        tokenCount = frm_strCountTokens(privateCompoundValue,",");
        // Only when the requested token's INDEX is less than tokenCount is this
        // an actionable request
        if(propOrderIdx < tokenCount)
        {
            // Load the resultBuffer with the contents of the 
            // 'ith' comma-delimited item as indicated by index 'propOrderIdx'
            frm_strGetToken(resultBuffer, privateCompoundValue, propOrderIdx,",");
        } else {
            xLogMsg("@E In(%s) Key(%s) Only %d tokens present. Invalid index(%d) requested! ~\n",
                fxName, targetKey_p, tokenCount, propOrderIdx);
        }
    }
    return resultBuffer;
} // end::nsbGetPropertyValueByKey()


// ---------------------------------------------------------------
// function  : nsbGetPropertyNumericByKey()
// purpose   : using keyword and a base0 index into a comma-delimited string
//             that is the 'value', this retrieves a float64 value associated
//             with 'targetKey_p' KEYWORD string from an NSB database.
// inputs    : NsbObj *nsb_p -- ptr to the NSB database container.
//             const char *targetKey_p -- ptr to the KEYWORD
//             const int propOrderIdx  -- integer base0 index of the item in the
//             a comma-delimited list of variable properties. Typically
//             an enum member is used as this index.
//             float *resultFloat64 -- ptr to the returned float64 value.
// returns   : float64, value as requested 
// notes     : Example of a compound 'value' string, always starting with
//             the name: 
//    L4C_OUTPUT_DATASET_NEE_MEAN = NEE/nee_mean,a,b,c,d,e,f
//
float64 nsbGetPropertyNumericByKey(NsbObj *nsb_p,const char *targetKey_p,const int propOrderIdx,float64 *resultFloat64_p)
{
 char fxName[256] = {0} ;
 char privateCompoundValue[FILENAME_MAX] = {0} ;
 float64 float64Value = 0.0 ;
 
 int  tokenCount = 0 ; // count of comma-delimited sub-properties in compound NVP string
 
    strcpy(fxName,__func__);
    float64Value = -99999999.0 ;
    
    if( nsbDoesKeyExist(nsb_p,targetKey_p) == 1)
    {
        nsbGetValueByKey(nsb_p,targetKey_p,privateCompoundValue);
        // fetch count of tokens to verify this is an actionable request
        tokenCount = frm_strCountTokens(privateCompoundValue,",");
        if(propOrderIdx < tokenCount)
        {
            // Load the resultBuffer with the contents of the 
            // 'ith' comma-delimited item as indicated by index 'propOrderIdx'
            float64Value = frm_strGetTokenAsFloat64(resultFloat64_p,privateCompoundValue,propOrderIdx,",");
        } else {
            xLogMsg("@E In(%s) Key(%s) Only %d tokens present. Invalid index(%d) requested! ~\n",
                fxName, targetKey_p, tokenCount, propOrderIdx);
        }
    }
    return float64Value;
} // end::nsbGetPropertyNumericByKey()




// ---------------------------------------------------------------
// function  : nsbGetNumericByKey()
// purpose   : retrieves (by key)an NVP value as a float64 
// inputs    : const char *targetKey_p,
//             float64 *valFlt64_p -- ptr to callers float64 to update too
// returns   : float64, numeric value
// notes     : Retrieves (by string keyword) a name-value-pair (string) value 
//          as a numeric float64 which can be cast by caller to any numeric type, 
//          like uint8,int16,uint32 etc.
//
float64 nsbGetNumericByKey(NsbObj *nsb_p,const char *targetKey_p,float64 *valFlt64_p)
{
 char valueBuffer[128] = {0} ;
 float64 resultFlt64 =0.0;
 
    // fetch value ASSUMED to be a numeric value convertible by atof()
    resultFlt64 = atof(nsbGetValueByKey(nsb_p,targetKey_p,valueBuffer));
    // also post the converted value as an OUTPUT valFlt64_p
    *valFlt64_p = resultFlt64 ;
    
 // Return generic numeric result regardless if it's an interger,float32 or float64
 // since we use a "preemptive up-cast" to normalize these intentionally
 return resultFlt64 ;
 
} // end:: nsbGetNumericByKey()



// ---------------------------------------------------------------
// function  : nsbGetNumericByIndex()
// purpose   : retrieves (by index)an NVP value as a float64 
// inputs    : const int ithNvp  --base0 index of this NVP in this registry
//             float64 *valFlt64_p -- ptr to callers float64 to update too
// returns   : float64, numeric value
// notes     : Retrieves (by index) a name-value-pair (string) value and  
//             a numeric float64 which can be cast by caller to any 
//             numeric type, like uint8,int16,uint32 etc.
float64 nsbGetNumericByIndex(NsbObj *nsb_p,const int ithNvp,float64 *valFlt64_p)
{
 char valueBuffer[128] = {0} ;
 float64 resultFlt64 =0.0;
 
    // fetch NVP value ASSUMED to be a numeric value convertible by atof()
    resultFlt64 = atof(nsbGetValueByIndex(nsb_p,ithNvp, valueBuffer));
    // also post the converted value as an OUTPUT valFlt64_p
    *valFlt64_p = resultFlt64 ;

 // return generic numeric result regardless if it's an interger,float32 or float64
 // since we use a "preemptive up-cast" to normalize these intentionally
 return resultFlt64 ;
} // end::nsbGetNumericByIndex


// ---------------------------------------------------------------
// function  : nsbSetValueByKey()
// purpose   : re-assigns an NVP value, targetKey_p string, at the 
//             base0 index position where the 'targetKey_p' exists
//             in the separate string dictonary naming all supported keys.
// notes     : LIMITATION: this method only works for single-valued NVPs,
//             it does NOT work (yet) for multi-valued keys, since each
//             key appears only ONCE in the supported key dictionary!
int nsbSetValueByKey(NsbObj *nsb_p,const char *targetKey_p,char *valueBuffer)
{
 size_t v =0;
 
 size_t byteOffsetToKey   = 0 ; 
 size_t byteOffsetToValue = 0 ; 
 
 char fxName[256] = {0} ;
 char *pWork_p = NULL;
 int atIndex = 0 ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    for(v=0; v < nsb_p->nsbActualCount; v++)
    {
        // fetch key from SUPPORTED KEYWORD dictionary, and if present, load
        // then LOAD VALUE into catalog, at this record
        if((atIndex = nsbSearchForSupportedKey(nsb_p,targetKey_p)) != -1)
        {
            byteOffsetToKey = NSB_OFFSET_TO_KEY(atIndex, nsb_p->nsbRecsize);
            
            // step PAST properties and key, to start of VALUE string itself
            byteOffsetToValue = NSB_OFFSET_TO_VAL(atIndex, nsb_p->nsbRecsize,
                                            nsb_p->nsbKeyMaxwidth);
            
            // TEST throughly prior to final implementation
#ifdef READY_TO_IMPLEMENT
            if(valLen > nsb_p->nsbValMaxwidth)
            {
              // add 1 for null termination of temporary working buffer  
              pWork_p = calloc(nsb_p->nsbValMaxwidth+1,sizeof(char));
              // AUTO-RESTRICT the copy to a maximum of of nsbValMaxwidth chars
              strncpy(pWork_p, valueStr,nsb_p->nsbValMaxwidth);
              pWork_p[nsb_p->nsbValMaxwidth] = '\0' ; // explicitly null terminate the src
              // finally, complete the LOAD, using the temporary (clipped) buffer as source
              memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToValue, pWork_p,valLen);      
              free(pWork_p); // free up the temporary buffer
             } else {
                // LOAD the value part of payload
                memcpy(nsb_p->nsbCoreStringblock_p + byteOffsetToValue,
                          valueBuffer, nsb_p->nsbValMaxwidth);      
             }
#endif    
            
            memcpy(nsb_p->nsbCoreStringblock_p + byteOffsetToValue,
                        valueBuffer, nsb_p->nsbValMaxwidth);
            break;
        } // end:: block v if keyword IS SUPPORTED
    } // end:: v iteration through list of all NVPs to find keyword
    
    if(v >= nsb_p->nsbActualCount)
    {
        xLogMsg("@F nsbSetNvpValueByKey: failed on key(%s)! ~\n",targetKey_p);
        exit(-1);
    }
   
   return 0;
} // end::nsbSetValueByKey()

// ---------------------------------------------------------------
// function  : nsbGetKeywordByIndex()
// purpose   : retrieves an NVP keyword, searched by base0 index, ithNvp
// inputs    : const int ithNvp,char *keyBuffer
// returns   : char *, the NVP "keyword" string mapped to base0 index 'ithNvp'
// notes     : Assuming ithNvp is an index that includes both single and
//              multivalued NPVs, and that nsbMasterStringblockRecordCount is a count
//              of ALL (e.g. both unique
//              and non-unique keyword NVPs, this function works for both
//              single-valued NVPs and multi-valued NVPs.
char *nsbGetKeywordByIndex(NsbObj *nsb_p,const int ithNvp,char *keyBuffer)
{
 
 size_t byteOffsetToKey        = 0 ;    
 
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    // Clear out the KEYWORD string first, also verifies it is 
    //    long enough, as it will force a seg fault here 
    //    if it is of insufficient length
    memset(keyBuffer,'\0', nsb_p->nsbKeyMaxwidth);

   if(ithNvp < (int)nsb_p->nsbActualCount)
   {
        // Offset to fetch just the keyword. This formulation works when the
        // ithNvp index is base0, such that an ithNvp of '4' points to the 5th physical NVP.
        byteOffsetToKey = NSB_OFFSET_TO_KEY(ithNvp, nsb_p->nsbRecsize);
        
        // Insert this latest keyword and valueToken into the master catalog, noting
        // how use of 'nsbKeyMaxwidth' "pushes" the data into a left-alined position
        memcpy(keyBuffer, nsb_p->nsbCoreStringblock_p + byteOffsetToKey,nsb_p->nsbKeyMaxwidth);
        
   } else {
       xLogMsg("@F GetNvpKeyByIndex: FATAL-ERROR invalid NVP index %d\n",ithNvp);
       exit(-1);
   }
   
   // return ptr to keyword string, or NULL if failed
   return keyBuffer;
   
} // end::nsbGetKeywordByIndex()


// ---------------------------------------------------------------
// function  : nsbGetValueByIndex()
// purpose   : retrieves an NVP value, searched by base0 index, ithNvp
// inputs    : const int ithNvp,char *valueBuffer
// returns   : char *, the NVP "value" string mapped to base0 index 'ithNvp'
// notes     : Assuming ithNvp is an index that includes both single and
//              multivalued NPVs, and that nsbMasterStringblockRecordCount is a count
//              of ALL (e.g. both unique
//              and non-unique keyword NVPs, this function works for both
//              single-valued NVPs and multi-valued NVPs.
//
char *nsbGetValueByIndex(NsbObj *nsb_p,const int ithNvp,char *valueBuffer)
{
 
 
 size_t byteOffsetToValue = 0 ; 
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    // clear out the value string first
    memset(valueBuffer,'\0',nsb_p->nsbValMaxwidth);
    
   if(ithNvp < (int)nsb_p->nsbActualCount)
   {
    // offset --STEP PAST the (properties+keyword) to fetch just the value string
    byteOffsetToValue = NSB_OFFSET_TO_VAL(ithNvp, nsb_p->nsbRecsize,
                                nsb_p->nsbKeyMaxwidth);
        
        // insert this latest keyword and valueToken into the master catalog
        memcpy(valueBuffer, nsb_p->nsbCoreStringblock_p + byteOffsetToValue, nsb_p->nsbValMaxwidth);
   } else {
       xLogMsg("@F nsbGetNvpValueByIndex: FATAL-ERROR invalid NVP index %d\n",ithNvp);
       exit(-1);
   }


   // <JMG> 2015-09-24 Auto-convert any quoted string content where
   //  a back-tick was substituted for a space in the default parser
   if(strchr(valueBuffer,'`')!=NULL)
   {
        frm_strReplaceChr("`",valueBuffer," ") ;
   }
   
   // return ptr to value string, or NULL if failed
   return valueBuffer;
} // end::nsbGetValueByIndex()


// ---------------------------------------------------------------
// function  : nsbGetRecordByIndex()
// purpose   : retrieves an entire NSB record, by base0 index, ithNvp
// inputs    : const int ithNvp,char *recordBuffer_p
// returns   : char *, the NSB 'record' string mapped to base0 index 'ithNvp'

// notes     : use int nsbGetStringblockRecsize(NsbObj *nsb_p) to determine
//             how large to size the 'recordBuffer_p' prior to calling this
// TBD: this does not work quite right due to null term at end of keyword
// in between key and value..so only the keyword appears. Fix later.
char *nsbGetRecordByIndex(NsbObj *nsb_p,const int ithNvp,char *recordBuffer_p)
{
 size_t byteOffsetToRecord = 0 ;
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    // clear out the RECORD string first
    memset(recordBuffer_p,'\0',nsb_p->nsbRecsize);
    
   if(ithNvp < (int)nsb_p->nsbActualCount)
   {
    // offset to start of ith record
    byteOffsetToRecord = NSB_OFFSET_TO_RECD(ithNvp, nsb_p->nsbRecsize);
    
    // Insert this latest keyword and valueToken into the master catalog
    memcpy(recordBuffer_p, nsb_p->nsbCoreStringblock_p + byteOffsetToRecord, nsb_p->nsbRecsize);
    
   } else {
       xLogMsg("@F nsbGetNvpRecordByIndex: FATAL-ERROR invalid NVP index %d\n",ithNvp);
       exit(-1);
   }
   
   // return ptr to (record) string, or NULL if failed
   return recordBuffer_p;
} // end::nsbGetRecordByIndex()


// ---------------------------------------------------------------
// function  : nsbSetValueByIndex()
// purpose   : assigns an NVP value, at base0 index ithNvp
// inputs    : const int ithNvp,char *valueBuffer
// returns   : int, string length of fetched value string, or -1 if failed
// notes     : Assuming nTotalNvp is a count of ALL (e.g. both unique
//              and non-unique keyword NVPs, this function works for both
//              single-valued NVPs and multi-valued NVPs.
//
// notes     : TBD, auto-limit to max value width, using strncpy() or equiv
//             method to refuse overwriting past end of allowed value width
int nsbSetValueByIndex(NsbObj *nsb_p,const int ithNvp,char *valueBuffer)
{
 int status = -1 ;
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
 
   if(ithNvp < (int)nsb_p->nsbActualCount)
   {
       status = nsbStoreValue(nsb_p, ithNvp, valueBuffer);
   } else {
       xLogMsg("@F nsbSetNvpValueByIndex: FATAL-ERROR invalid NVP index %d\n",ithNvp);
       exit(-1);
   }
   
  // return status, as value's string length or -1 if failed
  return status;
} // end::nsbSetValueByIndex()

// ---------------------------------------------------------------
// function  : nsbGetKeywordCount()
// purpose   : Retrieves the count of 'targetKey_p' keywords in database,
//             using either a DoCountStrictMatch(=0) approach or a 
//             DoCountPartialMatch(=1) matching policy regarding the keyword
// inputs    : NsbObj *nsb_p
//             const char *targetKey_p -- ptr to keyword (or keyword fragment)
//             to match on
//             const int matchPolicy 0=DoCountStrictMatch, 1=DoCountPartialMatch
//
// returns   : char **,strTbl_p, an array of strings holding ALL (multiple) values
//           :     mapped to a single common 'targetKey_p' string
//           : int *nElem -- count of strings in returned string table.
// notes     : When done with strTbl_p this returns be sure to
//             call freeStringTbl(s,nElem) to avoid orphaned memory
//             this createStringTbl() is not yet managed by memObjCalloc
int nsbGetKeywordCount(NsbObj *nsb_p,const char *targetKey_p,const int matchPolicy)
{
 size_t byteOffsetToKey   = 0 ; 
 size_t nMultiElements =0;
 int  v = 0 ;
 char keyword[1024] = {0} ;
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    for(v=0; v < (int)nsb_p->nsbActualCount; v++)
    {
        byteOffsetToKey = NSB_OFFSET_TO_KEY(v, nsb_p->nsbRecsize);
            
        memcpy(keyword, nsb_p->nsbCoreStringblock_p + byteOffsetToKey,nsb_p->nsbKeyMaxwidth);

        if( matchPolicy == 0) 
        {
            // DoCountStrictMatch
            if(strcmp(targetKey_p,keyword) ==0)
                    nMultiElements++ ;
        } else { 
            // DoCountPartialMatch, if a substring 'targetKey_p'
            // is found anywhere within stored keyword, consider it a match
            if(strstr(targetKey_p,keyword) != NULL)
                    nMultiElements++ ;
        }
    } // end:: v iteration through list of all NVPs to find keyword
    
    // return count of target keyword! 1 indicates a scalar.
    return nMultiElements;
    
} // end:: nsbGetKeywordCount()


// ---------------------------------------------------------------
// function  : nsbGetMultiValueByKey()
// purpose   : Retrieves a string-table of multiple values mapped to a 
//           : strict match on a single, shared 'targetKey_p' string key.
// inputs    : const char *targetKey_p,char *valueBuffer,int *nElem
// revised   : 2015-10-14T17:52, removed orphaned FILE *inf that was opened
//             but not used, and not closed (corrected now).
// revised   : 2015-09-24T11:21:00, started prototyping "!" macro expansion
//             syntax to allow a single KEY's value as a path to ONE text
//             file that itself contains a list of items (e.g. multiple paths).
// limitation: This is typically called ONCE ONLY, after which the string
//             table returned is used, passed around, freed. You CANNOT
//             call this function and not "catch" the returned stringtable!
// returns   : char **,strTbl_p, a "stringtable" -- an array of strings that
//             store ALL (multiple) values that are mapped to a single 
//             common 'targetKey_p' string. This is a multivalue object.
//           : int *nElem -- count of strings in returned string table.
// notes     : When done with the strTbl_p returned, the caller must be sure to
//             call freeStringTbl(s,nElem) to avoid orphaned memory
//             TBD, confirm if the createStringTbl() instance returned is 
//             managed by memCalloc() or not.
char **nsbGetMultiValueByKey(NsbObj *nsb_p,const char *targetKey_p,int *nElem)
{
 size_t v =0;
 
 size_t byteOffsetToKey   = 0 ; 
 size_t byteOffsetToValue = 0 ; 
 size_t nMultiElements =0;
 int ithNvp = 0 ;
 int maxWidthFound = 0;
 int nLines =0 ;
 char **strTbl_p = NULL; // ptr to string table of nMultiElements returned.
 char keyword[1024] = {0} ;
 char fxName[256] = {0} ;
 
 char listTextFilePath[FILENAME_MAX]={0};// ptr to VALUE as a text file to read K items from
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);

    // Prototype: new list-expansion macro, where if a keyword is prefaced
    // with a '!' we fetch its value (a text file path storing a collection of 
    // target strings)
    if( targetKey_p[0] == '!' )
    {
        // First, fetch the filename path storing the list of text lines
        // that will constitute the multivalue object...
        nsbGetValueByKey(nsb_p,targetKey_p,listTextFilePath);
        // Next, verify the list file actually exists, that stores the text lines
        // we will expand into the returned stringtable
        if(frm_ioDoesFileExist(listTextFilePath))
        {
            strTbl_p = frm_ioReadTextIntoTable(listTextFilePath,1,&maxWidthFound,&nLines);
            // <JMG> 2017-0-26T11:13 WARNING: the stringtable returned here may in fact contain line widths
            // that exceed the active nsb_p->nsbValMaxwidth, so more work needs to be done
            // to assure these are compatible for this use-case!!
        } else {
            xLogMsg("@W In(%s) Cannot expand keyword-macro (%s) value (%s) since NO valid file exists! ~\n",
               fxName, targetKey_p, listTextFilePath);
               // verify if a 'break' or 'continue' is needed here to 
               // branch out to attempt to perform traditional multi-value handling
               goto TRADITIONAL_MULTIVALUE; 
        }
        // return count of qualifying items in string table (n records)
        *nElem = nLines ;
        
        // Warn user if any constituent item string is wider than database width limit!
        // TBD: In future, we COULD just truncate the offending item value string to
        //  fit within the nsbValMaxwidth limit, but, since this is managed separately as
        //  an anoynmous string table that resides OUTSIDE the formal NSB database,
        //  this truncation is actually NOT necessary
        
        /* 
        // <JMG> 2017-04-26T11:32, removed this essentially SPURIOUS warning
        // since the memory for a stringtable resulting from a macro-expanded operation 
        // is NOT a part of the active NSB database itself, thus no width comparison is relevant.
        if(maxWidthFound >= nsb_p->nsbValMaxwidth)
        {
            xLogMsg("@A In(%s) Reminder: expanded keyword-macro (%s) item width %d exceeds NSB calculated soft-width VALUE %d!~ \n",
             fxName, targetKey_p, maxWidthFound, nsb_p->nsbValMaxwidth);
        }
        */
        
        // take EARLY return, list of items extracted from listTextFile_p
        return strTbl_p;
    } // end:: Prototype macro expansion logic block, when keyword prefaced with "!"

    // OTHERWISE, implements traditional repeating-keyword style of multi-value logic

    // Here we incrementally build up the list of multivalue NVP entries, by mapping
    // all text line VALUEs to a single KEYWORD. Important note: the
    // reason this works is that nsb_p->nsbKeyMaxwidth already represents the
    // length of the longest encountered VALUE string literally in the RIP file
    // irrespective of which KEY it maps to.
    
TRADITIONAL_MULTIVALUE:    
    // Pass 1: scan all keywords in existing database, to match ONE on 'targetKey_p'
    // and for each match, increment a counter of qualifying name-value-pairs
    for(v=0; v < nsb_p->nsbActualCount; v++)
    {
        byteOffsetToKey = NSB_OFFSET_TO_KEY(v, nsb_p->nsbRecsize);
            
        memcpy(keyword, nsb_p->nsbCoreStringblock_p + byteOffsetToKey,nsb_p->nsbKeyMaxwidth);
        // Test for exact KEYWORD match, between fetched keyword and targetKey_p
        // and where matched, increment the count of items in this collection
        if(strcmp(targetKey_p,keyword)==0)
            nMultiElements++ ;
    } // end:: v iteration through list of all NVPs to find keyword

    // ASSERT: count of the number of multivalue NVP is now available as 
    // (nMultiElements) where we assume it is >= 1, so then are able to
    // construct an (unregistered,ad-hoc) string table where we will store
    // ALL values mapped to this common key.  Note in this "edge" use-case the
    //  memory (for this stringtable) is NOT held within the NSB database
    //  itself; it is "foreign" or external memory, so technically this
    //  breaks the protocol that all data in one NSB database is literally
    //  stored WITHIN the NSB database. Arguably, this could or should be 
    //  reconciled by rebuilding the entire NSB to include this "foreign" stringtable.
    strTbl_p = memCreateStringTbl(nMultiElements, nsb_p->nsbValMaxwidth);
    
    // PASS 2: count is ready, so now fetch the multiple values
    for(v=0; v < nsb_p->nsbActualCount; v++)
    {
        byteOffsetToKey = NSB_OFFSET_TO_KEY(v, nsb_p->nsbRecsize);
        
        // Step PAST properties and key, to start of VALUE string itself
        byteOffsetToValue = NSB_OFFSET_TO_VAL(v, nsb_p->nsbRecsize,
                                                nsb_p->nsbKeyMaxwidth);
        
        // Fetch key from stringblock datastore, at this record, to test
        memcpy(keyword,nsb_p->nsbCoreStringblock_p + byteOffsetToKey, nsb_p->nsbKeyMaxwidth);
        
        // Proceed when a match exists between fetched keyword and targetKey_p
        if(strcmp(targetKey_p,keyword)==0)
        {
            // Retrieve value string at the record matching the targetKey_p string
            //  and increment count of multi-values fetched, loading the
            //  current string-table slot
            memcpy(strTbl_p[ithNvp++],
                nsb_p->nsbCoreStringblock_p + byteOffsetToValue, nsb_p->nsbValMaxwidth);
                
        } // end:: block if target KEY matches key in record...
        
    } // end:: v iteration through list of all NVPs to find keyword
    
    // return count of strings stored in resulting string table, strTbl_p.
    *nElem = ithNvp ;
    
   // return ptr to a string table storing ALL values mapped to targetKey_p
   return strTbl_p;
} // end::nsbGetMultiValueByKey()


// ---------------------------------------------------------------
// function  : nsbGetMultiValueByPartialKey()
// purpose   : Retrieves a string-table of multiple values mapped to a 
//           : "loose" partial match on 'targetKey_p' string. This 
//             implements a very limited form of keyword 'wildcard' semantics
// inputs    : const char *targetKey_p,char *valueBuffer,int *nElem
// returns   : char **,strTbl_p, a stringtable, which is an array of
//             strings holding ALL (multiple) values that are mapped to
//           : a single common KEY (e.g. 'targetKey_p') string.
//           : int *nElem -- count of strings in returned string table.
// revised   : 2015-09-23T10:58 added documentation clarification only
// notes     : We have NOT yet implemented then
// notes     : When the caller is done with the strTbl_p this returns, it
//             is the callers responsibility to call freeStringTbl(s,nElem) 
//             to avoid orphaned memory this createStringTbl() is not yet
//             managed by memObjCalloc() or memCalloc()
// notes     : <JMG> 2014-07-23, Modified to match on whenever "targetKey"
//             is found inside prior 'keyword' in database, e.g. allows
//             partial-substring to ALSO MATCH. Not yet tested.
// 
char **nsbGetMultiValueByPartialKey(NsbObj *nsb_p,const char *targetKey_p,int *nElem)
{
 size_t v =0;
 
 size_t byteOffsetToKey   = 0 ; 
 size_t byteOffsetToValue = 0 ; 
 size_t nMultiElements =0;
 int ithNvp = 0 ;
 char **strTbl_p = NULL; // ptr to string table of nMultiElements returned.
 char keyword[1024] = {0} ;
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    for(v=0; v < nsb_p->nsbActualCount; v++)
    {
        byteOffsetToKey = NSB_OFFSET_TO_KEY(v, nsb_p->nsbRecsize);
            
        memcpy(keyword, nsb_p->nsbCoreStringblock_p + byteOffsetToKey,nsb_p->nsbKeyMaxwidth);
        // Matches qualify as HIT when even PARTIAL-KEY like "L4C_OUTPUT_DATASET_" is found 
        // within a keyword like L4C_OUTPUT_DATASET_NEE_MEAN" etc.
        if(strstr(targetKey_p,keyword) != NULL)
            nMultiElements++ ;
    } // end:: v iteration through list of all NVPs to find keyword

    // Construct a (unregistered,ad-hoc) string table storing ALL values mapped to common key
    strTbl_p = memCreateStringTbl(nMultiElements, nsb_p->nsbValMaxwidth);
    
    // PASS 2: count is ready, so now fetch the multiple values
    for(v=0; v < nsb_p->nsbActualCount; v++)
    {
        byteOffsetToKey = NSB_OFFSET_TO_KEY(v, nsb_p->nsbRecsize);
        
        // Step PAST properties and key, to start of VALUE string itself
        byteOffsetToValue = NSB_OFFSET_TO_VAL(v, nsb_p->nsbRecsize,
                                                nsb_p->nsbKeyMaxwidth);
        
        // Fetch key from stringblock datastore, at this record, to test
        memcpy(keyword,nsb_p->nsbCoreStringblock_p + byteOffsetToKey, nsb_p->nsbKeyMaxwidth);
        
        // Match OK even if a PARTIAL-KEY like "L4C_OUTPUT_DATASET_" is found inside
        // prior keyword L4C_OUTPUT_DATASET_NEE_MEAN" etc.
        if(strstr(targetKey_p,keyword) != NULL)
        {
            // retrieve value string at the record matching the targetKey_p string
            //          and increment count of multi-values fetched
            memcpy(strTbl_p[ithNvp++],
                nsb_p->nsbCoreStringblock_p + byteOffsetToValue, nsb_p->nsbValMaxwidth);
                
        } // end:: block if target KEY matches key in record...
        
    } // end:: v iteration through list of all NVPs to find keyword
    
    // return count of strings stored in resulting string table, strTbl_p.
    *nElem = ithNvp ;
    
   // return ptr to a string table storing ALL values mapped to targetKey_p
   return strTbl_p;
} // end::nsbGetMultiValueByPartialKey()


// ---------------------------------------------------------------
// function  : AppendNvpIntoStringblock()
// purpose   : append latest {keyword,value} strings into a pre-sized NVP stringblock
//             taking care to make sure the stringblock can accomodate this append.
// inputs    : const char *keywordTok,const char *valuTok,const int ithNvp
// returns   : 0
// notes     : n/a
int nvpAppendIntoStringblock(NsbObj *nsb_p,const char *keywordTok,const char *valuTok,const int ithNvp)
{
    char fxName[256] = {0} ;    
    size_t byteOffsetToKey = 0;
    size_t byteOffsetToValue = 0;
    int    keyLen =0;
    int    valLen =0;
    
        // REGISTER fx 
        strcpy(fxName,(char *)__func__);

        keyLen = strlen(keywordTok);
        valLen = strlen(valuTok);
    
        byteOffsetToKey = NSB_OFFSET_TO_KEY(ithNvp, nsb_p->nsbRecsize);
        
        // Insert this latest keyword and valueToken into the master catalog
        memcpy(nsb_p->nsbCoreStringblock_p + byteOffsetToKey, keywordTok, keyLen);
        
        // When offsetting by kMaxwidth, we leave a constant gap between the
        // end of the actual keyword, and the start of the value string, so
        // that it represents a fixed-width record regardless of the length of
        // a given keyword and offset.
        byteOffsetToValue = NSB_OFFSET_TO_VAL(ithNvp, nsb_p->nsbRecsize,
                                            nsb_p->nsbKeyMaxwidth);

        memcpy( nsb_p->nsbCoreStringblock_p + byteOffsetToValue, valuTok, valLen);

 return 0;
} // end::nsbAppendIntoStringblock()
