// ---------------------------------------------------------------
// module       : nsb_api.c
// purpose      : Implements replicable (thread-safe) NVP "string-block" instances
// project      : Lupine Logic Inc, generalized name-value-pair stringblock package
// dependencies : n/a
// revision     : v2.0-2015-09-24T12:42, added capability to perform simple
//                macro expansion of multi-valued keywords via reading a
//                text file containing a list of items, triggered by a
//                KEYWORD prefaced with a '!' symbol, e.g. !MyKey, and added
//                 int nsbGetMaxKeyWidth(NsbObj *nsb_p);
//                 int nsbGetMaxValueWidth(NsbObj *nsb_p);
// revision     : v1.7-2014-02-03T14:14:00-7:00
// historyn     : 2014-01-06T10:45:00:00,jmg --migrated to use xLogMsg()
//              : 2014-01-05T13:32:00,jmg --refactored API to provide for 
//              :  multiple, concurrent, thread-safe of a NVP stringblock instances.
//              : 2013-12-20T12:22:00, pending, eliminated all SLL dependencies
//              : 2013-12-20T02:00:00,milestoned version
//              : 2013-11-19T12:33:00-07:00, original code. 
// author       : joe glassy
// notes        : 
// ---------------------------------------------------------------

#include "frm_api.h"

#include "nsb_api.h"
#include "nsb_api_prot.h"

// ---------------------------------------------------------------

// ---------------------------------------------------------------
// Define a default (test-only) table of supported keywords. Normally this would
// be read into a string-table in runtime, as populated from a special session
// config file or "master" RIP config instance.


/*
 // TBD: verify --DEPRECATED ? This was just an example fragment
 // for initial testing
static char *L4C_DAILY_DefaultKeyStrTbl[11] = {
    "L4C_CRID", 
    "L4C_MODEL_DAYOFYEAR", 
    "L4C_MODEL_PERIOD", 
    "L4C_MODEL_YEAR", 
    "L4C_MODIS_FPAR_GRANULE", 
    "L4C_MODIS_INPUT_ANCILLARY_PATH", 
    "L4C_MODIS_INPUT_ROOTDIR", 
    "L4C_MODIS_OUTPUT_PATH", 
    "L4C_MODIS_OUTPUT_ROOTDIR", 
    "L4C_MODIS_RELEASE_TAG", 
    "L4C_MODIS_SOFTWARE_VERSION" 
} ;
*/

// ---------------------------------------------------------------
// function: nsbDeclareSchemaPathToParse(()
// purpose : Load a user defined (optional) schema RIP path, such that if
//           this is defined prior to the call to nsbInitStringblock() the
//           resulting in-memory NSB stringblock database will be sized to
//           also accmodate this schema or ancillary RIP's name value pairs.
// returns : int, length of schemaRipPath_p file spec string, 
// notes   : The 'name' of the supported KeyStrTbl is formed from the 
//           NsbObj parent's name with a "|KEYS" qualifed concatenated to it.
int nsbDeclareSchemaPathToParse(NsbObj *nsb_p,const char *schemaRipPath_p)
{
    
 strcpy(nsb_p->nsbSchemaPath, schemaRipPath_p);

 // return length of path
 return (int)strlen(schemaRipPath_p);
    
} // end:: nsbDeclareSchemaPathToParse()


// ---------------------------------------------------------------
// function: nsbSetSupportedKeyDictionary()
// purpose : Load a user defined keyword dictionary (string-table) from a text file
//           whose path is supplied by the user at initialization time.
// returns : the count of supported keywords as read from the text file as a
//           simple list of strings. From here on, any keyword used must 
//           be matched by an entry in this keyStrTbl_p of type IsStringTable
// revised : 2014-02-23T11:21:00, migrated keyStrTbl_p to use memCreateStringTable()
//           as a private, anonymous, non-registered string table not in the registry
//           This simplifies the internals.
//           
// revised : 2014-02-16T16:26:00, note on nsbKeyPolicy...
// notes   : The 'name' of the supported KeyStrTbl is formed from the 
//           NsbObj parent's name with a "|KEYS" qualifed concatenated to it.
// notes   : The 'load' of a supported key table is now completely independent
//           of any requirement to actually use it. When nsbKeyPolicy==0, no
//           supported keytable is required or consulted, even if this function
//           was prior used to load one. If nsbKeyPolicy==1, then this fx must
//           be called, as the keyword table is required and used.
char **nsbSetSupportedKeyDictionary(NsbObj *nsb_p,const char *keyFilepath)
{
 FILE *keywrdf = NULL;
 char keywordLine[NSB_STRBLK_RIPLINE_MAXWIDTH] = {0} ;
 char **keywordStrTbl_p =NULL;
 char fxName[256] = {0} ;
 char nsbKeyTableName[512] = {0} ;
    
 int nLines =0 ;
 int ithKey =0;
 int keyLen =0;
 int nKeywords = 0 ;
 int maxwidth  = 0 ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);

    if((keywrdf=fopen(keyFilepath,"rt"))==NULL)
    {
        xLogMsg("@F nsbSetSupportedKeyDictionary: FAILED to open(%s)\n",
                    keyFilepath);
        exit(-1);
    }
 
    // parse RIP file containing USERS CUSTOM supported keyword catalog
    while(fgets(keywordLine,NSB_STRBLK_RIPLINE_MAXWIDTH-1, keywrdf)!= NULL)
    {
        keywordLine[strlen(keywordLine)-1]  = '\0'; // clip newline
        nLines++ ;
        if(keywordLine[0] != '#')  // if NOT a comment line, proceed
        {
            keyLen = strlen(keywordLine);
            if(nKeywords == 0)
            { // maxwidth does NOT include the terminating null here!!
                maxwidth = keyLen;
            } else {
                if( keyLen >  maxwidth)
                        maxwidth = keyLen;
            }
            nKeywords++ ; // increment count of valid keywords encountered
        } // end:: block if not a comment line
        
    } // end:: while on scan to count instances...

    // Create a fully qualified name of supported keyword dictionary (table) as 
    //  the existing name of the parent, plus the "|KEYS" qualifier
    memset(nsbKeyTableName, '\0', sizeof(nsbKeyTableName));
    sprintf(nsbKeyTableName, "%s|KEYS", nsb_p->nsbLabel);
    
    // NOTE: 2/9/2014: memObjCalloc() method initially chosen due to its support for
    // IsStringTable alloc's and corresponding frees.
    // keywordStrTbl_p= memObjCalloc(nsbKeyTableName, maxwidth, nKeywords, IsStringTable,"","");
    
    // NOTE: <JMG> 2/23/2014, migrated to use private, anonymous, non-registered string
    // table to store the supporting KEY keywords dictionary. This simplifies things a lot.
    // Add '2' here to the allocated string widths to ensure room for null term
    keywordStrTbl_p = memCreateStringTbl(nKeywords,maxwidth+2);
    // BIND this key stringtable to the parent
    nsb_p->nsbKeyStrTbl_p = keywordStrTbl_p ;

    // <JMG> 2/23/2014, now we explicitly store the WIDTH of the 
    //  strings in the supporting key dictionary IsSTringTable
    nsb_p->nsbKeyStrTblWidth = maxwidth ;
    // Store  the count of supported keywords in the table to the parent...
    nsb_p->nsbKeyStrTblCount = nKeywords;
    
    ithKey =0;
    rewind(keywrdf); // rewind the text file to start at top again
    while(fgets(keywordLine,NSB_STRBLK_RIPLINE_MAXWIDTH-1, keywrdf)!= NULL)
    {
        keywordLine[strlen(keywordLine)-1]  = '\0'; // clip newline
        
        if(keywordLine[0] != '#')  // if NOT a comment line, proceed
        {
            // TBD: migrate to use strncpy() to restrict length of
            // user supplied strings into this, to block any buffer overrun
            memset(keywordStrTbl_p[ithKey],'\0', maxwidth+1);
            // Enforce and restrict copy to at most 'maxwidth' characters per keyword
            strncpy(keywordStrTbl_p[ithKey], keywordLine, maxwidth);
            ithKey++ ; // increment count of keys processed
        } // end:: keywordLine[0] != '#' test
        
    } // end:: while loop, on final scan to load actual keywords
    
    // TBD: sort the table lexically, later...
    
    // Close the keyword catalog text file
    fclose(keywrdf);

    // BIND this key stringtable to the parent
    nsb_p->nsbKeyStrTbl_p = keywordStrTbl_p ;
    
    xLogMsg("@A nsbSetSupportedKeyDictionary: %d supported key entries loaded from ~(%s)\n",
                    nsb_p->nsbKeyStrTblCount, keyFilepath);
    
 // Return pointer to new keyword catalog (string-table)
 // although this may not be needed too often...
 return keywordStrTbl_p;    
 
} // end:: nsbSetSupportedKeyDictionary()


// ---------------------------------------------------------------
// function: nsbSearchForSupportedKey(keyword)
// purpose : given a "targetKey, search a pre-established catalog of 
//           supported keywords, for the node that matches this key.
// returns : The base0 index of the node matching the 'targetKey' or -1 if failed.
// notes   : TBD, for now we fix the search against just the "nsbMasterDefaultKeyStrTbl",
//         : but in the future, when user loads a custom catalog of supported
//           keys via nsbSetSupportedKeyDictionary(), we'll use THAT as the
//           target of searches supplying 'valids' to match against.
int nsbSearchForSupportedKey(NsbObj *nsb_p,const char *targetKey)
{
 int k= -1;
 char fxName[256] = {0} ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
 
    if( nsb_p->nsbKeyPolicy == 1)
    {
        // Must not be NULL, as it should have been bound to keyword 
        //  string table by the time this fx is called!
        assert(nsb_p->nsbKeyStrTblCount >0);
        assert(nsb_p->nsbKeyStrTbl_p != NULL);
    
        // Search table of recognized keys to determine actual index to reference
        for(k=0; k < (int)nsb_p->nsbKeyStrTblCount; k++)
        {
            // fprintf(stderr,"nsbSearch: comparing target(%s) with keytbl KEY(%s)\n",
            //  targetKey, nsb_p->nsbKeyStrTbl_p[k]);
                
            if(strcmp(targetKey, nsb_p->nsbKeyStrTbl_p[k]) == 0)
                break;
        } // end:: k loop through all possible supported keys

        // if K exceeds the count of supported keys, complain
        if(k >= (int)nsb_p->nsbKeyStrTblCount  || k < 0)
        {
            fprintf(stderr,"@E nsbSearchForSupportedKey ERROR on Unrecognized  key [%s] at index %d\n",
            targetKey,k);
            k = -1 ; // critical to set this to a fixed known sentinel value so that callers can reliably
            // distinguish status of lookup operation as succeeded (>=0) or failed = -1
        }
        // end:: block if nsbKeyPolicy ==1 and a supported keyword table was MANDATORY
    } else { 
        // block if nsbKeyPolicy==0, whereNO keyword table is required
        k = 0 ; // placeholder value to allow parser to proceed...
    }

 // Return base0 index in a predefined supported keywords table, where the
 //   where (keyword) found or -1 if NOT found. Note is nsbKeyPolic==0, then
 //   the 'k' returned is ALWAYS 0, so don't depend on this as a valid slot
 return k;
 
} // end::nsbSearchForSupportedKey()


// ---------------------------------------------------------------
// function : nsbCreateInstance(keyword)
// purpose  : Create a stand-alone concurrent instance of a NsbObj
// returns  : Fresh, uninitialized instance of a NsbObj stringblock parent
//            and, returns "thisUUID" to allow user to access fully qualified name
// revision : 2014-02-09T13:39:00-07:00, migrating to support @instance() properties..
// revision : 2014-01-03T15:57:00
// author   : joe glassy
// notes    : This fx does NOT yet allocate the stringblock itself.
// notes    : For users who need access to the fully qualified name, the
//            UUID is stored inside the parent object in member 'nsbUIUD'.
//            Having access to this is helpful when a fully qualified name
//            is need, as for memObjFree() for example.
NsbObj *nsbCreateInstance(const char *newNsbLabel)
{
    char fxName[256] = {0} ;
    char thisUUID[FRM_API_UUID_MAXLEN] = {0} ; // usually now 64
    char rsltUUID[FRM_API_UUID_MAXLEN] = {0} ; // now 64
    
    char nsbName[NSB_LBL_MAXLEN] = {0} ;
    char localName[NSB_LBL_MAXLEN]= {0}; 
    
    NsbObj *nsbObj_p = NULL;
    
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);

    // TBD: add to users name a UUID to augment identity of this instance. This
    // is done to make each instance unique
    strcpy(rsltUUID, frm_utilGenerateUUID(thisUUID,1));

    // Attach the UUID to the user supplied name to form a fully qualified name
    // for memObjCalloc() and friends...
    strcpy(nsbName,newNsbLabel);
    strcat(nsbName,"-");
    strcat(nsbName,rsltUUID);
    
    // Create one new instance of a NVP stringblock parent container, NsbObj
    // now using simple memCalloc...
    nsbObj_p = memCalloc(1, sizeof(NsbObj));
    
    // As a labelling convenience only, assign NAME of 
    //  overseer to the MemObj representation of this NsbObj.
    //  the MemObj is of course distinct from the nsbObj_p->nsbLabel
    //  which is a member of the NsbObj itsef
    strcpy(localName, newNsbLabel);
    memObjSetName(nsbObj_p, localName);
    // now also set placeholder (generic custom) datatype identifier
    memObjSetDatatype(nsbObj_p, IsCustomUserType);
    
    // Initialize, then mark parent's unique ID as this UUID
    memset(nsbObj_p->nsbUUID,'\0', FRM_API_UUID_MAXLEN); // ...64 to be safe
    strncpy(nsbObj_p->nsbUUID, rsltUUID, strlen(rsltUUID) );
    
    // <JMG> added 2-16/2014, nsbKeyPolicu option switch
    // When nsbKeyPolicy =1, a predefined table of supported keywords is required
    //    in order to use the main parser, nsbParseIntoStringblock() 
    // When nsbKeyPolicy =0, a supported keywords table IS NOT required . 
    nsbObj_p->nsbKeyPolicy = 0 ;
    
    // Set a copy of name inside the object, for self-referential use
    strcpy(nsbObj_p->nsbLabel, nsbName);
    
    // Set defaults...
    nsbObj_p->nsbCoreStringblock_p = NULL; // will allocate later!
    nsbObj_p->nsbKeyStrTbl_p      = NULL;  // will allocate later!
    nsbObj_p->nsbKeyStrTblCount = 0; 
    
    nsbObj_p->nsbActualCount  = 0 ; // count of NVPs parsed from RIP file
    
    // Future extension: when nsbMaxCount exceeds nsbActualCount, this indicates
    // that (delta) un-used (empty) records have been allocated for "growth" of the
    // stringblock database. This is not yet enabled or fully supported until tested.
    nsbObj_p->nsbMaxCount     = 0 ; // max allowable, may exceed nsbActualCount
    
    nsbObj_p->nsbRecsize      = 0 ; 
    nsbObj_p->nsbKeyMaxwidth  = 0 ;
    nsbObj_p->nsbValMaxwidth  = 0 ;
    nsbObj_p->nsbOwnerPID = -1 ; // unique ID of OS process owning this instance
    
    // <JMG> 2015-09-30, Store DEFAULT set of valid, KEY|VALUE delimiter symbols 
    // internally in parent object, such that user may override these in custom scenarios
    // later with nsbUpdateDelimiters(nsb_p,"\t=, "); etc
    strcpy(nsbObj_p->nsbDelimSymbols,"\t= ");
    
    // <JMG> 2015-09-30, store collection of delimiters that separate multi-valued items
    // such as the properties of a dataset, with a limit of 15 single-character delimiters.
    strcpy(nsbObj_p->nsbDelimCompound, ",|");
    
    // return pointer to new parent instance.
    return nsbObj_p ;
    
} // end::nsbCreateInstance()


// ---------------------------------------------------------------
// function: nsbReleaseInstance()
// purpose : release all memory used by an NsbObj instance, using nsb_p->nsbLabel
//           and fixed derivatives of this, e.g. suffix |KEYS for the key table
// inputs  : NsbObj *nsb_p -- pointer to the NsbObj object to free up
// returns : void
// revision: 2014-02-23T11:32, migrated nsbKeyStrTbl to use
//           memCreateStringTbl(), memFreeStringTbl() as an 
//           private, anonymous, non-registered IsStringTable instance
//           which simplifies logic and obviates need for memObjCalloc()
//           or memObjFree() which require 'names'
// author  : joe glassy
// notes   : 
void nsbReleaseInstance(NsbObj *nsb_p)
{
    // Only free the KEYS (supported valids) table if it has
    // been used.
    if(nsb_p->nsbKeyStrTbl_p != NULL && nsb_p->nsbKeyStrTblCount > 0)
    {
        // free up private, non-registered, IsStringTable 'count' KeyStrTbl entries
        memFreeStringTbl(nsb_p->nsbKeyStrTbl_p, nsb_p->nsbKeyStrTblCount);
    }

    // Free the "core" database (stringblock) itself, 
    //   identified only by its own DNA
    nsbFreeStringblock(nsb_p) ;
        
    // Last, release the encapsulating parent's object
    // which no longer requires a name..
    memFree(nsb_p);
    
} // end::nsbReleaseInstance() 


// ---------------------------------------------------------------
// function  : nsbInitStringblock()
// inputs    : const size_t nNvps,const int keyMaxwidth,const int valMaxwidth
// purpose   : uses memCalloc() to allocate memory for a master (CORE) stringblock 
// returns   : count of n. nvp "records" initialized
// notes     : DEPRECATED: nsb_p->nsbPropMaxwidth -- was fixed at compile time, as nsb_api_PROPMAXWIDTH 

// notes     : TBD, refine methods and options on exactly how the initialization at end
//             of this function is done, e.g. write a full-width line of blanks, write a single
//             blank, write NSB~REC-UNUSED sentinal, etc. Each has its advantages...
//
int nsbInitStringblock(NsbObj *nsb_p,const size_t nNvps,const int keyMaxwidth,const int valMaxwidth)
{
 size_t nBytes = 0 ;
 size_t recIdx = 0 ;
 char fxName[256] = {0} ;
 // int  fillCh = ' ' ;
 
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    // Register fixed widths for keywords and for values
    
    // nsb_p->nsbKeyMaxwidth dynamic, set at runtime as longest keyword encountered on 1st parse
    // nsb_p->nsbValMaxwidth dynamic, set at runtime as longest 'value' string encountered on 1st parse. 
    nsb_p->nsbRecsize = ( keyMaxwidth+ valMaxwidth );
    
    // TBD: update, assign... re-verify though..
    nsb_p->nsbKeyMaxwidth = keyMaxwidth;
    nsb_p->nsbValMaxwidth = valMaxwidth;
    
    // Register COUNT of name-value-pair entitites stored in this instance
    nsb_p->nsbActualCount = nNvps;
   
    // Set the n. of bytes required for entire stringblock, 
    // we always add 1 char for a null terminator so the whole
    //  stringblock itself can be treated like a string...
    nBytes = (nNvps * ( nsb_p->nsbKeyMaxwidth + nsb_p->nsbValMaxwidth +1));

    if(nBytes < 1 || nNvps < 1 )
    {
        fprintf(stderr,"@E In(%s) Insufficient number of name-value-pairs (%d) or bytes(%zu) detected!\n",
             fxName, (int)nNvps, nBytes) ;
        return 0 ;
    }
   
    // Allocate memory representing the "core" actual in-memory NSB database itself
    // sprintf(nsbCoreName,"%s|CORE",nsb_p->nsbLabel);
    // fprintf(stdout,"CORE-NAME [%s]\n",nsbCoreName);
    
    // <JMG> 2014-02-09, migrated to use memCalloc() now...
    nsb_p->nsbCoreStringblock_p = memCalloc(nBytes,sizeof(char));
    
    // <JMG> 2013-02-23. This initialization section works fine
    // as is, but could use further refinement, as we do not yet
    // explicitly mark each empty NSB record with NSB_RECORD_UNUSED
    
    for(recIdx=0; recIdx < nNvps; recIdx++)
    {
        // nsbSetKeywordByIndex(nsb_p, recIdx, (char *)NSB_RECORD_UNUSED);
        // nsbSetValueByIndex(nsb_p  , recIdx, (char *)NSB_RECORD_UNUSED);
        // Or, try setting to all blanks...
        // nsbSetValueByIndex(nsb_p  , recIdx, fillBuf_p);
        nsbSetValueByIndex(nsb_p  , recIdx, " ");
    } // end:: recIdx loop marking empty records as un-occupied
    
 return nNvps;
 
} // end::nsbInitStringblock()


// ---------------------------------------------------------------------
// function: nsbCreateDatabase()
// purpose : Creates a custom sized NSB directly from user specified
//           (3) parameters establishing: N-Records,MaxKeywordWidth,MaxValueWidth
// revised : 2015-09-28T16:44:00
// created : 2015-09-28T16:38:00
// author  : Joe Glassy
// notes   : Use with caution, user must verify own limits to keys and values
//             when making assignments via:
//     nsbStoreKeyAndValue(nsb_p, ithNvp, keywordToken, valueToken)  ..OR..
//     int nsbStoreKey(NsbObj *nsb_p,const int ithElement,const char *keyword)
//     int nsbStoreValue(NsbObj *nsb_p,const int ithElement,const char *valueStr)
//     int nsbSetValueByIndex(NsbObj *nsb_p,const int ithNvp,char *valueBuffer)
//     int nsbSetValueByKey(NsbObj *nsb_p,const char *targetKey_p,char *valueBuffer)
// 
// ---------------------------------------------------------------------
int nsbCreateDatabase(const size_t nvpMaxRecordCount,const int keyMaxWidth,const int valMaxWidth,const char *newNsbDatabaseLabel_p)
{
 NsbObj *newNsb_p = NULL;
     
    newNsb_p = nsbCreateInstance(newNsbDatabaseLabel_p); 
     
    nsbInitStringblock(newNsb_p, nvpMaxRecordCount, keyMaxWidth, valMaxWidth);
    
    // return pointer to new, user-specified custom sized NSB database
    return newNsb_p ;
    
} // end:: nsbCreateDatabase()


// ---------------------------------------------------------------
// function : nsbGetInstanceName()
// inputs   : NsbObj *nsb_p --ptr to instance of NsbObj
//            char *kTblName
//            char *coreName
// author   : joe glassy
// returns  : char *mainName -- primary instance name
//            char *kTblName -- name of keytable
//            char *coreName -- name of stringblock itself in memory
// purpose  : returns the name of the NsbObj_p passed in, as well as returning to 
//            the caller the keytable name and the (core) stringblock's name
// notes    : The fully qualified name of the core stringblock is constructed
//            internally here using the existing parent's name, plus "|" plus "CORE"
char *nsbGetInstanceName(NsbObj *nsb_p,char *keyTblName,char *stringblocksName)
{
    sprintf(keyTblName      , "%s|KEYS",nsb_p->nsbLabel);
    
    // NOTE The stringblocksName is a labelling convenience only, it is NOT
    //   used by constructor or deconstructor, since memCalloc(), memFree() are used
    sprintf(stringblocksName,"%s|CORE",  nsb_p->nsbLabel);
    
    // Return the fully qualified name of the NSB instance itself
    return nsb_p->nsbLabel ;
    
} // end:: nsbGetInstanceName()

// ---------------------------------------------------------------
// function : nsbGetVersion()
// inputs   : char *versBuffer_p, buffer to hold version string 
// author   : joe glassy
// returns  : char *, version string for this NSB sub-API
// purpose  : returns the version for this NSB sub-API.
// notes    : n/a
char *nsbGetVersion(char *versBuffer_p)
{
    strcpy(versBuffer_p,(char *)NSB_VERS_TAG);
    
    // Return the long-lived version string
    return versBuffer_p;
    
} // end:: nsbGetVersion()


// ---------------------------------------------------------------
// function : nsbFreeStringblock()
// inputs   : const char *customStringblockLabel, use NULL for default
// purpose  : explicitly free memory used by this parent object's stringblock 
// notes    : Users should normally never need to call this private function
//            and instead use nsbReleaseInstance() which does all this safely.
int nsbFreeStringblock(NsbObj *nsb_p)
{
  char fxName[256] = {0} ;
  
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    // free up all memory used by master stringblock
    memFree(nsb_p->nsbCoreStringblock_p);
    
 return 0;  
} // end::nsbFreeStringblock()

