// ---------------------------------------------------------------
// module       : nsb_dbsys.c
// purpose      : Implements NSB db system level services
// project      : Lupine Logic Inc, generalized name-value-pair stringblock package
// dependencies : n/a
// revision     : 2014-02-09T13:30:00, re-located DB level services here.
// history      : n/a 
// author       : joe glassy
// notes        : 
// ---------------------------------------------------------------

#include "nsb_api.h"
#include "nsb_api_prot.h"

// ---------------------------------------------------------------
// function  : nsbGetStringblockRecordCount(void)
// returns   : size_t, count of NVPs
// purpose   : fetch count of NVPs (n. records) defined in master stringblock
// notes     : this is the "public interface" accessor.
size_t nsbGetStringblockRecordCount(NsbObj *nsb_p)
{
    char fxName[256] = {0} ;
    
        // REGISTER fx 
    strcpy(fxName,(char *)__func__); 

    return nsb_p->nsbActualCount;
    
}// end:: nsbGetStringblockRecordCount()


// ---------------------------------------------------------------
// function  : nsbSetStringblockRecsize()
// inputs    : const size_t newRecsize
// purpose   : One time only, establish the master stringblocks record size
//             as the sum of a key's {property,keyword,value} subfields.
//             together these define a single virtual NVP "record"
//             in the stringblock, at offset =(ithNvp * nsbMasterStringblockRecsize)
void nsbSetStringblockRecsize(NsbObj *nsb_p,const size_t newRecsize)
{
    char fxName[256] = {0} ;
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);
    
    nsb_p->nsbRecsize = newRecsize ;
    
} // end:: nsbSetStringblockRecsize()

// ---------------------------------------------------------------
// function  : nsbSetStringblockRecordCount()
// inputs    : NsbObj *nsb_p -- ptr to NVP stringblock instance
//             size_t newNvpcount --the freshened count of NVPs (n. records) in the
//             master stringblock.
// returns   : void
// purpose   : set the count of NVPs (n. records) defined in the stringblock
// notes     : this is the "public" interface accessor fx.
void nsbSetStringblockRecordCount(NsbObj *nsb_p,const size_t newNvpCount)
{
    
    nsb_p->nsbActualCount = newNvpCount ;
    
} // end::nsbSetStringblockRecordCount()


// ---------------------------------------------------------------
// function  : nsbGetStringblockRecsize()
// inputs    : void
// purpose   : returns total stringblock implicit "record" size, as the sum
//             of the (3) fixed width {property,keyword,value} subfields.
// returns   : count of bytes required to support fixed width sum of both
//             one keyword and one value, e.g. defines one virtual record
//             in the stringblock, at offset =(ithNvp * nsbMasterRecsize)
int nsbGetStringblockRecsize(NsbObj *nsb_p)
{
  char fxName[256] = {0} ;
  
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);

  return nsb_p->nsbRecsize ;
    
} // end::nsbGetStringblockRecsize()

// ---------------------------------------------------------------
// function  : nsbGetMaxKeyWidth()
// inputs    : void
// purpose   : returns stringblock KEYWORD strings max width
// revised   : 2015-09-24T12:43
// author    : Joe Glassy
// returns   : int, current maximum width of a NSB database KEYWORD
int nsbGetMaxKeyWidth(NsbObj *nsb_p)
{
  char fxName[256] = {0} ;
  
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);

  return nsb_p->nsbKeyMaxwidth ;
    
} // end::nsbGetMaxKeyWidth()

// ---------------------------------------------------------------
// function  : nsbGetMaxValueWidth()
// inputs    : void
// purpose   : returns stringblock value strings max width
// author    : Joe Glassy
// revised   : 2015-09-24T12:43
// returns   : int, current maximum width of a NSB database VALUE string
int nsbGetMaxValueWidth(NsbObj *nsb_p)
{
  char fxName[256] = {0} ;
  
    // REGISTER fx 
    strcpy(fxName,(char *)__func__);

  return nsb_p->nsbValMaxwidth ;
    
} // end::nsbGetMaxValueWidth()


// ---------------------------------------------------------------
// function  : nsbExportDbToStringtable()
// inputs    : NsbObj *nsb_p,char *nsbOutputTextPath_p,const int delimChr
// purpose   : Exports a range of, or an entire NSB stringblock database to
//             a stringtable, optional delimiter between the key,value if delimChr >= 10 (linefeed)
//             of the (3) fixed width {property,keyword,value} subfields.
// returns   : sizeof db exported to text file
// revised   : 2015-09-24
// author    : Joe Glassy
// Notes     : Uses base0 indices!  When we want ALL records, then assign
//             userFirstIdx=0,userLastIdx=0.
//             If you want to fetch base1 records {2,3,4} then set
//             userFirstIdx=1 and userLastIdx=3
// Notes     : This does NOT yet support new experimental macro expansion
//             of multiValue NVPs, only scalar NVPs are supported thus far
char **nsbExportDbToStringtable(NsbObj *nsb_p,const int userFirstIdx,const int userLastIdx,const int delimChr)
{
 int recIdx = 0 ;
 int recWidth = 0;
 
 int keyLen = nsb_p->nsbKeyMaxwidth;
 int valLen = nsb_p->nsbValMaxwidth;
 
 // assign DEFAULT {range} as ALL via {0,n-1}.
 int  initialRecordIndex = 0 ;
 int  finalRecordIndex   = (nsb_p->nsbActualCount-1);
 
 char fxName[256] = {0} ;
 char *nsbKey_p = NULL;
 char *nsbVal_p = NULL;
 char **expStringTbl_p = NULL;
 char *currRecd_p = NULL ; // ptr to current record
 
    strcpy(fxName,__func__);
    
    // Establish all pertinent lengths or widths (keyword,value,record)
    //  and CREATE a string table of sufficient dimensions to hold the
    //  entire name-value-pair stringblock database. The string table is
    //  intentionally slightly oversized
    recWidth       =  nsbGetStringblockRecsize(nsb_p);
    expStringTbl_p = memCreateStringTbl(nsb_p->nsbActualCount+1,recWidth+2);
    // Allocate local working memory
    nsbKey_p         = memCalloc(keyLen+1,sizeof(char));
    nsbVal_p         = memCalloc(valLen+1,sizeof(char));
    
    if( userLastIdx >0 && userLastIdx < (int)nsb_p->nsbActualCount )
    {
        initialRecordIndex = userFirstIdx;
        finalRecordIndex   = userLastIdx;
    }
    
    // Perform a record by record (text) export
    for(recIdx = initialRecordIndex; 
        recIdx <= (int)finalRecordIndex; recIdx++)
    {
        currRecd_p = expStringTbl_p[recIdx] ;
        // fetch next record
        // nsbGetRecordByIndex(nsb_p,recIdx,nsbRecordBuffer_p);
        // fprintf(nsbf,"%s\n",nsbRecordBuffer_p);
        nsbGetKeywordByIndex(nsb_p,recIdx,nsbKey_p);
        nsbGetValueByIndex(nsb_p,recIdx,nsbVal_p);
        
        // Follows same semantics as for nsbExportDbToTextfile() where to
        // be valid, the delimChr must be >= a linefeed symbol. Typically
        //   this is a comma,tab,space,or special symbol like '|' or other
        //   user defined ASCII symbol
        if(delimChr >= 10)
            sprintf(currRecd_p,"%s%c%s\n",nsbKey_p, delimChr, nsbVal_p);
        else
            sprintf(currRecd_p,"%s %s\n",nsbKey_p,nsbVal_p);
        
    } // end:: recIdx loop exporting ALL records
    
    memFree(nsbKey_p);
    memFree(nsbVal_p);

    // returns pointer to populated stringtable
    return expStringTbl_p;
    
} // end:: nsbExportDbToStringtable()


// ---------------------------------------------------------------
// function  : nsbExportDbToTextfile()
// inputs    : NsbObj *nsb_p,char *nsbOutputTextPath_p,const int delimChr
// purpose   : Exports entire NSB stringblock database to a text file which
//             is created via nsbOutputTextPath_p, with 
//             optional delimiter between the key,value if delimChr >= 10 (linefeed)
//             of the (3) fixed width {property,keyword,value} subfields.
// returns   : sizeof db exported to text file
// revised   : 2015-09-24
// author    : Joe Glassy
// Notes     : This does NOT yet support new experimental macro expansion
//             of multiValue NVPs, only scalar NVPs are supported thus far
int nsbExportDbToTextfile(NsbObj *nsb_p,char *nsbOutputTextPath_p,const int delimChr)
{
 FILE *nsbf = stdout; // default text stream destination is stdout
 int recIdx = 0 ;
 int recWidth = 0;
 
 int keyLen = nsb_p->nsbKeyMaxwidth;
 int valLen = nsb_p->nsbValMaxwidth;
 char fxName[256] = {0} ;
 char *nsbKey_p = NULL;
 char *nsbVal_p = NULL;
 
    strcpy(fxName,__func__);

    // If an empty string (or NULL) is submitted as the path, we default
    // the output stream destination as stdout.
    if(nsbOutputTextPath_p != NULL && strlen(nsbOutputTextPath_p) >= 1)
      nsbf = frm_Openfile(nsbOutputTextPath_p,"wt",fxName);
    
    recWidth          =  nsbGetStringblockRecsize(nsb_p);
    
    nsbKey_p         = memCalloc(keyLen+1,sizeof(char));
    nsbVal_p         = memCalloc(valLen+1,sizeof(char));
    
    // Perform a record by record (text) export
    for(recIdx =0; recIdx < (int)nsb_p->nsbActualCount; recIdx++)
    {
        // fetch next record
        // nsbGetRecordByIndex(nsb_p,recIdx,nsbRecordBuffer_p);
        // fprintf(nsbf,"%s\n",nsbRecordBuffer_p);
        nsbGetKeywordByIndex(nsb_p,recIdx,nsbKey_p);
        nsbGetValueByIndex(nsb_p,recIdx,nsbVal_p);
        
        if(delimChr >= 10)
            fprintf(nsbf,"%s%c%s\n",nsbKey_p, delimChr, nsbVal_p);
        else
            fprintf(nsbf,"%s %s\n",nsbKey_p,nsbVal_p);
        
    } // end:: recIdx loop exporting ALL records
    
    memFree(nsbKey_p);
    memFree(nsbVal_p);
    
    if( nsbf != stdout && nsbf != stderr)
       frm_Closefile(nsbf,fxName);

    return 0;
} // end:: nsbExportDbToTextfile()


// ---------------------------------------------------------------
// function  : nsbExportDb()
// inputs    : void
// purpose   : exports one whole stringblock database, with NO formatting
//             into a raw binary file, using raw binary I/O
// returns   : size_t, n. of bytes of database as exported to a binary file
// author    : Joe Glassy
// notes     : n/a
size_t nsbExportDb(NsbObj *nsb_p,char *nsbOutputPath_p)
{
 FILE *nsbf = NULL;
 size_t nBytes =0 ;

 int recWidth = 0;
 
 
    nsbf      = frm_Openfile(nsbOutputPath_p,"wb", "nsbExportDb");

    // Reminder: recWidth includes the byte count of BOTH the 
    // keyword+value (and delimiter if any) that comprises a given record    
    recWidth  =  nsbGetStringblockRecsize(nsb_p);
    
    nBytes = nsb_p->nsbActualCount * recWidth ;
    
    if((fwrite(nsb_p->nsbCoreStringblock_p,sizeof(char),nBytes,nsbf))!= nBytes)
    {
         xLogMsg("@E nsbExportDb: error writing NSB DB to(%s)\n",
            nsbOutputPath_p);
    }
    
    frm_Closefile(nsbf,"nsbExportDb");

    return nBytes;
    
} // end:: nsbExportDb()


// ---------------------------------------------------------------
// function  : nsbParseSchema()
// purpose   : Parses a @schema() statement in an NSB database
// inputs    : NsbObj *nsb_p,
//             const char *keywordToken,
//             const char *valueToken
// returns   : int, 0
// notes     : Example @schema(database)  "myNsbDbName"
//             @schema(field)   keyword|IsString
//             This schema prototype does NOT yet support datatype-aware
//             NSB entities, nor all multivalue NVP handling semantics
int nsbParseSchema(NsbObj *nsb_p,const char *keywordToken, char *valueToken)
{
 char propertyTypeStr[64] = {0} ;
 char **refStrTbl_p = NULL;
 int nLen     =  0 ;
 int noOp     = -1 ; // placeholder only
 int nReferents= 0 ;
 
 
    // Set default structure as 2 fields, keyword,value
    strcpy(nsb_p->nsbSchema.keyField, "keyword");
    strcpy(nsb_p->nsbSchema.valField, "value");
    
    nsb_p->nsbSchema.nsbFieldcount = 2;
    
    // Derive sum of lengths of all pipe-sep referents, as nLen
    nLen = strlen(valueToken); 
    // Count the number of referent sub-values in pipe sep expression
    // add 1 since each pipe separates 2 referent value tokens
    nReferents = frm_strCountTokens(valueToken,"|") + 1;
    
    // Create a string table to store all nReferents
    refStrTbl_p = memCreateStringTbl(nReferents,nLen+2);
    
    // Extract all @schema() info between '(' and ')'
    frm_strCpBetween(keywordToken,propertyTypeStr,'(',')',&nLen);
    
    if(strcmp(propertyTypeStr       ,"database"   ) ==0) {
        strcpy(nsb_p->nsbLabel,valueToken);
    } else if(strcmp(propertyTypeStr,"delimiter " ) ==0) {
        strcpy(nsb_p->nsbSchema.nsbDbDelim, valueToken);
    } else if(strcmp(propertyTypeStr,"location") ==0)  {
        strcpy(nsb_p->nsbSchema.nsbLocationPath,valueToken);
    } else if(strcmp(propertyTypeStr,"recordcount") ==0) {
        nsb_p->nsbSchema.nsbRecordCount =(size_t)atoi(valueToken);
    } else if(strcmp(propertyTypeStr,"field"      ) ==0) {
        noOp = 1;  // placeholder for now. field names are fixed.
    } else if(strcmp(propertyTypeStr,"fieldwidth" ) ==0) {
        // @schema(fieldwidth)   0|256
        // @schema(fieldwidth)   1|1024
        if(     strcmp(refStrTbl_p[0], "0") == 0)
            nsb_p->nsbKeyMaxwidth = atoi(refStrTbl_p[1]) ;
        else if(strcmp(refStrTbl_p[0], "1") == 0)
            nsb_p->nsbValMaxwidth = atoi(refStrTbl_p[1]) ;
            
    } else if(strcmp(propertyTypeStr,"unoccupied" ) ==0) {
        strcpy(nsb_p->nsbSchema.nsbUnoccupiedMark, valueToken);
    } else if(strcmp(propertyTypeStr,"TBD"          ) ==0) {
        noOp = 1;
    }
        
    // free up local string table
    memFreeStringTbl(refStrTbl_p, nReferents);   
    return 0 ;
    
} // end:: nsbParseSchema()


// ---------------------------------------------------------------
// function  : nsbImportDbFromTextfile()
// purpose   : Retrieves data to populate an NSB database, from a 
//             text file containing name-value-pairs.
// inputs    : const char *nsbTextPath_p -- path to file containing 
//             NVPs to read into the NSB database
// returns   : NsbObj *nsb_p -- ptr to populated NSB parent object
// notes     : An internal FILE *handle is opened and closed here.
//           : TBD, consider how this differs from a RIP file, such that
//             we could import a collection of K "value-strings" with no
//             'keywords' per se, to auto assign a set of matching 'keywords'
//             using a simple integer sequence scheme, keys::{001,002,003,nnn} or
//             {preface}{1,2,3,...N} auto generated key identifier scheme.
NsbObj *nsbImportDbFromTextfile(const char *nsbTextPath_p,const char *keyValDelims_p,int *nRecordsFound)
{
 NsbObj *nsb_p = NULL; // ptr to NEW NSB database we're creating here
 FILE *nsbf = stdin; // default text stream source is stdin
 char fxName[256] = {0} ;
 // reminders: NSB_STRBLK_RIPLINE_MAXWIDTH is 1024
 //            NSB_STRBLK_KEYWORD_MAXWIDTH is  256
 int recIdx = 0 ;
 int recWidth = 0;
 int nRecordsImported = 0 ;
 int nLines = 0 ;
 int keyLen = nsb_p->nsbKeyMaxwidth;
 int valLen = nsb_p->nsbValMaxwidth;
 
    strcpy(fxName,__func__);


    // If an empty string (or NULL) is submitted as the path, we default
    // the output stream destination as stdout.
    if(nsbTextPath_p != NULL && strlen(nsbTextPath_p) >= 1)
      nsbf = frm_Openfile(nsbTextPath_p,"rt",fxName);
    
    // Using filename as a label, CREATE new empty NSB database parent
    nsb_p = nsbCreateInstance(nsbTextPath_p);

    if(keyValDelims_p != NULL && strlen(keyValDelims_p) >= 1)
    {
        // custom delimiter strings allowed now
        nsbSetDelimiters(nsb_p, keyValDelims_p);
    }
    
    // PASS 1 -- scan the text file to discover parameters for maximum
    // a) keyword length, b) value string length, c) total record length
    nsbScanRip(nsb_p,nsbf,&nLines);
    
    // go back to top to prepare for Pass 2
    rewind(nsbf);
    // PASS 2 -- rescan the text file, performing the active parse of
    // all keywords and value strings
    nRecordsImported = nsbParseIntoStringblock(nsb_p,"", nsbf);
    
    if( nsbf != stdout && nsbf != stderr)
       frm_Closefile(nsbf,fxName);

   *nRecordsFound = nRecordsImported;
   
   // returns pointer to populated NSB database parent object
   return nsb_p;
 
} // end:: nsbImportDbFromTextfile()


// ---------------------------------------------------------------------
// function : nsbGetDelimiters()
// inputs   : NsbObj *nsb_p,const char *newCustomDelimiterSymbols)
// returns  : char *, ptr to internally stored string of delimiter symbols
//            that separate a given KEYWORD from its VALUE, 
//            and returns a copy of these in userDelimiterBuffer_p too.
// purpose  : "Get" accessor to allow user to retrieve delimiter
//            symbols defined internally
// revised  : 2015-09-30T15:04:00
// author   : Joe Glassy
// notes    : In orig prior implementation, default delimiters were "\t= "
//            were hardwired, and not adjustable. Now we store these
//            explicitly inside the parent object, allowing them to be
//            retrieved as needed, using this function.
//
char *nsbGetDelimiters(NsbObj *nsb_p,char *userDelimiterBuffer_p)
{
    strcpy(userDelimiterBuffer_p,nsb_p->nsbDelimSymbols);
    
    // returns a pointer to internally stored collection of allowable delimiters
    // between KEYWORD <delim> VALUE
    return nsb_p->nsbDelimSymbols ;
    
} // end:: nsbGetDelimiters()

// ---------------------------------------------------------------------
// function : nsbSetDelimiters()
// inputs   : NsbObj *nsb_p,const char *newCustomDelimiterSymbols)
// purpose  : "Set" accessor to allow user to override default delimiter
//            symbols (that separate a KEYWORD from its VALUE)
//            with a custom collection of new delimiter symbols
// returns  : int, length of collection (e.g. n. of delimiters,one per char)
// revised  : 2015-09-30T15:04:00
// author   : Joe Glassy
// notes    : In orig prior implementation, default delimiters were "\t= "
//            were hardwired, and not adjustable. Now we store these
//            explicitly inside the parent object, allowing them to be
//            modified, as needed using this function.
//
int nsbSetDelimiters(NsbObj *nsb_p,const char *newCustomDelimiterSymbols)
{

    // Overwrite the defaults, or the current delimiter set,
    //  with a new user-defined set as stored in the
    //  string, nsb_p->newCustomerDelimiterSymbols.
    strcpy(nsb_p->nsbDelimSymbols, newCustomDelimiterSymbols);
    
    // returns length of set
    return (int)strlen(nsb_p->nsbDelimSymbols);
    
} // end:: nsbSetDelimiters()


// ---------------------------------------------------------------------
// function: nsbSetDelimiterMultiValue()
// inputs  : NsbObj *nsb_p,const char *compoundItemDelimter_p
// purpose : Assign a new set of delimiter symbols separating multi-value
//           items in a single string list, such as the properties of a
//           dataset, like: "myVarName,float32,minVal,maxVal,units"
// returns : int, 0 on success.
// revised : 2015-09-30T15:34:00
// author  : Joe Glassy
// notes   : Typical examples are "," comma and pipe "|" symbols.
int nsbSetDelimiterMultiValue(NsbObj *nsb_p,const char *compoundItemDelimiter_p)
{
    // Store a (new) string holding delimters for separating multi-valued items
    // such as a dataset's properties. Typical examples are "," comma and
    // pipe "|" symbols.
    strcpy(nsb_p->nsbDelimCompound, compoundItemDelimiter_p);

    return (int)strlen(compoundItemDelimiter_p);
    
} // end:: nsbSetDelimiterMultiValue()

// ---------------------------------------------------------------------
// function: nsbGetDelimiterMultiValue()
// inputs  : NsbObj *nsb_p,const char *compoundItemDelimter_p
// purpose : Retrieve string storing delimiter symbols (up to 15) for
//           separating multi-valued items in one VALUE string, such as
//           PROPERTIES = VALUE --> "myVar,float32,minVal,maxVal,units"
// returns : char *, string storing delimiters separating multi-value items
// revised : 2015-09-30T15:34:00
// author  : Joe Glassy
// notes   : Typical examples are "," comma and pipe "|" symbols.
//           for example:  "myVarName,float32|minVal|maxVal|units"
char *nsbGetDelimiterMultiValue(NsbObj *nsb_p,char *userBufferCompoundDelimiter_p)
{
    strcpy(userBufferCompoundDelimiter_p,nsb_p->nsbDelimCompound);
    
    // Returna a pointer to internal nsb_p object string that holds a 
    // set of delimiters between multi-valued or compound NVPs
    return nsb_p->nsbDelimCompound ;
    
} // end:: nsbGetDelimiterMultiValue()

