// ---------------------------------------------------------------------
// l4c_MET.c
//   Purpose    : l4c_MET.c/.ex is one of (2) L4_C pre-processors. This
//                pre-processor is used to pre-condition and transform
//                all meterology related inputs required by L4_C daily.
//                For each model day (1..366), this code reads 
//                a collection (24) GEOS-5 FP granules, transforming hourly variables to daily average variables,
//                for the variables T2M (yielding min,max,avg) and QV2M (yielding avg).  These daily average
//                variables are then output to a daily intermediate HDF5 file, L4_C_MET.h5 
//                (or l4c_MET_yyyymmdd.h5) daily met file
//                required by the L4_C daily model executable, l4c_daily.ex.  The development path for this
//                pre-processor code is to next extend it to also perform similar operation on the (8) 3-hr L4_SM
//                raw input granules, as well as the L3_SM_A variable, surface_flags.
//   Author     : joe glassy, Lupine Logic Inc, on contract to Univ. of Montana
//   Invocation : ./l4c_MET.ex <l4c_MET.rip>
//   Version    : 7.x, see AppVers macro in l4c_MET.h
//   
//   Revision   : 2015-09-07T00:36:00-07:00, corrected F/T bit wise frequency reporting
//   History    : 2014-05-20T09:56:00-07:00, sync, update for compatibility with SMAP_L4_C_Ancillary.h5 used
//                by l4c_DAILY v5.0.1+ and l4c_MODIS v5.0.1+
//   History    : v4.6.6-2013-10-29T11:06:00-07:00,nccs
//   History    : 2013-10-31, changed readPft to read PFT_M09_lc_mask from SMAP_L4_C_Ancillary.h5 instead of from a raw binary file
//              : 2013-10-29, added code to assure no memory leak on nvp strTbl_p[]
//              : 2013-09-26, and see AppVers macro in l4c_MET.h
//   Dependencies:  <uuid/uuid.h> and HDFv5 and szip libraries
//   Inputs     : l4c_met.rip, a text file of name-value-pair (NVP) entries, one per GOES-5 FP granule, naming
//              the unambiguous location of the current model day's GEOS-5 FP, L4_SM, and L3_SM_A input files.
//
//   Build      : gcc $(CFLAGS) -o l4c_MET.ex l4c_MET.c $(INC) $(LIBSHDF) $(LIB) --> -lhdf5_hl -lhdf5 -luuid -lsz -lz -ljpeg -lm
//
//   Notes      : Although the row-flip logic is available in this code, NO explict 'row-flip' operations
//                are currently required in this code, due to the way the geog_forwardtransform() is coded
//              with the latitude term.
//   Notes      : A sample of the GEOS-5 FP filename: GEOS.fp.asm.tavg1_2d_slv_Nx.20130508_0430.V01.nc4 
//   Notes      : Any raw binary I/O in this executable will generally be converted to use HDF5 retrieves or
//                or writes upon full maturity of this code.   
//

#include "l4c_MET.h"
#include "l4c_MET_prot.h"


// ---------------------------------------------------------------------

int main(int argc,char *argv[])
{
 // ptr master instance, session controls object
 SesCtl *sesCtl_p = l4c_GetSesCtl();  
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // local pointer to support image package    
 int status = -1 ;
 
    // -----------------------------------------------------------------
    // ACTION: Initialize l4c_MET session
    l4c_InitSession(argc,argv,sesCtl_p); 
    
    // -----------------------------------------------------------------
    // ACTION 1: pre-process all L4_SM granules, producing daily average form 
    //      variables and storing these in the new daily intermediate HDF5 file, 
    status = driverLogic_L4_SM(sesCtl_p,sesCtl_p->doy_Str);
    
    // -----------------------------------------------------------------
    // ACTION 2: pre-process all GEOS-5 FP granules, producing daily average form 
    //      variables and storing these in the new daily intermediate HDF5 file.
    // Now GEOS5 FP "TS" derived Freeze/thaw mode, DoUse_GEOS5 ..is also
    // handled here    
    status = driverLogic_GEOS5_FP_Meteorology(sesCtl_p,sesCtl_p->doy_Str);

    // -----------------------------------------------------------------    
    // ACTION 4: if available, preprocess one daily L3_SM_A granule, 
    //  producing daily average form variables and storing these in the 
    //  daily intermediate SOC file.
    if( sesCtl_p->ftSourceMode == DoUse_ACTIVE_RADAR || 
        sesCtl_p->ftSourceMode == DoUse_ACTIVE_TSURF)
    {
       status = driverLogic_L3_SM_A(sesCtl_p,sesCtl_p->doy_Str);

       // WRITE out FT_STATE_UM_M03 here
       write_FtStateUM_3km(sesCtl_p,"FT_STATE_UM_M03", imagePkg_p->ftStateUm_M03_p);
       
    } else if(sesCtl_p->ftSourceMode == DoUse_GEOS5_TS) {

        // WRITE out FT_STATE_UM_M03, noting that the
        // ftStateUm_M03_p[] array already pre-emptively populated 
        // above in geos5_fp_MeteorologyDrilldown, so we need to write it out here
        write_FtStateUM_3km(sesCtl_p,"FT_STATE_UM_M03", imagePkg_p->ftStateUm_M03_p);
        
    } else if(sesCtl_p->ftSourceMode == DoUse_PASSIVE_RADIOMETER)  {
       xLogMsg("@F F/T for PASSIVE RADIOMETER mode(%s) is NOT yet supported!",sesCtl_p->ftSourceStr);
    }

    // Close out the session gracefully...
    l4c_CloseSession(sesCtl_p);
 
 // return POSIX high-level status to calling script or OS
 return EXIT_SUCCESS;

} // end::main()


