// --------------------------------------------------------------
//   module     :  l4c_MET.h
//   purpose    : header for l4c_MET.c developmental prototype
//   invocation : ./l4c_MET.ex <l4c_MET.rip> 
//   revised    : v7.3.0-2017-06-07T18:35:00, sync with L4CMDL v7.3.0 and new BPLUT
//   revised    : v7.1.1-2017-03-20T17:29, add doc, revision info only, to sync up
//                with L4CMOD v7.1.1.
//   revised    : v7.1-2017-02-08T09:51:00, corrected bug where incorrect missing value was
//                assigned for L4_SM variables as 0.0, when it should be and now is -9999.0
//                and cleaned up logfile (min,max,mean) metrics for L4_SM variables by 
//                removing call to function calculateGranuleActualRange() which was NOT
//                working correctly for L4_SM variables while working right for GEOS5_FP var.
//   revised    :  v7.0-2015-09-08T12:06:00, now: ASSIGN all bits (0..7) in FT_STATE_UM_M03 
//                 as per definitions given below, and routed to file level attribute
//                 in SMAP_L4_C_MET.h5 file 
//                Re-verified the confusing use of InSwath=0 as per Barry
//                Weiss and JPL convention throughout
//   revised    : v7.0-2015-09-07T20:01:00, corrected F/T bit wise frequency reporting
//   revised    : v7.0-2015-08-28T18:23:00, refactored, blend TS processing in with rest of
//                GEOS5 FP variables, to pre-emptively process through to F/T regardless of
//                mode, allowing late-binding switch based on FT_METHOD etc.
//   revised    : v7.0-2015-08-19T11:59:00-07:00 v7 fork continued testing and refinements.
//   history    : v7.0-2015-07-31T11:52:00, V7 fork to support GEOS5 FP "TS" F/T and/or
//                pending new L3_SM_P passive radiometer derived F/T at 3KM.
//   history    : v6.0-2015-06-10T16:20, <JMG> changed L4C_MET_SOFTWARE_VERSION to L4C_SW_VERSION
//   history    : v6.0-2015-06-01T23:55:00-07:00, prep for V10002 delivery.
//   history    : v5.0.1-2014-11-17T13:07 began pre-assessment of changes
//                to support Rolf's proposed soil moisture wetness unit
//                changes, involving reading additional L4_SM "lmc" granule
//                to retrieve (1-3) new porosity variables. This change will
//                not likely be made until his new data is available on January 2015
//   history    : v5.0.1-2014-10-24T14:35:00, re-submitted to NCCS for ORT#2,ORR
//   history    : v5.0.1-2014-10-23T09:35:00, removed DiagImage call
//                to prep for delivery to NCCS which can't support this.
//   history    : v5.0.1-2014-10-221T23:16:00, added l4c_MET_sci.c and
//                corrected error in FT_STATE_UM_M03
//                where recent change in "surface_flag" bitflag mapping 
//                (Coastal prox move) shifted all bits by 1 so tsurf F/T
//                was incorrectly read at bit 7 when it SHOULD be read at bit 8
//                (same shift for others) in L3_SM_A "surface_flags" variable.
//   history    : v5.0.1-2014-10-15T16:55, added memory metrics to 
//                l4c_MET_ses.c CloseSession() for consistency .
//   history    : v5.0.1-2014-10-08T16:28, "R11E" for ORT#2. added
//                L4_C_MET_DOI, INPUTPOINTER_ANC, etc
//   history    : v5.0.1-2014-08-26T10:36:00, added l4c_GetExecutableVersionId()
//                and changed macro AppVers to L4C_MET_VERS_TAG for consistency
//   history    : v5.0.1-2014-08-20T18:27:00-07:00
//                including the 1km lat,lon now
//   history    : v5.0.1-2014-07-30T09:48:00, integration continued
//                into L4_C v5.0.1, split out logic orig in l4c_MET_fx 
//                into module files (add each to makefile etc):
//                -- l4c_MET_drivers.c
//                -- l4c_MET_drilldowns.c
//                -- l4c_MET_io.c
//                -- l4c_MET_grid_proj.c
//                -- l4c_MET_ses.c
//                 -- l4c_MET_PkgMgrInit.c
//
//   history    : 2014-07-28T15:21:00-07:00 further work on UM Framework
//                adoption and implementing recent changes required by
//                v5.0.1 of l4c_DAILY
//   history    : v5.0.1-2014-06-04T18:06:00,  goal: convert to use 
//                UM Framework services (NSB parser, logging, memObj) and
//                make conformal to the latest D5+ Ancillary and SOC files
//                used by l4c_DAILY v5.0+.
//   history    : v5.0-2014-02-12T12:08:00-07:00
//   history    : v5.0-2014-02-12T12:08:00-07:00
//   history    : v4.7-2014-02-12T12:08:00, added l4c_calc_vpd() into l4c_MET_fx.c to allow us to
//                calculate the required VPD needed by L4_C_DAILY.ex and store in SMAP_L4_CMET.h5 intermediate file
//   history    : v4.6.6-2013-10-31T12:35:00-07:00,changed readPFT to read the PFT 9km global mask out of a HDF5 dataset in SMAP_L4_C_Ancillary.h5
//   history    : 2013-10-29T11:16:00-07:00,added runModelYear, and corrected for strTbl_p[] potential memory leak.
//   history    : 2013-09-30T11:54:00, initial delivery
//   version    : 5.0x, see AppVers macro below
//   author     : joe glassy, Lupine Logic Inc.
//
//   notes      : CRITICAL NOTE: Recap of FT_STATE_UM_M03 bit definitions:
//      bit 0 -- defines the boolean freeze thaw state regardless which source defined it.
//               Note this is done for backwards compatibility in how L4CMDL consults bit0 as a
//               single location to always retrieve pixel F/T state, 1=Frozen, 0=Thawed.
//      bit 1 -- RADAR: defines the FT_RADAR boolean freeze thaw state, relevant ONLY when ACTIVE RADAR is driving logic
//      bit 2 -- QC: is the QualityControl est passed in by L3_SM_A. Its always TRUE if GEOS5 or TSURF used
//      bit 3 -- SWATH: is the In or Out of swath indicator, where IsInSwath==0 to retain the goofy JPL convention.         
//      bit 6 -- TSURF? Is set TRUE when L3_SM_A's tsurf is explicitly specified to drive F/T logic, false otherwise
//      bit 7 -- RADAR? Is set TRUE only when ACTIVE RADAR is explicitly specified to drive F/T logic, false otherwise
//     FOR GEOS5 TS use-case only:
//      bit 4 -- is set TRUE to explicitly indicate when the GEOS5 TS was explicitly specified
//      bit 5 -- is set TRUE only when the pixel is InTerrestrial land domain, or, otherwise is FALSE
//     Set bit1 to TRUE if isInTerrestrial, or FALSE if not
// --------------------------------------------------------------

#ifndef L4C_MET_Defs
#define L4C_MET_Defs  1

// changed to make consistent across all executables 
#define L4C_MET_APP_TAG  "l4c_MET.ex"
#define L4C_MET_VERS_TAG "v7.3.0-2017-06-07T18:35:00-07:00"

// <JMG> 2014-05-15 added, for UM Framework integration 
#include "frm_api.h"

#include <assert.h>  // for assert()

// Import header for client API.s of frm_api
#include "nsb_api.h"
#include "pkg_mgr.h"

// import HDF4 and HDFEOS required constants
#ifdef HAVE_HDF4
#include "hdf.h"
#include "mfhdf.h"
  // verify if HdfEosDef ref is really required any longer
  // #include "HdfEosDef.h"
#endif

// import HDF5 
#ifdef HAVE_HDF5
#include "hdf5.h"
#include "hdf5_hl.h"
#endif

//  CONSTANTS

// n. PkgMgr themes (verify in l4c_MET_PkgMgrInit.c)
#define  L4C_MET_PKGMGR_THEME_COUNT     6

#define N_HOURS     1
#define N_ROWS    721
#define N_COLS   1152
#define N_FP_NELEM (N_ROWS*N_COLS)

// DEPRECATED: literal dataset name "PFT_M09_lc_mask" ,now uses name below: 
#define L4C_PFT_9KM_LC_MASK    "MCD12Q1_M09_lc_mask_uint8"
#define L4C_PFT_3KM_LC_MASK    "MCD12Q1_M03_lc_mask_uint8"

// new SMAP_L4_C_MET.h5 missing value
#define L4C_MET_MISSING_VALUE    -9999.0 

// L4_SM missing value
#define L4_SM_MISSING_VALUE      -9999.0 

// FP variables missing value
#define GEOS5_FP_MISSING_VALUE   9.9999999E14

// --------------------------------------------------------------
// L3_SM_A Related Constants

// L3_SM_A RADAR frozen ground bit (7) in "surface_flag" (uses new Coastal definition)
#define L3_SM_A_FREEZE_THAW_BIT           7 

// <JMG> 7/22/2014, Tsurf frozen ground bit (8) holds the Freeze Thaw {0|1}
#define L3_SM_A_TSURF_FREEZE_THAW_BIT     8

// L3_SM_A stores RADAR F/T quality indicator in bit 4, If 0=good 1=bad
// in L3_SM_A uint32 dataset 'retrieval_qual_flag'
#define L3_SM_A_FREEZE_THAW_QA_BIT        4 

// L3_SM_A missing value. for flt32, its often -9999.0, but check for
// what it is for integer scalars
#define L3_SM_A_MISSING_VALUE    0   // TBD

// When whole integer value of "surface_flag" is read without regard 
// to bit-mapped definitions, the OUT-OF-SWATH value is 32769
#define L3_SM_A_SURFFLAG_OUTSIDE_SWATH  32769
// Sentinel value in EASE_row_index or EASE_col_index for OUT-OF-SWATCH pixels
#define L3_SM_A_OUTSIDE_SWATH           32769

// TBD: Migrate away from using these to use those in RIP file
// Define fields inside of L3_SM_A where the F/T and F/T QA values are stored
// where: in dataset 'surface_flag' bit 6 is the RADAR F/T
//        in dataset 'retrieval_qual_flag' bit 4 is F/T QA as 0=good  1=bad
#define L3_SM_A_SURF_FLAG_DATASETNAME   "surface_flag"
#define L3_SM_A_FT_DATASETNAME          "/Soil_Moisture_Retrieval_Data/surface_flag"
#define L3_SM_A_FT_QA_DATASETNAME       "/Soil_Moisture_Retrieval_Data/retrieval_qual_flag"

// <JMG> 2015-07-31, Future: name of passive radiometer 36KM dataset substitute dataset, for F/T calc
#define L3_SM_P_FT_DATASETNAME       "/Soil_Moisture_Retrieval_Data/TBD"
// <JMG> future, TBD
#define L3_SM_P_FREEZE_THAW_BIT         -1   
#define L3_SM_P_TSURF_FREEZE_THAW_BIT   -1   
#define L3_SM_P_FREEZE_THAW_QA_BIT      -1 

// --------------------------------------------------------------

// n. hours to aggregate into a single daily average per variable
#define N_HOURS_PER_DAILY_AVG      24

#define MAX_MEMOBJ_VNAME_LEN      256
#define MAX_DATASET_NAME_LEN      MAX_MEMOBJ_VNAME_LEN

#define    NROW_GEOG         721         // GEOS5 FP geographic grid North-South dimension 
#define    NCOL_GEOG        1152         // GEOS5 FP geographic grid Ease-West dimension 

// #define DLAT_GEOG        0.5          // orig, MERRA lat resolution
// #define DLON_GEOG        0.6666666    // orig, MERRA lon resolution

// These geographic projection DLAT,DLON parameters are SPECIFIC to
// the current GEOS5 FP 721 row x 1152 col input product, so we change
// to a higher resolution (newer) GEOS variant, these MUST also be changed.
#define    DLAT_GEOG        0.25         // GEOS5 FP  lat resolution
#define    DLON_GEOG        0.3125       // GEOS5 FP lon resoution

#define     NROW1KM         14616         // EASEv2 1km grid North-South dimension
#define     NCOL1KM         34704         // EASEv2 1km grid East-West dimension
#define     N_1KM_PIX       (NROW1KM*NCOL1KM)   // total EASEv2 1km grid pixels
// n. pixels in EASEGRID 1km global land domain (n. pixels processed by L4_C suite)
#define     LDOM1KM        134787240

#define    NROW3KM          4872         // EASEv2 3km grid North-South dimension
#define    NCOL3KM         11568         // EASEv2 3km grid East-West dimension
// N_3KM_PIX --> NROW3KM * NCOL3KM --> 56359296....
#define    N_3KM_PIX       (NROW3KM * NCOL3KM)   // total EASEv2 3km grid pixels
#define     LDOM3KM         -1 // TBD

#define    NROW9KM          1624         // EASEv2 9km grid North-South dimension
#define    NCOL9KM          3856         // EASEv2 9km grid East-West dimension
#define    N_9KM_PIX       (NROW9KM *NCOL9KM)  // total EASEv2 9km grid  pixels
// n. pixels in EASEGRID 9km global land domain (n. pixels processed by L4_C suite)
#define     LDOM9KM         1664040

// aliases, just for naming consistency
#define    NROWS_9KM        NROW9KM      // EASEv2 9km grid North-South dimension
#define    NCOLS_9KM        NCOL9KM      // EASEv2 9km grid East-West dimension

// SERIALIZATION OFFSETS
// 2D offsets: y is outermost, x is innermost (fastest varying) term
#define M_2D_B0(x,y,n_y) (((x)*n_y)+(y))
#define M_2DR_B0(x,y,n_x) (((y)*n_x)+(x))


// borrow PI from math.h's M_PI, if not otherwise defined.
#ifndef PI
#define PI M_PI
#endif


// <JMG> 2015-09-08T09:25:00 added enum to match definition used in
// in L4CMDL. Reminder: surface_flag bit {15,15} used to classify if 
// pixel in swath or not, where 0= IS IN SWATH and 1=NOT IN SWATH
enum FreezeThawSwathChoices {
  IsInSwath, IsNotInSwath
} ;


// Threshold rule: where pix <= 0.0 deg, pixFtState = IsFrozen, and
// where pix > 0.0 pixFtState = IsThawed
enum FreezeThawStates {
 IsFrozen, IsThawed, IsFtStateUndetermined
} ;


enum FreezeThawSources {
 DoUse_GEOS5_TS, DoUse_ACTIVE_RADAR, DoUse_ACTIVE_TSURF, DoUse_PASSIVE_RADIOMETER
} ;

enum ActionStatusValues {
    ActionSuccess, ActionFailure, ActionPending, ActionStatusUnknown
} ;

// used by l4c_GridManager() and friends, to call mkgrid style
// pre (DEFLATE) or post (INFLATE) grid  processing
enum AppPhaseIdentifiers {
  DoPreProcessing, DoPostProcessing , DoDaily_TCF_Model_Process
} ;

// For name-value-pairs, define more specific, commonly supported numeric types
// NOTE: Order is important here for fwd,inv projection transform code until
// it is modified to make order unimportant
enum EASEGRIDv2_Types {
  IsEGv2_Global_9km, IsEGv2_Global_3km, IsEGv2_Global_1km,
  IsEGv2_N_9km,      IsEGv2_N_3km,      IsEGv2_N_1km
} ;


// structure storing local, long-lived objects, 
// Framework API overseer pointers, and raster images
typedef struct tagImgPkg {
 uint8 *pftImg9km_p ;    // ptr to EASEGRID 9km global mask  
 uint8 *pftImg3km_p ;    // ptr to EASEGRID 3km global mask  

 // L4C_MET_SOC_DATASET_FT_STATE_UM_M03
 uint8 *ftStateUm_M03_p; // ptr to ft_state_um_m03

 // Lat,Lon 9km coordinate var buffers, read from SMAP_L4_C_Ancillary.h5
 float32 *latImg_M09_p ; // SMAP_L4_C_LAT_1624_x_3856_M09_flt32
 float32 *lonImg_M09_p ; // SMAP_L4_C_LON_1624_x_3856_M09_flt32
 
 // local VPD input arrays
 float32 *inpVpd_t2m_avg_p; // T2M_AVG used in VPD calc
 float32 *inpVpd_qv2m_p   ; // QV2M used in VPD calc
 float32 *inpVpd_ps_avg_p ; // PS used in VPD calc
 float32 *outVpd_p        ; // VPD output 9km VPD..add to SMAP_L4_C_MET.5 when ready

 char   pft9kmDatasetname[64];  // name of 9km PFT mask dataset in HDF5 file
 char   pft3kmDatasetname[64];  // name of 3km PFT mask dataset in HDF5 file
 
 size_t nElem1km;           // n. of 1km pixels across rows,columns
 size_t nElem3km;           // n. of 3km pixels across rows,columns
 size_t nElem9km;           // n. of 9km pixels 
 int   nPkgThemes ;         // count of PkgMgr Themes used in this app
 NsbObj *NsbMgrOverseer_p ; // overseer for NSB (NVP overseer)
 PkgMgr *PkgMgrOverseer_p ; // overseer for PkgMgr services
} ImgPkg ;



// granuleStrTbl_p = parseRipLines(pftPath,ripParametersPath,fpRootDir,&nGranules,&modelYearday,logf);
typedef struct tagSesCtl {
    char  logPath[FILENAME_MAX]     ;
    char  ripPath[FILENAME_MAX]     ;
    // ancillary root directory
    char  ancRootDir[FILENAME_MAX];
    // GEOS5 FP root directory
    char  fpRootDir[FILENAME_MAX]   ;
    // L4_SM root directory
    char  l4smRootDir[FILENAME_MAX] ;
    // L3_SM_A root directory
    char  l3smaRootDir[FILENAME_MAX];
    // SOC root input directory
    char  socRootDir[FILENAME_MAX]  ;   
    // OUTPUT root directory
    char  outputMetRootDir[FILENAME_MAX]  ; 

    // base paths
    char  outputMetPath[FILENAME_MAX];
    char  inputAncPath[FILENAME_MAX]; // will replace andHdf5Path...    
    char  inputSOCPath[FILENAME_MAX];  // SOC path to write rolling 3km F/T and its QA to
    char  inputL3_SM_A_Path[FILENAME_MAX];  // L3_SM_A path
    
    // Define a subset of standard L4C_ANC_DATASET_xxx names
    char  dsNameLatCoordvar[64] ; // SMAP_L4_C_LAT_1624_x_3856_M09_flt32
    char  dsNameLonCoordvar[64] ; // SMAP_L4_C_LON_1624_x_3856_M09_flt32
    // L4C_ANC_DATASET_M09LAND_COL_INDEX = MCD12Q1_M09land_row_uint16
    char  dsNameLandRowIndex[64]; // MCD12Q1_M09land_col_uint16
    char  dsNameLandColIndex[64]; // MCD12Q1_M09land_row_uint16
    
    // New, stringtables storing dataset names of MET variables,
    // one stringtable per HDF5 or NetCDF file
    char **varList_GEOS5_FP ;
    char **varList_L3_SM_A  ;
    char **varList_L4_SM    ;
    // Stringtable storing list of all L4_SM paths
    char **inputL4_SM_pathList; 
    // string table , stores ALL GEOS5 FP paths
    char **inputGEOS5_FP_pathList; 

    char  appAction[64]; // L4C_MET_RUN_MODE, which 'action' main performs
    char  thisHost[256]; // hostname from: gethostname()
    // <JMG> 2015-07-31, New to support GEOS5 TS derived F/T
    char  ftSourceStr[64];// ACTIVE_RADAR|ACTIVE_TSURF|GEOS5_TS|PASSIVE_RADIOMETER etc
        
    // COMMON VERSIONING objects
    char  SubsysReleaseTag[128] ; // shared subsys RELEASE identifier
    char  swVersion[64] ; // SOFTWARE Version (tag), local executable
    // char  **ripTbl_p ; // ptr to master RIP NVP string table
    char  l4c_Task_Label[256]; // production run identifier
    char  l4c_Task_UUID[64]  ; // L4C_TASK_UUID
    char  l4c_Platform[128]  ; // L4C_PLATFORM ..name of (host) running this job    
    char  ScienceVersID[64];   // Note: matches Rolf's L4_SM name scheme
    char  CRID[64] ;   // legacy CRID may not be needed?
    char  frmApiVersion[64]; // UM Framework version
    char  pkgApiVersion[64]; // PkgMgr version
    char  nsbApiVersion[64]; // NSB version
    char  granuleUUID[64] ; // only uses 37 bytes of this for UUID 
    char  datasetVersion[64];     // L4C_DATASET_VERSION    

    char  enableCompletionMsg; // boolean, 0=suppress it 1=enable
    char  enableSummarizeVar ; // boolean, 0=suppress it 1=enable
    
    // COMMON temporal session state terms
    char  doy_Str[8]     ; // day-of-year string
    char  year_Str[8]    ; // YEAR YYYY string, year pertaining to this run
    char  period_Str[8]  ; // 8-day composite period in string form {01..46}
    char  period16_Str[8]; // 16-day composite period in string form {01..23}
    char  geos5ProxyVar[64]; // HDF5 dataset name of proxy (default: "TS")
    
    char    currVarname[64]; // for hand off of granuleMin,granuleMax,granuleMean
    // to variable level handlers
    float32 granuleMin; // for hand off and HDF5 "obs" attribute generation
    float32 granuleMax; // for hand off and HDF5 "obs" attribute generation
    float32 granuleMean;// for hand off and HDF5 "obs" attribute generation
    // <JMG> 2017-02-08T09:00 added granuleSum for symmetry
    float32 granuleSum ;// for hand off and HDF5 attribute generation
    
    int   ftSourceMode ; // DoUse_GEOS5|DoUse_ACTIVE_RADAR|DoUse_PASSIVE_RADIOMETER etc
    int   isL3_SM_A_Available; // boolean, is True(1) only when L3_SM_A is available
    int   runDayOfYear;     // day-of-year index {1..366} applying to THIS run  
    int   runModelYear;     // model year YYYY
    int   compressLev;   // gzip compress level 
      
    int   nTotalNvp  ;       // count, total n. of qualified NVPs returned by NSB parser
    int   nFP_Granules;      // count, n. of FP granules 
    int   nL4_SM_Granules  ; // count, n. of L4_SM granules
    int   nL3_SM_A_Granules; // count, should be 1 -- of L3_SM_A
  
    int  nFP_Variables;      // count of GEOS5 FP variables used
    int  nL4_SM_Variables;   // count of L4_SM variables used
    int  nL3_SM_A_Variables; // count of L3_SM_A variables used
    
    size_t nFrozenPixels; // verification counts
    size_t nThawedPixels; // verification counts
  
    // FT_STATE_UM_M03 bit frequencies
    size_t ftStateFrq[2] ;  // bit0 --F/T status, if Tsurf or GEOS5 TS, bit0 freqs
    size_t ftBit1Frq[2]  ;  // bit1 --FT_RADAR's F/T status, bit1 freqs
    size_t ftQcFrq[2]    ;  // bit2 --QC bit2 freqs, proxy for QC
    size_t ftInSwathFrq[2]; // bit3 --(IsInSwath=0),not in swath is 1
    size_t ftBit4Frq[2] ;   // bit4 -- TRUE only when GEOS5 TS is used
    size_t ftBit5Frq[2] ;   // bit5 -- InTerrestrial=1, NotInTerrestrial=0
    size_t ftBit6Frq[2] ;   // bit6 -- TRUE only when L3_SM_A present and ACTIVE TSURF used
    size_t ftBit7Frq[2] ;   // bit7 -- TRUE only when L3_SM_A_ present and ACTIVE RADAR used
  
    // GLOBAL. Only used in vert drill down, as a test, when macro DIAG_ONLY is defined
    int32  *terrAddr ;   // DIAG ONLY, a 1D vector of valid terrestrial addresses
  
    hid_t  dcpl9km ;     // for all 9km vars in FP and L4_SM
    hid_t  dspace9km ;   // for all 9km vars in FP and L4_SM
    
    hid_t  dcpl3km ;     // for all 3km vars in L3_SM_A
    hid_t  dspace3km ;   // for all 3km vars in L3_SM_A
  
    hid_t  l3smaFile_id ;  // file_id for single L3_SM_A input file
    hid_t  l3smaQaFile_id; // file_id for single L3_SM_A QA input file
    hid_t  l4csmFile_id ;  // L4_SM file_id for current one at a time
    hid_t  soc_file_id  ;  // SOC file id 
    hid_t  anc_file_id  ;  // SMAP_L4_C_Ancillary.h5 file id 
    hid_t  met_file_id  ;  // for SMAP_L4_C_MET.h5, storing all: FP,L4_SM,L3_SM_A

    time_t sesStrtSec ;  // session starting time, in sec
    time_t sesEndiSec ;  // session ending time, in sec
    double sesElapSec ;  // elapsed seconds for session  
 
    FILE *logf      ;    // ptr to session log file handle
} SesCtl ;


// end::L4C_MET_Defs block
#endif
