// ---------------------------------------------------------------
// module     : l4c_MET_PkgMgrInit.c module file
// Purpose    : Initializes PkgMgr (master scoping tool index containers) and
//              a collection of child PkgThemes.
// Revised    : 2014-07-23T16:29:00
// Revised    : 2014-06-04T18:06:00
// History    : n/a
// Returns    : int, status
// Notes      : CALLED BY: l4c_SesInit()
// Author     : joe glassy

#include "l4c_MET.h"
#include "l4c_MET_prot.h"


// ---------------------------------------------------------------
// function   : l4c_PkgMgr_Init()
// Purpose    : Initializes PkgMgr (master scoping tool index containers) and
//              a collection of child PkgThemes.
// Inputs     : int nThemes (currently 22) - count of at least as many
//               PkgMgr themes we'll use, with one per major ADT 
//               in tcf_struct.h, and others used for pre,post processing
// revised    : 2014-08-19T15:43, updated for new ancillary contents
// Revised    : 2014-07-22T13:18:00, re-org theme ordering and added detail
// Revised    : 2014-06-04T18:06:00-07:00,
// History    : n/a
// Returns    : int, status
// Notes      : CALLED BY: l4c_SesInit()
// Author     : joe glassy
int l4c_PkgMgr_Init(const int nThemes)
{
 SesCtl *sesCtl_p     = l4c_GetSesCtl();
 ImgPkg *imagePkg_p   = l4c_GetImagePackage();
 PkgMgr  *pkgMgr_p    = NULL;  // convenience ptr to session-level OVERSEER
 PkgTheme *pkgTheme_p = NULL;  // convenience ptr to middle "parent"
 int status =0;
 
 // pObjCounts[] array stores individual counts of (child) PkgObj's that 
 // each defined theme {0..6} owns
 
 // ASSERT: we always size 'L4C_MET_THEME_COUNT' to be >= the n. of 
 // actually used.
 int  pObjCounts[L4C_MET_PKGMGR_THEME_COUNT] = {0} ;
 
    // fetch pointer to PkgMgr overseer
    pkgMgr_p = imagePkg_p->PkgMgrOverseer_p ;
    
    // <JMG> TBD 7/17/2014, these PkgObj counts here are PLACEHOLDERS until refined...
    
    pObjCounts[ 0] = 22 ;  // SMAP_L4_C_Ancillary, 8/19/2014 updated
    pObjCounts[ 1] = 2  ;  // SMAP_L4_C_SOC.h5 datasets, now with file-level info
    // INDIVIDUAL MET INPUTS
    pObjCounts[ 2] = 2  ;  // SMAP_L4_SM datasets, now with file-level info
    pObjCounts[ 3] = 2  ;  // SMAP_L3_SM_A datasets, now with file-level info
    pObjCounts[ 4] = 2  ;  // GEOS5_FP datasets, now with file-level info
    pObjCounts[ 5] = 2  ;  // SMAP_L4_C_MET datasets, now with file-level info
    
    // Reminder: pkgMgrCreateThemes uses the above counts, in the
    //   pObjCounts[] array, to assign each of nThemes with this number of
    //   child (PkgObj) entries carried in the pObjCounts[] slot
    pkgMgr_CreateThemes(pkgMgr_p, nThemes, pObjCounts, "L4_C_MET");
    
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 0);
    l4c_PkgInit_SMAP_L4_C_Ancillary(  pkgTheme_p,"SMAP_L4_C_Ancillary");    

    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 1);
    l4c_PkgInit_SMAP_L4_C_SOC(        pkgTheme_p,"SMAP_L4_C_SOC");

    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 2);
    l4c_PkgInit_SMAP_L4_SM(        pkgTheme_p,"SMAP_L4_SM");

    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 3);
    l4c_PkgInit_SMAP_L3_SM_A(  pkgTheme_p,"SMAP_L3_SM_A");
    
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 4);
    l4c_PkgInit_GEOS5_FP( pkgTheme_p,"GEOS5_FP");

    // L4CMET preprocessor OUTPUTS definitions go here
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 5);
    l4c_PkgInit_L4_C_MET(pkgTheme_p,"SMAP_L4_C_MET");

    
    l4c_EventCompletionMsg((char *)__func__);

    return status;
    
} // end:: l4c_PkgMgr_Init()


// ---------------------------------------------------------------
// function : l4c_PkgInit_SetFileProperties()
// purpose  : For theme "SMAP_INPUT_HDF5_FILES" set key file properties
// inputs   : char *themeName_p
// notes    : CRITICAL: wait to call THIS function until AFTER all
//            themes have been named, and until after all HDF5 files
//            have been opened, so their file_id's are available
// notes    : CALLED BY: l4c_SesInit() 
// notes    : As long as the file_id's don't change this function may 
//             safely be called multiple times, where perhaps not all the
//             files were opened the first time, but were opened by
//             subsequent calls...
int l4c_PkgInit_SetFileProperties(void)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl(); // ptr to session controls structure
  ImgPkg *imagePkg_p = l4c_GetImagePackage();
  // declare the PkgMgr mechanism parts we need
  PkgMgr *pkgMgr_p  = NULL; // ptr to overseer
  PkgTheme *pThme_p = NULL; // ptr to theme (parent) instance
  PkgObj   *pObj_p  = NULL; //  ptr to child instances
 
    // set pointer to master instance of the PkgMgr overseer
    pkgMgr_p = imagePkg_p->PkgMgrOverseer_p;

    // pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_INPUT_HDF5_FILES");
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_Ancillary");
    // <JMG> 2014-05-19, set SELF-REFERENTIAL file info into its own theme
    pObj_p = pkgMgr_GetObjByName(pThme_p,"SMAP_L4_C_Ancillary");
    // Use owner id to store this HDF5 file's file_id. when it is available
    pkgMgr_SetObjOwner(pObj_p   , sesCtl_p->anc_file_id);
    pkgMgr_SetObjMetadata(pObj_p, sesCtl_p->inputAncPath);
    pkgMgr_SetObjDatatype(pObj_p, IsString);    
    // I use Nelem here store count of DATASETS in file
    // pkgMgr_SetObjNelem(pObj_p, 9);
    // resolution, rank, Fillvalue not relevant here

    // DAILY INTERMEDIATE and DAILY OUTPUT
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"SMAP_L4_C_SOC");  
    pObj_p = pkgMgr_GetObjByName(pThme_p,"SMAP_L4_C_SOC");
    pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);  
    pkgMgr_SetObjMetadata(pObj_p, sesCtl_p->inputSOCPath);
    
    // DAILY OUTPUT 
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"SMAP_L4_C_MET");
    pObj_p = pkgMgr_GetObjByName(pThme_p,"SMAP_L4_C_MET");
    pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id) ;
    pkgMgr_SetObjMetadata(pObj_p, sesCtl_p->outputMetPath);
    
    // INDIVIDUAL INPUTS
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"SMAP_L4_SM"); 
    pObj_p = pkgMgr_GetObjByName(pThme_p,"SMAP_L4_SM");
    // Since there a multiple L4_SM, we don't define ObjOwner as a
    // single file_id as we did in L4_C
    pkgMgr_SetObjOwner(pObj_p,-1);  
    // pkgMgr_SetObjMetadata(pObj_p, sesCtl_p->inputFparPath);
    
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"SMAP_l3_SM_A");
    pObj_p = pkgMgr_GetObjByName(pThme_p,"SMAP_L3_SM_A");
    pkgMgr_SetObjOwner(pObj_p, sesCtl_p->l3smaFile_id );
    pkgMgr_SetObjMetadata(pObj_p, sesCtl_p->inputL3_SM_A_Path);
    
    
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"GEOS5_FP");
    pObj_p = pkgMgr_GetObjByName(pThme_p,"GEOS5_FP");
    // pkgMgr_SetObjMetadata(pObj_p, sesCtl_p->inputGMetClimatology);
    // placeholder since there are multiple FP granules
    pkgMgr_SetObjOwner(pObj_p, -1);

    l4c_EventCompletionMsg( (char *)__func__);

 return 0;  
    
} // end:: l4c_PkgInit_SetFileProperties()


// ---------------------------------------------------------------
// function : l4c_PkgInit_SMAP_L4_C_Ancillary()
// purpose  : Set properties of indicated theme
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p
// revised  : 2014-08-19T15:42:00, updated to reflect new Ancillary contents
// notes    : TBD double check that the datasets here exactly MATCH
//            those present in the latest Ancillary, as these do NOT
//            match exactly now...

// notes    : CRITICAL: wait to call THIS function until AFTER all
//            themes have been named, and until after all HDF5 files
//            have been opened, so their file_id's are available
// notes    : CALLED BY: l4c_SesInit() 
// notes    : As long as the file_id's don't change this function may 
//             safely be called multiple times, where perhaps not all the
//             files were opened the first time, but were opened by
//             subsequent calls...
/*
   DATASET "L4C_M01_PixFilter_uint8" {
   DATASET "L4C_M01land_PixFilter_uint8" {
   DATASET "MCD12Q1_M01_lc_dom_uint8" {
   DATASET "MCD12Q1_M01_lc_mask_uint8" {
   DATASET "MCD12Q1_M01land_lc_dom_uint8" {
   DATASET "MCD12Q1_M03_lc_mask_uint8" {
   DATASET "MCD12Q1_M09_lc_mask_uint8" {
   DATASET "MCD12Q1_M09land_col_uint16" {
   DATASET "MCD12Q1_M09land_row_uint16" {
   DATASET "MOD12Q1_M09_lc_dom_uint8" {
   DATASET "SMAP_L4_C_LAT_14616x34704_M01_flt32" {
   DATASET "SMAP_L4_C_LAT_1624_x_3856_M09_flt32" {
   DATASET "SMAP_L4_C_LON_14616x34704_M01_flt32" {
   DATASET "SMAP_L4_C_LON_1624_x_3856_M09_flt32" {
   DATASET "tile_c_M01_1km_ease2_template_int16" {
   DATASET "tile_h_M01_1km_ease2_template_uint8" {
   DATASET "tile_r_M01_1km_ease2_template_int16" {
   DATASET "tile_v_M01_1km_ease2_template_uint8" {

 */
void l4c_PkgInit_SMAP_L4_C_Ancillary(PkgTheme *pkgTheme_p,char *themeLabel_p)
{

    pkgMgr_SetThemeName(pkgTheme_p, themeLabel_p);  
    
    // these are all HDF5 dataset names in the ancillary .h5 
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "L4C_M01_PixFilter_uint8");
    // note: Currently for pixel filtering THIS IS THE ONE WE USE, M01land_PixFilter 
    // since it is already deflated...
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "L4C_M01land_PixFilter_uint8");
    
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "MCD12Q1_M01_lc_dom_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "MCD12Q1_M01_lc_mask_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "MCD12Q1_M01land_lc_dom_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "MCD12Q1_M03_lc_mask_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "MCD12Q1_M09_lc_mask_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, "MCD12Q1_M09_lc_nmargveg_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, "MCD12Q1_M09_lc_nnonveg_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "MCD12Q1_M09_lc_nveg_uint8" );
    pkgMgr_SetThemes_ChildName(pkgTheme_p,10, "MCD12Q1_M09land_col_uint16");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,11, "MCD12Q1_M09land_row_uint16");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,12, "MCD12Q1_M09_lc_dom.uint8");

    pkgMgr_SetThemes_ChildName(pkgTheme_p,13, "SMAP_L4_C_LAT_1624_x_3856_M09_flt32");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,14, "SMAP_L4_C_LON_1624_x_3856_M09_flt32");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,15, "SMAP_L4_C_LAT_14616_x_34704_M01_flt32");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,16, "SMAP_L4_C_LON_14616_x_34704_M01_flt32");
    
    pkgMgr_SetThemes_ChildName(pkgTheme_p,17, "tile_c_M01_1km_ease2_template_int16");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,18, "tile_h_M01_1km_ease2_template_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,19, "tile_r_M01_1km_ease2_template_int16");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,20, "tile_v_M01_1km_ease2_template_uint8"); 
    // <JMG> 2014-05-19, added one to represent self-referential, file level info
    pkgMgr_SetThemes_ChildName(pkgTheme_p,21, "SMAP_L4_C_Ancillary"); 
    
    l4c_EventCompletionMsg( (char *)__func__);
    
} // end:: l4c_PkgInit_SMAP_L4_C_Ancillary()


// ---------------------------------------------------------------
// function : l4c_PkgInit_xxx
// purpose  : Set properties of indicated theme
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p
// revised  : 2014-06-04
// notes    : n/a
void l4c_PkgInit_SMAP_L4_C_SOC(PkgTheme *pkgTheme_p,char *themeLabel_p)
{
 char fxName[256] = {0} ;
 
    strcpy(fxName,__func__);
    xLogMsg("@A In(%s) NOT-YET-REQUIRED-OR-IMPLEMENTED\n",fxName);
    
} // end::l4c_PkgInit_SMAP_L4_C_SOC()
    

// ---------------------------------------------------------------
// function : l4c_PkgInit_xxx
// purpose  : Set properties of OUTPUT SMAP_L4_C_MET.h5 theme
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p
// revised  : 2014-06-04
// notes    : n/a
void l4c_PkgInit_L4_C_MET(PkgTheme *pkgTheme_p,char *themeLabel_p)
{
 char fxName[256] = {0} ;
 
    strcpy(fxName,__func__);
    xLogMsg("@A In(%s) NOT-YET-REQUIRED-OR-IMPLEMENTED\n",fxName);
    
} // end::l4c_PkgInit_L4_C_MET()


    
// ---------------------------------------------------------------
// function : l4c_PkgInit_xxx
// purpose  : Set properties of indicated theme
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p
// revised  : 2014-06-04
// notes    : n/a
void l4c_PkgInit_SMAP_L4_SM(PkgTheme *pkgTheme_p,char *themeLabel_p)
{
 char fxName[256] = {0} ;
 
    strcpy(fxName,__func__);
    xLogMsg("@A In(%s) NOT-YET-REQUIRED-OR-IMPLEMENTED\n",fxName);
    
} // end::l4c_PkgInit_SMAP_L4_SM()


// ---------------------------------------------------------------
// function : l4c_PkgInit_xxx
// purpose  : Set properties of indicated theme
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p
// revised  : 2014-06-04
// notes    : n/a
void l4c_PkgInit_SMAP_L3_SM_A(PkgTheme *pkgTheme_p,char *themeLabel_p)
{
 char fxName[256] = {0} ;
 
    strcpy(fxName,__func__);
    xLogMsg("@A In(%s) NOT-YET-REQUIRED-OR-IMPLEMENTED\n",fxName);
    
} // end::l4c_PkgInit_SMAP_L3_SM_A()


// ---------------------------------------------------------------
// function : l4c_PkgInit_xxx
// purpose  : Set properties of indicated theme
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p
// revised  : 2014-06-04
// notes    : n/a
void l4c_PkgInit_GEOS5_FP(PkgTheme *pkgTheme_p,char *themeLabel_p)
{
 char fxName[256] = {0} ;
 
    strcpy(fxName,__func__);
    xLogMsg("@A In(%s) NOT-YET-REQUIRED-OR-IMPLEMENTED\n",fxName);
    
} // end::l4c_PkgInit_GEOS5_FP()
