// ---------------------------------------------------------------------
//   module     : l4c_MET_fx.c
//   Purpose    : implement support functions for l4c_MET.c
//   Author     : joe glassy
//   Version    : v7 see AppVers macro in l4c_MET.h
//   Revision   : 2015-08-25T15:02, added calc_NativeRange() for diags.
//   Revision   : 2014-10-21, moved l4c_calc_vpd() to new l4c_MET_sci.c module
//   Revision   : 2014-07-22, re-integration into L4_C v5.0.1 in progress
//                and split out logic orig in l4c_MET_fx into:
//                -- l4c_MET_drivers.c
//                -- l4c_MET_drilldowns.c
//                -- l4c_MET_io.c
//                -- l4c_MET_grid_proj.c
//                -- l4c_MET_ses.c
//                 -- l4c_MET_PkgMgrInit.c

//   History    : NEW 2/12/2014,, added uint8, 3km 2D "freezeThawAge_p" layer, 
//                which tracks how many days OLD the Freeze Thaw 
//              value "LAST-GOOD" was, as an integer count {0,1,2,3,4...}
//
//   Revision   : 2014-07-22, see AppVers macro in l4c_MET.h
//   Note       : moved l4c_calc_vpd() to l4c_MET_sci.c to match how l4c_DAILY is done
//   Notes      : Several new fields added for D5 representing either new
//                capabilities needed to support the use of F/T in L4_C_DAILY,
//                or (VPD) that were omitted earlier.  New fields are:
//                -- FT_ROLLING_AGE_3KM  ...from Freeze Thaw AGE (uint8,3km)
//                -- vpd (float32, 9KM, "units=Pa") in SMAP_L4_C_MET.h5
//                -- smroot_min (float32, 9KM) in SMAP_L4_C_Ancillary.h5
//                -- smroot_max (float32, 9KM) in SMAP_L4_C_Ancillary.h5
//                 and new static 3KM float32 layers "smroot_min","smroot_max"
//              in SMAP_L4_C_Ancillary.h5
//

// module level scope variables...
// NOTE: <JMG> 2014-07-22, fetch all input dataset names now 
// from RIP file using nsbGetValueByKey() calls
// NOTE: DEPRECATED, endian labels. Now use UM Frameworks..
// char *endianLabels[] = { "big", "little", "vaxEndian","unknownEndian"} ;  
// Note: moved l4c_calc_vpd() to l4c_MET_sci.c to match how l4c_DAILY is done

#include "l4c_MET.h"
#include "l4c_MET_prot.h"


                      
// -----------------------------------------------------------------------------
// function: rescaleFloatImage(float32 *srcFltImg_p,
//           const size_t srcRowCount,const size_t srcColCount,const size_t destRowCount,size_t destColCount)
// Purpose: Rescale a float32 source image to a destination image where
//          the raster dimensions of both are supplied, returning a pointer
//          to the new destination (float32) image
// Inputs : float32 *srcFltImg_p -- ptr to original source image
//          const size_t srcRowCount -- source image number of rows
//          const size_t srcColCount,-- source image number of columns
//          const size_t destRowCount --destination image number of rows
//          size_t destColCount       -- destination image number of columns
// returns: float32 *newDestFlt32Img_p  -- ptr to new rescaled image 
// revised: 2015-09-04T10:07
// author : Joe Glassy
// NOTES  : Uses nearest neighbor rescale method.
// notes  : TBD, re-write using TIV datatypes as a single universal
//          type independent implementation. This could also be
//          refactored to use all integer operations, using x<<16 hack etc.
float32 *rescaleFloatImage(float32 *srcFltImg_p,
         const size_t srcRowCount,const size_t srcColCount,
         const size_t destRowCount,const size_t destColCount)
{
 char fxName[256] = {0};
 float32 *destFltImg_p= NULL;
 size_t ithRow=0;
 size_t ithCol=0;
 size_t nDestGridElem = (destRowCount * destColCount);
 size_t nSrcGridElem  = (srcRowCount * srcColCount);
 size_t srcIndex = 0 ; // base0 1D serialized index into 2D source image
 size_t destIndex= 0 ; // base0 1D serialized index into 2D destination image
 // scaling ratios in both axes for nearest neighbor
 double x_ratio = (double)srcColCount/(double)destColCount ;
 double y_ratio = (double)srcRowCount/(double)destRowCount ;
 double px = 0.0, py = 0.0 ; 
 
 
    // CREATE buffer for 2D new destination raster (image)
    destFltImg_p = memCalloc(nDestGridElem,sizeof(float32));
    
    // POPULATE new image using original image data, rescaling as we go   
    for(ithRow=0;ithRow < destRowCount;ithRow++)
    {
        for(ithCol=0; ithCol < destColCount;ithCol++)
        {
            px = floor(ithCol * x_ratio) ; // coerces ithCol to float in expression
            py = floor(ithRow * y_ratio) ; // coerces ithRow to float in expression
            // verify in C implementation if (py) needs to be rounded prior
            srcIndex = (size_t)((py * srcColCount) + px); // integer truncation...
            destIndex= (ithRow*destColCount) + ithCol;
            destFltImg_p[ destIndex ] = srcFltImg_p[ srcIndex ] ;
        } // end:: ithCol loop
    } // end:: ithRow loop
 // return pointer to float32 destination 2D raster grid (image)
 return destFltImg_p ;    
} // end:: rescaleFloatImage()         

// -----------------------------------------------------------------------------
// function: rescaleByteImage(uint8 *srcFltImg_p,
//           const size_t srcRowCount,const size_t srcColCount,const size_t destRowCount,size_t destColCount)
// Purpose: Rescale a uint8 (byte) source image to a destination image where
//          the raster dimensions of both are supplied, returning a pointer
//          to the new destination (uint8) image
// Inputs : uint8 *srcByteImg_p -- ptr to original (byte) source image
//          const size_t srcRowCount -- source image number of rows
//          const size_t srcColCount,-- source image number of columns
//          const size_t destRowCount --destination image number of rows
//          size_t destColCount       -- destination image number of columns
// returns: uint8 *newDestByteImg_p  -- ptr to new rescaled image 
// revised: 2015-09-04T10:07
// author : Joe Glassy
// NOTES  : Uses nearest neighbor rescale method
uint8 *rescaleByteImage(uint8 *srcByteImg_p,
         const size_t srcRowCount,const size_t srcColCount,
         const size_t destRowCount,const size_t destColCount)
{
 uint8 *destByteImg_p= NULL;
 size_t ithRow=0;
 size_t ithCol=0;
 size_t nDestGridElem = (destRowCount * destColCount);
 size_t nSrcGridElem  = (srcRowCount * srcColCount);
 size_t srcIndex = 0 ; // base0 1D serialized index into 2D source image
 size_t destIndex= 0 ; // base0 1D serialized index into 2D destination image

 double px = 0.0, py = 0.0 ; 
 // establish scaling ratios in both axes for nearest neighbor
 double x_ratio = (double)srcColCount/(double)destColCount ;
 double y_ratio = (double)srcRowCount/(double)destRowCount ;
 
    // CREATE buffer for 2D new destination raster (image)
    destByteImg_p = memCalloc(nDestGridElem,sizeof(uint8));
    // POPULATE new image using original image data, rescaling as we go   
    for(ithRow=0;ithRow < destRowCount;ithRow++)
    {
        for(ithCol=0; ithCol < destColCount;ithCol++)
        {
            px = floor(ithCol * x_ratio) ; // coerces ithCol to float in expression
            py = floor(ithRow * y_ratio) ; // coerces ithRow to float in expression
            srcIndex = (size_t)((py * srcColCount) + px); // integer truncation...
            destIndex= (ithRow * destColCount) + ithCol;
            destByteImg_p[destIndex] = srcByteImg_p[srcIndex] ;
        } // end:: ithCol loop
    } // end:: ithRow loop
 // return pointer to byte (uint8) destination 2D raster grid (image)
 return destByteImg_p ;    
} // end:: rescaleByteImage()

// -----------------------------------------------------------------------------
// function: granuleHHMM_to_HrIndex()
// Purpose : extract the 'HHMM' timestamp from the current days FP granule pathname
//          and return this as a base0 integer, as the daily hour index {0..24-1}
// Inputs  : granulePath -- string containing representative GEOS-5 FP path to use as
//          seed material to retrieve the YYYY and HHMM info from. An example filename
//          is:  "GEOS.fp.asm.tavg1_2d_slv_Nx.20130611_0030.V01.nc4"
// Outputs : hhmmString -- a new hhmm string
//          yyyyStr    -- a new YYYY string
// NOTES  : This function originally intended only for GEOS-5 FP context, so 
//          TBD to verify if filename parsing works for L4_SM and L3_SM_A (probably not)
int32 granuleHHMM_to_HrIndex(char *granulePath,char *hhmmString,char *yyyyStr)
{
 int32 k =0,hourIndex= -1;
 char *p = NULL;
 char hhmmStr[8] = {0} ;
 char hhStr[4]= {0};
 
    // refine method to extact out the hh:mm identifier in the FP granule
    // GEOS.fp.asm.tavg1_2d_slv_Nx.20130508_0430.V01.nc4 
    k = strlen(granulePath); // capture the initial total length of the granule
    
    p = strrchr(granulePath,'_'); // find the right-most instance of navigator token, '_' here.
    if(k >= 8 && p != NULL)
    {
        // WARNING: if granulePath is not exactly formed as specified, this will
        // cause a segfault, so TBD--> enclose in a more robust check...
        strncpy(hhmmStr,granulePath+(k-strlen(p)+1),4);
        // fprintf(stdout,"@A granuleHHMM_to_HrIndex:: granule-path(%s) --> hhmm(%s) \n",granulePath,hhmmStr);
    } else {
        fprintf(stderr,"@E granuleHHMM_to_HrIndex::invalid granule-path(%s)\n",
            granulePath);
        exit(-1);
    }
    
    if((p = strstr(granulePath,"_Nx."))!=NULL)
    {
        strcpy(yyyyStr,"");
        // fprintf(stdout,"@A granuleHHMM_to_HrIndex:: _Nx p(%s)\n",p);
        strncpy(yyyyStr,granulePath+(k-strlen(p))+4,4);
        // fprintf(stdout,"@A granuleHHMM_to_HrIndex:: yyyyStr (%s)\n",yyyyStr);
        strncpy(hhStr,hhmmStr,2);
        hourIndex = atoi(hhStr); // now, is {0..23}
        strcpy(hhmmString,hhmmStr);
        // fprintf(stdout,"@A granuleHHMM_to_HrIndex:: granule-path(%s) --> yyyy(%s) hh(%s)\n",
        //  granulePath,yyyyStr,hhStr);
    } else {
        fprintf(stderr,"@E granuleHHMM_to_HrIndex::invalid granule-path(%s) missing _Nx\n",
            granulePath);
        exit(-1);
    }

#ifdef DIAG_ONLY    
    fprintf(stdout,"@A HHMM-DIAG Granule[%s] hhmmStr(%s) hhStr (%s) yr(%s) hourIndex %d\n",
        granulePath,hhmmStr,hhStr,yyyyStr,hourIndex);
#endif
    
 return hourIndex;
} // end::granuleHHMM_to_HrIndex()


// -----------------------------------------------------------------------------
// l4_sm_granuleHHMM_to_HrIndex()
// Purpose: extract the 'HHMM' timestamp from the current days FP granule pathname
//          and return this as a base0 integer, as the daily hour index {0..24-1}
// Inputs : granulePath -- string containing representative GEOS-5 FP path to use as
//          seed material to retrieve the YYYY and HHMM info from. An example filename
//          is:  "GEOS.fp.asm.tavg1_2d_slv_Nx.20130611_0030.V01.nc4"
// Outputs: hhmmString -- a new hhmm string
//          yyyyStr    -- a new YYYY string
// NOTES  : This function originally intended only for GEOS-5 FP context, so 
//          TBD to verify if filename parsing works for L4_SM and L3_SM_A (probably not)
int32 l4_sm_granuleHHMM_to_HrIndex(char *granulePath,char *hhmmString,char *yyyyStr)
{
 int32 k =0,hourIndex= -1;
 char *p = NULL;
 char hhmmStr[8] = {0} ;
 char hhStr[4]= {0};
 
    // refine method to extact out the hh:mm identifier in the L4_SM granule
    // SMAP_L4_SM_gph_20130720T013000_V10001_000.h5
    k = strlen(granulePath); // capture the initial total length of the granule
    
    // TASK 1: isolate the hhmm portion of filename, store in hhmmStr.
    p = strstr(granulePath,"T") ;
    if(k >= 8 && p != NULL)
    {
        // WARNING: if granulePath is not exactly formed as specified, this will
        // cause a segfault, so TBD--> enclose in a more robust check...
        strncpy(hhmmStr,granulePath+(k-strlen(p)+1),4);
        // fprintf(stdout,"@A l4_sm_granuleHHMM_to_HrIndex:: granule-path(%s) --> hhmm(%s) \n",granulePath, hhmmStr);
    } else {
        // fprintf(stderr,"@E l4_sm_granuleHHMM_to_HrIndex::invalid granule-path(%s)\n",granulePath);
    }

    // TASK 2: isolate yyyy from filename
    // SMAP_L4_SM_gph_20130720T013000_V10001_000.h5 
    if((p = strstr(granulePath,"gph_"))!=NULL)
    {
        strcpy(yyyyStr,"");
        // fprintf(stdout,"@A granuleHHMM_to_HrIndex:: _Nx p(%s)\n",p);
        strncpy(yyyyStr,granulePath+(k-strlen(p))+4,4);
        // fprintf(stdout,"@A granuleHHMM_to_HrIndex:: yyyyStr (%s)\n",yyyyStr);
        strncpy(hhStr,hhmmStr,2);
        hourIndex = atoi(hhStr); // now, is {0..23}
        strcpy(hhmmString,hhmmStr);
        // fprintf(stdout,"@A granuleHHMM_to_HrIndex:: granule-path(%s) --> yyyy(%s) hh(%s)\n",granulePath,yyyyStr,hhStr);
        
    } else {
        fprintf(stderr,"@E granuleHHMM_to_HrIndex::invalid granule-path(%s) missing _gph_\n",
            granulePath);
    }
    
#ifdef DIAG_ONLY    
    fprintf(stdout,"@A HHMM-DIAG Granule[%s] hhmmStr(%s) hhStr (%s) yr(%s) hourIndex %d\n",
        granulePath,hhmmStr,hhStr,yyyyStr,hourIndex);
#endif
    
 return hourIndex;
} // end::l4_sm_granuleHHMM_to_HrIndex()

// ----------------------------------------------------------------
// Purpose: l4c_to_upper(char *targString)
char *l4c_to_upper(char *targString)
{
 int i, sLen = strlen(targString);
 
    for(i=0; i < sLen; i++)
    {
        targString[i] = toupper(targString[i]);
    }

    // return transformed string, as a convenience 
    return targString;

} // end::l4c_to_upper()


// ----------------------------------------------------------------
// Function : rowReversal()
// Purpose  : perform a North-to-South "vertical-flip" on rows, since
//            the FP datasets are natively all upside down. 
//            Currently this function is not needed due to
//            the way the related reprojection transform logic works.
// Inputs   : float32 *src_p,const int32 nRows,const int32 nCols
// Outputs  : void *, ptr to adjusted buffer
void *rowReversal(float32 *src_p,const int32 nRows,const int32 nCols)
{
 int r;
 int offset1=0;    // initialize this offset for ascending row sequence
 int offset2=(nRows-1)*nCols ; // init for descending row sequence
 int32 nBytes = nCols * sizeof(float32); // establish one row worth of bytes
 float32 *dest_p = NULL;
 
 dest_p=(float32 *)memCalloc((nRows *nCols),sizeof(float32)) ;
 memObjSetName(dest_p,"rowReversal_GEOS5_FP_Buffer");
 
    for(r=0; r < nRows; r++)
    {
        memcpy(dest_p+offset2, src_p+offset1, nBytes);
        // only offset1 is currently used...
        offset1 += (nCols); // adjust, to move in ascending row order
        offset2 -= (nCols); // adjust, to move in descending row order
    }
    
    // RESULT: adjusted grid in dest_p, so copy back to original source now
    memcpy(src_p,dest_p,nBytes);
    
    // free up row-reversal working array
    memFree(dest_p);
 
 // return ptr to adjusted source grid 
 return src_p;
} // end::rowReversal()


// ------------------------------------------------------------------
// function: fp_summarizeVariable()
// inputs  : const float32 allowableMin,
//           const float32 allowableMax,
//           char *datasetName,
//           hsize_t *dims,
//           float32 *dsNativeInputArray_p,
//           float32 *hourlyMin,
//           float32 *hourlyMax
// Purpose : calculate min,max,mean,sum for each 'float32' datasetName..
// Returns : mean value across all spatial pixels for THIS hourly dataset
// notes   : Current implementation is kind of a messy hack right now
//           since it is dependent on one type of input data (GEOS5 FP)
float32 fp_summarizeVariable(const float32 allowableMin,
                             const float32 allowableMax,
                             char *datasetName,hsize_t *dims,
                             float32 *dsNativeInputArray_p,
                             float32 *hourlyMin,float32 *hourlyMax)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 char fxName[256] = {0} ;
 
 float32 pix,vMin,vMax,vSum,vMean;
 float32 missing_value = GEOS5_FP_MISSING_VALUE ; // 9.9999999E14 
 size_t t,r,c,n =0;
 size_t offset = 0;
 
 vMin = vMax = vSum = vMean = 0.0;
 
 strcpy(fxName,__func__);
 
 for(n=0,t=0; t < dims[0]; t++)
 {
    for(r=0; r < dims[1]; r++)
    {
        for(c=0; c < dims[2]; c++)
        {
            // WARNING offset scheme here ONLY for GEOS5 3D data arrays
            // that include time as 1st dim [1x721x1152]
            offset = ((t*dims[1]*dims[2]) + ((r*dims[2])+c)); 
            
            pix = dsNativeInputArray_p[offset];
         
            // Select pixel only if in valid range, where allowableMin
            // may be a negative number if known. Here we assume that
            // missing value is assigned to be VERY HIGH such that valid
            // values are always below it. This works for GEOS5 variables
            // but NOT when the missing_value of input is like -9999.0 etc!
            if(pix >= allowableMin && pix < missing_value)
            {
                vSum += pix;         
                if(n == 0)
                {
                    vMin = vMax = pix;
                } else {
                    if(pix > vMax) vMax = pix;
                    if(pix < vMin) vMin = pix;
                } // end:: n min,max block
                n++ ; // increment count of qualified pixels
            } // end:: if pix qualified to track
        } // end:: c loop
    } // end:: r loop
 }
 // store  hourly (min,max) across whole spatial domain for later
 *hourlyMin = vMin;
 *hourlyMax = vMax;

 vMean = vSum / (float32)n;
 
 if(sesCtl_p->enableSummarizeVar)
    xLogMsg("@A In(%s) STATS ON %8s N %8zu Min %7.3f Max %7.3f mean %12.5f Sum %.2f ~\n",
        fxName,datasetName, n, vMin,vMax, vMean,vSum);
        
 return vMean;
} // end::fp_summarizeVariable()


// ------------------------------------------------------------------
// l4_sm_summarizeVariable()
// Purpose: calculate min,max, mean,sum for each 'float32' datasetName..
// Return: mean value across all spatial pixels for THIS hourly dataset
float32 l4_sm_summarizeVariable(FILE *logf,char *datasetName,hsize_t *dims,float32 *dsData_p,float32 *hourlyMin,float32 *hourlyMax)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 
 float32 pix,vMin,vMax,vSum,vMean;
 float32 missing_value = L4_SM_MISSING_VALUE ; // L4_SM_MISSING_VALUE -9999.0
 size_t  r,c,n  = 0;
 size_t  offset = 0;
 size_t  nMissing=0;
 char fxName[256] = {0} ;
 
  strcpy(fxName,__func__);
 
 vMin = vMax = vSum = vMean = 0.0;
 
    for(r=0; r < dims[0]; r++)
    {
        for(c=0; c < dims[1]; c++)
        {
            offset = (r * dims[1])+c ;
            pix = dsData_p[offset];
            // missing value -9999.0 so valid qualifiers must be > than this
            if(pix > missing_value)
            {
                vSum += pix;         
                if(n == 0)
                {
                    vMin = vMax = pix;
                } else {
                    if(pix > vMax) vMax = pix;
                    if(pix < vMin) vMin = pix;
                } // end:: n min,max block
                n++ ; // increment count of qualified pixels
            } else { // end:: if pix qualified to track
                nMissing++ ;
            }
        } // end:: c loop
    } // end:: r loop
 
 // store  hourly (min,max) across whole spatial domain for later
 *hourlyMin = vMin;
 *hourlyMax = vMax;

 vMean = vSum / (float32)n;
 // output DIAG
 if(sesCtl_p->enableSummarizeVar)
    xLogMsg("@A In(%s) STATS ON %8s N %8zu Min %7.3f Max %7.3f mean %12.5f Sum %.2f (Missing %zu)~\n",
        fxName,datasetName, n, vMin,vMax, vMean,vSum, nMissing);

    // store in sesCtl for subsequent retrieval
    // for hand off of granuleMin,granuleMax
    strcpy(sesCtl_p->currVarname, datasetName);
    sesCtl_p->granuleMin = vMin;
    sesCtl_p->granuleMax = vMax;
    sesCtl_p->granuleMean= vMean; 
    sesCtl_p->granuleSum = vSum ; 
 
 return vMean;
} // end::l4_sm_summarizeVariable()


// ----------------------------------------------------------------
int32 showHourly_stats(const char *datasetName,const int32 nGranules,
                float32 *hrMinAry,float32 *hrMaxAry,float32 *hrMeanAry)
{
 int32 h = 0;

    for(h=0; h < nGranules; h++)
    {
        xLogMsg("%s HOUR %-2.2d min %.4f max %.4f mean %.4f\n",
            datasetName, h+1,hrMinAry[h], hrMaxAry[h], hrMeanAry[h]);
    }
 return h;
} // end::showHourly_stats()



// DEPRECATED bit fx, now uses UM Framework's
// -----------------------------------------------------------------------------
//   module     : l4c_GetBits.c
//   revision   : Updated 05/04/1993 @ 15:03:53 
//   purpose    : getbits() from K&R C Language text is used to fetch
//                n bits from position pStart in a 16 bit word.
//   limitations: argument unsigned int' is 32 bits on many if not most 32 and 64 bit platforms,
//                and should NOT be relied upon to represent 16 bits as it used to on some PCs.
// unsigned int l4c_GetBits(unsigned int xTarget, const unsigned int pStart,const unsigned int nBits)
// uint32 l4c_SetBits(int16 startPos, int16 endPos, void *data, size_t dataSize)
// uint32 l4c_RetrieveBits(const int16 startPos,const int16 endPos, void *data,size_t dataSize)


// -----------------------------------------------------------------------------
// function: l4c_GenerateUUID()
// Purpose : generate a type-1 UUID using libuuid's services
// Inputs  : char *resultUUID_String
//         : displayMode (boolean, if 0=displays as packed string, if 1=displays as
//           output from uuid_unparse(), into couplets etc.
//
// Dependencies: -luuid must be linked in and header <uuid/uuid.h> must be available
// on the build platform!
char * l4c_GenerateUUID(char *resultUUID_String,const int displayMode)
{
 uuid_t uuid ;
 int i,k=0;

    strcpy(resultUUID_String,"no-uuid-avail");

// for NCCS had to temporarily comment out...
#ifdef UUID_AVAIL
    uuid_generate(uuid);
    uuid_unparse(uuid, resultUUID_String);
#endif

    if(displayMode==0)
    {
     for(i=0; i < (int)strlen(resultUUID_String);i++)
      if(resultUUID_String[i] != '-')
        resultUUID_String[k++] = resultUUID_String[i];
     // re-insert term ch
     resultUUID_String[k] = '\0';
    }
    
    return(resultUUID_String);
} // end::l4c_GenerateUUID()

// ---------------------------------------------------------------------
// function:calculateGranuleActualRange()
// inputs  :SesCtl *sesCtl_p,char *datasetName,float32 *varImg_p,size_t nElem
// revised : 2015-09-04T13:16
// author  : joe glassy
// outputs : Loads sesCtl_p->granuleMin,granuleMax,granuleMean with most
//           recently named dataset 'datasetName''s min,max and mean values
//           that qualify as non-missing (valid) values
// returns : size_t nQual --number of qualifying pixels
// notes   : Intended primarily for GEOS5 style fields where all pixels
//           have a biophysical valid value
size_t calculateGranuleActualRange(SesCtl *sesCtl_p,char *datasetName,
                const float32 missingVal,float32 *varImg_p,size_t nElem)
{
 char fxName[256] = {0};
 size_t pix =0;
 size_t serialIdx =0;
 
 size_t nQual =0;
 size_t nDisqualified=0; // count of pixels excluded due to failed doAllow test
 
 size_t nLowMissVal  =0 ; // count, where a low missing value hit -9999 or lower
 size_t nHighMissVal =0 ; // count, where a high missing value hit E14 etc
 float32 minGranule  = 0.0;
 float32 maxGranule  = 0.0;
 float32 meanGranule = 0.0;
 float32 sumGranule  = 0.0;
 
 int doAllow = 0;
 int hitFirst=0 ; // boolean TRUE when first qualifying pixel is encountered
   
  strcpy(fxName,__func__);
  
    // Next, now that a geographic vMean[] is available on a pixel basis
    // calculate the granule wise min,max for actual_min,actual_max attributes
    for(serialIdx=0; serialIdx < nElem; serialIdx++)
    {
        pix = varImg_p[serialIdx];
        doAllow = 0 ;  // reset conditional boolean switch
        
        // Check for missing values here. Challenge is that a missing
        // value might be very LOW -9999.0, or very high 9.9999999E14
        // so MUST ADJUST test as less than or greater than to accomodate this
        if(missingVal <= -9999.0)
        {
            doAllow = (pix > missingVal) ? 1 : 0 ;
            nLowMissVal++;
        } else if(missingVal > 20000.0) {
            doAllow = (pix < missingVal) ? 1 : 0 ;
            nHighMissVal++;
        }
        
        // based on the acceptance filter, exclude all missing value pix
        if(doAllow)
        {
            // update sum of values of qualifying pixels
            sumGranule += pix ;
            if(!hitFirst) // trip the binary flag
            {
               hitFirst = 1;
               minGranule = maxGranule = pix ;
            } else {
               if(pix < minGranule) minGranule = pix;
               if(pix > maxGranule) maxGranule = pix;
            }
            nQual++ ;  // and count of qualifiers
        } else { // end:: blockif doAllow
            nDisqualified++ ;
        }
        
    } // end::serialIdx loop 
 
    if(nQual >0)   
       meanGranule = sumGranule / (float)nQual;
    else
       meanGranule = -9999.0 ; // set as sentinel to flag bad mean
    
    xLogMsg("@A In(%s) GRANULE-RANGE calculated for %s with min %f max %f mean %f sum %f Count %zu~\n",
     fxName, datasetName, minGranule,maxGranule,meanGranule, sumGranule, nQual);
    xLogMsg("@A In(%s) GRANULE-RANGE for %s missing value is %g Count of missingValue tests low: %zu high: %zu (%zu disqualified)~\n",
     fxName, datasetName, missingVal, nLowMissVal, nHighMissVal, nDisqualified);

    // <JMG> 2017-02-08T09:01 this is about to be deprecated in favor of using 
    // equiv variables from xxx_summarizeVariable() functions.
    
    // Orig logic: store in sesCtl for subsequent retrieval
    // for hand off of granuleMin,granuleMax
    strcpy(sesCtl_p->currVarname, datasetName);
    sesCtl_p->granuleMin = minGranule ;  
    sesCtl_p->granuleMax = maxGranule ;  
    sesCtl_p->granuleMean= meanGranule; 
    
 return 0 ;    
} // end::calculateGranuleActualRange()

// ---------------------------------------------------------------------
// function  : calc_NativeRange()
// inputs    : SesCtl *sesCtl_p -- pointer to session controls
//             char *datasetName -- name of variable to process
//             const size_t nPixels
//             const float32 missingValue -- missing value to ignore
//             const float32 lowestValid -- lowest valid value expected
//             float32 *fltRaster_p -- ptr to native float32 dataset array
// revised   : 2015-08-25T11:59:00
// author    : Joe Glassy
// purpose   : For one float32 variable, calculates the native data range
//             and mean value (min,max,mean)
//
// returns   : size_t nQualifed -- count of qualifying (non missing) pixels

size_t calcNativeRange(SesCtl *sesCtl_p,
    char *datasetName, const size_t nPixels,
    const float32 missingValue, const float32 lowestValid,
    float32 *fltRaster_p)
{
 char fxName[256] = {0} ;
 int hitFirstQualified = 0 ; // boolean, for when first qualified pix hit
 size_t ithPix     = 0 ;
 size_t serialIdx  = 0 ;    
 
 size_t nQualified = 0 ;    
 size_t nMinAdjusts= 0 ;
 size_t nMaxAdjusts= 0 ;
 
 float32 vSum  = 0.0 ;
 float32 vMin  = 0.0 ;
 float32 vMax  = 0.0 ;
 float32 vMean = 0.0 ;
 float32 thisPix     = 0.0;
 
    strcpy(fxName,__func__);
    
    // Here we treat 2D raster as a simple serialized 1D vector 
    for(ithPix = 0 ; ithPix < nPixels; ithPix++)
    {
        thisPix = fltRaster_p[ithPix] ;
        
        if( thisPix >= lowestValid && thisPix < missingValue)
        {
             nQualified++ ; // increment count of qualified pixels
             vSum += thisPix;
             if( hitFirstQualified ==0)
             {
                 hitFirstQualified = 1;
                 vMin = vMax = thisPix ;
             } else {
                 if( thisPix < vMin )
                 {
                   vMin = thisPix;
                   nMinAdjusts++ ;
                 }
                 if( thisPix > vMax ) 
                 { 
                   vMax = thisPix ;
                   nMaxAdjusts++ ;
                 }
             }
        } // end:: thisPix qualified block
        
    } // end:: ithPix
    
    xLogMsg("@A In(%s) NATIVE %s missingVal %f min %f max %f mean %f ~\n",fxName,
     datasetName, missingValue, vMin,vMax,vMean);
    
 return nQualified ;
} // end:: calcNativeRange()                        

// ---------------------------------------------------------------------
// function: l4c_SummarizeFreezeThaw(sesCtl_p,ftData_p,nelem)
// purpose : Scan FT_STATE_UM_M09 as a 2D (3KM) raster, and calc
//           frequencies of bit0 (assuming tsurf used if derived from L3_SM_A,
//           or, assuming bit0 populated by GEOS5 FP "TS" proxy approach
//           where 1=Frozen,0=Thawed.
//
// notes   : We retain L3_SM_A logic where 1=Frozen, 0=Thawed
//
// ASSIGNMENT LOGIC for L3_SM_A usage for FT_STATE_UM_M03 is:
//             if(ft_tsurf)
//                 ftStateUm_M03 = (uint8)frm_utilSetBits(0,0, &ftStateUm_M03,sizeof(uint8));
//             if(ft_radar)
//                 ftStateUm_M03 = (uint8)frm_utilSetBits(1,1, &ftStateUm_M03,sizeof(uint8));
//             if(ft_radar_qc)
//                 ftStateUm_M03 = (uint8)frm_utilSetBits(2,2, &ftStateUm_M03,sizeof(uint8));
//             if(ft_in_swath)
//                 ftStateUm_M03 = (uint8)frm_utilSetBits(3,3, &ftStateUm_M03,sizeof(uint8));
//
// ASSIGNMENT LOGIC when GEOS5 FP "TS" is used:
//             Bit0 set 1=TRUE when frozen, 0=false=thawed from GEOS5 FP "TS" as deg C.
//                 ftStateUm_M03 = (uint8)frm_utilSetBits(0,0, &ftStateUm_M03,sizeof(uint8));
//             Bit1 set 1=TRUE when InTerrestrial, 0=FALSE when outside land mask (in ocean,rock etc)
//                 ftStateUm_M03 = (uint8)frm_utilSetBits(1,1, &ftStateUm_M03,sizeof(uint8));
//             Bit2 set 1=TRUE ALWAYS if GEOS5 FP "TS" used, to match radar QC being GOOD all the time
//                 ftStateUm_M03 = (uint8)frm_utilSetBits(2,2, &ftStateUm_M03,sizeof(uint8));
//             Bit3 (ft_in_swath) set 1=TRUE always 
//                 ftStateUm_M03 = (uint8)frm_utilSetBits(3,3, &ftStateUm_M03,sizeof(uint8));
//
int l4c_SummarizeFreezeThaw(SesCtl *sesCtl_p, uint8 *ftData_p,const size_t nElem3km)
{
 // 56359296 pixels in 2D full (4872x11568) raster, or LDOM3KM=14976360 in 1D sparse
 uint8  ftState3kmUint8 =0 ; // whole uint8 pixel for FT_STATE_UM_M03
 uint8  ftBit0State = 0 ;    // returned boolean {1|0} 1=frozen, 0=thawed
 uint8  ftBit1State = 0 ;    // bit1 is InTerrestrial or not. Substitutes for ft_radar (0|1) when L3_SM_A used
 uint8  ftBit2State = 0 ;    // bit2 is ALWAYS FALSE if GEOS5 FP "TS" used, otherwise may be true or false if L3_SM_A used
 uint8  ftBit3State = 0 ;    // bit3 InSwath always TRUE if GEOS FP "TS" used otherwise is (0|1)
 // <JMG> 2015-09-08 added explicit assignment to bits 4,5,6,7
 uint8  ftBit4State =0;  // set TRUE only if GEOS5 TS used
 uint8  ftBit5State =0;  // set to F/T state when GEOS5 TS used
 uint8  ftBit6State =0;  // set TRUE only when L3_SM_A present and TSURF used
 uint8  ftBit7State =0;  // set TRUE only when L3_SM_A present and FT RADAR used
 
 size_t pix3km =0;
 size_t ftStateFreq[2]   = {0};  // F/T status, if Tsurf or GEOS5 TS, bit0 freqs
 size_t ftBit1Freq[2]    = {0} ; // FT_RADAR's F/T status, bit1 freqs
 size_t ftQcFreq[2]      = {0} ; // QC bit2 freqs, proxy for QC
 size_t ftInSwathFreq[2] = {0} ; // bit3 freqs, remember that (IsInSwath=0),not in swath is 1
 size_t ftBit4Freq[2] = {0} ; 
 size_t ftBit5Freq[2] = {0} ; 
 size_t ftBit6Freq[2] = {0} ; 
 size_t ftBit7Freq[2] = {0} ; 

 char fxName[256] = {0} ;

   strcpy(fxName,__func__);
   
   xLogMsg("@A In(%s) FT_STATE_UM_M03 freeze thaw summary of %zu pixels ~\n",fxName,nElem3km);
   xLogMsg("@W In(%s) CRITICAL Reminder: InSwath is defined as 0 not 1 ~ \n",fxName);
   xLogMsg("@A In(%s) RIP defined FT policy %s ~ \n",fxName, sesCtl_p->ftSourceStr);
   
   for(pix3km=0; pix3km < nElem3km; pix3km++)
   {
       // Retrieve whole uint8 3KM sparse pixel 
       ftState3kmUint8 = ftData_p[pix3km];
       
       // Alternatives:
       // uint32 frm_utilRetrieveBits(const int16 startPos,const int16 endPos, void *data,size_t dataSize)
       // frm_util_GetBits(0,0,&ftState3kmUint8,sizeof(uint8));
       
       // 1) Extract the bit0 value. Others are NOT yet scanned
       ftBit0State   = frm_utilRetrieveBits(0,0,&ftState3kmUint8,sizeof(uint8));
       
       // 2) Extract the bit1, as 1=TRUE when in land domain, false otherwise
       ftBit1State   = frm_utilRetrieveBits(1,1,&ftState3kmUint8,sizeof(uint8));
       
       // 3) Extract the bit2, as always true, matching a radar_qc of always good
       ftBit2State   = frm_utilRetrieveBits(2,2,&ftState3kmUint8,sizeof(uint8));
       
       // 4) extract bit3 (InSwath), which should ALWAYS be true if GEOS5 FP "TS" used, otherwise
       // may be true or false if L3_SM_A used. 
       ftBit3State   = frm_utilRetrieveBits(3,3,&ftState3kmUint8,sizeof(uint8));
       
       ftBit4State   = frm_utilRetrieveBits(4,4,&ftState3kmUint8,sizeof(uint8));
       ftBit5State   = frm_utilRetrieveBits(5,5,&ftState3kmUint8,sizeof(uint8));
       ftBit6State   = frm_utilRetrieveBits(6,6,&ftState3kmUint8,sizeof(uint8));
       ftBit7State   = frm_utilRetrieveBits(7,7,&ftState3kmUint8,sizeof(uint8));
                                   
       
       // NEXT, collect frequencies for each bit 0,1,2,3 used in FT_STATE_UM_M03's 8-bit byte
       // noting that bits 4,5,6,7 are NOT yet used. Tests here shouldn't be needed
       // as we pulling off only 1 bit valued(0|1) but just to be sure
       if( ftBit0State <= 1)
                     ftStateFreq[ftBit0State]++ ;
       if( ftBit1State <= 1)
                     ftBit1Freq[ftBit1State]++ ;
       if( ftBit2State <= 1)
                     ftQcFreq[ftBit2State]++;
       // Remember that InSwath=0 NOT 1 as per Barry Weiss convention                     
       if( ftBit3State <= 1)
                     ftInSwathFreq[ftBit3State]++ ;
    
       if( ftBit4State >= 1) ftBit4Freq[ftBit4State]++ ;
       if( ftBit5State >= 1) ftBit5Freq[ftBit5State]++ ;
       if( ftBit6State >= 1) ftBit6Freq[ftBit6State]++ ;
       if( ftBit7State >= 1) ftBit7Freq[ftBit7State]++ ;
                     
   } // end:: pix3km 1D sparse loop

   // NEXT, report on frequency results, by bit {0,1,2,3}   
   xLogMsg("@A In(%s) bit0 Freeze-Thaw State         frozen(%9zu) thawed(%9zu) ~\n",fxName,ftStateFreq[1]  ,ftStateFreq[0]  );
   xLogMsg("@A In(%s) bit1 ft_radar                  frozen(%9zu) thawed(%9zu) ~\n",fxName,ftBit1Freq[1]   ,ftBit1Freq[0]   );
   xLogMsg("@A In(%s) bit2 radar_qc Good=1, Bad=0    Good  (%9zu)   Bad (%9zu) ~\n",fxName,ftQcFreq[1]     ,ftQcFreq[0]     );
   xLogMsg("@A In(%s) bit3 InSwath (0=In 1=Out-swath)   In (%9zu)   Out (%9zu) ~\n",fxName,ftInSwathFreq[0],ftInSwathFreq[1]);
   // <JMG> 2015-08-08 report on the rest of the status bits
   xLogMsg("@A In(%s) bit4 GEOS5 used YES=1,NO=0        YES (%9zu)   NO (%9zu) ~\n",fxName,ftBit4Freq[1],ftBit4Freq[0]);
   xLogMsg("@A In(%s) bit5 GEOS5 Only: In-Terrestrial?   In (%9zu)  Out (%9zu) ~\n",fxName,ftBit5Freq[1],ftBit5Freq[0]);
   xLogMsg("@A In(%s) bit6 TSURF used YES=1,NO=0        YES (%9zu)   NO (%9zu) ~\n",fxName,ftBit6Freq[1],ftBit6Freq[0]);
   xLogMsg("@A In(%s) bit7 RADAR used YES=1,NO=0        YES (%9zu)   NO (%9zu) ~\n",fxName,ftBit7Freq[1],ftBit7Freq[0]);   
   xLogMsg("~\n",fxName);
   
   // transfer results to sesCtl for later reporting out as file level attributes
   memcpy(sesCtl_p->ftStateFrq  ,ftStateFreq  ,(sizeof(size_t)*2)); // bit0 --F/T status, if Tsurf or GEOS5 TS, bit0 freqs
   memcpy(sesCtl_p->ftBit1Frq   ,ftBit1Freq   ,(sizeof(size_t)*2)); // bit1 --FT_RADAR's F/T status, bit1 freqs
   memcpy(sesCtl_p->ftQcFrq     ,ftQcFreq     ,(sizeof(size_t)*2)); // bit2 --QC bit2 freqs, proxy for QC
   memcpy(sesCtl_p->ftInSwathFrq,ftInSwathFreq,(sizeof(size_t)*2)); // bit3 --(IsInSwath=0),not in swath is 1
   memcpy(sesCtl_p->ftBit4Frq   ,ftBit4Freq   ,(sizeof(size_t)*2)); // bit4 -- TRUE only when GEOS5 TS is used
   memcpy(sesCtl_p->ftBit5Frq   ,ftBit5Freq   ,(sizeof(size_t)*2)); // bit5 -- InTerrestrial=1, NotInTerrestrial=0
   memcpy(sesCtl_p->ftBit6Frq   ,ftBit6Freq   ,(sizeof(size_t)*2)); // bit6 -- TRUE only when L3_SM_A present and ACTIVE TSURF used
   memcpy(sesCtl_p->ftBit7Frq   ,ftBit7Freq   ,(sizeof(size_t)*2)); // bit7 -- TRUE only when L3_SM_A_ present and ACTIVE RADAR used
    
 return 0 ;    
} // end:: l4c_SummarizeFreezeThaw()
