// ---------------------------------------------------------------------
//   module     : l4c_MET_grid_proj.c
//   Purpose    : implement gridding and projection support functions for l4c_MET.c
//   Author     : joe glassy
//   Version    : v5.0.1, see AppVers macro in l4c_MET.h
//   Revision   : 2014-07-24T12:48, re-integration into L4_C v5.0.1 in progress
//   Notes      : Several new fields added for D5 representing either new
//                capabilities needed to support the use of F/T in L4_C_DAILY,
//                or (VPD) that were omitted earlier.  New fields are:
//                -- FT_ROLLING_AGE_3KM  ...from Freeze Thaw AGE (uint8,3km)
//                -- vpd (float32, 9KM, "units=Pa") in SMAP_L4_C_MET.h5
//                -- smroot_min (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                -- smroot_max (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                 and new static 3KM float32 layers "smroot_min","smroot_max"
//              in SMAP_L4_C_SOC.h5
//

#include "l4c_MET.h"
#include "l4c_MET_prot.h"

// ----------------------------------------------------------------
// genTerrestrialAddresses
// Inputs  : pftImg_p -- buffer containing EASEGRIDv2 2D global 9km
//           MASK {0|1} uint8 image.
//           nElem9km -- n. of pixels in the EASEGRIDv2 9km 2D extent
// Returns : int32 *terrestrialMaskImg_p -- ptr to indirection-index
//           containing the serialized 2D indices where terrestrial
//           pixels reside.
// Outputs : pointer to a 1D sparse vector (simple array) containing the 
//           serialized form of the 2D raster {row,col} indices associated
//           with JUST the global 9km pixels classified as terrestrial,
//           ordered in row-major-order.
// Notes   : This is currently used only for diagnostic purposes. Be sure
//           to free buffer via memFree() later to avoid an orphan.
// Notes   : to UNPACK a serialized index into constituent {row,col} form, 
//           use the following customary 1D --> 2D identities:
//             row = (serialIndex % nCols9km )
//             col = (serialIndex / nCols9km )
int32 *genTerrestrialAddresses(uint8 *pftImg_p,const int32 nElem9km)
{
 int32 *terr_pool = NULL;
 int32 serialIndex =0;
 int32 k;
 
    // PASS 1: scan the global Plant Function Type mask {0|1} image,
    //  and generate a list of 1,664,040) 1D serial indices (addresses)
    //  corresponding to the pixels in the terrestrial classification,
    //  and count up these valid addresses, storing count in k
    for(k=0,serialIndex=0; serialIndex < nElem9km; serialIndex++)
        if(pftImg_p[serialIndex]== 1) k++;
        
    // Construct a (int32) 1D vector exactly this size
    terr_pool=(int32 *)memCalloc(k,sizeof(int32));
    
    // PASS 2, Now populate the new vector with the valid addresses ;
    for(k=0,serialIndex=0; serialIndex < nElem9km; serialIndex++)
        if(pftImg_p[serialIndex] == 1)
        {
            terr_pool[k++] = serialIndex ;
        }
 // Return ptr to new vector containing the serial indices where
 // terrestrial pixels reside.
 return terr_pool;      
} // end:: genTerrestrialAddresses()


// ----------------------------------------------------------------
// function : l4c_easegrid2_wgs84_fwd() 
// purpose  : Given a projection identifier for a supported projection and
//            resolution, perform a pixel wise forward transform,  
//             converting one x,y coordinate FROM geographical {lat,lon} 
//             projection TO the ease-grid2 {row,col} projection 
// returns  : status, ActionSuccess, and, ptrs to EASEGRID ithRow,ithCol
// inputs   : mapProjectionCode
//          : float64 lat
//          : float64 lon
//          : float64 *ithRow
//          : float64 *ithColumn
// outputs  : *ithRow,*ithColumn as EASEGRID raster {row,col} indices, in row-major-order
// history  : original name: wgs84_convert() 
// notes    : n/a
/*
  info[ 19 ] = { gpd: 'EASE2_M09KM', projection: '', map_scale_m: 9008.055210146D, $
                 cols: 3856L, rows: 1624L, r0: 0.0D, s0: 0.0D }
  info[ 20 ] = { gpd: 'EASE2_M03KM', projection: '', map_scale_m: 3002.6850700487D, $
                 cols: 11568L, rows: 4872L, r0: 0.0D, s0: 0.0D }
*/
int32 l4c_easegrid2_wgs84_fwd(const int32 mapProjectionCode,const float64 lat,const float64 lon, float64 *ithRow, float64 *ithColumn)
{
 char fxName[MAX_MEMOBJ_VNAME_LEN]    = {0} ;
 char moduleFile[MAX_MEMOBJ_VNAME_LEN]= {0} ;
 int32 status = -1 ;
 
    // variables removed from forward transform, not used: {beta,e4,e6}
    // parameter definition 
    float64 map_reference_latitude =0.0, 
            map_reference_longitude =0.0;
    float64 map_second_reference_latitude =0.0,
            map_eccentricity =0.0;
    float64 kz =0.0,
            qp =0.0;
    float64 rho =0.0;
    float64 phi =0.0,
            lam =0.0;
    float64 resolution =0.0, 
            Radius =0.0;
    float64 e2 =0.0 ;
    float64 x =0.0,
            y =0.0;
    float64 epsilon =0.0;
    float64 dlon =0.0;
    float64 q =0.0;
    float64 sin_phi1 = 0.0, 
            cos_phi1 = 0.0;
    float64 sin_phi  = 0.0;

    int32 nRows = 0 ; // varies by 1km vs. 9km
    int32 nCols = 0 ;

    float64 r0 = 0.0, 
        s0 = 0.0;
    float64 r_out = 0.0, 
        s_out= 0.0;

    float64 r =0.0; // preserve original semantics for now
    float64 s =0.0; // preserve original semantics for now

    // first define the projection parameters 
    // for global projection, define the geodetic radius and eccentricity  
    Radius = 6378137.0 ; // radius in meters of WGS84 sphere 
    map_eccentricity = 0.081819190843; // WGS84 ecc factor
    e2 = pow(map_eccentricity, 2.0);
    epsilon = 1.e-6;

    // register this module and function
    strcpy(fxName,(char *)__func__);
    strcpy(moduleFile,(char *)__FILE__);
    
    // branch on global vs. Northern hemisphere projections, to set up the basics..
    if(mapProjectionCode <= IsEG_Global_1km)
    {
        map_reference_latitude = 0.0 ;
        map_reference_longitude = 0.0;
        map_second_reference_latitude = 30.0;
        sin_phi1 = sin(30.0 *PI/180.0);
        cos_phi1 = cos(30.0 *PI/180.0);
        kz = cos_phi1/sqrt(1.0 - e2 * sin_phi1 * sin_phi1 );
        // override resolution for 9km case
    } else { // handle case of 'N' North hemisphere
        map_reference_latitude = 90.0 ;
        map_reference_longitude = 0.0 ;
    }

    switch(mapProjectionCode)
    {
    case IsEG_Global_1km:   resolution = 1000.89502334956; // km 
                            nRows = 14616;
                            nCols = 34704;
                            break;
    case IsEG_Global_3km: // projection info: reference lat,long
                            resolution = 3002.6850700487 ; // for 3km
                            nRows = 4872  ;
                            nCols = 11568 ; 
                            break;
    case IsEG_Global_9km: // projection info: reference latitudes and longitudes 
                            resolution = 9008.055210146; // for 9km 
                            nRows = 1624 ;
                            nCols = 3856 ;
                            break;
    case IsEG_N_1km     :   resolution = 1000.0; // 1km 
                            nRows      = 18000;
                            nCols      = 18000;
                            break;
    case IsEG_N_9km     :   resolution = 9000.0; // 9km km 
                            nRows = 2000;
                            nCols = 2000;
                            break ;
    default: fprintf(stderr,"@F ~map-projection code %d is NOT yet supported\n",
                        mapProjectionCode);
                        break;
    } // end:: switch on mapProjectionCode

    /* r0: column that is mapped to map_reference_latitude; 
       s0: row that is mapped to map_reference_latitude */
    r0 = (float64)(nCols - 1)/2.0;
    s0 = (float64)(nRows - 1)/2.0;

    dlon = lon - map_reference_longitude; 
    if( fabs(dlon) > 180.0) 
    {
        fprintf(stderr,"@E ERROR:: dlon %lf out-of-valid-range. Verify longitude\n",dlon);
    }
    
    phi = lat  * PI / 180.0;
    lam = dlon * PI / 180.0;

    sin_phi = sin(phi);

    q = ( 1.0 - e2 ) * ( ( sin_phi / ( 1.0 - e2 * sin_phi * sin_phi ) ) \
                        - ( 1.0 / ( 2.0 * map_eccentricity ) ) \
                        * log( ( 1.0 - map_eccentricity * sin_phi ) \
                                / ( 1.0 + map_eccentricity * sin_phi ) ) );

    qp = 1.0 - ((1.0 - e2)/(2.0 * map_eccentricity) \
                * log((1.0 - map_eccentricity) \
                        /(1.0 + map_eccentricity) ) );

    //  Reminder: we support just these EASEGRID v2 types: 
    //  { IsEGv2_Global_9km, IsEGv2_Global_3km, IsEGv2_Global_1km,
    //    IsEGv2_N_9km,      IsEGv2_N_3km,      IsEGv2_N_1km }
    if(mapProjectionCode >= IsEG_N_9km)
    {
        if( fabs(qp-q) < epsilon ) 
            rho = 0.0;
        else
            rho = Radius * sqrt(qp-q);
            
        x = rho * sin(lam);
        y = -1.0 * rho * cos(lam);  
        
    } else { // for all GLOBAL cases... rho is not defined...
        // IsEG_Global_9km, isEG_Global_3km, IsEG_Global_1km.. handled here
        x = Radius * kz * lam;
        y = Radius * q/(2.0 * kz);
    }
    
    r_out = r0 + x/resolution ;
    s_out = s0 - y/resolution ;

    if( (r_out < -0.5) || (r_out >= nCols) || (s_out < -0.5) || (s_out >= nRows) )
    {
        fprintf(stderr,
            "@E ERROR:: coord transform lat: %f, lon: %f, r_out: %f, s_out: %f\n",
                lat, lon, r_out, s_out);
        (*ithColumn) = -1.0;
        (*ithRow) = -1.0;
    } 

    // preserve original semantics, used cartographic 'r','s' terms to pass output terms
    r = r_out;
    s = s_out;
    
    // pass results back now to real numbers (float64's) --> ithColumn and ithRow 
    status = ActionSuccess ;
    // <JMG> added fabs() since we only want positive row,col anyway
    *ithColumn = fabs(r_out);
    *ithRow    = fabs(s_out);
    
    // return final status 
    return status;
} // end::void l4c_easegrid2_wgs84_fwd()


// ----------------------------------------------------------------
// easegrid2_wgs84_inversetransform()
// Purpose: perform inverse transform, given a map projection integer code, and an 
//  EASE-GRID v2 {row,column} coordinate, derive a decimal {latitude,longitude} coordinate.
// Inputs : int32 mapProjectionCode (from enum..)
//        : float64 ithRow,
//        : float64 ithColumn
//        : float64 *lat
//        : float64 *lon
// Outputs: lat,lon
// history: adapted from wgs84_inverse()
// Returns: status, ActionSuccess, and, ptrs to EASEGRID *lat,*lon
int32 l4c_easegrid2_wgs84_inv(const int32 mapProjectionCode,float64 ithRow, float64 ithColumn,float64 *lat,float64 *lon)
{
 char fxName[MAX_MEMOBJ_VNAME_LEN]    = {0} ;
 char moduleFile[MAX_MEMOBJ_VNAME_LEN]= {0} ;

    /* parameter definition */
    float64 map_reference_latitude        = 0.0 ;
    float64 map_reference_longitude       = 0.0;
    float64 map_second_reference_latitude = 0.0;
    float64 map_eccentricity              = 0.0;
    float64 sin_phi1 = 0.0;
    float64 cos_phi1 = 0.0;
    float64 kz = 0.0;
    float64 qp = 0.0;
    float64 beta = 0.0;
    float64 rho = 0.0;
    float64 phi = 0.0;
    float64 lam = 0.0;
    float64 resolution = 0.0;
    float64 Radius = 0.0;
    float64 e2 = 0.0;
    float64 e4 = 0.0;
    float64  e6 = 0.0;
    float64 x = 0.0;
    float64 y = 0.0;
    float64 epsilon = 0.0;

    int32 nRows =0;
    int32 nCols =0;
    int32 status = -1 ;
    
    float64 r0 = 0.0, s0 = 0.0;
    float64 lat_out = 0.0;
    float64 lon_out = 0.0;

    float64 r = ithColumn;// preserve original goofy semantics
    float64 s = ithRow   ;// preserve original goofy semantics

    // Define the radius and eccentricity of the geodetic model used
    Radius              = 6378137.0 ;      // WGS84 sphere radius in decimal meters
    map_eccentricity    = 0.081819190843;  // WGS84 */
    e2 = pow(map_eccentricity, 2.0);
    e4 = pow(map_eccentricity, 4.0);
    e6 = pow(map_eccentricity, 6.0);
    epsilon = 1.e-6 ;

    // register this module and function
    strcpy(fxName,(char *)__func__);
    strcpy(moduleFile,(char *)__FILE__);
    
    // Branch on global vs. Northern hemisphere to set up the basics..
    //  Reminder: we support just these EASEGRID v2 types: 
    //  { IsEGv2_Global_9km, IsEGv2_Global_3km, IsEGv2_Global_1km,
    //    IsEGv2_N_9km,      IsEGv2_N_3km,      IsEGv2_N_1km }
    if(mapProjectionCode <= IsEG_Global_1km)
    {
        map_reference_latitude  = 0.0;
        map_reference_longitude = 0.0;
        map_second_reference_latitude = 30.0;
        sin_phi1 = sin(30.0 *PI /180.0);
        cos_phi1 = cos(30.0 *PI /180.0);
        kz = cos_phi1/sqrt(1.-e2*sin_phi1*sin_phi1);
    } else { // handle case of 'N' North hemisphere
        map_reference_latitude = 90.;
        map_reference_longitude = 0.;
    }

    // note: resolution here is "map scale"
        switch(mapProjectionCode)
        {
        case IsEG_Global_1km:   resolution = 1000.89502334956; // km 
                                nRows = 14616;
                                nCols = 34704;
                                break;
        case IsEG_Global_3km:   resolution = 3002.6850700487 ; // for 3km
                                nRows = 4872  ;
                                nCols = 11568 ; 
                                break;
        case IsEG_Global_9km:   resolution = 9008.055210146; // for 9km 
                                nRows = 1624 ;
                                nCols = 3856 ;
                                break;
        case IsEG_N_1km     :   resolution = 1000.0; // 1km 
                                nRows      = 18000;
                                nCols      = 18000;
                                break;
        case IsEG_N_9km     :   resolution = 9000.0; // 9km km 
                                nRows = 2000;
                                nCols = 2000;
                                break ;
    default: fprintf(stderr,"@F map-projection code %d is not yet supported\n",
                        mapProjectionCode);
                        break;
    } // end:: switch on mapProjectionCode

    // r0: column that is mapped to map_reference_latitude; 
    //   s0: row that is mapped to map_reference_latitude. 
    r0 = (float64)(nCols - 1)/2.0;
    s0 = (float64)(nRows - 1)/2.0;

    /* convert (col, row) to xy coordinates for this projection */
    x = (r-r0) * resolution;
    y = (s0-s) * resolution;

    qp = ( 1.0 - e2 ) * ( ( 1.0 / ( 1.0 - e2 ) ) \
                        - ( 1.0 / ( 2.0 * map_eccentricity ) ) \
                        * log( ( 1.0 - map_eccentricity ) \
                                / ( 1.0 + map_eccentricity ) ) );

    //  Reminder: we support just these EASEGRID v2 types: 
    //  { IsEGv2_Global_9km, IsEGv2_Global_3km, IsEGv2_Global_1km,
    //    IsEGv2_N_9km,      IsEGv2_N_3km,      IsEGv2_N_1km }
    if(mapProjectionCode >= IsEG_N_9km)
    {
        rho = sqrt( x * x + y * y);
        beta = asin(1. - (rho * rho/( Radius * Radius * qp)));
        lam = atan(x/(-1.0 * y));
    } else { // for all Global cases...(no rho is defined)
        beta = asin(2.0 * y * kz / (Radius * qp));
        lam  = x/(Radius * kz);
    }

    phi = beta \
        + ( ( ( e2 / 3.0 ) + ( ( 31.0 / 180.0 ) * e4 ) \
              + ( ( 517.0 / 5040.0 ) * e6 ) ) * sin( 2.0 * beta ) ) \
        + ( ( ( ( 23.0 / 360.0 ) * e4) \
              + ( ( 251.0 / 3780.0 ) * e6 ) ) * sin( 4.0 * beta ) ) \
        + ( ( ( 761.0 / 45360.0 ) * e6 ) * sin( 6.0 * beta ) );

    lat_out = 180.0 * phi/PI;
    lon_out = 180.0 * lam/PI + map_reference_longitude;

    if( (fabs(lat_out) > 90.0) || (fabs(lon_out) > 180.0) )
    {
        fprintf(stderr,"@F out-of-range latitude (%lf {-90<lat<90}) or longitude (%lf {-180<lon<180}) transformation!\n",
            lat_out,lon_out);
    } // end:: fabs block

    // Perform explict cast from float64 precision, to float64 precision
    // acknowledging we may lose some precision for some point translations
    // TBD: verify if we need a R_ROUND() call here...
    status = ActionSuccess ;
    (*lat) = (float64)lat_out;
    (*lon) = (float64)lon_out;
    
    // return final status
    return status;
} // end::l4c_easegrid2_wgs84_inv()


// -----------------------------------------------------------------------------
// geog_forwardtransform()
// Purpose      : given a {lat,lon} derive the raster {row,col} position in a geographic grid
//                tested specifically here for GEOS5 FP of 721x1152 pixels, 
//                or .25 degree by .3125 degree resolution
// Inputs       : float64 lat
//              : float64 lon
//              : int32 *ithRow
//              : int32 *ithCol 
// Notes        : this performs an implicit row-reversal
// Dependencies : DLAT_GEOG here is 0.25, and DLON_GEOG is 0.3125. These constants are not yet 
//  full parametrized (obviously) to work for other geographic resolutions than GEOS-5
void geog_forwardtransform(const float64 lat,const float64 lon,int32 *ithRow,int32 *ithCol)
{
 float32 row_flt = 0.0;
 float32 col_flt = 0.0;
 
  row_flt = (lat + 90.0)  / (float64)DLAT_GEOG; // verify: this implements an implicit row reversal
  col_flt = (lon + 180.0) / (float64)DLON_GEOG; 

  // translate real number {row,col} to integer row,col
  *ithRow = (int32)FRM_API_ROUND(row_flt);
  *ithCol = (int32)FRM_API_ROUND(col_flt);
} // end:: geog_forwardtransform()


// -----------------------------------------------------------------------------
// geog_inverse()
// Purpose: given a raster {row,col}, derive and return the equivalent {lat,lon}
// Inputs : int32 ithRow -- raster row index
//        : ithCol -- raster col index
//        : float64 *lat -- ptr to latitude, derived and returned
//        : float64 *lon -- ptr to longitude, derived returned
// Outputs: *lat,*lon -- ptr to new lat, lon
void geog_inverse(const int32 ithRow,const int32 ithCol,float64 *lat,float64 *lon)
{
 int32 nRowsGeog = 721 ;
 int32 nColsGeog = 1152;
 float64 latFrac = (float64)ithRow/(float64)nRowsGeog ;
 float64 lonFrac = (float64)ithCol/(float64)nColsGeog ;
 float64 latitude   = (latFrac * 180.0) - 90.0 ; 
 float64 longitude  = (lonFrac * 360.0) - 180.0;
 
        if(latitude < -90.0) latitude = -90.0;
        if(latitude > 90.0 ) latitude = 90.0 ;
        if(longitude < -180.0) longitude = -180.0;
        if(longitude > 180.0)  longitude = 180.0 ;
        
        *lat = latitude;
        *lon = longitude;
 } // end:: geog_inverse()


#ifdef IS_3KM_REPROJECTION_ENABLED
// ---------------------------------------------------------------------
// function :regrid_FP_to_EG_3km()
// inputs   : const int32 nColsGeog --n. columns in geographic GEOS5 
//            dataset, where dimensions of 721 rows x 1151 cols are the default
//            uint8 *pftImg3km_p -- ptr to a (3KM) pft image
//            uint8 *destEgImg_p -- this is a uint8 NOT a float32 here
//            float32 *srcGeogImg_p
// returns  : int32, nPixQualified count
// revised  : 2015-07-31T10:46:00
// author   : Joe Glassy
// notes    : Note change to UINT8 From FLOAT32 here
int32 regrid_FP_to_EG_3km(const int32 nColsGeog,uint8 *pftImg3km_p,
                        uint8 *destEgImg_p, uint8 *srcGeogImg_p)
{
 char fxName[256] = {0} ;
 // float32 fillValue = GEOS5_FP_MISSING_VALUE; // 9.9999999E14 ; from GEOS5 FP _slv_ T2M and QV2M
 uint8 fillValue = 254 ;
 
 int32 r,c;
 int32 nRows3km =  4872 ;
 int32 nCols3km = 11568 ;
 // int32 nRows9km = 1624;
 // int32 nCols9km = 3856;
 
 int32 geog_row;
 int32 geog_col;
 int32 ease_serialIdx =0;
 int32 geog_serialIdx =0;
 
 uint8 binIdx = 0 ;
 uint8 test_pix = 0;
 int32 pixAbsFrq[512] = {0};
 
 int32 absFreq = 0 ;
 float32 relFreq=0.0,
         cumFreq=0.0 ;
 
#ifdef VERBOSE_MODE 
 int32 i=0;
#endif
 
 int32 nPixQualified =0;
 int32 nPixDisqualified = 0;
 int32 nPixMaskedOut = 0;
    
 uint8 pft_pix =0; // current PFT mask pixel value{0|1}
 uint8 pftIdx = 0;
  
 float64 ease_row,ease_col; 
 float64 lat,lon;
    
    strcpy(fxName,__func__);
    // Start with an explict zero-out of 3KM raster first...
    memset(destEgImg_p, '\0', N_3KM_PIX * sizeof(uint8) );
    
    // Outer loop, 3km grid loop
    for(r=0; r < nRows3km; r++)
    {
        for(c=0; c < nCols3km; c++)
        {
            // derive the serialized 1D index corresponding to this {row,column}
            ease_serialIdx = (r * nCols3km) + c;
            
            // fetch the next 3KM 2D Plant Function Type classifier pixel
            pft_pix   = pftImg3km_p[ease_serialIdx];
            
            if(pft_pix == 0)
            {
                // Use uint8 fillvalue NOT float32 (-9999) fillvalue here
                destEgImg_p[ease_serialIdx] = 254;
                nPixMaskedOut++ ;
                continue;
            } else {
                
                ease_row = (float64)r ; // get lat,long for destination 9km grid
                ease_col = (float64)c ;
                // Given this (pixel's) base0 EASEGRID row and column indices, derive 
                // closest {latitiude,longitude} coordinate in decimal degrees
                // l4c_easegrid2_wgs84_inv(IsEG_Global_9km, ease_row, ease_col,&lat,&lon);
                l4c_easegrid2_wgs84_inv(IsEG_Global_3km, ease_row, ease_col,&lat,&lon);
                
                // Given this {latitude,longitude} derive geog_row,geog_column indices
                geog_forwardtransform(lat,lon, &geog_row,&geog_col);
                // Given these {geog_row,geog_col} and number of columns in geographic space
                // derive the serial (1D) index into 2D geographic space where this pixel lives
                
                // Thus, transfer pixel's DATA from the GEOS5 FP (721x1152) geographic grid
                //  to ease v2 (1624x3856) 9km grid, as expressed via its serial 1D offset...
                geog_serialIdx = M_2D_B0(geog_row,geog_col,nColsGeog);
                
                // Finally, regrid this pixel, if it is NOT a fillvalue. Make sure to
                // adjust the two opposing cases: when _Fillvalue== -9999, the test must
                // allow valid pixels > _Fillvalue to qualify, BUT when _Fillvalue == 254,
                // the test must allow valid pixels < _Fillvalue as those which qualify
                if(srcGeogImg_p[geog_serialIdx] < fillValue)
                {
                    // From within 721x1152 geographic grid, retrieve source 
                    //   uint8 pixel at ['geog_serialIdx'] and insert it into
                    //   EASEGRID destination (uint8) image at index 'ease_serialIdx'
                    destEgImg_p[ease_serialIdx] = srcGeogImg_p[geog_serialIdx];
                    nPixQualified++ ;
                } else {
                    // Insert our own fill value, 254 
                    destEgImg_p[ease_serialIdx] = 254; 
                    nPixDisqualified++ ;
                } // end:: if src pix >= fillValue
                
            } // end:: if pft_pix WAS a terrestrial pixel

            // Update frequency bins for validation below
            test_pix  = destEgImg_p[ease_serialIdx] ;
            // increment bin count for this 3KM value
            pixAbsFrq[test_pix]++ ;
            
        } // end:: c 3km column loop
  
    } // end:: r 3km row loop

    // DIAG only, validate results, produce histogram
    xLogMsg("@A In(%s) Regridding Validation~\n",fxName);
    for(binIdx=0; binIdx < 255; binIdx++)
    {
        if( pixAbsFrq[binIdx] > 0)
        {
           absFreq = (int32)pixAbsFrq[binIdx] ;
           relFreq = ((float)pixAbsFrq[binIdx]/(float)N_3KM_PIX)*100.0 ;
           cumFreq += relFreq ;
           xLogMsg("@A In(%s) GridVerification Bin %3d abs-freq %9d rel-freq %7.2f cum-freq %7.2f~\n",
           fxName,  binIdx, absFreq, relFreq, cumFreq );
        } // end:: if a pos freq hit
    } // end:: binIdxloop

  // ASSERT: ujint8 destEgImg_p[ 4872 x 11568 ] at 3KM 
  //   stores simple uint8 F/T (0|1) states
 
 // return count of qualified pixels
 return nPixQualified ;    
} // end:: regrid_FP_to_EG_3km()
 #endif
 
// -----------------------------------------------------------------------------
// regrid_FP_to_EG_9km()
// purpose: reproject GEOS-5 FP granule variable 'srcGeogImg_p' into EASEGRID 9km
//          using nearest neighbor logic.
// inputs : nRowsGeog -- n. rows present in GEOS-5 FP layer, (0.25)x(0.31) deg resolution
//          or 721 x 1152 in raster coords.
//          nColsGeog   -- n. columns present in GEOS-5 FP layer
//          pftImg_p    -- ptr to 2D full global 9km EASEGRID mask (uint8) 0|1 layer
//          destEgImg_p -- ptr to destination 2D raster (float32) 9km image
//          srcGeogImg_p-- ptr to source raster (float32) image
int32 regrid_FP_to_EG_9km(const int32 nColsGeog,uint8 *pftImg_p,
                        float32 *destEgImg_p, float32 *srcGeogImg_p)
{
 float32 fillValue = GEOS5_FP_MISSING_VALUE; // 9.9999999E14 ; from GEOS5 FP _slv_ T2M and QV2M
 
 int32 r,c;
 int32 nRows9km = 1624;
 int32 nCols9km = 3856;
 
 int32 geog_row;
 int32 geog_col;
 int32 ease_serialIdx =0;
 int32 geog_serialIdx =0;
 
#ifdef VERBOSE_MODE 
 int32 i=0;
#endif
 
 int32 nPixQualified =0;
 int32 nPixDisqualified = 0;
 int32 nPixMaskedOut = 0;
    
 uint8 pft_pix =0;// current PFT mask pixel value{0|1}
 
 float64 ease_row,ease_col; 
 float64 lat,lon;
 
    // Start with an explict zero-out of dest first...
    memset(destEgImg_p, '\0', (nRows9km * nCols9km) * sizeof(float32) );
    
    // Outer loop, 9km grid loop
    for(r=0; r < nRows9km; r++)
    {
        for(c=0; c < nCols9km; c++)
        {
            // derive the serialized 1D index corresponding to this {row,column}
            ease_serialIdx = (r * nCols9km) + c;
            
            // fetch the next 2D Plant Function Type classifier pixel
            pft_pix   = pftImg_p[ease_serialIdx];
            
            if(pft_pix == 0)
            {
                destEgImg_p[ease_serialIdx] = L4C_MET_MISSING_VALUE; // -9999.0 
                nPixMaskedOut++ ;
                continue;
            } else {
                
                ease_row = (float64)r ; // get lat,long for destination 9km grid
                ease_col = (float64)c ;
                // Given this (pixel's) base0 EASEGRID row and column indices, derive 
                // closest {latitiude,longitude} coordinate in decimal degrees
                l4c_easegrid2_wgs84_inv(IsEG_Global_9km, ease_row, ease_col,&lat,&lon);
                
                // Given this {latitude,longitude} derive geog_row,geog_column indices
                geog_forwardtransform(lat,lon,&geog_row,&geog_col);
                // Given these {geog_row,geog_col} and number of columns in geographic space
                // derive the serial (1D) index into 2D geographic space where this pixel lives
                
                // Thus, transfer pixel's DATA from the GEOS5 FP (721x1152) geographic grid
                //  to ease v2 (1624x3856) 9km grid, as expressed via its serial 1D offset...
                geog_serialIdx = M_2D_B0(geog_row,geog_col,nColsGeog);
                
                // Finally, regrid this pixel, if it is NOT a fillvalue. This logic
                //  counts on the fact that all valid pixel values MUST be less than
                //  the indicated 'fillValue' In some cases, this may not be a very robust
                //  assumption.
                if(srcGeogImg_p[geog_serialIdx] < fillValue)
                {
                    // reach into geographic grid, pull out pixel value at 'geog_serialIdx]
                    // and poke it into EASEGRID destination image at index 'ease_serialIdx'
                    destEgImg_p[ease_serialIdx] = srcGeogImg_p[geog_serialIdx];
                    nPixQualified++ ;
                } else {
                    // Insert our own fill value, -9999.0 
                    destEgImg_p[ease_serialIdx] = L4C_MET_MISSING_VALUE ; 
                    nPixDisqualified++ ;
                } // end:: if src pix >= fillValue
                
            } // end:: if pft_pix WAS a terrestrial pixel
            
        } // end:: c 9km column loop
  
    } // end:: r 9km row loop

 // return count of qualified pixels
 return nPixQualified ;
} // end::regrid_FP_to_EG_9km()


