// ---------------------------------------------------------------------
//   module     : l4c_MET_io.c
//   Purpose    : implement I/O support functions for l4c_MET.c
//   Author     : joe glassy
//   Version    : v5.0.1, see AppVers macro in l4c_MET.h
//   Revision   : 2014-07-29T11:26, re-integration into L4_C v5.0.1 in progress
//   Notes      : Several new fields added for D5 representing either new
//                capabilities needed to support the use of F/T in L4_C_DAILY,
//                or (VPD) that were omitted earlier.  New fields are:
//                -- FT_ROLLING_AGE_3KM  ...from Freeze Thaw AGE (uint8,3km)
//                -- vpd (float32, 9KM, "units=Pa") in SMAP_L4_C_MET.h5
//                -- smroot_min (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                -- smroot_max (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                 and new static 3KM float32 layers "smroot_min","smroot_max"
//              in SMAP_L4_C_SOC.h5
//

#include "l4c_MET.h"
#include "l4c_MET_prot.h"


// ----------------------------------------------------------------
// function : l4c_PrepOutput_HDF5_Datasets() 
// Purpose  : 
// Inputs   : 
// Returns  : 
// Outputs  : 
void l4c_PrepOutput_HDF5_Datasets(SesCtl *sesCtl_p) 
{
 char fxName[256];
 
   strcpy(fxName,__func__);
   
  xLogMsg("@A In(%s) IS NOT YET implemented!~ \n",fxName);
  
  l4c_EventCompletionMsg(__func__);
    
} // end::l4c_PrepOutput_HDF5_Datasets()


// -----------------------------------------------------------------------------
// function : createMetHdf5()
// purpose  : create a 9km MET file (SMAP_L4_C_MET.h5) once only per session
// inputs   : hdf5Path -- name of new HDF5 file (SMAP_L4_C_MET.h5)
//            dspace -- current dataspace (note: this is for default 9km dimensions, so must add
//                   another 3km specific dataspace when L3_SM_A is added...
//            dcpl -- Data control property list
//            compressLevel -- GZIP level 4 or 5 is all that's needed.
hid_t createMetHdf5(SesCtl *sesCtl_p)
{
 int32 rank = 2; // shared rank, all are 2D
 
 float32 fltFill = L4C_MET_MISSING_VALUE; // -9999
 
 uint8   byteFill= 254 ;
 hid_t   file_id = -1 ;
 hid_t   plist9km = -1;     // property list for majority, 9km variables
 hid_t   plist3km = -1;     // property list for 3km variable
 hsize_t dims[2]    ={0}; // dims for 9km
 hsize_t dims3km[2] ={0}; // dims for 3km 
 hsize_t chunk_size[2] ;  // shared chunks
 int32   compressLevel = sesCtl_p->compressLev ;
 hid_t   dspace9km = -1;   // for all 9km vars in FP and L4_SM
 hid_t   dspace3km = -1;   // for all 3km vars in L3_SM_A
 
 char *hdf5Path = sesCtl_p->outputMetPath; // SMAP_L4_C_MET.h5 ....

    // -------------------------------------------------------------------------------------
    // TASK 1: set the dataspace dims for the majority of the 9km output variables
    dims[0] = NROW9KM ;
    dims[1] = NCOL9KM ;
    
    // create the empty shell of the new file
    if((file_id = H5Fcreate(hdf5Path, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT))<0)
    {
        fprintf(stderr,"@F createMetHdf5:: FAILED creating new HDF5 [%s]\n",
                    hdf5Path);
        exit(-1);
    } 
    
    // Create the first of two common 2D dataspaces, this one for 9km
    dspace9km  = H5Screate_simple(rank, dims, dims);
    // set the chunk size, specific to a 9km 
    chunk_size[0] = (dims[0] / 10);
    chunk_size[1] = (dims[1] / 10);
    // create a dcpl, a Dataset-Creation-Property-List instance
    plist9km = H5Pcreate(H5P_DATASET_CREATE);
    H5Pset_chunk (plist9km, rank, chunk_size);
    // Now, use the dcpl (plist) also request GZIP LZ77 compression at level 4 or 5
    H5Pset_deflate(plist9km,compressLevel);

    // Set the default fill value, -9999.0 . 
    //  One fillvalue WILL NOT APPLY to all datasets, so
    //  a per-dataset plist and fillvalue are created, esp for L3_SM_A
    fltFill = L4C_MET_MISSING_VALUE;   // -9999
    H5Pset_fill_value(plist9km, H5T_NATIVE_FLOAT, &fltFill);
 
    // -------------------------------------------------------------------------------------
    // TASK 2: create a 3 KM (e.g. L3_SM_A resolution) specific dataspace
    dims3km[0] = NROW3KM ;
    dims3km[1] = NCOL3KM ;

    dspace3km = H5Screate_simple(rank, dims3km, dims3km);
    // set the chunk size, specific to a 9km 
    chunk_size[0] = (dims3km[0] / 10);
    chunk_size[1] = (dims3km[1] / 10);
    // create a dcpl, a Dataset-Creation-Property-List instance
    plist3km = H5Pcreate(H5P_DATASET_CREATE);
    H5Pset_chunk (plist3km, rank, chunk_size);
    // Next, use the dcpl (plist) for same GZIP LZ77 compression
    H5Pset_deflate(plist3km,compressLevel);

    // Set the default fill value, 254 
    H5Pset_fill_value(plist3km, H5T_NATIVE_UINT8, &byteFill);
 
    xLogMsg("@A ~createMetHdf5:: CREATED NEW HDF5 [%s] file_id %d\n",hdf5Path, file_id);
    xLogMsg("@A ~createMetHdf5:: Compression level %s %d\n",hdf5Path, compressLevel);
    
    xLogMsg("@A ~createMetHdf5::9km fillvalue %.2f\n",fltFill);
    xLogMsg("@A ~createMetHdf5::9km dataspace properties dspace %d dcpl-9km %d\n\n",
            dspace9km, plist9km);
    
    xLogMsg("@A ~createMetHdf5::3km fillvalue %3d\n",byteFill);
    xLogMsg("@A ~createMetHdf5::3km dataspace properties dspace %d dcpl-3km %d\n",
            dspace3km, plist3km);
 
    // pass back the new constructs made for the 9km and 3km use-cases
   sesCtl_p->dcpl9km   =  plist9km  ;   // for all 9km vars in FP and L4_SM
   sesCtl_p->dspace9km =  dspace9km ;   // for all 9km vars in FP and L4_SM
   
   sesCtl_p->dcpl3km   =  plist3km  ;   // for all 3km vars in L3_SM_A
   sesCtl_p->dspace3km =  dspace3km ;   // for all 3km vars in L3_SM_A
   
   // TBD: add file-level attributes here
   // "title","author","source", software versions, etc
   
 // return unique file_id
 return file_id;    
} // end::createMetHdf5()


// ---------------------------------------------------------------------
// function : createMetHdf5Dataset()
// purpose  : Create (WRITE) a given dataset into SMAP_L4_C_MET.h5 
//            (type: float32), using a common 
//            dataspace and plist properties, owned by file_id,
//            whose name is datasetName, using data in dsData_p
//            the dataset_id is NOT closed by this function, so must 
//            be done externally
// inputs   : SesCtl *sesCtl_p,
//          : int32 dsGridRes -- enum {IsEG_Global_9km, isEG_Global_3km}
//                  indicating grid, to identify dataspace
//          : const char *datasetName
//          : const int32 dsType_id -- H5T_NATIVE_FLOAT32, ..UINT32, etc.
//          : void *dsData_p -- data passed in, to write out here
// returns: hid_t,  file handle
// Notes  : As of v5.0.1, we now write out uint8 FT_STATE_UM_M03 dataset in place
//          of legacy "SURFACE_FLAG_M03" prev written.

//
hid_t createMetHdf5Dataset(SesCtl *sesCtl_p,const int32 dsGridRes,const char *datasetName,
            const hid_t dsType_id,void *dsData_p)
{
 hid_t   dataset_id =-1;
 herr_t  ret = -1 ;
 hid_t   dcpl   = -1;
 hid_t   dspace = -1 ;
 hid_t   file_id = sesCtl_p->met_file_id ; // handle to new HDF5 file
 char fxName[256] = {0} ;
 char *inputGranulePath = sesCtl_p->outputMetPath ;

char dsName[FILENAME_MAX] = {0}; // local datasetName adjusted to include any prefix GROUP

    strcpy(fxName,__func__);
    
    strcpy(dsName,"/");
    strcat(dsName,datasetName);
    
    if(dsGridRes == IsEG_Global_3km)
    {
        // <JMG> 2014-07-25, now create FT_STATE_UM_M03 uint8 dataset 
        // to replace the old uint16 SURFACE_FLAG_M03 previously used
        
        // L3_SM_A "surface_flag" uint8 definitions go here....
        dcpl   = sesCtl_p->dcpl3km ;
        dspace = sesCtl_p->dspace3km;
        // Convert local instance of dataset name to all uppercase to match names
        l4c_to_upper(dsName);
        // xLogMsg("@A In(%s) 3km dataspace properties dspace %d dcpl-9km %d for file_id %d \n\n",
        //    fxName, sesCtl_p->dspace3km, sesCtl_p->dcpl3km, file_id);
        fflush(sesCtl_p->logf);
        
        if(dspace <= 0)
        {
            xLogMsg("@A In(%s) INVALID 3km dataspace properties dspace %d dcpl-9km %d for file_id %d ~ \n\n",
                fxName, sesCtl_p->dspace9km, sesCtl_p->dcpl9km, file_id);
        }
        
        dataset_id = H5Dcreate(file_id, dsName, dsType_id, dspace, H5P_DEFAULT, dcpl, H5P_DEFAULT);
        
        H5Fflush(file_id, H5F_SCOPE_LOCAL); // flush...

        // xLogMsg("@A In(%s): CREATED-DATASET 3KM [%s] Id(%d) to [%s] file ID(%d) ~ \n",
        //        fxName, dsName, dataset_id, sesCtl_p->outputMetPath, file_id);

        if(dataset_id < 0)
            xLogMsg("@F In(%s): FAILED-DATASET-CREATION 3KM [%s] Id(%d) to [%s] file ID(%d) ~ \n",
                fxName, dsName, dataset_id, sesCtl_p->outputMetPath, file_id);
        
    } else if(dsGridRes == IsEG_Global_9km) {
        
        dcpl   = sesCtl_p->dcpl9km   ;
        dspace = sesCtl_p->dspace9km ;

        // xLogMsg("@A In(%s) 9km dataspace properties dspace %d dcpl-9km %d  file_id %d ~ \n\n",
        //    fxName, sesCtl_p->dspace9km, sesCtl_p->dcpl9km, file_id);
        fflush(sesCtl_p->logf);

        if(dspace <= 0)
        {
            xLogMsg("@A In(%s) INVALID 9km dataspace properties dspace %d dcpl-9km %d ~ \n\n",
                fxName, sesCtl_p->dspace9km, sesCtl_p->dcpl9km);
            dataset_id = -1 ;
        }
        // convert local instance of dataset name to all uppercase to match FP's names
        l4c_to_upper(dsName);
        
        // all L4_SM and GEOS-5 "float32" variables are handled in this block
        dataset_id = H5Dcreate(file_id, dsName, dsType_id, dspace, H5P_DEFAULT, dcpl, H5P_DEFAULT);         
        H5Fflush(file_id, H5F_SCOPE_LOCAL); // flush...

        // xLogMsg("@A In(%s): CREATED-DATASET 9KM [%s] Id(%d) to [%s] file ID(%d) ~ \n",
        //   fxName, dsName, dataset_id, sesCtl_p->outputMetPath, file_id);

        
        if(dataset_id <= 0)
            xLogMsg("@F In(%s): FAILED-DATASET-CREATION 9KM [%s] Id(%d) to [%s] file ID(%d) ~ \n",
                fxName, dsName, dataset_id, sesCtl_p->outputMetPath, file_id);

    } else {
        xLogMsg("@F ~createMetHdf5Dataset:: unsupported grid-res %d!\n",dsGridRes);
    }


    if(dataset_id >= 0)
    {
        // NOTE: we MAY likely need to supply grid specific dcpl and/or dataspace here and NOT take the default!!
        ret = H5Dwrite(dataset_id, dsType_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, dsData_p);

        // xLogMsg("@A In(%s) H5Dwrite dataset[%s] file_id %d dsetID %d dcpl %d dspace %d ret %d\n",
        //    fxName, dsName,file_id, dataset_id,dcpl,dspace,ret);
        
        if(ret < 0)
        {
            xLogMsg("@E In(%s) FAILED on dataset[%s] owner %d\n",fxName,datasetName,file_id);
        }
        
        // Keep and return dataset_id to later pass to create- attribute for this dataset!
    } else {
        xLogMsg("@F In(%s) FAILED on dataset[%s] owner %d bad dataset ID (%d)~ \n",
                fxName, datasetName,file_id, dataset_id);
    } 

    // <JMG> 2014-10-21T14:37:00, reorganized
    // <JMG> 2015-09-04, reorganized to update valid_min,valid_max explicitly by variable
    xLogMsg("@A In(%s) WROTE dataset[%s] owner %d calling WriteDataset_metadata now\n",
         fxName, datasetName,file_id); 
         
    l4c_WriteDataset_metadata(sesCtl_p,dsName, dataset_id);
    
    l4c_EventCompletionMsg(__func__);
    
 // return the datasets unique id
 return dataset_id;
} // end:: createMetHdf5Dataset()


// -----------------------------------------------------------------------------
// Purpose: Write to SOC file, creating a HDF5 dataset (type: float32), using a common
//          dataspace and plist properties, owned by file_id, whose name is datasetName, 
//          using data in dsData_p, the dataset_id is NOT closed by this function, 
//          so must be done externally.
//
// Inputs : SesCtl *sesCtl_p,
//        : int32 dsMGridRes -- enum {IsEG_Global_9km, isEG_Global_3km} indicating grid, to identify dataspace
//        : const char *datasetName
//        : const int32 dsType_id -- H5T_NATIVE_FLOAT32, ..UINT32, etc.
//        : void *dsData_p -- data to write out
// Outputs:
hid_t createSOCHdf5Dataset(SesCtl *sesCtl_p,const int32 dsGridRes,const char *datasetName,const hid_t dsType_id,void *dsData_p)
{
hid_t   dataset_id =-1;

herr_t  ret = -1 ;
hid_t   dcpl   = -1;
hid_t   dspace = -1 ;
hid_t   file_id = sesCtl_p->soc_file_id ; // handle to existing SOC HDF5 file

char *inputGranulePath = sesCtl_p->outputMetPath ;
char fxName[256] ;
char dsName[FILENAME_MAX] = {0}; // local datasetName adjusted to include any prefix GROUP

    strcpy(fxName,__func__);
    
    strcpy(dsName,"/");
    strcat(dsName,datasetName);
    
    if(dsGridRes == IsEG_Global_3km)
    {
        // L3_SM_A "ft_state_um_m03" (prior: "surface_flag")  definitions go here....
        dcpl   = sesCtl_p->dcpl3km ;
        dspace = sesCtl_p->dspace3km;
        
        // convert local instance of dataset name to all uppercase to match FP's names
        l4c_to_upper(dsName);
        
        // <JMG> 09.30.2013--first check if dataset already exists, fetch valid ID if it does
        // and only if it doesn't do we attempt to create it...
        if((dataset_id=H5Dopen(file_id,dsName,H5P_DEFAULT)) >0)
        {
            // create a new HDF5 dataset in the root level
            xLogMsg("@A In(%s)  ~createSOCHdf5Dataset::H5Dopen:: OPEN-EXISTING-DATASET 3km [%s]\nfrom granule:[%s] id %d\n",
                fxName, dsName, inputGranulePath, dataset_id);      
            // note: may have to explicitly re-fetch dcpl and/or dspace again here? verify!
        } else {
            // ADD a custom dataspace creation property list for the 3KM uint8 freeze-thaw instance
            dataset_id = H5Dcreate(file_id, dsName, dsType_id, dspace, H5P_DEFAULT, dcpl, H5P_DEFAULT);
            // create a new HDF5 dataset in the root level
            xLogMsg("@A ~createSOCHdf5Dataset::H5Dcreate:: CREATED-DATASET 3km [%s]\nfrom granule:[%s] id %d\n",
                dsName, inputGranulePath, dataset_id);
        }
        
    } else if(dsGridRes == IsEG_Global_9km) {
    
        dcpl   = sesCtl_p->dcpl9km   ;
        dspace = sesCtl_p->dspace9km ;
        // convert local instance of dataset name to all uppercase to match FP's names
        l4c_to_upper(dsName);
        
        if((dataset_id=H5Dopen(file_id,dsName,H5P_DEFAULT)) >0)
        {
            // create a new HDF5 dataset in the root level
            xLogMsg("@A ~createSOCHdf5Dataset::H5Dopen:: OPEN-EXISTING-DATASET 9km [%s]\nfrom granule:[%s] id %d\n",
                dsName, inputGranulePath, dataset_id);      
        } else {
            // all L4_SM and GEOS-5 "float32" variables are handled in this block
            dataset_id = H5Dcreate(file_id, dsName, dsType_id, dspace, H5P_DEFAULT, dcpl, H5P_DEFAULT);         
            // create a new HDF5 dataset in the root level
            xLogMsg("@A ~createSOCHdf5Dataset::H5Dcreate:: CREATED-DATASET 9km [%s]\nfrom granule:[%s] id %d\n",
                dsName, inputGranulePath, dataset_id);
        }
        
    } else {
        xLogMsg("@F ~createSOCHdf5Dataset:: unsupported grid-res %d!\n",dsGridRes);
    }
        
    if(dataset_id >= 0)
    {
        xLogMsg("@A In(%s) ~ H5Dwrite dataset [%s] file_id %d dsetID %d dcpl %d dspace %d\n",
            fxName, dsName,file_id, dataset_id, dcpl, dspace);

        ret = H5Dwrite(dataset_id, dsType_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, dsData_p);         
        
        if(ret < 0)
        {
            xLogMsg("@E In(%s) FAILED on dataset[%s] owner %d ret %d ~ \n",
                    fxName, datasetName,file_id, ret);
        } else {
            xLogMsg("@A In(%s) OK - WROTE dataset[%s] owner %d ret %d\n",
                fxName, datasetName,file_id, ret );     
        }
        
        frm_hdf5_CreateStringAttribute(dataset_id,"units", "<undefined!>");
        frm_hdf5_CreateStringAttribute(dataset_id,"short_name" ,datasetName);
        frm_hdf5_CreateStringAttribute(dataset_id,"long_name"  ,datasetName);
        
        // Keep and return dataset_id to later pass to create- attribute for this dataset!
        
    } else {
        xLogMsg("@E In(%s): FAILED on dataset[%s] owner %d bad-dataset id %d\n",
                fxName, datasetName,file_id, dataset_id);
    } 
    
 // return the datasets unique id
 return dataset_id;
} // end:: createSOCHdf5Dataset()

// ----------------------------------------------------------------
// function :int32 readPftMask3km(ImgPkg *imagePkg_p)
// Purpose: open,read Plant Function Type (PFT) 3km pixel mask {0|1} image
// Inputs : ImgPkg *imagePkg_p -- ptr to container holding PFT and
//          related ancillary image arrays (pointers)
// Returns: status, no. of pixels read (nElem_9km on success)
// Outputs: ptr to same.
// notes  : modified call interface to match lat,lon reader
// revised: 2015-07-31T11:14, added to support new GEOS FP TS derived FreezeThaw,
//          so expects to read 3KM pft from SMAP_L4_C_Ancillary.h5 .
// History: 2013-10-31T12:37 (v4.6.6), modified to read pftImg_p out of PFT_M09_mask" dataset inside
//      the SMAP_L4_C_Ancillary.h5 instead of as a raw binary file.
int32 readPftMask3km(ImgPkg *imagePkg_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();    
 char *hdf5Path   = sesCtl_p->inputAncPath ;
 char fxName[256] = {0};
  
 int32 status = -1 ;
 herr_t  ret  = -1 ;
 hid_t dsTypeId = H5T_NATIVE_UINT8 ;
 
 size_t nRows3km =  4872 ;
 size_t nCols3km = 11568 ;
 size_t ease_serialIdx = 0 ;
 size_t r,c, pftIdx =0 ;
 size_t pftPixAbsFrq[512] = {0} ;
 uint8  pft_pix = 0 ;
 int32   absFreq = 0 ;
 float32 relFreq = 0.0,
         cumFreq = 0.0 ;
  
 
    strcpy(fxName,__func__);
    
    // Create buffer and READ EASEGRID 3km (4872 x 11568) mask image
    imagePkg_p->pftImg3km_p = (uint8*)memCalloc(imagePkg_p->nElem3km,sizeof(uint8));

    // From SMAP_L4_C_Ancillary.h5, READ uint8 EG 9km {0|1} PFT ask image
    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, dsTypeId, 
        imagePkg_p->pft3kmDatasetname, (uint8 *)imagePkg_p->pftImg3km_p,fxName);

    xLogMsg("@A In(%s): READ dataset (%s) from (%s) file_id %d Status %d~ \n",
                fxName, imagePkg_p->pft3kmDatasetname, 
                hdf5Path, sesCtl_p->anc_file_id, status);
    
    xLogMsg("@A In(%s) PFT_MASK_3KM Verification~\n",fxName);
    // DIAG only, verify contents of pftImg3km_p[] raster
    for(r=0; r < nRows3km; r++)
    {
        for(c=0; c < nCols3km; c++)
        {
            ease_serialIdx = (r * nCols3km) + c;
            pft_pix = imagePkg_p->pftImg3km_p[ease_serialIdx] ;
            // increment bin count for this PFT 3KM value
            pftPixAbsFrq[pft_pix]++ ;
        } // end:: c column loop
    } // end:: r row loop
    
    for(pftIdx=0; pftIdx < 255; pftIdx++)
    {
        if(pftPixAbsFrq[pftIdx] > 0)
        {
           absFreq = (int32)pftPixAbsFrq[pftIdx] ;
           relFreq = ((float)pftPixAbsFrq[pftIdx]/(float)N_3KM_PIX)*100.0 ;
           cumFreq += relFreq ;
           xLogMsg("@A In(%s) PFT-MASK 3KM Bin %3d abs-freq %d rel-freq %7.2f cum-freq %7.2f~\n",
           fxName, pftIdx,
           absFreq, relFreq, cumFreq );
        } // end:: if a positive frequency hit
    } // end:: pfxIdx loop through all possible frequency bin indices
    
    
  // return status
  return status;
  
} // end::readPftMask3km()


// ----------------------------------------------------------------
// function :int32 readPftMask9km(ImgPkg *imagePkg_p)
// Purpose: open,read Plant Function Type (PFT) 9km pixel mask {0|1} image
// Inputs : ImgPkg *imagePkg_p -- ptr to container holding PFT and
//          related ancillary image arrays (pointers)
// Returns: status, no. of pixels read (nElem_9km on success)
// Outputs: ptr to same.
// notes  : modified call interface to match lat,lon reader
// History: 2013-10-31T12:37 (v4.6.6), modified to read pftImg_p out of PFT_M09_mask" dataset inside
//      the SMAP_L4_C_Ancillary.h5 instead of as a raw binary file.
int32 readPftMask9km(ImgPkg *imagePkg_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();    
 char *hdf5Path = sesCtl_p->inputAncPath ;
 char fxName[256] = {0};
  
 int32 status = -1 ;
 herr_t  ret  = -1 ;
 hid_t dsTypeId = H5T_NATIVE_UINT8 ;
    
 size_t ease_serialIdx = 0 ;
 size_t r,c, pftIdx = 0 ;
 size_t pftPixAbsFrq[512] = {0} ;
 size_t nRows9km = 1624 ;
 size_t nCols9km = 3856 ;
 uint8 pft_pix = 0 ;
 int32   absFreq = 0 ;
 float32 relFreq = 0.0,
         cumFreq = 0.0 ;
        
    strcpy(fxName,__func__);
    
    // create buffer and READ EASEGRID 9km (1624 x 3856) mask image
    imagePkg_p->pftImg9km_p = (uint8*)memCalloc(imagePkg_p->nElem9km,sizeof(uint8));

    // From SMAP_L4_C_Ancillary.h5, READ uint8 EG 9km {0|1} PFT ask image
    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, dsTypeId, 
        imagePkg_p->pft9kmDatasetname, (uint8 *)imagePkg_p->pftImg9km_p,fxName);

    xLogMsg("@A In(%s): READ dataset (%s) from (%s) file_id %d Status %d~ \n",
                fxName, imagePkg_p->pft9kmDatasetname, 
                hdf5Path, sesCtl_p->anc_file_id, status);
    
    xLogMsg("@A In(%s) PFT_MASK_3KM Verification~\n",fxName);
    // DIAG only, verify contents of pftImg3km_p[] raster
    for(r=0; r < nRows9km; r++)
    {
        for(c=0; c < nCols9km; c++)
        {
            ease_serialIdx = (r * nCols9km) + c;
            pft_pix = imagePkg_p->pftImg9km_p[ease_serialIdx] ;
            // increment bin count for this PFT 3KM value
            pftPixAbsFrq[pft_pix]++ ;
        } // end:: c column loop
    } // end:: r row loop
    
    for(pftIdx=0; pftIdx < 255; pftIdx++)
    {
        if(pftPixAbsFrq[pftIdx] > 0)
        {
           absFreq = (int32)pftPixAbsFrq[pftIdx] ;
           relFreq = ((float)pftPixAbsFrq[pftIdx]/(float)N_9KM_PIX)*100.0 ;
           cumFreq += relFreq ;
           xLogMsg("@A In(%s) PFT-MASK 9KM Bin %3d abs-freq %d rel-freq %7.2f cum-freq %7.2f~\n",
           fxName, pftIdx,
           absFreq, relFreq, cumFreq );
        } // end:: if a pos freq hit
    } // end:: pfxIdx loop
    
  // return status
  return status;
  
} // end::readPftMask9km()


// ----------------------------------------------------------------
// function : readAndWriteLatLon_9kmCoordVars()
// Purpose: open,read lat,lon coordinate vars for EASEGRID 9km global grid
// Inputs : ImgPkg *imagePkg_p -- ptr to container holding lat,lon images
// Returns: int32, status from frm_hdf5_WriteData()
// Outputs: ptr to same.
// History: 2014-07-30T10:32:00
//
int32 readAndWriteLatLon_9kmCoordVars(ImgPkg *imagePkg_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();    
 char *hdf5Path   = sesCtl_p->inputAncPath ;
 char fxName[256] = {0};
 size_t nElem9km  = imagePkg_p->nElem9km ;
 
 int32 status = -1 ;
 herr_t  ret  = -1 ;
 hid_t dsTypeId = H5T_NATIVE_FLOAT ;
 hsize_t lldims[2] = {0} ;
 
    strcpy(fxName,__func__);
    
    // create image buffers for lat,lon
    imagePkg_p->latImg_M09_p = (float32*)memCalloc(imagePkg_p->nElem9km, sizeof(float32));
    imagePkg_p->lonImg_M09_p = (float32*)memCalloc(imagePkg_p->nElem9km, sizeof(float32));
    
    // READ LAT, from SMAP_L4_C_Ancillary.h5, READ float32 lat 9km bin-center image
    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, dsTypeId,
      sesCtl_p->dsNameLatCoordvar, (float32 *)imagePkg_p->latImg_M09_p,fxName);

    xLogMsg("@A In(%s): READ latitude 9km dataset (%s) from (%s) file_id %d Status %d~ \n",
                fxName, sesCtl_p->dsNameLatCoordvar,
                hdf5Path, sesCtl_p->anc_file_id, status);

    // READ LON, from SMAP_L4_C_Ancillary.h5, READ float32 lon 9km bin-center image
    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, dsTypeId,
      sesCtl_p->dsNameLonCoordvar, (float32 *)imagePkg_p->lonImg_M09_p,fxName);

    xLogMsg("@A In(%s): READ longitude 9km dataset (%s) from (%s) file_id %d Status %d~ \n",
                fxName, sesCtl_p->dsNameLonCoordvar,
                hdf5Path, sesCtl_p->anc_file_id, status);

    // next, write out
    lldims[0] = NROW9KM ;
    lldims[1] = NCOL9KM ;

    status = frm_hdf5_WriteData(sesCtl_p->met_file_id, dsTypeId,
        "cell_lat", imagePkg_p->latImg_M09_p, 
        2,lldims,fxName);
    
    status = frm_hdf5_WriteData(sesCtl_p->met_file_id, dsTypeId,
        "cell_lon", imagePkg_p->lonImg_M09_p, 
        2,lldims,fxName);
        
    H5Fflush(sesCtl_p->met_file_id, H5F_SCOPE_LOCAL); // flush...
    
    // Free lat,lon bin center coordinate variable images
    memFree(imagePkg_p->latImg_M09_p);
    memFree(imagePkg_p->lonImg_M09_p);
    
  // return status
  return status;
  
} // end::readAndWriteLatLon_9kmCoordVars()


// -----------------------------------------------------------------------------
// function: fp_get_dataset()
// inputs  : hid_t file_id,char *datasetName,int *dRank,hsize_t *dims
// returns : void *image_p -- ptr to (datasetName) image read
// outputs : rank and dims[] ..array of dimensions
// author  : Joe Glassy
// revised : 2015-08-24, added documentation
// Purpose : Read a GEOS-5 FP variable 'datasetName' find its properties and read
//          the variable into a buffer, returning a pointer to the buffer.
// notes   : <JMG> 2015-08-24, note FP variables are typically 3D {1,721,1152}
//           noting that H5LT
// notes   : <JMG> 7/24/2014, migrated to use frm_hdf5_ReadData() etc.
//sesCtl_p->met_file_id
void *fp_get_dataset(hid_t file_id,char *datasetName,int *dRank,hsize_t *dims)
{
 herr_t status,ret    = -1;
 hid_t  dsTypeId = H5T_NATIVE_FLOAT ;

 size_t d,nElem =0;
 size_t sizOne = 0;
 char fxName[256];
 
 void *dsData_p= NULL;
 
 strcpy(fxName,__func__);
  
 if((ret=H5LTfind_dataset (file_id,datasetName))< 0)
 {
   xLogMsg("@F In(%s) GEOS5 FP required mandatory dataset %s is NOT present in %d!\n",
     fxName, datasetName,file_id);
   
 } else {
   ret = 1;
 }
 
 // fetch (minimalist properties) --> the dimensions of the dataset and size of it 
 ret  = H5LTget_dataset_info(file_id,datasetName,dims,NULL,&sizOne);
 // fprintf(stderr,"@E H5LTget_dataset_info status %d for dataset %s sizOne %zu\n",ret,datasetName,sizOne);
 
 // fetch this dataset's properties:: rank, then calc dims
 status = H5LTget_dataset_ndims(file_id,datasetName,dRank);
 
 // Establish total n. of pixels for this dataset. Here we account
 //  for typical GEOS5 FP dataset geometry of {1x721x1152} where time is
 //  the first dimension. Since time is 1 (unity) it doesn't change the overall
 //  interpretation of the dataset as a 2D dataset, read by frm_hdf5_ReadData()
 //  as 1D serial stream via a single buffer read, then is implicitly 
 //  interpreted later as a 2D matrix or grid.
  for(nElem=1,d=0; d < (size_t)*dRank ; d++)
  nElem *= dims[d];
  
  // Here we assume we know the datatype directly, as float32
  dsData_p = (float32 *)memCalloc(nElem,sizeof(float32));

  // now use frm_hdf5_ReadData()
  ret =  frm_hdf5_ReadData(file_id, dsTypeId, datasetName, dsData_p, fxName);

  // NOTE: using H5LTxxx family fx we don't retrieve the dataset_ID explicitly, 
  // so this is why we don't explicitly close it here
 
 // return pointer to a 1D serial stream of data, later to be interpreted 
 // as a 2D row-major order dataset
 return dsData_p;
} // end::fp_get_dataset()


// -----------------------------------------------------------------------------
// l4_sm_get_dataset()
// Purpose: retrieve L4_SM dataset named by 'datasetName' from open HDF5 file identified by file_id
// NOTE: currently, all datasets required here are float32. But if any are not, this function must
//       then be made type aware.
void *l4_sm_get_dataset(hid_t file_id,char *datasetName,int *dRank_p,hsize_t *dims,int32 *nPix)
{
 herr_t ret    = -1;
 hid_t  dsetID = -1 ;
 size_t d,nElem =0;
 size_t sizOne = 0;
 int    dsRank = 0, i =0;
 float32 chkSum =0.0, chkMean =0.0 ;
 float32 *chkFlt_p = NULL;
 char fxName[256] = {0} ;
 
 void *dsData_p= NULL;
 
    strcpy(fxName,__func__);
    
  // Open the dataset the long (explicit) way. Returns a neg value if fails
  // in order to obtain a proper dataset ID, needed below...
  if((dsetID = H5Dopen(file_id,datasetName,H5P_DEFAULT)) <= 0)
  {
   xLogMsg("@E In(%s) H5Dopen::SMAP L4_SM dataset [%s] is NOT present in %d.  Ret %d~\n",
        fxName, datasetName, file_id,ret);  
  } else {
   // xLogMsg("@A In(%s) H5Dopen::SMAP L4_SM file_id %d dataset [%s] opened with dsetID %d~\n", fxName, file_id, datasetName, dsetID);
  }
  
 // fetch dims and rank
 if((ret= H5LTget_dataset_ndims(file_id,datasetName,dRank_p)) < 0)
 {
    xLogMsg("@E In(%s) H5LTget_dataset_ndims:: failed on %s! rank %d~ \n",
        fxName, datasetName,*dRank_p);
 } else {
    // xLogMsg("@A In(%s) H5LTget_dataset_ndims:: dataset(%s) rank %d~ \n", fxName,datasetName,*dRank_p);
    ret = 0;
 }
 dsRank = *dRank_p;
 
 // fetch (minimalist properties) --> the dimensions of the dataset and size of it 
 if((ret  = H5LTget_dataset_info(file_id, datasetName, dims,NULL,&sizOne)) < 0 )
 {
    xLogMsg("@F In(%s) H5LTget_dataset_info FAILED %d for dataset %s sizOne %zu~ \n",
        fxName, ret, datasetName,sizOne);
 }
 
 // Confirm total n. of pixels for this dataset
 for(nElem=1,d=0; d < (size_t)dsRank ; d++)
        nElem *= dims[d];
 *nPix = nElem;
 // Check to make sure...
 assert(nElem == N_9KM_PIX ) ;
 assert(*nPix == N_9KM_PIX ) ;
  
 // here we assume we know the datatype directly
 dsData_p=(float32 *)memCalloc(nElem,sizeof(float32)) ;
 
    // READ the datasetName into buffer dsData_p... using a simplified "LT" call
    if((ret=H5LTread_dataset_float(file_id, datasetName,(float32 *)dsData_p))<0)
    {
     fprintf(stderr,"@F l4_sm_get_dataset::H5LTread_dataset_float FAILED on [%s] for %zu float32\n",datasetName,nElem);
     exit(-1);
    } else {
        // fprintf(stdout,"@A l4_sm_get_dataset::H5LTread_dataset_float:: [%s] READ\n",datasetName);
        ret = 0;
    }
    
#ifdef DIAG_ONLY    
    // DIAG only
    chkFlt_p = (float32 *)dsData_p;
    for(chkSum=0.0,i=0;i < N_9KM_PIX; i++)
       chkSum += chkFlt_p[i] ;
    chkMean = chkSum / (float32)N_9KM_PIX;
    xLogMsg("@A ~l4_sm_get_dataset:: H5LTread_dataset_float::DIAG-CHECK [%s] Sum %.2f Mean %.2f\n",
        datasetName, chkSum,chkMean);   
#endif      

 // verify where dataset is closed if not here...
 ret = H5Dclose(dsetID);
 // xLogMsg("@A In(%s) ~l4_sm_get_dataset::H5Dclose:: DIAG ret %d\n",fxName,ret);
 
 // return pointer to data read
 return dsData_p;
} // end::l4_sm_get_dataset()


// -----------------------------------------------------------------------------
// function: l3_sm_a_get_surface_flag()
// inputs  : SesCtl *sesCtl_p,char *datasetName,int *dRank,hsize_t *dims
// purpose : retrieve L3_SM_A dataset 'datasetName' from the open 
//          L3_SM_A HDF5 file identified by file_id
// returns : uint16 *, pointer to (1) surface_flag uint16 image.
// revised : 2015-07-30, changed name to l3_sm_a_get_surface_flag() since
//           this is all this really does now.
// NOTE    : datasetName --> "surface_flag" field, is uint16, dims[0]= 4872, dims[1]=11568
//
uint16 *l3_sm_a_get_surface_flag(SesCtl *sesCtl_p,char *datasetName,int *dRank,hsize_t *dims)
{
 herr_t ret    = -1;
 hid_t  dsetID = -1 ;
 hid_t  dsType_id = H5T_NATIVE_UINT16 ;
 hid_t  file_id = -1;
 size_t d,nElem =0;
 size_t sizOne = 0;
 int    dsRank = 0, i =0;
 int32  serialIdx = 0 ;   // frq bin slot index
 int32  nBins = 65536 ;   // max no. of candidate frequency bins
 int32  nOccurs= 0    ;   // count of pos valued frequency bins
 uint16 *uPix_p = NULL;  // typed ptr to surface_flag buffer dsData_p
 uint16 *frqBins = NULL; // for DIAG context, read verification only
 char   fxName[256];
 void   *dsData_p = NULL;
 
    // fetch file_id handle to pre-opened L3_SM_A
    strcpy(fxName,__func__);
    file_id = sesCtl_p->l3smaFile_id ;
  
    // Variable needed is: "surface_flags" in L3_SM_A
    // fetch rank, then using H5LTget_dataset_info() Fetch dims[]
    if((ret= H5LTget_dataset_ndims(file_id,datasetName,dRank)) < 0)
    {
        xLogMsg("@F In(%s) ~ H5LTget_dataset_ndims:: failed on file_id %d (%s) rank %d\n",
            fxName, file_id, datasetName,*dRank);
    } 
    dsRank = *dRank;
 
    // fetch (minimalist properties) --> the dimensions of the dataset and size of it 
    if((ret  = H5LTget_dataset_info(file_id, datasetName, dims,NULL,&sizOne)) < 0 )
    {
        xLogMsg("@F In(%s) ~ H5LTget_dataset_info FAILED %d for dataset (%s) sizOne %zu\n",
            fxName, ret, datasetName,sizOne);
    }
 
    // Confirm total n. of pixels for this dataset
    for(nElem=1,d=0; d < (size_t)*dRank ; d++)
                nElem *= dims[d];
 
    // Check to make sure of right resolution
    assert(nElem == N_3KM_PIX ) ;
  
    // here we assume we know the datatype (uint16) directly
    dsData_p=(uint16 *)memCalloc(nElem,sizeof(uint16)) ;
    
    // READ uint16 surface_flag variable from L3_SM_A
    if((ret=H5LTread_dataset(file_id,datasetName, dsType_id,(uint16 *)dsData_p)) < 0)
    {
        xLogMsg("@F %(s)  ~ H5LTread_dataset FAILED on [%s] for %zu float32 \n",
                fxName, datasetName,nElem);
    } 
        
    xLogMsg("@A In(%s) H5LTread_dataset:: [%s] (uint16) READ~ \n",
        fxName, datasetName);
    ret = 0;
            
#ifdef DIAG_ONLY            
            // DIAG ONLY -- READ VERIFICATION
            nBins = 65536 ;
            frqBins=(uint16 *)memCalloc(nBins,sizeof(uint16));
            
            uPix_p = (uint16 *)dsData_p;
            for(serialIdx=0; serialIdx < N_3KM_PIX; serialIdx++)
            {
             frqBins[ uPix_p[serialIdx]]++ ; // increment occurrence count of this pixel value
            }
            fprintf(stdout,"@A \nDIAG surface_flag frequency-bin verification\n");
            nOccurs = 0;
            for(serialIdx=0; serialIdx < nBins; serialIdx++)
            {
                if(frqBins[serialIdx]> 0)
                {
                    nOccurs++ ; // increment count of positive count bins
                    xLogMsg(
                        "@A ~l3_sm_a_get_dataset:: surface_flag pix value %6d occurred %6u times\n",
                        serialIdx,frqBins[serialIdx]);
                } // end:: block if pixel valued at 'serialIdx' occurred at least once
            }
            memFree(frqBins);
            xLogMsg("@A ~l3_sm_a_get_dataset::DIAG complete\n");
#endif          

 xLogMsg("@A In(%s) ~ H5LTread_dataset status:: DIAG ret %d\n",fxName, ret);
 
 // return pointer to data read
 return dsData_p;
} // end::l3_sm_a_get_dataset()


// ---------------------------------------------------------------------
// function  : write_FtStateUM_3km(SesCtl *sesCtl_p)
// purpose   : Write out the FT_STATE_UM_3KM dataset to the SMAP_L4_C_MET.h5 output file
// inputs    : SesCtl *sesCtl_p
//             const char *datasetName, "FT_STATE_UM_M03" by default
//             uint8 *ftStateUm_M03_Img_p
// returns   : int , 0 , status
// revised   :2015-07-31T12:24:00
// author    : joe glassy
// notes     : n/a
int write_FtStateUM_3km(SesCtl *sesCtl_p,const char *datasetName, uint8 *ftStateUm_M03_Img_p)
{
 herr_t  status      = -1 ;
 hid_t   dsType_id   = -1 ;
 hid_t   dset_id = -1 ;
 
 int g =0;
 
 size_t nElem3km = N_3KM_PIX ; // 4872 x 11568 
 
 char fxName[256] = {0} ;
 char hDatasetname[256] = {0};
 char datasourceLbl[256] = {0} ;
 char attribStr[256] = {0} ;
  
    strcpy(fxName,__func__);

    dsType_id = H5T_NATIVE_UINT8 ;


    // Perform a bitwise analysis of FT_STATE_UM_M03 and report it out
    l4c_SummarizeFreezeThaw(sesCtl_p,ftStateUm_M03_Img_p,nElem3km);

    // WRITE: FT_STATE_UM_M03, New consolidated F/T info passed through to L4_C 
    // <JMG> 2014-07-25T18:45, added "FT_STATE_UM_M03" as a 2D 3KM uint8 field...
    
    dset_id = createMetHdf5Dataset(sesCtl_p, IsEG_Global_3km, datasetName, dsType_id, ftStateUm_M03_Img_p);
    
    // Referencing each of (24) hourly GEOS5 granules, add provenance metadata, indicating specific source of this data
    for(g= 0; g < sesCtl_p->nFP_Granules; g++)
    {
         sprintf(datasourceLbl,"INPUTPOINTER_GEOS5_FP_%-3.3d",g+1) ;
         frm_hdf5_CreateStringAttribute(dset_id, datasourceLbl,sesCtl_p->inputGEOS5_FP_pathList[g]);
    }
    
   sprintf(attribStr,"bit0 Freeze-Thaw State         frozen(%9zu) thawed(%9zu) ~\n", sesCtl_p->ftStateFrq[1]  ,sesCtl_p->ftStateFrq[0] );    
   frm_hdf5_CreateStringAttribute(dset_id,"BIT0_FREQ",attribStr);
   
   sprintf(attribStr,"bit1 ft_radar                  frozen(%9zu) thawed(%9zu) ~\n", sesCtl_p->ftBit1Frq[1]   ,sesCtl_p->ftBit1Frq[0]   );   
   frm_hdf5_CreateStringAttribute(dset_id,"BIT1_FREQ",attribStr);
   
   sprintf(attribStr,"bit2 radar_qc Good=1, Bad=0    Good  (%9zu)   Bad (%9zu) ~\n", sesCtl_p->ftQcFrq[1],sesCtl_p->ftQcFrq[0]     );   
   frm_hdf5_CreateStringAttribute(dset_id,"BIT2_FREQ",attribStr);

   sprintf(attribStr,"bit3 InSwath (0=In 1=Out-swath)   In (%9zu)   Out (%9zu) ~\n", sesCtl_p->ftInSwathFrq[0],sesCtl_p->ftInSwathFrq[1]);   
   frm_hdf5_CreateStringAttribute(dset_id,"BIT3_FREQ",attribStr);

   sprintf(attribStr,"bit4 GEOS5 used YES=1,NO=0        YES (%9zu)   NO (%9zu) ~\n", sesCtl_p->ftBit4Frq[1],sesCtl_p->ftBit4Frq[0]);   
   frm_hdf5_CreateStringAttribute(dset_id,"BIT4_FREQ",attribStr);
   
   sprintf(attribStr,"bit5 GEOS5 Only: In-Terrestrial?   In (%9zu)  Out (%9zu) ~\n", sesCtl_p->ftBit5Frq[1],sesCtl_p->ftBit5Frq[0]);   
   frm_hdf5_CreateStringAttribute(dset_id,"BIT5_FREQ",attribStr);   

   sprintf(attribStr,"bit6 TSURF used YES=1,NO=0        YES (%9zu)   NO (%9zu) ~\n", sesCtl_p->ftBit6Frq[1],sesCtl_p->ftBit6Frq[0]);   
   frm_hdf5_CreateStringAttribute(dset_id,"BIT6_FREQ",attribStr);   

   sprintf(attribStr,"bit7 RADAR used YES=1,NO=0        YES (%9zu)   NO (%9zu) ~\n", sesCtl_p->ftBit7Frq[1],sesCtl_p->ftBit7Frq[0]);
   frm_hdf5_CreateStringAttribute(dset_id,"BIT7_FREQ",attribStr);      
     
    H5Dclose(dset_id);
      
    xLogMsg("@A In(%s): (%s) dataset written out~\n",fxName,datasetName);
    
 return 0 ;    
} // end:: write_FtStateUM_3km()

