// ---------------------------------------------------------------------
//   module     : l4c_MET_rip.c
//   purpose    : Implement runtime input parameter (RIP) and
//                name-value-pair related processing for l4c_MET
//   revised    : 2015-08-24T15:52, now defines L4C_MET_INPUT_VAR_GEOS5_FP_FT keyword
//   history    : 2015-08-06T15:43, added robust pre-check on keyword existance
//   history    : 2015-06-10T16:19,  changed L4C_SOFTWARE_VERSION to L4C_SW_VERSION
//   history    : 2014-07-30T15:22, split out "core" from app-specific NVPs
//   history    : v5.0.1-2014-07-29T11:26:00-07:00, D5+
//   version    : 6.x, see AppVers macro in l4c_MET.h
//   author     : joe glassy, Lupine Logic Inc.
//   notes      : 7/29/2014, use caution when modifying an NSB entry which
//            uses one-time-only pre-allocated width stringtables. These
//            string widths are custom tailored by the FRM API NSB facility 
//            exactly sized to fit just the initial payload. To mitigate, use
//            frm_strExpandStringTbl() which adjusts (widens)
//            a string table so it can accomodate a pre-pended entry like ROOTDIR.
// ---------------------------------------------------------------------

#include "l4c_MET.h"
#include "l4c_MET_prot.h"


// ---------------------------------------------------------------------
// function :l4c_ApplyRootDirToPaths()
// Purpose  : Apply standard root-directory prepend rules, and apply 
//            any optional file-specific or generic ROOTDIR to any
//            associated BASE PATHS following these rules. All pre-qualifies
//            existence of any input files.
// Inputs   : SesCtl *sesCtl_p -- ptr to session controls
// returns  : void
// notes    : This must be called AFTER l4c_SesCollectNvps() has parsed
//            the RIP file and stored all releavant paths in the 
//            appropriate sesCtl (sesCtl_p) members.
void l4c_ApplyRootDirToPaths(SesCtl *sesCtl_p)
{
 char fxName[256] = {0};
 char *currPath_p = NULL;
 int ithGrn = 0 ;
  
 strcpy(fxName,__func__);

    xLogMsg("@A In(%s) %d GEOS5 FP and %d L4_SM granules identified\n",
        fxName,sesCtl_p->nFP_Granules,sesCtl_p->nL4_SM_Granules);
        
    // Step 1: Do one-time-only path qualifications
    frm_ioQualifyPathWithRoot(sesCtl_p->ancRootDir      , sesCtl_p->inputAncPath);
    frm_ioQualifyPathWithRoot(sesCtl_p->socRootDir      , sesCtl_p->inputSOCPath);
    
    frm_ioQualifyPathWithRoot(sesCtl_p->outputMetRootDir, sesCtl_p->outputMetPath);
  
    //int   ftSourceMode ; // DoUse_GEOS5_TS|DoUse_ACTIVE_RADAR|DoUse_PASSIVE_RADIOMETER etc
    if(strlen(sesCtl_p->l3smaRootDir) >= 2 && sesCtl_p->ftSourceMode != DoUse_GEOS5_TS)
    {
        frm_ioQualifyPathWithRoot(sesCtl_p->l3smaRootDir    , sesCtl_p->inputL3_SM_A_Path);
    }
    
    if(strlen(sesCtl_p->l3smaRootDir) >= 2 && sesCtl_p->ftSourceMode != DoUse_GEOS5_TS)
    {
       if( !frm_ioDoesFileExist(sesCtl_p->inputL3_SM_A_Path))
        xLogMsg("@E In(%s) L3_SM_A input file(%s) does NOT exist at named location~ \n",
            fxName, sesCtl_p->inputL3_SM_A_Path);
    }
    
    // Step 3: Verify (singletons) input files exist at the indicated path locations 
    if( !frm_ioDoesFileExist(sesCtl_p->inputAncPath))
        xLogMsg("@F In(%s) input file(%s) does NOT exist at named location~ \n",
            fxName, sesCtl_p->inputAncPath);

    if( !frm_ioDoesFileExist(sesCtl_p->inputSOCPath))
        xLogMsg("@F In(%s) input file(%s) does NOT exist at named location~ \n",
            fxName, sesCtl_p->inputSOCPath);

    
    // Step 2: Perform path pre-qualifications on multiple-granule file types

    for(ithGrn=0; ithGrn < sesCtl_p->nFP_Granules;  ithGrn++)
    {
        frm_ioQualifyPathWithRoot(sesCtl_p->fpRootDir, sesCtl_p->inputGEOS5_FP_pathList[ithGrn]);
        
        if( !frm_ioDoesFileExist(sesCtl_p->inputGEOS5_FP_pathList[ithGrn]))
        {
            xLogMsg("@F In(%s) FAILED-EXIST at location (%s)~ \n",
                fxName,sesCtl_p->inputGEOS5_FP_pathList[ithGrn]);
        }
    }
     
    for(ithGrn=0; ithGrn < sesCtl_p->nL4_SM_Granules;  ithGrn++)    
    {
        frm_ioQualifyPathWithRoot( sesCtl_p->l4smRootDir, sesCtl_p->inputL4_SM_pathList[ithGrn] );
        if( !frm_ioDoesFileExist(sesCtl_p->inputL4_SM_pathList[ithGrn]))
        {
            xLogMsg("@F In(%s) FAILED-EXIST at location (%s)~ \n",
                fxName,sesCtl_p->inputL4_SM_pathList[ithGrn]);
        }
    }
    
    frm_utilEventCompletionMsg(__func__, sesCtl_p->sesStrtSec, sesCtl_p->enableCompletionMsg);
    
} // end::l4c_ApplyRootDirToPaths()


// ---------------------------------------------------------------------
// function :l4c_SesCollectCoreNvps()
// Purpose  : collect "CORE" standardized NVPs from RIP file, loading
//             into destination vars
// Inputs   : SesCtl *sesCtl_p -- ptr to session controls
//          : NsbObj *nsb_p    -- ptr to NSB overseer
// returns  : NsbMgr *, ptr to name-value-pair string-block NSB overseer
// revision : 2015-08-06T15:43, added robust pre-check on keyword existance
// revision : 2014-07-30T15:22, split out "core" from app-specific NVPs
// notes    : TBD, wrap in a mutex lock to make thread safe
void  l4c_SesCollectCoreNvps(SesCtl *sesCtl_p)
{
// inherit NSB overseer
 NsbObj *nsbMgr_p   = l4c_GetNsbOverseer();
 // inherit pkgMgr_p 
 PkgMgr   *pkgMgr_p = l4c_GetPkgMgrOverseer();
 
 // ...and pThme_p used for FparClimatology special case handling 
 PkgTheme *pThme_p  = NULL;
 PkgObj   *pObj_p   = NULL;
 char    nvpBuf[FILENAME_MAX] = {0};    
 char    *dsetName_p= NULL; // ptr to RIP keyword for a NVP to rerieve...
 char    fxName[256] = {0} ; float64 resultFlt64 =0.0 ;

    // register this entry
    strcpy(fxName,(char *)__func__);
    xLogMsg("@A In(%s) organizing CORE runtime input parameters (RIP)~ \n",fxName);
    
    strcpy(sesCtl_p->frmApiVersion, frm_GetApiVersion(nvpBuf));
    strcpy(sesCtl_p->pkgApiVersion, pkgMgr_GetVersion(nvpBuf));
    strcpy(sesCtl_p->nsbApiVersion, nsbGetVersion(nvpBuf));
    
    // fetch: Science Version ID, CRID, release tag, software sw version, "action" operator etc
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_SCIENCE_VERSION_ID"))
    {
        strcpy(sesCtl_p->ScienceVersID, nsbGetValueByKey(nsbMgr_p,"L4C_SCIENCE_VERSION_ID",nvpBuf));
    } else {
        xLogMsg("@E In(%s) L4C_SCIENCE_VERSION_ID not found!~\n",fxName);
    }
    
    strcpy(sesCtl_p->CRID ,nsbGetValueByKey(nsbMgr_p,"L4C_CRID",nvpBuf));

    // Note: a "RELEASE_TAG" here refers to the combination of a 
    // tightly-coupled collection of required ANCILLARY DATA 
    // (BPLUT,SMAP_L4_C_Ancillary.h5,etc) making up a release instance
    // L4C_SUBSYSTEM_RELEASE
    strcpy(sesCtl_p->SubsysReleaseTag,"");
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_SUBSYSTEM_RELEASE"))
    {
        strcpy(sesCtl_p->SubsysReleaseTag,nsbGetValueByKey(nsbMgr_p,"L4C_SUBSYSTEM_RELEASE",nvpBuf));
    } else {
        xLogMsg("@E In(%s) L4C_SUBSYSTEM_RELEASE tag was not found!~\n",fxName);
    }
    
    // <JMG> 2015-06-10T16:18, changed L4C_SOFTWARE_VERSION to L4C_SW_VERSION 
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_SW_VERSION"))
    {
        strcpy(sesCtl_p->swVersion,nsbGetValueByKey(nsbMgr_p,"L4C_SW_VERSION",nvpBuf));
    } else {
        xLogMsg("@E In(%s) L4C_SW_VERSION tag was not found!~\n",fxName);        
    }

    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MET_RUN_MODE"))
    {
        strcpy(sesCtl_p->appAction, nsbGetValueByKey(nsbMgr_p,"L4C_MET_RUN_MODE",nvpBuf));
    } else {
        xLogMsg("@E In(%s) L4C_MET_RUN_MODE tag was not found!~\n",fxName);        
    }
    
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MODEL_YEARDAY"))
    {
        strcpy(sesCtl_p->doy_Str,nsbGetValueByKey(nsbMgr_p,"L4C_MODEL_YEARDAY",nvpBuf));
        sesCtl_p->runDayOfYear = (int)atoi(nvpBuf);
    } else {
        xLogMsg("@E In(%s) L4C_MODEL_YEARDAY tag was not found!~\n",fxName);        
    }

    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MODEL_YEAR"))
    {
      nsbGetValueByKey(nsbMgr_p,"L4C_MODEL_YEAR",nvpBuf);
      sesCtl_p->runModelYear = (int)atoi(nvpBuf);
      strcpy(sesCtl_p->year_Str, nvpBuf);
    } else {
        xLogMsg("@E In(%s) L4C_MODEL_YEAR tag was not found!~\n",fxName);        
    }

    strcpy(sesCtl_p->l4c_Task_Label, nsbGetValueByKey(nsbMgr_p,"L4C_MET_TASK_LABEL",nvpBuf));
    strcpy(sesCtl_p->l4c_Task_UUID , nsbGetValueByKey(nsbMgr_p,"L4C_MET_TASK_UUID",nvpBuf));

    // xLogMsg("@A In(%s) L4C Subsystem release tag %s~ \n",fxName,sesCtl_p->SubsysReleaseTag);

    frm_utilEventCompletionMsg(__func__, sesCtl_p->sesStrtSec, sesCtl_p->enableCompletionMsg);
    
} // end:: l4c_SesCollectCoreNvps()


// ---------------------------------------------------------------------
// function :l4c_SesCollectNvps()
// Purpose  : collect all non-core (application-specific) NVPs from
//            RIP file, loading into destination vars in SesCtl and ImgPkg
// Inputs   : SesCtl *sesCtl_p -- ptr to session controls
//          : NsbObj *nsb_p    -- ptr to NSB overseer
// returns  : NsbMgr *, ptr to name-value-pair string-block NSB overseer
// revision : 2014-07-30T15:22, split out "core" from app-specific NVPs
// notes    : Call l4c_SesCollectCoreNvps() first to retrieve a 
//            standardized set of session parameters.
//
void  l4c_SesCollectNvps(SesCtl *sesCtl_p)
{
 ImgPkg *imagePkg_p = l4c_GetImagePackage(); // local pointer to support image package  
 // inherit NSB overseer
 NsbObj *nsbMgr_p = l4c_GetNsbOverseer();
 // inherit pkgMgr_p 
 PkgMgr *pkgMgr_p = l4c_GetPkgMgrOverseer();
 
 // ...and pThme_p used for FparClimatology special case handling 
 PkgTheme *pThme_p  = NULL;
 PkgObj   *pObj_p   = NULL;
 char    nvpBuf[FILENAME_MAX] = {0};    
 char    *dsetName_p= NULL; // ptr to RIP keyword for a NVP to rerieve...
 char    fxName[256] = {0} ; float64 resultFlt64 =0.0 ;
 int g =0 ;

 
    // register this entry
    strcpy(fxName,(char *)__func__);

    xLogMsg("@A In(%s) organizing APPLICATION-SPECIFIC runtime input parameters (RIP)~ \n",fxName);
    
    // L4C_MET_ENABLE_STATS  = YES
    sesCtl_p->enableSummarizeVar = 0 ;
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MET_ENABLE_STATS"))
    {
        nsbGetValueByKey(nsbMgr_p,"L4C_MET_ENABLE_STATS",nvpBuf);
        if(strcmp(nvpBuf,"YES")==0 || strcmp(nvpBuf,"yes")==0 || strcmp(nvpBuf,"ON")==0)
        {
            sesCtl_p->enableSummarizeVar = 1 ;
            xLogMsg("@A In(%s) Statistical summarizations are enabled ~\n",fxName);
        }
    } else {
         xLogMsg("@F In(%s Missing Mandatory keyword L4C_MET_FT_SOURCE!~\n",fxName);
    }
    
    
    // Remember to have caller (InitSession) call l4c_SesCollectCoreNvps() usually first

    // NEW, freeze-thaw datasource (GEOS5_TS,ACTIVE_RADAR,PASSIVE_RADIOMETER)
    // RIP Choices : "GEOS5_TS","ACTIVE_RADAR","ACTIVE_TSURF","PASSIVE_RADIOMETER"
    // Enum Choices: DoUse_GEOS5_TS, DoUse_ACTIVE_RADAR, DoUse_ACTIVE_TSURF, DoUse_PASSIVE_RADIOMETER
    strcpy(sesCtl_p->ftSourceStr,"NO-FT-SOURCE-DETERMINED");
    
    // Set default to use GEOS5 substitute for SMAP F/T
    sesCtl_p->ftSourceMode = DoUse_GEOS5_TS;
    
    strcpy(sesCtl_p->ftSourceStr,"FT_SOURCE-is-undefined!");
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MET_FT_SOURCE"))
    {
        strcpy(sesCtl_p->ftSourceStr,nsbGetValueByKey(nsbMgr_p,"L4C_MET_FT_SOURCE",nvpBuf));
    } else {
         xLogMsg("@F In(%s Missing Mandatory keyword L4C_MET_FT_SOURCE!~\n",fxName);
    }
    
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MET_FT_GEOS5VAR"))
    {
        strcpy(sesCtl_p->geos5ProxyVar,nsbGetValueByKey(nsbMgr_p,"L4C_MET_FT_GEOS5VAR",nvpBuf));
        xLogMsg("@A In(%s) GEOS5 user defined F/T proxy var: %s~ \n",fxName,sesCtl_p->geos5ProxyVar);
    }
        
    if(strcmp(sesCtl_p->ftSourceStr,"ACTIVE_RADAR")==0)
    {
        sesCtl_p->ftSourceMode = DoUse_ACTIVE_RADAR;
    } else if(strcmp(sesCtl_p->ftSourceStr,"ACTIVE_TSURF")==0) {
        sesCtl_p->ftSourceMode = DoUse_ACTIVE_TSURF;
    } else if(strcmp(sesCtl_p->ftSourceStr,"PASSIVE_RADIOMETER")==0) {
        sesCtl_p->ftSourceMode = DoUse_PASSIVE_RADIOMETER;
    } else if(strcmp(sesCtl_p->ftSourceStr,"GEOS5_TS")==0) {
        sesCtl_p->ftSourceMode = DoUse_GEOS5_TS;
    } else {
        // FALLBACK: can always rely on using this as a last resort
        sesCtl_p->ftSourceMode = DoUse_GEOS5_TS;
    }
    xLogMsg("@A In(%s) FreezeThaw data source mode is %s~ \n",fxName,sesCtl_p->ftSourceStr);

    // New, for GEOS5 derived F/T, uses L4C_ANC_DATASET_PFT_MASK_3KM =  MCD12Q1_M03_lc_mask_uint8
    strcpy(imagePkg_p->pft3kmDatasetname, nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_PFT_MASK_3KM",nvpBuf));
    // L4C_ANC_DATASET_PFT_MASK_9KM =  MCD12Q1_M09_lc_mask_uint8
    strcpy(imagePkg_p->pft9kmDatasetname, nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_PFT_MASK_9KM",nvpBuf));
    
    // latitude, longitude coord vars
    strcpy(sesCtl_p->dsNameLatCoordvar,nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_LATITUDE_9KM_COORD",nvpBuf));
    strcpy(sesCtl_p->dsNameLonCoordvar,nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_LONGITUDE_9KM_COORD",nvpBuf));
    // MCD12Q1_M09land_col_uint16
    strcpy(sesCtl_p->dsNameLandRowIndex,nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_M09LAND_ROW_INDEX",nvpBuf)); 
    // MCD12Q1_M09land_row_uint16
    strcpy(sesCtl_p->dsNameLandColIndex,nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_M09LAND_COL_INDEX",nvpBuf)); 

    strcpy(sesCtl_p->ancRootDir   ,nsbGetValueByKey(nsbMgr_p,"L4C_MET_INPUT_ANCILLARY_ROOTDIR",nvpBuf));    
    // Retrieve all ROOTDIR name-value-pairs
    strcpy(sesCtl_p->outputMetRootDir ,nsbGetValueByKey(nsbMgr_p, "L4C_MET_OUTPUT_ROOTDIR",nvpBuf));
    strcpy(sesCtl_p->socRootDir   ,nsbGetValueByKey(nsbMgr_p,"L4C_MET_INPUT_SOC_ROOTDIR",nvpBuf));
    
    strcpy(sesCtl_p->fpRootDir    ,nsbGetValueByKey(nsbMgr_p,"L4C_MET_INPUT_FP_ROOTDIR",nvpBuf));
    strcpy(sesCtl_p->l4smRootDir  ,nsbGetValueByKey(nsbMgr_p,"L4C_MET_INPUT_L4_SM_ROOTDIR",nvpBuf));
    
    // <JMG> 2015-08-03, Added check to handling missing L3_SM_A params
    strcpy(sesCtl_p->l3smaRootDir,".");
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MET_INPUT_L3_SM_A_ROOTDIR"))
    {
        strcpy(sesCtl_p->l3smaRootDir ,nsbGetValueByKey(nsbMgr_p,"L4C_MET_INPUT_L3_SM_A_ROOTDIR",nvpBuf));
    }
    
    // Retrieve all basepath name value pairs
    strcpy(sesCtl_p->inputAncPath ,nsbGetValueByKey(nsbMgr_p, "L4C_MET_INPUT_ANCILLARY_PATH",nvpBuf));
    strcpy(sesCtl_p->inputSOCPath ,nsbGetValueByKey(nsbMgr_p, "L4C_MET_INPUT_SOC_PATH",nvpBuf));
    strcpy(sesCtl_p->inputAncPath ,nsbGetValueByKey(nsbMgr_p, "L4C_MET_INPUT_ANCILLARY_PATH",nvpBuf));

    strcpy(sesCtl_p->outputMetPath ,nsbGetValueByKey(nsbMgr_p, "L4C_MET_OUTPUT_PATH",nvpBuf));

    // Fetch (3) lists of MET INPUT-VARIABLE NAMEs , one stringtable per input file
    sesCtl_p->varList_GEOS5_FP = nsbGetMultiValueByKey(nsbMgr_p,"L4C_MET_INPUT_VARLIST_FP"     ,&sesCtl_p->nFP_Variables);
    sesCtl_p->varList_L4_SM    = nsbGetMultiValueByKey(nsbMgr_p,"L4C_MET_INPUT_VARLIST_L4_SM"  ,&sesCtl_p->nL4_SM_Variables);
       
    // <JMG> 2015-08-03, initialize stringtable holding L3_SM_A variables as NULL for now
    sesCtl_p->varList_L3_SM_A = NULL;
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MET_INPUT_VARLIST_L3_SM_A"))
    {
        sesCtl_p->varList_L3_SM_A  = nsbGetMultiValueByKey(nsbMgr_p,
            "L4C_MET_INPUT_VARLIST_L3_SM_A",&sesCtl_p->nL3_SM_A_Variables);
    } else {
        xLogMsg("@W %s No L3_SM_A variable list encountered - must default to use GEOS5 FP TS!~\n",fxName);
        strcpy(sesCtl_p->ftSourceStr,"GEOS5_TS");
        sesCtl_p->ftSourceMode = DoUse_GEOS5_TS;
    }
    
    // Retrieve all MET INPUTS as specified
    strcpy(sesCtl_p->inputL3_SM_A_Path,"L3_SM_A_PATH-is-undefined");
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_MET_INPUT_L3_SM_A_PATH"))
    {
        strcpy(sesCtl_p->inputL3_SM_A_Path,nsbGetValueByKey(nsbMgr_p,"L4C_MET_INPUT_L3_SM_A_PATH",nvpBuf));
    } else {
        xLogMsg("@W %s No L3_SM_A path was defined - must default to use GEOS5 FP TS!~\n",fxName);
        strcpy(sesCtl_p->ftSourceStr,"GEOS5_TS");
        sesCtl_p->ftSourceMode = DoUse_GEOS5_TS;
        strcpy(sesCtl_p->geos5ProxyVar,"TS");
    }
    
    // Fetch LIST of all (hourly) FP input paths
    sesCtl_p->inputGEOS5_FP_pathList = nsbGetMultiValueByKey(nsbMgr_p,"L4C_MET_INPUT_FP_PATH"   , &sesCtl_p->nFP_Granules);
    
    // hard-verify valid count right away
    if(sesCtl_p->nFP_Granules < 24)
    {
        xLogMsg("@E %s Inadequate count (%d) of GEOS5 FP input granules! 24 expected~\n",
         fxName,sesCtl_p->nFP_Granules);
    } else {
        xLogMsg("@A %s Adequate count (%d) of GEOS5 FP input granules submitted ~\n",
         fxName,sesCtl_p->nFP_Granules);
    }
    assert(sesCtl_p->nFP_Granules >= 24);
    
    // Explicitly expand the orig string table to allow pre-pend of root dir
    // sesCtl_p->inputGEOS5_FP_pathList = l4c_ExpandStringTbl(sesCtl_p->nFP_Granules,
    //  strlen(sesCtl_p->fpRootDir),sesCtl_p->inputGEOS5_FP_pathList );
    
    // <JMG> 2015-10-15, TBD, frm_strExpandStringTbl() as used here causes valgrind
    //  to report "still reachable" bytes after returning, where there should not 
    //  technically be any 'still reachable' bytes. Final leak counts show 0 though.
    //  Investigate later..
    sesCtl_p->inputGEOS5_FP_pathList = frm_strExpandStringTbl(sesCtl_p->nFP_Granules,
        strlen(sesCtl_p->fpRootDir),sesCtl_p->inputGEOS5_FP_pathList );
    
    // Fetch LIST of all (3-hour timestep) L4_SM input paths
    sesCtl_p->inputL4_SM_pathList = nsbGetMultiValueByKey(nsbMgr_p,"L4C_MET_INPUT_L4_SM_PATH",&sesCtl_p->nL4_SM_Granules);
    // hard-verify valid count right away
    assert(sesCtl_p->nL4_SM_Granules >= 8);
    
    // Explicitly expand the orginal string table to allow it to accomodate
    // the prepend of a l4smRootDir string in front of existing info
    
    // sesCtl_p->inputL4_SM_pathList = l4c_ExpandStringTbl(sesCtl_p->nL4_SM_Granules,
    //  strlen(sesCtl_p->l4smRootDir),sesCtl_p->inputL4_SM_pathList);
    sesCtl_p->inputL4_SM_pathList = frm_strExpandStringTbl(sesCtl_p->nL4_SM_Granules,
        strlen(sesCtl_p->l4smRootDir),sesCtl_p->inputL4_SM_pathList);

    // Adjust ALL PATHS to PRE-PEND ROOTDIR if a relevant ROOT is supplied!
    // open SOC file in READ+WRITE mode
    // sprintf(socPath,"%s/%s",sesCtl_p->socRootDir, sesCtl_p->inputSOCPath);
    l4c_ApplyRootDirToPaths(sesCtl_p);
    
    xLogMsg("@A %s All L4C_MET Input Parameters have been processed~\n",fxName);
    
    frm_utilEventCompletionMsg(__func__, sesCtl_p->sesStrtSec, sesCtl_p->enableCompletionMsg);
    
} // end::l4c_SesCollectNvps()

