// ---------------------------------------------------------------------
//   module     : l4c_MET_sci.c
//   Purpose    : implement optional MET science logic for l4c_MET.c
//   Author     : joe glassy
//   Version    : v5.0.1, see AppVers macro in l4c_MET.h
//   Revision   : 2014-10-21T21:59:00, for ORT,ORR if possible
// ---------------------------------------------------------------------


#include "l4c_MET.h"
#include "l4c_MET_prot.h"


// --------------------------------------------------------------------
// function : l4c_MET_CalcVPD()
// purpose  :  -- compute vapor pressure deficit (in Pa units) from 
//           MERRA's QV2M specific humidity and PS (surface pressure) in Pa
// Inputs   : 
//   tavg -- daily average temperature in degC (if in Kelvin, must convert first)
//   qv2m -- specific humidity, in MERRA units
//   ps   -- surface pressure, in Pa (in MERRA units)
// Revision : 2014-10-21T22:30:00, adapted from L4CMDL
// Returns  : vpd, in Pa units

float32 l4c_MET_CalcVPD(const float32 tavg,const float32 qv2m,const float32 ps)
{
 float32 tavg_degC = tavg - 273.15; // Convert from K to deg C
 float32 vpd = 0.0; // vapor pressure deficit
 float32 esat= 0.0; // saturated vapor pressure
 float32 avp = 0.0; // ambient (actual) vapor pressure
 
    // <Yonghong>: used this logic in MERRA_hdf_process.c (as shown here)
    avp = (qv2m * ps) / (0.622 + (0.378 * qv2m)); // here, in Pa  units
    
    // Note: replaced orig:  avp = avp * 1000.0 ;  to convert kpa to pa 
    esat =  610.7 * exp(17.38 * tavg_degC/(239.0 + tavg_degC));
    
    vpd = (esat - avp);
    
    // bound it...
    vpd = (vpd < 0.0)? 0.0 : esat - avp;
  // return VPD in Pa units 
 return vpd ;
} // end:: l4c_MET_CalcVPD()

// ---------------------------------------------------------------------
// function     : l4c_MET_Init_VPD()
// purpose      : allocate input variables required by VPD
// inputs       : 
// returns      : size_t, count of pixels processed
// notes        : Sources: 
//                inpVpd_t2m_avg --  
//                inpVpd_qv2m    -- 
//                inpVpd_ps_avg  -- 
//     

int l4c_MET_InitForVPD(void)
{
ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; 
size_t ithPix =0 ;

 imagePkg_p->inpVpd_t2m_avg_p = (float32 *)memCalloc(N_9KM_PIX,sizeof(float32));
 imagePkg_p->inpVpd_qv2m_p    = (float32 *)memCalloc(N_9KM_PIX,sizeof(float32));
 imagePkg_p->inpVpd_ps_avg_p  = (float32 *)memCalloc(N_9KM_PIX,sizeof(float32));

    for(ithPix = 0; ithPix < N_9KM_PIX ; ithPix++)
    {
        imagePkg_p->inpVpd_t2m_avg_p[ithPix] = L4C_MET_MISSING_VALUE ;
        imagePkg_p->inpVpd_qv2m_p[  ithPix]  = L4C_MET_MISSING_VALUE ;
        imagePkg_p->inpVpd_ps_avg_p[ithPix]  = L4C_MET_MISSING_VALUE ;
    } // end:: ithPix loop

 return 0;

} // end:: l4c_MET_InitForVPD()



// ---------------------------------------------------------------------
// function     : l4c_MET_ProcessVPD(vpd_M09_ary_p)
// purpose      : allocate vpd array and then calculate pixel wise VPD
//                to store in 1D-sparse vector space metArrayObj_p->vpd array
// inputs       : float32 *vpd_M09_ary_p --ptr to VPD array
// returns      : size_t, count of pixels processed
// notes        : Sources: 
//                t2m_avg --  
//                qv2m    -- 
//                ps_avg  -- 
//
size_t  l4c_MET_ProcessVPD(float32 *vpd_M09_ary_p)
{
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // image pkg containing inputs
 float32 *in_t2m_p = imagePkg_p->inpVpd_t2m_avg_p ;
 float32 *in_qv2m_p= imagePkg_p->inpVpd_qv2m_p ;
 float32 *in_ps_p  = imagePkg_p->inpVpd_ps_avg_p;
 char fxName[256];
 size_t pix = 0;
 time_t secNow = time(0);

    strcpy(fxName,__func__);

    xLogMsg("@A In(%s) calculating VPD at (%s)~\n",fxName, ctime(&secNow));

        // t2m_avg, qv2m_p, and ps_avg_p

        // NCOL9KM,NROW9KM..
        for(pix=0; pix < (size_t)N_9KM_PIX; pix++)
        {
            // float32 l4c_calc_vpd(const float32 tavg,const float32 qv2m,const float32 ps)
            vpd_M09_ary_p[pix] = l4c_MET_CalcVPD(in_t2m_p[pix],in_qv2m_p[pix],in_ps_p[pix]);
            
        } // end:: for pix...

    // l4c_EventCompletionMsg(__func__);
    
  return pix ;      
  
} // end:: l4c_PreProc_Calculate_VPD()



