// ---------------------------------------------------------------------
//   module     : l4c_MET_ses.c
//   Purpose    : store session initialization and related code
//   Revised    : v5.0.1-2014-07-22T10:10:00-07:00, D5+
//   History    : n/a
//   Version    : 5.0x, see AppVers macro in l4c_MET.h
//   Author     : joe glassy, Lupine Logic Inc.
//   Notes      : n/a
// ---------------------------------------------------------------------

#include "l4c_MET.h"
#include "l4c_MET_prot.h"

static SesCtl sesCtl ;  // master instance, session controls object

// Contains long-lived images and NsbMgrOverseer and PkgMgrOverseer
static ImgPkg imgPackageMaster ;


// ---------------------------------------------------------------------
// function : l4c_GetExecutablVersionId()
// purpose  : Retrieve public ptr to executables datetime stamped version ID
// inputs   : void
// returns  : char *, static (long-lived) string containing executables version id
// notes    : Implemented across all components as a macro {L4_C_|L4_C_MODIS_|L4C_MET_}
char *l4c_GetExecutableVersionId(void)
{
 static char execSwVersion[64]= {0} ;
 
    // technically, this assignment only needs to be done once
    strcpy(execSwVersion,(char *)L4C_MET_VERS_TAG);

        // return pointer to long-lived version identifier
    return execSwVersion ;
    
} // end:: l4c_GetExecutableVersionId()

// ---------------------------------------------------------------
// l4c_GetSesCtl()
// purpose: return a pointer to the master (private) session control structure
// revision: 2013-11-04T14:55
// history : 2013-10-17T15:40:00, original refactoring
// note    : in progress...
SesCtl *l4c_GetSesCtl(void)
{
 return &sesCtl  ;
 
} // end:: l4c_GetSesCtl()


// ---------------------------------------------------------------------
// function :l4c_GetNsbOverseer()
// Purpose  : provides public access to private, NSB overseer 
// Inputs   : void
// returns  : NsbMgr *, ptr to name-value-pair string-block NSB overseer
// notes    : TBD, wrap in a mutex lock to make thread safe
NsbObj *l4c_GetNsbOverseer(void)
{
  ImgPkg *imagePkg_p = &imgPackageMaster;
  
  return imagePkg_p->NsbMgrOverseer_p ;

} // end::l4c_GetNsbOverseer()


// ---------------------------------------------------------------------
// function :l4c_GetPkgMgrOverseer()
// Purpose  : provides public access to private, PkgMgr overseer
// Inputs   : void
// returns  : NsbMgr *, ptr to name-value-pair string-block NSB overseer
// notes    : TBD, wrap in a mutex lock to make thread safe
PkgMgr *l4c_GetPkgMgrOverseer(void)
{
  ImgPkg *imagePkg_p = &imgPackageMaster;
  
  return imagePkg_p->PkgMgrOverseer_p ;

} // end::l4c_GetPkgMgrOverseer()


// ---------------------------------------------------------------------
// function :l4c_GetImagePackage()
// Purpose  : provides public access to private, ImgPkg container, 
//            which stores EASEGRID 9km mask (*pftImg_p) and related
//            other image resources.
// Inputs   : void
// returns  : ImgPkg *, pointer to master ImgPkg.
// notes    : TBD, wrap in a mutex lock to make thread safe
ImgPkg *l4c_GetImagePackage(void)
{

 return &imgPackageMaster ;
    
} // end:: l4c_GetImagePackage()


// ---------------------------------------------------------------------
// function : l4c_GetEasegrid9kmMaskImage()
// Purpose  : Provides public access to private, pftImg_p image pointer
//            for EASEGRID 9km mask (*pftImg_p)
// Inputs   : void
// returns  : uint8 * pointer to EASEGRID 9km PFT mask {0|1} image
// notes    : TBD, wrap in a mutex lock to make thread safe
uint8 *l4c_GetEasegrid9kmMaskImage(void)
{
 ImgPkg *imgPkg_p = &imgPackageMaster ;
 
 // return pointer to EASEGRID 9km mask image
 return  imgPkg_p->pftImg9km_p ;
    
} // end:: l4c_GetEasegrid9kmMaskImage()

// ---------------------------------------------------------------------
// function : l4c_GetEasegrid3kmMaskImage()
// Purpose  : Provides public access to private, pftImg_p image pointer
//            for EASEGRID 3km mask (*pftImg_p)
// Inputs   : void
// returns  : uint8 * pointer to EASEGRID 3km PFT mask {0|1} image
// notes    : TBD, wrap in a mutex lock to make thread safe
uint8 *l4c_GetEasegrid3kmMaskImage(void)
{
 ImgPkg *imgPkg_p = &imgPackageMaster ;
 
 // return pointer to EASEGRID 3km mask image
 return  imgPkg_p->pftImg3km_p ;
    
} // end:: l4c_GetEasegrid3kmMaskImage()


// ---------------------------------------------------------------
// function   : EventCompletionMsg()
// Purpose    : Announces that function 'thisFunc_p' has
//              completed, with a snapshot of current time,
// Inputs     : const char *thisFunc_p
// Returns    : double, cumulative sec since session began
// Notes      :  n/a
// Author     : joe glassy
double l4c_EventCompletionMsg(const char *thisFunction_p)
{
 SesCtl *sesCtl_p  = &sesCtl ;
 time_t secNow = time(0);
 char timeStr[32] = {0};
 
 // update a snapshot of cumulative running time
 double cumulativeSec = difftime(secNow,sesCtl_p->sesStrtSec) ;
 
    // <JMG> 4/28/2014, only display completion-message when
    // explicitly enabled, to reduce clutter...
    if(sesCtl_p->enableCompletionMsg == 1)
        xLogMsg("@A <%s> COMPLETED %s (%.2f elap sec)~\n",
            thisFunction_p,  frm_utilGetCtime(timeStr,0), cumulativeSec );
        
    return cumulativeSec;
    
} // end:: l4c_EventCompletionMsg()


// ---------------------------------------------------------------
// function   : l4c_EventHighWaterMarkMsg()
// Purpose    : Announces the point at which current 
//              increment build (high-water-mark) testing has
//              been completed to.
// Inputs     : const char *thisFunc_p
// Returns    : double, cumulative sec since session began
// Notes      :  n/a
// Author     : joe glassy
double l4c_EventHighWaterMarkMsg(const char *thisFunction_p)
{
 SesCtl *sesCtl_p  = &sesCtl ;
 char timeStr[32] = {0};
 
 time_t secNow = time(0);
 // snapshot cumulative running time
 double cumulativeSec = difftime(secNow,sesCtl_p->sesStrtSec) ;
 
    xLogMsg("@A <%s> HIGH-WATER-MARK-COMPLETION %s (%.2f elap sec)~\n",
        thisFunction_p,  frm_utilGetCtime(timeStr,0), cumulativeSec );
    
    // PERFORM INTENTIONAL, ORDERLY EARLY SHUTDOWN OF SESSION   
    l4c_CloseSession(sesCtl_p);
    
    // logic will never get to here but this is shown for completeness
    return cumulativeSec ;
    
} // end:: l4c_EventHighWaterMarkMsg()



// ---------------------------------------------------------------------
// function : l4c_InitSession(argc,argv,sesCtl_p)
// Purpose  : initialize the session
// Inputs   : argc --count of command line args
//            argv --vector of command line arg strings
//            sesCtl_p -- ptr to master session control object
// Actions  : initialize the session and read PFT 9km global mask in
int32 l4c_InitSession(int argc,char *argv[],SesCtl *sesCtl_p)
{
 int32 status = -1 ;
 char  fxName[256]= {0} ;
 char  versTagBuffer[64]= {0};
 char  timeBuffer[32]   = {0};
 char  tmpStr[128] = {0} ;
 char  nvpBuf[FILENAME_MAX] = {0};
 FILE *ripf = NULL;
 
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // local pointer to support image package
 NsbObj *nsbMgr_p    = l4c_GetNsbOverseer(); 
 size_t nElem3km     = N_3KM_PIX ; // 4872 x 11568
 
 pid_t thisPID = 0 ;
 pid_t sesParentPID = 0 ;


    // Step 1: Set hard defaults
    strcpy(fxName,(char *)__func__);
    
    FRM_API_RANDOMIZE() ; // for DIAG_ONLY, for random sampling of test results

    sesCtl_p->sesStrtSec  = time(0L);
    sesCtl_p->compressLev = 3;     // fallback default
    sesCtl_p->enableCompletionMsg = 1;
    sesCtl_p->enableSummarizeVar = 0 ; // boolean, default, 0=suppress it, later; 1=enable
    
    strcpy(sesCtl_p->frmApiVersion, frm_GetApiVersion(nvpBuf));
    strcpy(sesCtl_p->pkgApiVersion, pkgMgr_GetVersion(nvpBuf));
    strcpy(sesCtl_p->nsbApiVersion, nsbGetVersion(nvpBuf));
    // <JMG> 2015-08-19, name of GEOS5 FP variable serving as F/T proxy
    // defaulting to "TS", but allows user to optionally override
    strcpy(sesCtl_p->geos5ProxyVar, "TS");
    
    sesCtl_p->runDayOfYear = 1;
    sesCtl_p->runModelYear = 2014 ;
    
    // Generate a unique Task ID
    frm_utilGenerateUUID(sesCtl_p->l4c_Task_UUID,1);
    
    imagePkg_p->nElem1km = N_1KM_PIX;
    imagePkg_p->nElem3km = N_3KM_PIX;
    imagePkg_p->nElem9km = N_9KM_PIX;
    
    strcpy(sesCtl_p->inputAncPath , "") ; // initialize, to allow sanity check...
    strcpy(sesCtl_p->fpRootDir    , "")  ; // init, to allow for sanity check
    strcpy(sesCtl_p->l4smRootDir  , "")  ; // L4_SM root directory
    strcpy(sesCtl_p->l3smaRootDir , "")  ; // L3_SM_A root directory
    
    sesCtl_p->nFP_Granules      = 0 ; 
    sesCtl_p->nL4_SM_Granules   = 0 ;
    sesCtl_p->nL3_SM_A_Granules = 0 ;
    
    memset(sesCtl_p->thisHost,'\0', 256);
    
    if(argc >= 2)
    {
        strcpy(sesCtl_p->ripPath,argv[1]);
        // DEPRECATED OLD: status = parseRipLines(sesCtl_p);
    } else {
        fprintf(stderr,"@F l4c_MET invocation: <l4c_MET.rip> missing required RIP file!\n");
        exit(-1);
    }

    // for now, turn off HDF5's internal error handler
    // H5Eset_auto(error_stack, NULL, NULL), but use Framework's implementation
    // TEST with stack enabled first
    //    frm_hdf5_ErrorStackDisable();

    // Initialize the local memory manager and garbage collector
    // TBD: fix up which default log instance to give it ..later
    memObjInit(stdout);

    // Toggle OFF the verbose memObj messaging for now
    memObjAdjustVerbose(0);
    
    // Register which function to call atexit. Optional.
    // atexit(memObjGarbageCollect);

    // Capture name of host we're running on
    gethostname(sesCtl_p->thisHost,255) ;
    // Capture PID and any other diagnostic info
    thisPID = frm_utilGetPid(&sesParentPID);    

    // TBD: use definition from RIP file, to later relabel this if feasible
    strcpy(sesCtl_p->logPath,"SMAP_L4_C_MET.log");
    xOpenLogfile(sesCtl_p->logPath, sesCtl_p->logf);

    // -------------------- SESSION STARTUP MESSAGE -------------
    xLogMsg("\n~");
    xLogMsg("=====================================================================~\n");
    xLogMsg("@A L4_C task started %s on host [%s] at pid %zu parents-pid %zu ~\n",
            frm_utilGetCtime(timeBuffer, sesCtl_p->sesStrtSec),
            sesCtl_p->thisHost, (size_t)thisPID,(size_t) sesParentPID);
    
    xLogMsg("\n@A Executable %s version ID %s ~ \n",argv[0], l4c_GetExecutableVersionId() );

    xLogMsg("@A size of SesCtl is: %d ~\n", (int)sizeof(SesCtl));   

    // CREATE PKG MGR master instance as a scoping tool. 
    //  Create we only declare 0 themes since we'll create the actual
    //  themes using pkgMgr_CreateThemes() below
    imagePkg_p->nPkgThemes = L4C_MET_PKGMGR_THEME_COUNT ; 
    imagePkg_p->PkgMgrOverseer_p = pkgMgr_CreateMgr("L4_C_MET_PkgMgr", 0, "L4_C_MET_Data");
    
    // Initialize all Themes with info available so far
    l4c_PkgMgr_Init( imagePkg_p->nPkgThemes );
    
    // ----------------------------------------------------------------
    // OPEN and PARSE the main RIP file, retrieve ALL runtime inputs
    if((ripf=fopen(sesCtl_p->ripPath,"rt"))==NULL)
    {
        xLogMsg("@F In(%s) missing required RUNTIME INPUTS (RIP) FILE (%s)~\n",
            fxName, sesCtl_p->ripPath);
    } 
    
    // CREATE NSB overseer (name-value-pair stringblock datahbase) instance 
    imagePkg_p->NsbMgrOverseer_p = nsbCreateInstance("L4_C_MET_NSB_Manager");

    // -----------------------------------------------------------------
    // RIP -- Retrieve all RUNTIME INPUT PARAMETERS from RIP file
    // Note: the "" for delimiter string enables default delimiters ("\t= ")
    sesCtl_p->nTotalNvp = nsbParseIntoStringblock(imagePkg_p->NsbMgrOverseer_p,"",ripf);
    
    // DIAG ONLY: if need to verify that all keywords have been retrieved, stored etc.
    // nsbExportDbToText(sesCtl_p->sesNsbDb_p,"L4C_RIP_KeyVerification.txt",' ');
        
    xLogMsg("@A l4c_InitSession: %d name-value-pairs stored from RIP (%s)~\n",
            sesCtl_p->nTotalNvp, sesCtl_p->ripPath);
    // CLOSE the runtime input parameters text file
    fclose(ripf);
    
    // COLLECT all CORE Nvps first
    l4c_SesCollectCoreNvps(sesCtl_p);
    
    // Next, COLLECT all NVPs into local variables, using expected keyword set
    l4c_SesCollectNvps(sesCtl_p);

    xLogMsg("@A Production task-run tag  (%s)~ \n", sesCtl_p->l4c_Task_Label);
    xLogMsg("@A Production task-run UUID (%s)~ \n", sesCtl_p->l4c_Task_UUID );
    xLogMsg("@A Initial session log      (%s)~ \n", sesCtl_p->logPath);
    
    nsbMgr_p = l4c_GetNsbOverseer(); 
    // xLogMsg("@A Subsystem Release tag    (%s)~\n",nsbGetValueByKey(nsbMgr_p,"L4C_SUBSYSTEM_RELEASE",nvpBuf));
    xLogMsg("@A Subsystem Release tag    (%s)~\n",sesCtl_p->SubsysReleaseTag);

    // ACTIVE_RADAR|ACTIVE_TSURF|GEOS5_TS|PASSIVE_RADIOMETER etc
    xLogMsg("@A FreezeThaw source (%s)~ \n", sesCtl_p->ftSourceStr);
    
    xLogMsg("@A Executable %s AppVers %s ~ \n",argv[0], l4c_GetExecutableVersionId() );
    xLogMsg("@A API: UM Framework version (%s)~ \n", sesCtl_p->frmApiVersion);
    xLogMsg("@A API: PkgMgr       version (%s)~ \n", sesCtl_p->pkgApiVersion);
    xLogMsg("@A API: NSB          version (%s)~ \n", sesCtl_p->nsbApiVersion);

    xLogMsg("@A In(%s) using RIP file assigned GRANULE-LIST (%s)~ \n",
            fxName,sesCtl_p->ripPath);
            
    xLogMsg("@A In(%s) Creating new MET output: [%s]~ \n",
            fxName, sesCtl_p->outputMetPath);

    // Do NOT use generic frm_hdf5_Createfile() since it doesn't
    // pass through the dspace's etc.
    // sesCtl_p->met_file_id  = frm_hdf5_Createfile(sesCtl_p->outputMetPath,fxName);
    sesCtl_p->met_file_id  = createMetHdf5(sesCtl_p);
    
    // Create FILE LEVEL level attributes right away
    l4c_WriteFile_metadata(sesCtl_p, sesCtl_p->met_file_id) ;
    
    // H5Fflush(sesCtl_p->met_file_id,H5F_SCOPE_GLOBAL); // flush...
    H5Fflush(sesCtl_p->met_file_id,H5F_SCOPE_LOCAL); // flush...

    // <JMG> open all (singleton) INPUT HDF5 files once
    // Note: must open ancillary file prior to l4c_Hdf5_CreateOutput() call since
    //       creating output requires access to Ancillary (lat,lon) vars...
    sesCtl_p->anc_file_id  = frm_hdf5_Openfile(      sesCtl_p->inputAncPath ,fxName);
    sesCtl_p->soc_file_id  = frm_hdf5_OpenfileAppend(sesCtl_p->inputSOCPath, fxName);
    
    sesCtl_p->l3smaFile_id = -1;
    if( !frm_ioDoesFileExist(sesCtl_p->inputL3_SM_A_Path))
    {
        xLogMsg("@E In(%s) ERROR: cannot open non-existant L3_SM_A granule [%s]~ \n",
            fxName, sesCtl_p->inputL3_SM_A_Path);        
    } else {
        sesCtl_p->l3smaFile_id = frm_hdf5_Openfile(sesCtl_p->inputL3_SM_A_Path ,fxName);
        xLogMsg("@A In(%s) Found and opened L3_SM_A granule [%s]~ \n",
            fxName, sesCtl_p->inputL3_SM_A_Path);        
        sesCtl_p->nL3_SM_A_Granules = 1 ;
    }

    // allocate the ftStateUm_M03_p array managed in imagePkg_p
    xLogMsg("@A In(%s) FT_STATE_UM_M03 array of %zu pixels allocated here ~ \n",
                     fxName,nElem3km);
                     
    imagePkg_p->ftStateUm_M03_p  =(uint8 *)memCalloc(nElem3km,sizeof(uint8));
    memObjSetName(imagePkg_p->ftStateUm_M03_p, "imagePkg_p->ftStateUm_M03_Img_p");


    // <JMG> 2015-07-31T11:13, to support new GEOS5 derived Freeze Thaw at 3KM,
    // READ EASEGRID 3km mask image
    readPftMask3km(imagePkg_p);

    // READ EASEGRID 9km mask image
    readPftMask9km(imagePkg_p);
    
    // READ and WRITE lat,lon coordinate variables in, 
    //  write to SMAP_L4_C_MET.h5
    readAndWriteLatLon_9kmCoordVars(imagePkg_p);
    
    // Define ALL relevant DATASET properties for OUTPUT L4C HDF5
    // datasets, including: datatype,rank,fillvalue, etc and be sure
    // to do this BEFORE CreateOutput and writes are called.
    //     l4c_PrepOutput_HDF5_Datasets(sesCtl_p) ;
    
    xLogMsg("@A ~New HDF5 %s met_file_id %d   Session-log(%s)\n",
            sesCtl_p->outputMetPath, sesCtl_p->met_file_id,sesCtl_p->logPath);
        
 return EXIT_SUCCESS;
 
} // end::l4c_InitSession()


// -----------------------------------------------------------------------------
// function :l4c_CloseSession(SesCtl *sesCtl_p)
// purpose  : perform all actions required to gracefully close out the session
// Inputs   : SesCtl *sesCtl_p
// returns  : int32
// notes    : n/a
int32 l4c_CloseSession(SesCtl *sesCtl_p)
{
 int32 status = -1 ;
 char fxName[256] = {0};
 ImgPkg *imagePkg_p = l4c_GetImagePackage(); // local pointer to support image package
 
 size_t highestSlot = 0;
 size_t totalBytesTracked = 0;
 size_t hwmMem = 0;
 
    // detailed memory report, before any frees
    strcpy(fxName,__func__);
    
    // memObjReport(stdout, "l4c_MET memory usage -- detail report");
    // memObjReport(sesCtl_p->logf, "l4c_MET memory usage to session log\n");

    // close the HDF5 file and the 9km parts
    status = H5Pclose(sesCtl_p->dcpl9km);
    status = H5Sclose(sesCtl_p->dspace9km);
    // close out the 3km parts
    status = H5Pclose(sesCtl_p->dcpl3km);
    status = H5Sclose(sesCtl_p->dspace3km);

    // close the ancillary assuming no further access needed 
    H5Fclose(sesCtl_p->anc_file_id);
    
    // close out the new HDF5 daily intermediate file
    status = H5Fclose(sesCtl_p->met_file_id);
    // CLOSE the SOC file
    H5Fclose(sesCtl_p->soc_file_id);
    
    // CLOSE the L3_SM_A file, but only if it was defined
    if(sesCtl_p->ftSourceMode == DoUse_ACTIVE_RADAR ||
       sesCtl_p->ftSourceMode == DoUse_ACTIVE_TSURF)
       {
           H5Fclose(sesCtl_p->l3smaFile_id );
       }
    

    // Release 3KM FT_STATE_UM_M03 raster
    xLogMsg("@A In(%s) Releasing FT_STATE_UM_M03 and PFT mask raster memory ~\n",fxName);
    memFree(imagePkg_p->ftStateUm_M03_p);
    // Free PFT global 3km mask image 
    memFree(imagePkg_p->pftImg3km_p);
    // Free PFT global 9km mask image
    memFree(imagePkg_p->pftImg9km_p);


#ifdef NOT_YET
    // free the (3) variable-list string tables, one per input file
    memFreeStringTbl(sesCtl_p->varList_GEOS5_FP, sesCtl_p->nFP_Variables);
    memFreeStringTbl(sesCtl_p->varList_L4_SM   , sesCtl_p->nL4_SM_Variables);
    memFreeStringTbl(sesCtl_p->varList_L3_SM_A , sesCtl_p->nL3_SM_A_Variables);
    memFreeStringTbl(sesCtl_p->inputL4_SM_pathList, sesCtl_p->nL4_SM_Granules); // string table, stores ALL L4_SM paths
    memFreeStringTbl(sesCtl_p->inputGEOS5_FP_pathList,sesCtl_p->nFP_Granules); // string table , stores ALL GEOS5 FP paths
#endif
 
    sesCtl_p->sesEndiSec = time(0L);
    sesCtl_p->sesElapSec = difftime(sesCtl_p->sesEndiSec,sesCtl_p->sesStrtSec);
    
    xLogMsg("\n@A produced new daily intermediate MET output [%s]~\n",
        sesCtl_p->outputMetPath);

    xLogMsg("@A released %d nvp string-table entries~\n",sesCtl_p->nTotalNvp);

 // close out the PkgMgr facility using overseer handle
 pkgMgr_ReleaseMgr( imagePkg_p->PkgMgrOverseer_p);
 
 // close out the NSB facility, using overseer handle
 nsbReleaseInstance(imagePkg_p->NsbMgrOverseer_p);
 
 // memObjReport(stdout, "l4c_MET memory usage -- detail report");
    xLogMsg("@A \n");
    memObjGetMetrics(&highestSlot,&totalBytesTracked,&hwmMem);
    xLogMsg("@A In(%s): session memory resource: high-slot %zu total-bytes %zu high-water-mark %zu (%.1f Gb)~\n",
        fxName, highestSlot, totalBytesTracked, hwmMem, ((float)hwmMem/1024000000.0) );

 xLogMsg("\n@A L4_C preprocessor task [%s] vers[%s] run on [%s]\n Started %s completed %s requiring %.2lf sec~\n",
    (char *)L4C_MET_APP_TAG, l4c_GetExecutableVersionId(), sesCtl_p->thisHost,
     ctime(&sesCtl_p->sesStrtSec),ctime(&sesCtl_p->sesEndiSec),sesCtl_p->sesElapSec);
 
 // Close out the memObj internal journaled memory management subsystem
 memObjGarbageCollect();

 // And finally, close session log file
 xCloseLog(sesCtl_p->logf);

 return EXIT_SUCCESS ;
 
} // end::l4c_CloseSession()

