// ---------------------------------------------------------------------
//   module     : l4c_MET_util.c
//   Purpose    : Sidebar utility for L4C MET, for assessing bit-packed FT_STATE_UM_M03
//                and related datasets
//   Author     : joe glassy
//   Invocation :  <inputHdf5file> <datasetName> <bitStart> <bitEnd> {outfile}
//   Version    : 1.0
//   revised    : 2015-08-10T11:57:00
//   created    :
//   notes      : This stand-alone MET utility 'borrows' all symbols from the l4c_MET.h
//              : header(s) for convenience.
//

/*
 (Macro defined)  FRM_API_RANDOMIZE() srand((unsigned)time(NULL))

// return random probability in range {0.0<=r<=1.0} 
(Macro defined)  FRM_API_RANDOM_PROB() ((double)rand()/(double)RAND_MAX)
(Macro defined)  FRM_API_RANDOM_INT(limit) ((rand()%(abs(limit)+1)))
(Macro defined)  FRM_API_RANDOM_RANGE_1(lo,hi) ((lo)+(rand()%(abs((hi)-(lo))+1)))
// base0 index style range-wise PRNG
(Macro defined)  FRM_API_RANDOM_RANGE_0(lo,hi) (((lo)-1)+(rand()%(abs((hi)-(lo))+1)))
// base1 index style range-wise PRNG
(Macro defined)  FRM_API_RANDOM_RANGE_1(lo,hi) ((lo)+(rand()%(abs((hi)-(lo))+1)))
// floating point round() macro from "The Standard C", P.J.Plaugher 
(Macro defined)  FRM_API_ROUND(x) (((x)<0.50)?ceil(x-0.5):floor(x+0.50))

*/

#include "l4c_MET.h"
// #include "l4c_MET_prot.h"

typedef struct tagMetUtilCtl {
  char  inputUtilPath[FILENAME_MAX] ;
  char  outputUtilPath[FILENAME_MAX];
  char  inputAncPath[FILENAME_MAX] ; 
  char  logPath[FILENAME_MAX];
  char  pft3kmDatasetname[256]; // name of 3KM mask dataset
  char  targDatasetName[256]; // text name of dataset to analyze
  int   targDatasetType;  // TIV style datatype identifier DFNT_UINT8, DFNT_UINT16 etc
  hid_t inputUtilHObjId ; // HDF5 parent object
  hid_t anc_file_id ; // ANC file id
  int   inputResolutionKm; // {1|3|9|36} input datasets resolution in integer KM
  int   bitStart; // base0 bit in (targDatasetName) to start extraction at
  int   bitEnd  ; // base0 bit in (targDatasetName) to end extraction at
  int   bitCount; // total number of bits in (targDatasetName) to extract
  int   doProbSelection; // boolean. Each pixel is qualified stochastically if this is true.
  float selectProb; // probability of pixel-wise selection, defaults to 1.0
  size_t nElemInputPixels; // n. pixels in input image, based on resolution
  void  *datasetBuffer_p;  // ptr to memory for analyzed dataset itself
  uint8 *pftImg1km_p; // ptr to PFT 1KM mask, enabled only if used
  uint8 *pftImg3km_p; // ptr to PFT 3KM mask, enabled only if used
  uint8 *pftImg9km_p; // ptr to PFT 9KM mask, enabled only if used
  FILE  *logf ;
} MetUtilCtl ;


// ----------------------------------------------------------------
// function :int32 metUtilReadPftMask3km(metUtilCtl_p *metUtilCtl_p)
// Purpose: open,read Plant Function Type (PFT) 3km pixel mask {0|1} image
// Inputs : ImgPkg *imagePkg_p -- ptr to container holding PFT and
//          related ancillary image arrays (pointers)
// Returns: status, no. of pixels read (nElem_9km on success)
// Outputs: ptr to same.
// notes  : modified call interface to match lat,lon reader
// revised: 2015-07-31T11:14, added to support new GEOS FP TS derived FreezeThaw,
//          so expects to read 3KM pft from SMAP_L4_C_Ancillary.h5 .
// History: 2013-10-31T12:37 (v4.6.6), modified to read pftImg_p out of PFT_M09_mask" dataset inside
//      the SMAP_L4_C_Ancillary.h5 instead of as a raw binary file.
int metUtilReadPftMask3km(MetUtilCtl *metUtilCtl_p)
{
 char *hdf5Path   = metUtilCtl_p->inputAncPath ;
 char fxName[256] = {0};
  
 int32 status = -1 ;
 hid_t dsTypeId = H5T_NATIVE_UINT8 ;
 
 size_t nRows3km =  4872 ;
 size_t nCols3km = 11568 ;
 size_t ease_serialIdx = 0 ;
 size_t r,c, pftIdx =0 ;
 size_t pftPixAbsFrq[512] = {0} ;
 uint8  pft_pix = 0 ;
 int32   absFreq = 0 ;
 float32 relFreq = 0.0,
         cumFreq = 0.0 ;
  
 
    strcpy(fxName,__func__);
    
    // Create buffer and READ EASEGRID 3km (4872 x 11568) mask image
    fprintf(stdout,"@A In(%s): Allocating 3KM mask of %zu pixels\n",
                fxName,metUtilCtl_p->nElemInputPixels);
    
    metUtilCtl_p->pftImg3km_p = (uint8*)memCalloc(metUtilCtl_p->nElemInputPixels,
                                                  sizeof(uint8));
                                                  
    // ... L4C_PFT_3KM_LC_MASK    "MCD12Q1_M03_lc_mask_uint8"
    // From SMAP_L4_C_Ancillary.h5, READ uint8 EG 9km {0|1} PFT ask image
    status = frm_hdf5_ReadData(metUtilCtl_p->anc_file_id, dsTypeId, 
          metUtilCtl_p->pft3kmDatasetname, (uint8 *)metUtilCtl_p->pftImg3km_p, fxName);

    fprintf(stdout,"@A In(%s): READ dataset (%s) from (%s) file_id %d Status %d~ \n",
                fxName, metUtilCtl_p->pft3kmDatasetname, 
                hdf5Path, metUtilCtl_p->anc_file_id, status);
    
    fprintf(stdout,"@A In(%s) PFT_MASK_3KM Verification~\n",fxName);
    // DIAG only, verify contents of pftImg3km_p[] raster
    for(r=0; r < nRows3km; r++)
    {
        for(c=0; c < nCols3km; c++)
        {
            ease_serialIdx = (r * nCols3km) + c;
            pft_pix = metUtilCtl_p->pftImg3km_p[ease_serialIdx] ;
            // increment bin count for this PFT 3KM value
            pftPixAbsFrq[pft_pix]++ ;
        } // end:: c column loop
    } // end:: r row loop
    
    for(pftIdx=0; pftIdx < 255; pftIdx++)
    {
        if( pftPixAbsFrq[pftIdx] > 0)
        {
           absFreq = (int32)pftPixAbsFrq[pftIdx] ;
           relFreq = ((float32)pftPixAbsFrq[pftIdx] / (float32)N_3KM_PIX ) * 100.0 ;
           cumFreq += relFreq ;
           fprintf(stdout,"@A In(%s) PFT-MASK 3KM Bin %3zu abs-freq %d rel-freq %7.2f cum-freq %7.2f~\n",
           fxName, 
           pftIdx,
           absFreq, relFreq, cumFreq );
        } // end:: if a positive frequency hit
    } // end:: pfxIdx loop through all possible frequency bin indices
    
    
  // return status
  return status;
  
} // end::metUtilReadPftMask3km()


// ---------------------------------------------------------------------
// function  :int initSessionMetUtil()
// inputs    : MetUtilCtl *metUtilCtl_p --ptr to stand-alone utility session controls
// purpose   : Initialize a stand-alone MET utility session
// author    : joe glassy
// revised   : 2015-08-10T12:08:00
// returns   : int, status 0=good
// notes     : n/a
int initSessionMetUtil(MetUtilCtl *metUtilCtl_p)
{
 char fxName[256] = {0};
 int status = 0 ; 
 time_t strtSec = time(0);
 
     strcpy(fxName,__func__);
     
    // set hard defaults    
    FRM_API_RANDOMIZE() ;
    
    // init local buffers for PFT terrestrial land masks
    metUtilCtl_p->pftImg1km_p = NULL;
    metUtilCtl_p->pftImg3km_p = NULL;
    metUtilCtl_p->pftImg9km_p = NULL;
    metUtilCtl_p->selectProb = 1.0 ; // default selection probability is ALL

    strcpy(metUtilCtl_p->logPath,"l4c_MET_UTIL.log"); 
    strcpy(metUtilCtl_p->pft3kmDatasetname, "MCD12Q1_M03_lc_mask_uint8");
    strcpy(metUtilCtl_p->inputAncPath,"/pilv2/SMAP/ANC/SMAP_L4_C_Ancillary.h5");    

    fprintf(stdout,"@A In(%s) %s started %s\n",fxName,
                     "l4c_MET_UTIL.ex", ctime(&strtSec));
     
    // Initialize the memory manager
    memObjInit(stdout);
    memObjAdjustVerbose(0);

    // open log file, currently mandatory for frm_Hdf5_xx I/O and memObj usage
    fprintf(stdout,"@A In(%s) Opening session log %s\n",fxName,
                 metUtilCtl_p->logPath);

    xOpenLogfile(metUtilCtl_p->logPath,NULL);

    // Open the Ancillary file to read PFT mask out of
    metUtilCtl_p->anc_file_id  = frm_hdf5_Openfile(metUtilCtl_p->inputAncPath ,fxName);
    
    fprintf(stdout,"@A In(%s) Opened ancillary %s ID %d\n",
        fxName,metUtilCtl_p->inputAncPath,(int)metUtilCtl_p->anc_file_id);
          
     if(!frm_ioDoesFileExist(metUtilCtl_p->inputUtilPath))
     {
         fprintf(stderr,"@F In(%s) Required input HDF5 file(%s) does not exist where indicated!\n",
          fxName, metUtilCtl_p->inputUtilPath);
         status = -1;
     } else {
       // Open the prmary input HDF5 data file, store its file_id
       metUtilCtl_p->inputUtilHObjId  = frm_hdf5_Openfile(metUtilCtl_p->inputUtilPath,fxName);
       fprintf(stdout,"@A In(%s) Opened primary input granule %s ID %d\n",
        fxName,metUtilCtl_p->inputUtilPath,(int)metUtilCtl_p->inputUtilHObjId);
     }

     // hard wire for testing only. Normally parse out of metadata or file
     metUtilCtl_p->inputResolutionKm = 3 ;
     // hard wired as 3KM for testing
     switch(metUtilCtl_p->inputResolutionKm)
     {
         case  1: metUtilCtl_p->nElemInputPixels = 14616 * 34704 ; break;
         case  3: metUtilCtl_p->nElemInputPixels = 4872  * 11568 ; break;
         case  9: metUtilCtl_p->nElemInputPixels = 1624  * 3856  ; break;
         case 36: break;
     } // end:: switch 
     
     fprintf(stdout,"@A In(%s) Resolution KM = %d N Pixels %zu\n",
      fxName, metUtilCtl_p->inputResolutionKm, metUtilCtl_p->nElemInputPixels);
     
     // ALLOCATE and then READ the 3KM terrestrial mask
     fprintf(stdout,"@A In(%s) reading 3KM terrestrial mask %s\n",fxName,
       metUtilCtl_p->pft3kmDatasetname);
       
     metUtilReadPftMask3km(metUtilCtl_p);
     
 return 0 ;    
} // end::initSessionMetUtil()

// ---------------------------------------------------------------------
// function  :int metUtilDriver_Assessment()
// inputs    : MetUtilCtl *metUtilCtl_p --ptr to stand-alone utility session controls
// purpose   : Perform a pixel level retrieval on the requested dataset,
//             extracting a bit-range if specified, and then calculating
//             a set of basic assessment statistics {min,max,mean} along
//             with a frequency table (abs,rel,cum) on each qualifying pixel
//             value, if the datatype is integer etc.
// author    : joe glassy
// revised   : 2015-08-10T16:03:00-07:00
// returns   : int, status 0=good
// notes     : This initial implementation favors integer level pixel datatypes,
//             though it may be extended to support real number pixel datatypes
//             later.
// notes     : The critiera for selecting a given pixel involves a multiple
//             level set of tests:
//             a) the pixel must spatially qualify as a terrestrial pixel
//             b) if relevant, the pixel must be "in-swath"
//             c) the pixel cannot be a _Fillvalue or MISSING_VALUE
//             d) If probability selection mode is enabled, the fixed 
//                user-supplied probability-of-selection {0.0<=u<=1.0} 
//                qualify relative to the pixel level PRNG derived 
//                probability {0.0<=p<=1.0}
//             e) TBD -- other tests
//
int metUtilDriver_Assessment(MetUtilCtl *metUtilCtl_p)
{
 char fxName[256]= {0};
 int status = -1;
 int inputResKm = 3 ; // hard wire as 3KM for early testing
 // TBD generalize later...
 size_t nElem1km = 14616 * 34704 ;
 size_t nElem3km =  4872 * 11568 ;
 size_t nElem9km =  1624 *  3856 ;
 size_t pixIdx =0;
 
 hid_t dsTypeId = H5T_NATIVE_UINT8 ;
 
 
 uint8   pixPft   = 0 ;
 uint8   pixUint8 = 0 ;
 uint8   pft_pix  = 0 ;
 uint8  *data_p   = NULL ; // ptr to primary DATASET to analyze
 
 int     pixBitsReturned  = 0 ;
 
 int     doAnalyze = 0 ; // boolean, for pixel-wise inclusion in analysis
 int     firstObs  = 1 ;
 
 size_t  pftPixAbsFrq[512] = {0} ;
 size_t  nObsCount = 0 ;
 size_t  nQualified = 0 ;
 int32   absFreq = 0 ;
 
 float32 prob = 0.0 ;
 float32 thisValu=  0.0,
         meanValu = 0.0,
         sumValu = 0.0,
         sumSq   = 0.0,
         minValu = 0.0,
         maxValu = 0.0,
         relFreq = 0.0,
         cumFreq = 0.0 ;
  
    strcpy(fxName,__func__);

    fprintf(stdout,"@A In(%s) Assessing %s now\n",
                  fxName, metUtilCtl_p->targDatasetName);
 
    // Allocate memory for the dataset, using worst-case sizing
    metUtilCtl_p->datasetBuffer_p = (uint8 *)memCalloc(metUtilCtl_p->nElemInputPixels,
                                         sizeof(uint8));
    
    // typed ptr to primary dataset
    data_p = (uint8 *)metUtilCtl_p->datasetBuffer_p ;
    
    strcpy(metUtilCtl_p->targDatasetName,"FT_STATE_UM_M03");
    
    fprintf(stdout,"@A In(%s) Allocated buffer for %zu data 3KM pixels in %s\n",
      fxName, metUtilCtl_p->nElemInputPixels,metUtilCtl_p->targDatasetName);
    
    // Read the requested dataset. Note this opens, performs read, and
    // then closes the dset_id
    status = frm_hdf5_ReadData(metUtilCtl_p->inputUtilHObjId, dsTypeId, 
                               metUtilCtl_p->targDatasetName,
                               metUtilCtl_p->datasetBuffer_p, fxName);

    fprintf(stdout,"@A In(%s) read target %zu pixels from dataset %s\n",
      fxName, metUtilCtl_p->nElemInputPixels,metUtilCtl_p->targDatasetName);
                               
    if(metUtilCtl_p->inputResolutionKm == inputResKm)
    {
      fprintf(stdout,"@A In(%s) Now scanning bits %d to %d of resolution %d KM\n",
         fxName, metUtilCtl_p->bitStart, metUtilCtl_p->bitEnd, metUtilCtl_p->inputResolutionKm);
        
        for(pixIdx=0; pixIdx < metUtilCtl_p->nElemInputPixels; pixIdx++)
        {
            // retrieve the terrestrial mask value for this pixel
            pixPft = metUtilCtl_p->pftImg3km_p[pixIdx] ;
            
            // Only if a terrestrial pixel, then analyze
            if(pixPft >= 1 && pixPft <= 8)
            {
              // Fetch next pixel from requested HDF5 dataset
              pixUint8 = data_p[pixIdx];
              // Retrieve requested bit range from start to end inclusive,
              // normalized to a 32 bit signed integer to allow for a wider
              // range of input dataset datatypes
              pixBitsReturned = (int)frm_utilGetBits(metUtilCtl_p->bitStart,
                                        metUtilCtl_p->bitEnd, pixUint8); 

              doAnalyze = 1 ; // boolean screen, at pixel level
              prob    = 1.0 ; // set default selection filter
              // Compare PRNG generated prob against user threshold, where
              // the user threshold (selectProb) MUST be < 1.0 in order to
              // enable this style of pixel level stochastic screening
              if(metUtilCtl_p->selectProb < 1.0)
              {
                // Draw a random probability of selection (0<=p<=1.0)
                prob = FRM_API_RANDOM_PROB() ;
                doAnalyze = (prob <= metUtilCtl_p->selectProb)? 1 : 0 ;
              }

              // Analyze, assess this pixels retrieved bits
              if(doAnalyze)
              {
                  nQualified++ ; // count...
                  thisValu = (float32)pixBitsReturned;
                  sumValu  += thisValu ;
                  sumSq    += (thisValu * thisValu);
                  nObsCount++ ;
                  // Evaluate for dataset wise minimum,maximum, taking care
                  // to filter out MISSING_VALUE and/or _Fillvalue's!
                  if(firstObs) 
                  {
                      minValu = maxValu = pixBitsReturned;
                      firstObs = 0 ;
                  } else {
                      if(pixBitsReturned < minValu) minValu = pixBitsReturned;
                      if(pixBitsReturned > maxValu) maxValu = pixBitsReturned;
                  }
              
              } // end:: if block where doAnalyze was TRUE 

              
            } // end:: pixPft if block
        } // end:: pixIdx loop
        
        // calculate univariate statistics, and any other central tendencies
        meanValu = sumValu / (float32)nObsCount;
        
        status = 0 ;
    } else {
        fprintf(stdout,"@E In(%s) Resolution not yet supported!\n",fxName);
    }

  fprintf(stdout,"@A In(%s) N %zu pixels qualified\n",fxName,nQualified);
  
  fprintf(stdout,"@A In(%s) N Obs %zu Mean %.4f Sum %.1f\n"  ,fxName,nObsCount,meanValu,sumValu);
  fprintf(stdout,"@A In(%s) Min %.4f Max %.4f\n",fxName,minValu,maxValu);
  // TBD: std dev, std error, etc

 return 0 ;    
} // end::metUtilDriver_Assessment()


// ---------------------------------------------------------------------
// function  :int closeSessionMetUtil()
// inputs    : MetUtilCtl *metUtilCtl_p --ptr to stand-alone utility
//             session controls
// purpose   : close out a stand-alone MET utility session
// author    : joe glassy
// revised   : 2015-08-10T12:08:00
// returns   : int, status 0=good
// notes     : n/a
int closeSessionMetUtil(MetUtilCtl *metUtilCtl_p)
{
 char fxName[256] = {0};
 int status = 0 ;  
 time_t endiSec = time(0);
 
  strcpy(fxName,__func__);
  
 
 fprintf(stdout,"@A In(%s) Closing HDF5 ancillary and data files\n",fxName);
 
 H5Fclose(metUtilCtl_p->anc_file_id);
 
 H5Fclose(metUtilCtl_p->inputUtilHObjId);
 
 fprintf(stdout,
  "@A In(%s) releasing memory for PFT mask and primary data\n",fxName);

 // Release memory for 3KM PFT mask
 memFree(metUtilCtl_p->pftImg3km_p); 
 // Release memory for the analyzed dataset
 memFree(metUtilCtl_p->datasetBuffer_p);
    
 // Close out the memObj internal journaled memory management subsystem
 memObjGarbageCollect();
 
 // Close session log file
 // xLogClose(metUtilCtl_p->logf);    
 xLogClose(NULL);

 fprintf(stdout,"@A In(%s) %s ended %s\n",fxName,
        "l4c_MET_UTIL.ex", ctime(&endiSec));
  
 return status ; 
 
} // end::closeSessionMetUtil()



// ---------------------------------------------------------------------
int main(int argc,char *argv[])
{
 MetUtilCtl metUtilCtl ;
   
   
   strcpy(metUtilCtl.outputUtilPath,"NO-OUTPUT-SPECIFIED-YET");
   // set default bit-range to extract
   metUtilCtl.bitStart = 0 ;
   metUtilCtl.bitEnd   = 0 ;
   
   if( argc >= 2)
   {
       strcpy(metUtilCtl.inputUtilPath,argv[1]);
       if(argc >= 3)
         strcpy(metUtilCtl.targDatasetName,argv[2]);
       if(argc >= 5)
         metUtilCtl.bitStart = (int)atoi(argv[3]);
         metUtilCtl.bitEnd   = (int)atoi(argv[4]);
       if(argc >= 6)
         strcpy(metUtilCtl.outputUtilPath,argv[5]);
   } else {
       fprintf(stderr,"Invalid invocation: <infile> <datasetName> <bitStart><bitEnd> {outfile}\n");
   }
   // initialize stand-alone MET utility session  
   initSessionMetUtil(&metUtilCtl);
   
   // call assessment logic
   metUtilDriver_Assessment(&metUtilCtl);

   // close out MET utility session
   closeSessionMetUtil(&metUtilCtl);

 return(0);
} // end::main()
