// ---------------------------------------------------------------------
// module       : l4c_MODIS_FparQaBin.c , was orig fpar_qa_bin_util.c
// purpose      : Implement utility functions for modifying MODIS fpar qc bit fields
//                e.g MOD15A2 FPAR QA/QC and similar style bitfields
// project      : SMAP L4C science development.
// dependencies : HDF4, HDF5, FRM API libfrm.a
// revised      : 2021-07-18, EVAL/DIAG coding in anticipation of VIIRs integration
// revised      : 2017-02-09T16:13, adapted l4c_MODIS_FparQaBin.[ch]
//                from L4CMDL code set, now required for new MOD15A2H 500m processing.
// revision     : Initial reformating -- laj 09/10/13
// author       : Lucas A. Jones; lucas@ntsg.umt.edu
// maint. author: Joe Glassy
// notes        : 2017-02-09, Added l4c_MODIS_FparQaBin.[ch] for new MOD15A2H processing as
//                these were previously never part of the L4CMOD code set.


#include "l4c_MODIS.h"
// <JMG> 2017-02-09T16:22, new header, l4c_MODIS_FparQaBin.h is included in l4c_MODIS_prot.h now
#include "l4c_MODIS_prot.h"


// ---------------------------------------------------------------------
// fpar_qa_bin_reclas()
// Purpose: High-level function to reclassify MOD15A2 FPAR QA/QC bit-field into four, ranked integer classes (0-3).
// Inputs: MOD15A2 FPAR QA/QC one byte bit-field.
// Outputs: Ranked integer FPAR quality class.
// Revised: date undocumented
// revised: 2021-07-23, logic review only.
// Notes  : Calls low-level function 'set_qa_state()'. The ranked integer class primarily depends on if clouds
//        were present, and if clouds were not present, which algorithm was used for FPAR retrieval.
uint8 fpar_qa_bin_reclas(uint8 fpar_qa)
{
  uint8     rank;
  QcBitAdt   QCBIT;

  // Extract bits from QC field
  QCBIT.qc = fpar_qa;

  // break into the various QC fields
  bitword_parse(&QCBIT);

  // Re-classify qc conditions
  set_qa_state(QCBIT,&rank);

  return(rank);

} // end :: fpar_qa_bin_reclas()


// --------------------------------------------------------------------
// getBitsFromUint32()
// Adapted from: MODIS code.
//
// Purpose: this is an example of bit retrieval logic
//   using uint32 arguments. To work with uint8 (pixel) variables,
//   first load the uint8 value into a uint32 variable and pass in the
//   uint32 variable.

/*----------------------------------------------------------------------
!M  module  : l4c_FparQaBinUtil.c
!F  function: getBitsFromUint32(startBits,endBits,dataPixel)
!R  returns : uint32 -- up to 32 bits extracted, as an unsigned 32 bit int
!I  inputs  :int32 strt_at --starting base0 bit index, like 0.
             int32 end_at  --ending base0 bit index, like 1
             uint32 dta    -- a single raw pixel, expressed as a uint32 integer
!P  purpose : getBitsFromUint32() uses standard K&R ANSI C bit
    extraction logic. The function extracts an embedded
    bitfield extending from base-0 bit position "strt_at"
    to "end_at" inclusive, from the variable "dta" and returns
    this result as a right-adjusted decimal value. On Intel architectures
    we assume a little-endian byte-ordering.
    The standard word layout we assume is depicted below, where
    MSB is 'most-significant-bit' and LSB is 'least-significant-bit)
    regardless of whether the input word "dta" is a uint8, uint16 or
    uint32:

         MSB end                                 LSB end
         15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
          0  0  0  0  0  0  0  0  0  0  0  0  1  1  0  1

        Example: we begin with a (uint32) word whose
        value in binary is '1101'. Assuming we wish to extract
        bits {0,1}, e.g. the pattern '01', valued at decimal =1, we call
         modland_qc = getBitsFromUint32(0,1,thisPixelData)
!X--------------------------------------------------------------------*/
uint32 getBitsFromUint32(const int32 strt_at,const int32 end_at,uint32 dta)
{
 int32 p = end_at,             // anchor the start here
       n = (end_at-strt_at)+1; // sum of bits (width) to retrieve

 return (dta >> (p+1-n)) & ~(~0 << n);

} // getBitsFromUint32()


// ---------------------------------------------------------------------
// function : set_qa_state()
// purpose  : Low-level function to reclassiy MOD15A2 FPAR QA/QC
//            bit-field into four, ranked integer classes,valued at (0-3).
// Inputs   : QcBitAdt qcBitObj,
//            uint8 fpar_rank, a (5-level) rank {0,1,2,3,4}
// Outputs  : via call argument pointer, Ranked integer FPAR quality class.
// Notes    : The ranked integer class primarily depends on
//            if clouds were present, and if clouds were not present,
//            which algorithm was used for FPAR retrieval.
//
void set_qa_state(QcBitAdt qcBitObj, uint8 *fpar_rank)
{

  // Worst conditions (QA=4): gapfill with climatology
  if (qcBitObj.fillval == 1 || qcBitObj.cloud != 0)
  {

    *fpar_rank = 4;

  } else if (qcBitObj.scf == 2 || qcBitObj.scf == 3) {
    // Poor conditions (QA=3): fpar backup alg.
    *fpar_rank = 3;

  } else if (qcBitObj.scf == 1)  {
    // Fair conditions (QA=2): fpar main RT with saturation
    *fpar_rank = 2;

  } else if (qcBitObj.scf == 0) {
    // Good conditions (QA=1): fpar main RT with saturation
    *fpar_rank = 1;
  }  else {
    *fpar_rank = 0;
  }

} // end :: set_qa_state()

// ---------------------------------------------------------------------
// function : bitword_fpar_qc_parse()
// purpose  : replaces inspecific Specific to legacy fpar_qc uint8 bit definitions, this
//            low-level (single pixel) function splits bit string into sub-units
//            ('bitwords') aligned with FPAR QC fields
// inputs   : QcBitAdt *qcBitObj
// history  : new FparQcBitColl6Adt used here, supporting Collection 6 MODIS or VIIRs
// history  : Replaces legacy qcBitObj which defines bit wise
//              subfields only specific to legacy to fpar_qc variable in
//              collection 5 and earlier definitions.
// revised  : 2021-07-24T12:28, renamed as specific to
//            Collection 5 and earliler legacy fpar_qc ONLY
// Outputs  : populates one pixel's FPAR (legacy fpar_qc) bitword structure.
// notes    : Bitword structure fields contain descriptive names for
//            each QC situation.
// notes    : MOD15A2 FPAR QA/QC one byte bit-field.
// notes    : other L4_C functions use l4c_GetBits(),l4c_AssignBits()
//            l4c_SetBits() etc.
void bitword_fpar_qc_parse(FparQcBitAdt *fparQcBitObj_p)
{
 // algor_path, 3-bits used, LSB bits 0-2
 // 000=0 Main (RT) algorithm used, best result possible (no saturation)
 //	001=1 Main (RT) algorithm used, saturation occured. Good, very usable.
 //	010=2 Main algorithm failed due to bad geometry, empirical algorithm used
 //	011=3 Main algorithm faild due to problems other than geometry, empirical algorithm used
 //	100=4 Pixel not produced at all, value coudn't be retrieved (possible reasons: bad L1B data, unusable VNP09GA data, water pixel)
 fparQcBitObj_p->scf  = (uint8)getBitsFromUint32(0,2,fparQcBitObj_p->qc);
 // dead detector status, 1 bit used, LSB bit 3
 //	0 = Both red and NIR detectors are fine
 //	1 = At least one band has dead detector
 fparQcBitObj_p->dead_detect = (uint8)getBitsFromUint32(3,3,fparQcBitObj_p->qc);
 // biome type, 4 bits used, LSB bits 4-7
 fparQcBitObj_p->biome_type  = (uint8)getBitsFromUint32(4,7,fparQcBitObj_p->qc);

} // end :: bitword_fpar_qc_parse()

// ---------------------------------------------------------------------
// function : bitword_fpar_qc_parse()
// purpose  : replaces inspecific Specific to legacy fpar_qc uint8 bit definitions, this
//            low-level (single pixel) function splits bit string into sub-units
//            ('bitwords') aligned with FPAR QC fields
// inputs   : QcBitAdt *qcBitObj
// history  : new FparQcBitColl6Adt used here, supporting Collection 6 MODIS or VIIRs
// history  : Replaces legacy qcBitObj which defines bit wise
//              subfields only specific to legacy to fpar_qc variable in
//              collection 5 and earlier definitions.
// revised  : 2021-07-24T12:28, renamed as specific to
//            Collection 5 and earliler legacy fpar_qc ONLY
// Outputs  : populates one pixel, Collection 6 fpar_extra_qc bitword structure
// notes    : If there is no good estimation in this composite, whole byte set to 255!
// notes    : ASSERT: no use is made of LSB bit {7-7} so its value is undefined.
void bitword_fpar_extra_qc_parse(FparExtraQcBitAdt *fparExtraQcBitObj_p)
{
 // LSB bits 0-1, Cloud detection and confidence, 2 bits used,
 //  00=(0)Confident pixel is clear, 01=(1)Probably clear, 10=(2)Probably Cloudy, 11=(3)Confident its cloudy on pixel
 fparExtraQcBitObj_p->cloud_detect_conf = (uint8)getBitsFromUint32(0,1,fparExtraQcBitObj_p->qc);
 // LSB bits 2-2, cloud shadow {0|1},0=NO cloud shadow,1=YES Cloud shadow detected for pixel
 fparExtraQcBitObj_p->cloud_shadow      = (uint8)getBitsFromUint32(2,2,fparExtraQcBitObj_p->qc);
 // LSB bits 3,3, thin cirrus, 0=NO thin cirrus detected, 1=YES thin cirrus detected for pixel
 fparExtraQcBitObj_p->thin_cirrus       = (uint8)getBitsFromUint32(3,3,fparExtraQcBitObj_p->qc);
 // LSB bits 4,5, aerosol quantity code 00=(0)Climatology, 01=(0)Low, 10=(1)Average, 11=(3)High
 fparExtraQcBitObj_p->aerosol_quant     = (uint8)getBitsFromUint32(4,5,fparExtraQcBitObj_p->qc);
 // LSB bits 6,6, Snow-Ice present, {0|1}, 0=NO Snow or ice detected, 1=YES Snow or ice present on pixel.
 fparExtraQcBitObj_p->snow_ice          = (uint8)getBitsFromUint32(6,6,fparExtraQcBitObj_p->qc);

} // end :: bitword_fpar_extra_qc_parse()

// ---------------------------------------------------------------------
// function : bitword_parse()
// purpose  : LEGACY COLLECTION 5 ONLY, Low-level function splits
//            bit string into sub-units
//            ('bitwords') aligned with FPAR QC fields
// inputs   : QcBitAdt *qcBitObj
// revised  : 2021-07-23T18:24, review only.
// notes    : This function SHOULD BE RENAMED to reflect that it ONLY
//            applies to legacy fpar_qc bit mapping.
// Outputs  : populated FPAR bitword structure.
// notes    : Bitword structure fields contain descriptive names for
//            each QC situation.
// notes    : MOD15A2 FPAR QA/QC one byte bit-field.
// notes    : other L4_C functions use l4c_GetBits(),l4c_AssignBits()
//            l4c_SetBits() etc.
void bitword_parse(QcBitAdt *qcBitObj)
{

  // MODLAND_QC field
  qcBitObj->modland = (uint8)getBitsFromUint32(0,0,qcBitObj->qc); // <laj 06/02/2014> updated fpar qa bit read

  // Sensor (Terra/Aqua) field
  qcBitObj->sensor = (uint8)getBitsFromUint32(1,1,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read

  // DeadDetector field
  qcBitObj->detector = (uint8)getBitsFromUint32(2,2,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read

  // CloudState field
  qcBitObj->cloud = (uint8)getBitsFromUint32(3,4,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read

  // SCF_QC field
  qcBitObj->scf = (uint8)getBitsFromUint32(5,6,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read

  // Fill values (from SFC_QC leftmost bit)
  qcBitObj->fillval = (uint8)getBitsFromUint32(7,7,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read

} // end :: bitword_parse()


/*Retrieves 'n' bits from the bit string buffer 'srcbuff' with location
 *determined by 'offset' and puts them in string 'desbuff'  */

// -------------------------------------------------------------------------------------------------------------
// getbits()
// Purpose: Retrieves bits from a bit string buffer.
// Notes:
//
// Inputs: Number of bits (n) to retrieve from the bit string buffer ('srcbuff') with array location ('offset').
// Outputs: Retrieved bits returned in string ('desbuff')
void getbits(char srcbuff[], char desbuff[], int offset, int n)
{

  strcpy(desbuff, "0\0");
  strncpy(desbuff,srcbuff+offset,n);
  desbuff[n+1] = '\0';

} // end :: getbits()

/*Extracts a string of bits from a single byte integer*/

// -------------------------------------------------------------------------------------------------------------
// printbits()
// Author: Lucas Jones (presumed)
// Purpose: Extracts a string of bits from a single byte integer
void printbit(uint8 a, char buff[])
{
  int   i;
  int   n = CHAR_BIT;
  int   mask = 1 << (n-1);
  char  cbuff[MXSTRLEN] = {0};
  strcpy(buff, "\0");

  for (i = 1; i<=n; ++i)
  {
    sprintf(cbuff,"%c",((a & mask) == 0) ? '0' : '1');
    strcat(buff,cbuff);
    a <<= 1;
  }

  strcat(buff,"\0");
} // end :: printbits()

// -------------------------------------------------------------------------------------------------------------
// printf_qcBitObjs()
// Purpose: Prints QCBIT structure fields.
// Notes:
// Inputs: QCBIT structure fields.
// Outputs:
// Revised: 2017.
void printf_qcBitObjs(QcBitAdt qcBitObj)
{

  /*print complete binary representation to string
   *(for viewing) */
  printbit(qcBitObj.qc, qcBitObj.bits);

  printf("\n");
  printf("Entire QC: %s\n", qcBitObj.bits);
  printf("Modland:   %d\n", qcBitObj.modland);
  printf("Sensor:    %d\n", qcBitObj.sensor);
  printf("Detector:  %d\n", qcBitObj.detector);
  printf("Cloud:     %d\n", qcBitObj.cloud);
  printf("Scf:       %d\n", qcBitObj.scf);
  printf("Fillval:   %d\n", qcBitObj.fillval);
  printf("\n");
} // end :: printf_qcBitObjs()
