// --------------------------------------------------------------
// module       : fpar_qa_bin_util.h for tcf_spin v3.2, with ADT objects as documented.
// purpose      : Data structures and function prototypes for 'fpar_qa_bin_util.c' utilties module.
// project      : SMAP L4C science development.
// dependencies : HDF4, and UM 'Framework API' e.g. FRM_API  libfrm.a
// revision     : 2021-08-09, mods per Arthur meeting, remove Coll6 qualifier from new VIIRS vs. MODIS QC adts
// revision     : 2021-07-30, additions for Collection 6 MODIS and VIIRs support
// revision     : 2021-07-24T12:26, modify and add for Collection 6 MODIS and VIIRs support
// revision     : 2014-02-15T11:55:00-07:00
// author       : joe glassy , Lupine Logic Inc.
// history      : <JMG> 2021-07-24T12:35, Add modernized QA object for Collection 6 VIIRs and MODIS

/*
 module       : fpar_qa_bin_util.h
 purpose      : Data structures and function prototypes for 'fpar_qa_bin_util.c' utilties module.
 project      : SMAP L4C science development.
 dependencies :
 revision     : v3.1, Initial reformatting -- laj 09/10/13
 original author: Lucas A. Jones; lucas@ntsg.umt.edu
 notes        : goal was to support only Collection 5 and earlier datasets.

*/

#ifndef QA_BIN_UTIL_DEF
#define QA_BIN_UTIL_DEF 1

/* FparLai_QC 3 discrete (algor_path,dead_detect,biome_type) bitfields IN one 8 BIT WORD
0-2	Algorithm path
	000=0 Main (RT) algorithm used, best result possible (no saturation)
	001=1 Main (RT) algorithm used, saturation occured. Good, very usable.
	010=2 Main algorithm failed due to bad geometry, empirical algorithm used
	011=3 Main algorithm faild due to problems other than geometry, empirical algorithm used
	100=4 Pixel not produced at all, value coudn't be retrieved (possible reasons: bad L1B data, unusable VNP09GA data, water pixel)
3-3	Dead detector
	0 = Both red and NIR detectors are fine
	1 = At least one band has dead detector
4-7	Biome type
	0000=0  Water
	0001=1  Grasses/cereal crops
	0010=2  Shrubs
	0011=3  Broadleaf crops
	0100=4  Savanna
	0101=5  Evergreen broadleaf forest
	0110=6  Deciduous broadleaf forest
	0111=7  Evergreen needleleaf forest
	1000=8  Deciduous needleleaf forest
	1001=9  Unvegetated
	1010=10 Urban
	1011=11 Unclassified
	1100=12 Filled Value
*/

// Fpar_qc (byte) bitfield definitions. Deprecated Coll6 qualifier 8/9/2021.
typedef struct tagFparQcBitAdt {
  char          bits[9]; // 9 bytes allows for a null term, for readability
  uint8         qc;
  unsigned char scf        : 3 ; // 3-bits,{000..100}
  unsigned char dead_detect: 1 ; // 1-bit, {0|1},0=Both RED,NIR detectors FINE,1=At least one detector is DEAD
  unsigned char biome_type : 4 ; // 4-bits,{0000-1100},upper range to 1111 are NOT defined.
} FparQcBitAdt;


//  Fpar_extra_qc (byte) bitfield definitions, applicable to VIIRS
/* FparExtra_QC 7 BITFIELDS IN 8 BITWORD (Extracted from VNP09GA)
   ASSERT: no use is made of LSB bit 7-7 so its value is undefined.
0-1 Cloud detection and confience
	00 = 0 Confident clear
	01 = 1 Probably clear
	10 = 2 Probably cloudy
	11 = 3 Confident cloudy
2-2 Cloud shadow
	0 = 0 No cloud shadow
	1 = 1 Shadow
3-3 Thin cirrus flag
	0 = 0 No
	1 = 1 Yes
4-5 Aerosol quantity
	00 = 0 Climatology
	01 = 1 Low
	10 = 2 Average
	11 = 3 High
6-6 Snow/Ice
	0 = 0 No
	1 = 1 Yes
If there is no good estimation in this composite, this value is set to 255!
*/
// Deprecated Coll6 qualifier 8/9/2021.
typedef struct tagFparExtraQcBitAdt {
  char          bits[9]; // 9 bytes allows for a null term, for readability
  uint8         qc; // whole 8-bits <11111111> = 255, indicate NO GOOD EST available
  unsigned char cloud_detect_conf: 2; // 2-bits,{00,01,10,11} 4 discrete values
  unsigned char cloud_shadow     : 1; // 1-bits,{0|1}, 2 discrete values
  unsigned char thin_cirrus      : 1; // 1-bit, {0|1}, 2 discrete values 0=No thin cirrus,1=Yes thin cirrus detected
  unsigned char aerosol_quant    : 2; // 2-bits,{00,01,10,11}, 4 discrete values
  unsigned char snow_ice         : 1; // 1-bit, {0|1}, 2 discrete values
} FparExtraQcBitAdt;

// <JMG> 2021-07-24T12:35, LEGACY ONLY (e.g. Collection 5 and earlier)
// note: this is an explicit bitfield mapping, within a typedef
typedef struct tagQcBitAdt {
  // holds full uint8 qc field
  char             bits[9];
  uint8            qc;
  // holds bit wise subfields, ('bitword') assumed Intel LSB ordering,
  // read  *from right to left in the bitbuffer
  unsigned char  modland  : 1;  // MODLAND_QC field
  unsigned char  sensor   : 1;  // Sensor (Terra/Aqua) field
  unsigned char  detector : 1;  // DeadDetector field
  unsigned char  cloud    : 2;  // CloudState field
  unsigned char  scf      : 2;  // SCF_QC field (minus left-most bit)
  unsigned char  fillval  : 1;  // Fill values, left-most bit of SFC_QC (my own designation)
} QcBitAdt;


// FUNCTION PROTOTYPES specific to QA/QC bit handling
uint8   fpar_qa_bin_reclas(uint8 fpar_qa);
void    set_qa_state(QcBitAdt qcBitObj, uint8 *fpar_rank);

// <JMG> LEGACY QA/QC handler left intact and unmodified
void    bitword_parse(QcBitAdt *qcBitObj_p);

// <JMG> 2021-07-24T13:39, new Collection 6 specific bit parsing fx added
// <KAE> 2021-08-23 Clarified that this is not specific to Collection 6 but to VIIRS vs. MODIS
void    bitword_fpar_qc_parse(FparQcBitAdt *fparQcBitObj_p);
void    bitword_fpar_extra_qc_parse(FparExtraQcBitAdt *fparExtraQcBitObj_p) ;

void    printbit(uint8 a, char buff[]);
void    getbits(char srcbuff[], char desbuff[], int offset, int n);
void    printf_qcbits(QcBitAdt qcBitObj);

uint32  getBitsFromUint32(const int32 strt_at,const int32 end_at,uint32 dta);

#endif
