// --------------------------------------------------------------
// function: l4c_MODIS_bplut.h
// purpose : Cloned from L4CMDL, implements function prototypes for l4c_MODIS_bplut.c
// author  : joe glassy
// revised : v5.0.1-2014-10-14T14:25:00, cloned, built, but is NOT yet used in l4c_MODIS
//           This is included in anticipation of future use
// revised : v5.0.1-2014-08-06T17:00:00-07:00, <LAJ>, Add FT-related BPLUT columns.
// revised : 4.2.7-2014-02-22T12:26, split out scan and allocation of BPLUT
//           to better facilitate multiple uses (or READS) of BPLUT within
//           a complex application, without the risk of redundant allocations
//
// revised : 4.2.6-2014-02-19T23:44:00-07:00 (RunParamAdt added, and fix-ups)
// history : adapted from tcf_spin v3.2
// notes   : Version enforced on fixed UUID 6640b8c0-9a01-11e3-a29d-000c29ddf1f0
// notes   : This module implements the text I/O method on SMAP_BPLUT_2014-02-11.csv
//           Later in production the BPLUT will be re-deployed in the SMAP_L4_C_Ancillary.h5
//
// notes   : NPFT is now 8.  The BPLUT ADT itself is defined in tcf_struct.h
//

#ifndef L4C_BPLUT_DEF
#define L4C_BPLUT_DEF 1

#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"

// include BPLUT ADTs and protos 

// from: tcf_const.h in L4CMDL, required directives L4C_BPLUT function
#define MAX_LINE_BUFFER_LEN       512
#define STRING_MATCHED            0
#define BPLUT_N_EXPECTED_RECORDS  8
#define BPLUT_N_EXPECTED_FIELDS   37    // LAJADD


typedef struct tagBPLUT
{
  // key for lookups...
  char lutKey[32] ;  // landcover or PFT abbreviated unique labels, e.g. EBF, ENF,...
  int  lutIntegerIndex ; // added for base0 index retrievals

  /*  parameters for NDVI-FPAR conversion  */
  char     useNDVItoFPAR[32];  // TRUE: use linear model between NDVI and FPAR;  Otherwise: use FPAR with standard gain and offset
  float64  NDVItoFPAR_scale;   // linear model scale coefficient used to convert NDVI to FPAR
  float64  NDVItoFPAR_offset;  // linear model offset coefficient used to convert NDVI to FPAR

  /* parameters for MODIS GPP/NPP algorithm */
  float64  LUEmax;                  /* [g C m-2 d-1 MJ-1] Maximum light use efficiency for GPP */

  /* meteorological constraint parameters */
  float64  Tmin_min;                // [K] Minimum value where minimum daily air temperature constrains GPP
  float64  Tmin_max;                // [K] Maximum value where minimum daily air temperature constrains GPP
  float64  VPD_min;                 // [Pa] Minimum value where average daily vapor pressure deficit constrains GPP
  float64  VPD_max;                 // [Pa] Maximum value where average daily vapor pressure deficit constrains GPP
  float64  SMrz_min;                // [% Sat.] Minimum value where root zone soil moisture constrains GPP
  float64  SMrz_max;                // [% Sat.] Maximum value where root zone soil moisture constrains GPP

  // <LAJ> 08/06/2014, Addded FT_min/max fields for incorporating FT into model forward step and QA calculations
  float64  FT_min;                  // [dim.] Minimum value where freeze-thaw constrains GPP (Value if Frozen, FT==1)
  float64  FT_max;                  // [dim.] Maximum value where freeze-thaw constrains GPP (Value if non-frozen, FT==0)

  float64  SMtop_min;               // [% Sat.] Minimum value where surface soil moisture constrains Rh
  float64  SMtop_max;               // [% Sat.] Maximum value where surface soil moisture constrains Rh
  float64  Tsoil_beta0;             // [dim.] Arrhenius temperature [K] constraint curve shape parameter for Rh
  float64  Tsoil_beta1;             // [dim.] Arrhenius temperature [K] constraint curve shape parameter for Rh
  float64  Tsoil_beta2;             // [dim.] Arrhenius temperature [K] constraint curve shape parameter for Rh

  /* Fractionation parameters for the ecosystem respiration model component */
  float64   fraut;  // [dim.] Fraction of GPP respired as autotrophic respiration
  float64   fmet;   // [dim.] Fraction of annual NPP allocated to soil organic C metabolic pool
  float64   fstr;   // [dim.] Fraction of structural soil organic C pool allocated to slow soil organic C pool

  // Rate constants for soil decompoisiton model component
  float64   kopt;   // [g C m-2 d-1] Optimal decomposition rate for metabolic soil organic C pool
  float64   kstr;   // [dim.] Structural soil organic C pool decomposition optimal rate as a fraction of the metabolic optimal rate
  float64   kslw;   // [dim.] Slow soil organic C pool decomposition optimal rate as a fraction of the metabolic optimal rate

  // QA parameter fields
  uint32    Nee_QA_Rank_min;    // [dim.] Integer rank representing highest quality (lowest error level) Net Ecosystem CO_2 Exchange
  uint32    Nee_QA_Rank_max;    // [dim.] Integer rank representing lowest quality (highest error level) Net Ecosystem CO_2 Exchange
  float64   Nee_QA_Error_min;   // [g C m-2 d-1] Floating point lowest error corresponding to highest quality rank Net Ecosystem CO_2 Exchange
  float64   Nee_QA_Error_max;   // [g C m-2 d-1] Floating point highest error corresponding to lowest quality rank Net Ecosystem CO_2 Exchange

  uint32    Fpar_QA_Rank_min;   // [dim.] Integer rank representing highest quality (lowest error level) Fpar input
  uint32    Fpar_QA_Rank_max;   // [dim.] Integer rank representing lowest quality (highest error level) Fpar input
  float64   Fpar_QA_Error_min;  // [dim.] Floating point lowest error corresponding to highest quality rank Fpar input
  float64   Fpar_QA_Error_max;  // [dim.] Floating point highest error corresponding to lowest quality rank Fpar input

  // <LAJ> 08/06/2014, Addded FtAge fields for incorporating FT into QA calculation
  float64   FtMethod_QA_mult;   // [dim.] Multiplier representing additional error for Radar vs. Tsurf FT input (TsurfError = RadarError + RadarError * FtMethod_QA_mult)
  uint32    FtAge_QA_Rank_min;  // [dim.] Integer rank representing highest quality (lowest error level) FT Age input
  uint32    FtAge_QA_Rank_max;  // [dim.] Integer rank representing lowest quality (highest error level) FT Age input
  float64   FtAge_QA_Error_min; // [dim.] Floating point lowest error corresponding to highest quality FT Age input
  float64   FtAge_QA_Error_max; // [dim.] Floating point highest error corresponding to lowest quality FT Age input

  float64   Par_QA_Error;       // [MJ m-2 d-1] Model input static photosynthetically active radiation error level
  float64   Tmin_QA_Error;      // [K] Model input static minimum daily air temperature error level
  float64   Tsoil_QA_Error;     // [K] Model input static rootzone soil moisture error level
  float64   Vpd_QA_Error;       // [Pa] Model input static average daily vapor pressure deficit error level
  float64   Smrz_QA_Error;      // [% Sat.] Model input static rootzone soil moisture error level
  float64   Smtop_QA_Error;     // [% Sat.] Model input static surface soil moisture error level

} BPLUT;


// <JMG> added 2014-02-22, to facilitate multiple uses in one app
int   l4c_BPLUT_Scan(FILE *bplutf);
// separated out the allocate step to avoid this occuring multiple times

// Allocates a module-level-static bpLUT_Master_p so does NOT return it
// since it is already visible to all functions in the module file
int l4c_BPLUT_Allocate(const int nValidRecs,const char *bpLUT_Label_p);

// Performs read only, now separated from pre-scan to count, and 
// separated out from the allocation...
BPLUT *l4c_BPLUT_Read(const char *bpLutPath,int *nBplutRecords);

BPLUT *l4c_GetBPLUT_Private(void);
BPLUT *l4c_GetBPLUT_Record(const char *landcoverLutKey_p,const int nValidBPLUT_Records);
BPLUT *l4c_GetBPLUT_ByIndex(const int landcoverIntegerKey,const int nValidBPLUT_Records);
void l4c_UnitTest_BPLUT(const int nRec);

int l4c_BPLUT_Report(FILE *outf,const int nPftClasses);


#endif
