// ----------------------------------------------------------------------------
/*
 subsystem           : SMAP L4_C
 module              : l4c_MODIS_easegrid.c
 purpose             : implement EASEGRID 2 forward and inverse transforms
 revision            : 2013-11-04 (see VERS_TAG)
 version             : v4.1 (D4.1)
 invocation          : n/a
 author              : Lucas Jones
 L4_C integration POC: Joe Glassy
 notes               : n/a
*/

#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"

// ----------------------------------------------------------------
// l4c_easegrid2_wgs84_fwd() 
// Purpose: Given a projection identifier, perform a forward transform,  
//           converting one x,y coordinate FROM geographical {lat,lon} 
//           projection TO the ease-grid2 {row,col} projection 
// Returns: status, ActionSuccess, and, ptrs to EASEGRID ithRow,ithCol
// Inputs : mapProjectionCode
//        : float64 lat
//        : float64 lon
//        : float64 *ithRow
//        : float64 *ithColumn
// Outputs: *ithRow,*ithColumn as EASEGRID raster {row,col} indices, in row-major-order
// history: original name: wgs84_convert() 
/*
  info[ 19 ] = { gpd: 'EASE2_M09KM', projection: '', map_scale_m: 9008.055210146D, $
                 cols: 3856L, rows: 1624L, r0: 0.0D, s0: 0.0D }
  info[ 20 ] = { gpd: 'EASE2_M03KM', projection: '', map_scale_m: 3002.6850700487D, $
                 cols: 11568L, rows: 4872L, r0: 0.0D, s0: 0.0D }
*/
int32 l4c_easegrid2_wgs84_fwd(const int32 mapProjectionCode,const float64 lat,const float64 lon, float64 *ithRow, float64 *ithColumn)
{
 char fxName[MAX_MEMOBJ_VNAME_LEN]    = {0} ;
 char moduleFile[MAX_MEMOBJ_VNAME_LEN]= {0} ;
 int32 status = -1 ;
 
	// variables removed from forward transform, not used: {beta,e4,e6}
	// parameter definition 
	float64 map_reference_latitude =0.0, 
			map_reference_longitude =0.0;
	float64 map_second_reference_latitude =0.0,
			map_eccentricity =0.0;
	float64 kz =0.0,
			qp =0.0;
	float64 rho =0.0;
	float64 phi =0.0,
			lam =0.0;
	float64 resolution =0.0, 
			Radius =0.0;
	float64 e2 =0.0 ;
	float64 x =0.0,
			y =0.0;
	float64 epsilon =0.0;
	float64 dlon =0.0;
	float64 q =0.0;
	float64 sin_phi1 = 0.0, 
			cos_phi1 = 0.0;
	float64 sin_phi  = 0.0;

	int32 nRows = 0 ; // varies by 1km vs. 9km
	int32 nCols = 0 ;

	float64 r0 = 0.0, 
		s0 = 0.0;
	float64 r_out = 0.0, 
		s_out= 0.0;

	float64 r =0.0; // preserve original semantics for now
	float64 s =0.0; // preserve original semantics for now

	// first define the projection parameters 
	// for global projection, define the geodetic radius and eccentricity  
	Radius = 6378137.0 ; // radius in meters of WGS84 sphere 
	map_eccentricity = 0.081819190843; // WGS84 ecc factor
	e2 = pow(map_eccentricity, 2.0);
	epsilon = 1.e-6;

	// register this module and function
	strcpy(fxName,(char *)__func__);
	strcpy(moduleFile,(char *)__FILE__);
	
	// branch on global vs. Northern hemisphere projections, to set up the basics..
	if(mapProjectionCode <= IsEG_Global_1km)
	{
		map_reference_latitude = 0.0 ;
		map_reference_longitude = 0.0;
		map_second_reference_latitude = 30.0;
		sin_phi1 = sin(30.0 *PI/180.0);
		cos_phi1 = cos(30.0 *PI/180.0);
		kz = cos_phi1/sqrt(1.0 - e2 * sin_phi1 * sin_phi1 );
		// override resolution for 9km case
	} else { // handle case of 'N' North hemisphere
        map_reference_latitude = 90.0 ;
		map_reference_longitude = 0.0 ;
	}

	switch(mapProjectionCode)
	{
	case IsEG_Global_1km:	resolution = 1000.89502334956; // km 
							nRows = 14616;
							nCols = 34704;
							break;
	case IsEG_Global_3km: // projection info: reference lat,long
							resolution = 3002.6850700487 ; // for 3km
							nRows = 4872  ;
							nCols = 11568 ;	
							break;
	case IsEG_Global_9km: // projection info: reference latitudes and longitudes 
							resolution = 9008.055210146; // for 9km 
							nRows = 1624 ;
							nCols = 3856 ;
							break;
	case IsEG_N_1km		:	resolution = 1000.0; // 1km 
							nRows      = 18000;
							nCols      = 18000;
							break;
	case IsEG_N_9km	 	:	resolution = 9000.0; // 9km km 
							nRows = 2000;
							nCols = 2000;
							break ;
	default: fprintf(stderr,"@F ~map-projection code %d is NOT yet supported\n",
						mapProjectionCode);
						break;
	} // end:: switch on mapProjectionCode

	/* r0: column that is mapped to map_reference_latitude; 
	   s0: row that is mapped to map_reference_latitude */
	r0 = (float64)(nCols - 1)/2.0;
	s0 = (float64)(nRows - 1)/2.0;

	dlon = lon - map_reference_longitude; 
	if( fabs(dlon) > 180.0) 
	{
		fprintf(stderr,"@E ERROR:: dlon %lf out-of-valid-range. Verify longitude\n",dlon);
	}
	
	phi = lat  * PI / 180.0;
	lam = dlon * PI / 180.0;

	sin_phi = sin(phi);

	q = ( 1.0 - e2 ) * ( ( sin_phi / ( 1.0 - e2 * sin_phi * sin_phi ) ) \
                        - ( 1.0 / ( 2.0 * map_eccentricity ) ) \
                        * log( ( 1.0 - map_eccentricity * sin_phi ) \
                                / ( 1.0 + map_eccentricity * sin_phi ) ) );

    qp = 1.0 - ((1.0 - e2)/(2.0 * map_eccentricity) \
                * log((1.0 - map_eccentricity) \
                        /(1.0 + map_eccentricity) ) );

	// IsEG_Global_9km, IsEG_Global_1km, IsEG_N_9km, IsEG_N_1km
	if(mapProjectionCode >= IsEG_N_9km)
	{
		if( fabs(qp-q) < epsilon ) 
			rho = 0.0;
		else
			rho = Radius * sqrt(qp-q);
			
		x = rho * sin(lam);
		y = -1.0 * rho * cos(lam);	
		
	} else { // for GLOBAL cases... rho is not defined...
		// IsEG_Global_9km, isEG_Global_3km, IsEG_Global_1km.. handled here
		x = Radius * kz * lam;
		y = Radius * q/(2.0 * kz);
	}
	
	r_out = r0 + x/resolution ;
	s_out = s0 - y/resolution ;

	if( (r_out < -0.5) || (r_out >= nCols) || (s_out < -0.5) || (s_out >= nRows) )
	{
		fprintf(stderr,
			"@E ERROR:: coord transform lat: %f, lon: %f, r_out: %f, s_out: %f\n",
				lat, lon, r_out, s_out);
		(*ithColumn) = -1.0;
		(*ithRow) = -1.0;
	} 

	// preserve original semantics, used cartographic 'r','s' terms to pass output terms
	r = r_out;
	s = s_out;
	
	// pass results back now to real numbers (float64's) --> ithColumn and ithRow 
	status = ActionSuccess ;
	// <JMG> added fabs() since we only want positive row,col anyway
	*ithColumn = fabs(r_out);
	*ithRow    = fabs(s_out);
	
	// return final status 
	return status;
} // end::void l4c_easegrid2_wgs84_fwd()


// ----------------------------------------------------------------
// easegrid2_wgs84_inversetransform()
// Purpose: perform inverse transform, given a map projection integer code, and an 
//  EASE-GRID v2 {row,column} coordinate, derive a decimal {latitude,longitude} coordinate.
// Inputs : int32 mapProjectionCode (from enum..)
//        : float64 ithRow,
//        : float64 ithColumn
//        : float64 *lat
//        : float64 *lon
// Outputs: lat,lon
// history: adapted from wgs84_inverse()
// Returns: status, ActionSuccess, and, ptrs to EASEGRID *lat,*lon
int32 l4c_easegrid2_wgs84_inv(const int32 mapProjectionCode,float64 ithRow, float64 ithColumn,float64 *lat,float64 *lon)
{
 char fxName[MAX_MEMOBJ_VNAME_LEN]    = {0} ;
 char moduleFile[MAX_MEMOBJ_VNAME_LEN]= {0} ;

	/* parameter definition */
	float64 map_reference_latitude        = 0.0 ;
	float64 map_reference_longitude       = 0.0;
	float64 map_second_reference_latitude = 0.0;
	float64	map_eccentricity              = 0.0;
	float64 sin_phi1 = 0.0;
	float64 cos_phi1 = 0.0;
	float64 kz = 0.0;
	float64 qp = 0.0;
	float64 beta = 0.0;
	float64 rho = 0.0;
	float64 phi = 0.0;
	float64 lam = 0.0;
	float64 resolution = 0.0;
	float64 Radius = 0.0;
	float64 e2 = 0.0;
	float64 e4 = 0.0;
	float64  e6 = 0.0;
	float64 x = 0.0;
	float64 y = 0.0;
	float64 epsilon = 0.0;

	int32 nRows =0;
 	int32 nCols =0;
	int32 status = -1 ;
	
	float64 r0 = 0.0, s0 = 0.0;
	float64 lat_out = 0.0;
	float64 lon_out = 0.0;

	float64 r = ithColumn;// preserve original goofy semantics
	float64 s = ithRow   ;// preserve original goofy semantics

	// Define the radius and eccentricity of the geodetic model used
	Radius 				= 6378137.0 ;      // WGS84 sphere radius in decimal meters
	map_eccentricity	= 0.081819190843;  // WGS84 */
	e2 = pow(map_eccentricity, 2.0);
	e4 = pow(map_eccentricity, 4.0);
	e6 = pow(map_eccentricity, 6.0);
	epsilon = 1.e-6 ;

	// register this module and function
	strcpy(fxName,(char *)__func__);
	strcpy(moduleFile,(char *)__FILE__);
	
	// branch on global vs. Northern hemisphere to set up the basics..
	if(mapProjectionCode <= IsEG_Global_1km)
	{
		map_reference_latitude  = 0.0;
		map_reference_longitude = 0.0;
		map_second_reference_latitude = 30.0;
		sin_phi1 = sin(30.0 *PI /180.0);
		cos_phi1 = cos(30.0 *PI /180.0);
		kz = cos_phi1/sqrt(1.-e2*sin_phi1*sin_phi1);
	} else { // handle case of 'N' North hemisphere
		map_reference_latitude = 90.;
		map_reference_longitude = 0.;
	}

	// note: resolution here is "map scale"
		switch(mapProjectionCode)
		{
		case IsEG_Global_1km:	resolution = 1000.89502334956; // km 
								nRows = 14616;
								nCols = 34704;
								break;
		case IsEG_Global_3km: 	resolution = 3002.6850700487 ; // for 3km
								nRows = 4872  ;
								nCols = 11568 ;	
								break;
		case IsEG_Global_9km: 	resolution = 9008.055210146; // for 9km 
								nRows = 1624 ;
								nCols = 3856 ;
								break;
		case IsEG_N_1km		:	resolution = 1000.0; // 1km 
								nRows      = 18000;
								nCols      = 18000;
								break;
		case IsEG_N_9km	 	:	resolution = 9000.0; // 9km km 
								nRows = 2000;
								nCols = 2000;
								break ;
	default: fprintf(stderr,"@F map-projection code %d is not yet supported\n",
						mapProjectionCode);
						break;
	} // end:: switch on mapProjectionCode

	// r0: column that is mapped to map_reference_latitude; 
	//   s0: row that is mapped to map_reference_latitude. 
	r0 = (float64)(nCols - 1)/2.0;
	s0 = (float64)(nRows - 1)/2.0;

	/* convert (col, row) to xy coordinates for this projection */
	x = (r-r0) * resolution;
	y = (s0-s) * resolution;

	qp = ( 1.0 - e2 ) * ( ( 1.0 / ( 1.0 - e2 ) ) \
                        - ( 1.0 / ( 2.0 * map_eccentricity ) ) \
                        * log( ( 1.0 - map_eccentricity ) \
                                / ( 1.0 + map_eccentricity ) ) );

	// IsEG_Global_9km, IsEG_Global_1km, IsEG_N_9km, IsEG_N_1km
	if(mapProjectionCode >= IsEG_N_9km)
	{
		rho = sqrt( x * x + y * y);
		beta = asin(1. - (rho * rho/( Radius * Radius * qp)));
		lam = atan(x/(-1.0 * y));
	} else { // for Global cases...
		beta = asin(2.0 * y * kz / (Radius * qp));
		lam  = x/(Radius * kz);
	}

	phi = beta \
        + ( ( ( e2 / 3.0 ) + ( ( 31.0 / 180.0 ) * e4 ) \
              + ( ( 517.0 / 5040.0 ) * e6 ) ) * sin( 2.0 * beta ) ) \
        + ( ( ( ( 23.0 / 360.0 ) * e4) \
              + ( ( 251.0 / 3780.0 ) * e6 ) ) * sin( 4.0 * beta ) ) \
        + ( ( ( 761.0 / 45360.0 ) * e6 ) * sin( 6.0 * beta ) );

    lat_out = 180.0 * phi/PI;
	lon_out = 180.0 * lam/PI + map_reference_longitude;

	if( (fabs(lat_out) > 90.0) || (fabs(lon_out) > 180.0) )
	{
		fprintf(stderr,"@F out-of-range latitude (%lf {-90<lat<90}) or longitude (%lf {-180<lon<180}) transformation!\n",
			lat_out,lon_out);
	} // end:: fabs block

	// Perform explict cast from float64 precision, to float64 precision
	// acknowledging we may lose some precision for some point translations
	// TBD: verify if we need a R_ROUND() call here...
	status = ActionSuccess ;
	(*lat) = (float64)lat_out;
	(*lon) = (float64)lon_out;
	
	// return final status
	return status;
} // end::l4c_easegrid2_wgs84_inv()

