// ----------------------------------------------------------------
// function: l4c_MODIS_logic.c
// purpose : Implement logic l4c_MODIS, process a single 8-day period
// author  : joe glassy
// research code : lucas jones
// revised : 2021-10-15T16:08 <JMG> replaced frm_utilSetBits() with frm_utilAssignBits()
//           in reorg_VIIRS_Qc() function
// revised : 2017-03-09T16:46 <LAJ> fixed fill value bug in resample_fpar_window()
//         : NOTE that fpar has multiple fill values (different fill values for different land cover types, 
//           urban, permanent ice/snow, water, etc.).  Fpar now considered fill if it has value >99.
// revised : 2017-02-09T16:46, modified ImgPkg struct in l4c_MODIS.h to 
//           supported qualified 1km vs. 500 tile containers
// revised : 2017-02-01 <LAJ>, added support code for MODIS Collection 6 500m MOD15A2H
// revised : 2017-01-26T12:43, adding support for MODIS Collection 6 500m MOD15A2H
// revised : 2014-08-19, cleaned up for NDVI vs. FPAR use cases
// revised : 2014-07-29T13:32, integration, sync up with v5.0.1 release
// history : 2021-10-15T16:08 <JMG> replaced frm_utilSetBits() with frm_utilAssignBits()
//           in reorg_VIIRS_Qc() function.
// history : 2013-11-11T16:02:00-07:00, cleaned up, removed un-used functions
// history : 2013-11-01
// notes   : n/a

// ------------- LEGACY BLOCK ------------------------------------

// -----header files----------------------------------------------
#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"


// <LAJ> 2017-02-01 ADD: 
// ---------------------------------------------------------------------
// function : resample_fpar_window()
// purpose  : Contains logic to resample MOD15A2H (collection 6) FPAR & QC into the same 
//            format as MOD15A2 (collection 5) FPAR & QC for a single 1 km grid cell. 
//            Wrapped by resample_tile_500m_to_1km().
//            
// Inputs   : Allocated uint8 fpar and uint8 fpar QC arrays for 500 m (MOD15A2H c6) grid cells within each 1 km window.
// Outputs  : Returns 1 km tile fpar and fpar QC scalar pointers
// Notes    : The QC bitfields are translated into a 5-class rank.  The 500 m grid cells are then binned by rank.  The 
//            fpar average of the highest rank 500 m grid cells are then assigned to the 1 km grid cell.  The fpar QC of 
//            the last 500 m grid cell containing the highest rank is then assigned to the 1 km grid cell. This preserves
//            the rank interpretation for ingestion into the L4C and TCF science codes. If the ranking system changes, the 
//            scheme used in this function should be likewise changed.
// revised  : 2021-07-23T17:22, modify call sig to include new extra_qc1km scalar           
// History  : Created laj 01/25/2017

uint8 resample_fpar_window(uint8 *fpar_window4x4, uint8 *qc_window4x4, uint8 *fpar1km, uint8 *qc1km)
{
  int k = 0;
  uint8    fpar_uint8 = FILL_VALUE;
  uint8    qc_rank = 0;
  uint8    rank_min = 4;
  
  uint8    last_qcBitfieldByRank[N_QC_RANK] = {0};

  uint8     fparRank_cnt[N_QC_RANK] = {0};
  float32   fparRank_sum[N_QC_RANK] = {0};

  float32   fparRank = 0.0; // <JMG> added 0.0 init, was undefined prior

  // Initialize sums and counts
  for (k=0; k<MODIS_1KM_NESTED_500M_PIXELS; k++)
  {
    fparRank_sum[k] = 0.0;
    fparRank_cnt[k] = 0;
    last_qcBitfieldByRank[k] = FILL_VALUE; // Set all bits to 1 by default
  } // end :: initialize sums and counts

  // Loop through nested 500 m grid cells, bin by QC rank
  for (k=0; k < MODIS_1KM_NESTED_500M_PIXELS; k++)
  {
    fpar_uint8 = fpar_window4x4[k];
    // Bin only non-fill values (> 100 gets all fpar fill values)
    // NOTE: FPAR has multiple fill values, each which denote a different surface type
    if (fpar_uint8 < 100)
    {
        // Rank fpar qc using the usual 4-class system
        qc_rank = fpar_qa_bin_reclas(qc_window4x4[k]);
        // Bin based on 500 m grid cell QC rank
        fparRank_sum[qc_rank] += (float32) fpar_uint8;
        // Otherwise fill values: do not bin these
      } else {
        qc_rank = 0;
    } // end :: check non-fill values

    // Tally here to include missing values (qc_rank = 4) too
    fparRank_cnt[qc_rank] ++;

    // Track current qc bitfield in each rank
    last_qcBitfieldByRank[qc_rank] = qc_window4x4[k];

    // Find max QC rank for window
    // <JMG> 2021-07-23 ASSERT: actual logic here refers
    // to rank_min NOT rank_max
    if ((qc_rank != 0) && (qc_rank < rank_min))
    {
        rank_min = qc_rank; 
    } // end :: max QC rank check

  } // end :: nested 500 m grid cell loop

  // Assign average of highest-QC fpar to 1km grid cell
  if (fparRank_cnt[0] != 4)
  {
      // Assign average of highest-QC fpar to 1km grid cell
      // <JMG> ASSERT: comment refers to "highest" 
      //              yet actually, rank_min is used. TBD VERIFY
      fparRank = (float32) fparRank_sum[rank_min]/fparRank_cnt[rank_min];
  
      // <JMG> 2017-02-09T15:57 resolved missing macro, now using FRM_API_ROUND()
      // following other L4_C conventions to use UM Framework API
      *fpar1km = (uint8)FRM_API_ROUND(fparRank);

      // Assign most current (last read) fpar QC with max QC rank to 1km grid cell
      // NOTE & DISCLAIMER: The foremost consideration here is that this QC field will evaluate to the correct QC rank.
      // This gives a fair representation of the QC of the averaged grid cells from the rank perspective; however, if the
      // user/developer wishes to use the QC in another way, say re-rank based on other criteria, then this scheme should
      // be modified for the appropriate objective.
      *qc1km = last_qcBitfieldByRank[rank_min];

      // All 500 m values missing within window
    } else {
      *fpar1km = FILL_VALUE;
      // <JMG> 2021-07-23, review if just assign qc1km here,
      // OR qc1km & extra_qc1km depending on how this fx is called
      *qc1km = FILL_VALUE;
  } // end :: check all values missing
   
  // Return overall QC rank, noting this is rank_min and NOT rank_max here 
  return rank_min;

} // end :: resample_fpar_window()


// <LAJ> 2017-02-01 ADD: 
// ---------------------------------------------------------------------
// function : resample_tile_500m_to_1km()
// purpose  : Enclosing loop to resample MOD15A2H (collection 6) FPAR & QC into the same 
//            format as MOD15A2 (collection 5) FPAR & QC. Wraps resample_fpar_window().
//            
// Inputs   : Allocated tile structures for 500 m (MOD15A2H c6) and 1 km (MOD15A2 c5) tiles
//            containing uint8 fpar and fpar QC arrays
// Outputs  : Populates 1 km tile fpar and fpar QC arrays
// Notes    : See resample_fpar_window()
// revised  : 2021-07-24, start VIIRs integration modifications.
// revised  : 2017-02-10T17:05, joe corrected tile_500m to be pointer to struct
// History  : Created laj 01/25/2017

int32 resample_tile_500m_to_1km(tile_struct *tile_500m_p, tile_struct *tile_1km_p)
{
  int   i  = 0;
  int   j  = 0;
  int   in = 0;
  int   jn = 0;
  int   ij_window =0;
  int   rtnStatus =0;
  int32 kCompleted = 0 ;
  char fxName[256] = {0}; // fxName to register with __func__
  uint8     fpar1km =0;
  uint8     qc1km =0;  //  legacy qc at 1km, storing the post-500 handling
  uint8     extra_qc1km = 0; // <JMG> 2021-07-21 add for extra_qc handling
  uint8     extra_qc_rank=0;
  uint8     qc_rank =0;

  // reminder: MODIS_1KM_NESTED_500M_PIXELS is only 4...
  uint8     fpar_window4x4[MODIS_1KM_NESTED_500M_PIXELS] = {0};
  uint8     qc_window4x4[MODIS_1KM_NESTED_500M_PIXELS]   = {0};
  // <JMG> 2021-07-23, add extra_qc window to match other qc
  uint8     extra_qc_window4x4[MODIS_1KM_NESTED_500M_PIXELS]   = {0};

  // REGISTER fx 
  strcpy(fxName,(char *)__func__); 
  
  // xLogMsg("@A In(%s) tile_loop: RESAMPLING 500m to 1km ~\n",fxName);

  // Loop through MODIS 1 km tile
  for (i=0; i < MODIS_1KM_EW_PIXELS; i++) 
  {
      
    for (j=0; j < MODIS_1KM_NS_PIXELS; j++) 
    {
      // <JMG> 2017-02-08 changed to use MODIS_1KM_EW_PIXELS);
      tile_1km_p->ij = M_2D_B0(i,j,MODIS_1KM_EW_PIXELS);

       // Unpack nested 500 m window from MODIS 500 m tile (there's 4 500 m grid cells per 1 km grid cell)
       for (in=0; in < MODIS_1KM_NESTED_500M_EW_PIXELS; in++)
       {
           for (jn=0; jn < MODIS_1KM_NESTED_500M_NS_PIXELS; jn++)
           { 
             tile_500m_p->ij = M_2D_NESTED_B0(i,j,in,jn,MODIS_1KM_NESTED_500M_EW_PIXELS, MODIS_500M_EW_PIXELS);
             
             ij_window = M_2D_B0(in,jn,MODIS_1KM_NESTED_500M_EW_PIXELS);
        
             fpar_window4x4[ij_window]     = tile_500m_p->fpar[tile_500m_p->ij ];
             qc_window4x4[ij_window]       = tile_500m_p->qc[tile_500m_p->ij ];
             // <JMG> 2021-07-23, added to parallel logic for extra_qc assuming its needed,useful
             extra_qc_window4x4[ij_window] = tile_500m_p->extra_qc[tile_500m_p->ij ];
             // increment count of completed inner-most operations
             kCompleted++ ;
           } // end :: inner 500 m tile NS loop
       } // end :: inner 500 m tile EW loop

      // Resample the nested 500 m window (4 grid cells) into a single 1 km grid cell 
      // return current qc rank for convenience (not further needed, except for diagnostics).
      
      // <JMG> EXPERIMENTAL, need to evaluate what is needed and what's best
      // <JMG> ASSERT: both extra_qc1km AND fpar1km ARE ASSIGNED HERE, so we put 
      //  this call to resample_fpar_window() first, in case we temporarily
      //  want the final fpar1km assignment (that sticks) to be driven from legacy logic...
      //  noting also that "fpar_window4x4 is read-only,not altered, allowing 2 independent calls
      // NOT YET IMPLEMENTED
      // extra_qc_rank = resample_fpar_window(fpar_window4x4, extra_qc_window4x4, &fpar1km, &extra_qc1km);
      extra_qc1km = 0 ;
      
      
      // <JMG> 2021-07-24, NOTE that pixel wise qc1km is assigned here, so this requires 
      // changes to support Coll 6 VIIRs and MODIS alike.
      // ASSERT: in this scheme, only ONE rank may be returned as driving logic, e.g. qc_rank
      // which NOW derives from qc1km and does not yet use "extra_qc1km" so verify 
      // that extra_qc1km is correctly defined from re-sampling all the way through to assignment.
      qc_rank = resample_fpar_window(fpar_window4x4, qc_window4x4, &fpar1km, &qc1km);

      // Assign resampled values to desintation 1km (result) tile
      // <JMG> ASSERT: study carefully to incorporate new extra_qc influenced logic
      tile_1km_p->fpar[      tile_1km_p->ij ] = fpar1km;
      tile_1km_p->qc[        tile_1km_p->ij ] = qc1km;
      // <JMG> ADD in anticipation of potential extra_qc handling to match regular qc
      // ASSERT: we DO NOT yet write out extra_qc to the OUTPUT HDF5 file at this time
      // so this in-memory change (step) does NOT YET propogate to a change in OUTPUT HDF5 file structure.
      tile_1km_p->extra_qc[  tile_1km_p->ij ] = extra_qc1km;
    } // end :: 1 km tile NS loop
    
  } // end :: 1 km tile EW loop
  
  rtnStatus = kCompleted;
  // xLogMsg("@A In(%s) RESAMPLING Completed %d pixels ~\n",fxName,kCompleted);
  // enable return of completion status
  return rtnStatus;
  
} // end::resample_tile_500m_to_1km()


// ----------------------------------------------------------------------------
// function: pop_tile_to_1kmease2
// returns : int, count of all pixels mosaicked
// purpose : Implement mosaic operation. Inserts each reprojected tile, {tile_p->fpar|tile_p->qc} 
//           into the (2) resulting EASEGRID destination images, {ease_p->fpar,ease_p->qc}
// revision: 2014-08-19, <JMG> modified for NDVI vs. FPAR distinction
//           (deprecated): based on sesCtl_p->gppMethodCode (0=NDVI, 1=FPAR, 2=FPARCLIM, 3=UNSPECIFIED)
// revision: 2013-11-07
// author  : joe glassy
// history : Changed args to pointers (cont --> now cont_p etc)
// notes   : TBD future refactoring: there is no reason for local scalar 
//             variables ease_p->ij, or tile_p->h,tile_p->v
//             to be members of ease_p or tile_p. Factor these out after confirming no side-effects...

int pop_tile_to_1kmease2(tileref_struct *tileref_p, tile_struct *tile_p, ease_struct *ease_p)
{
    SesCtl *sesCtl_p   = l4c_GetSesCtl(); // import session controls    
    
    int istop  =0,
        istart =0,
        jstop  =0,   
        jstart =0; 
        
    int32   ithRow =0,
            ithCol =0,
            serialIdx =0,
            nSumPixelsMosaiced = 0 ;

  int16 thisTile_vert = 0;
  int16 thisTile_hori = 0;
  
    // Determine coords for a rectangular ease grid patch entirely 
    //     covering the roughly trapezoidal-shaped tile region
    get_patch_dims(tile_p, &istart, &istop, &jstart, &jstop);
  
    // loop though ease grid patch rows
    for (ithRow = istart; ithRow < istop; ithRow++)
    {
        // loop through ease grid patch columns
        for (ithCol = jstart; ithCol < jstop; ithCol++)
        {
            // reference full ease grid
            ease_p->ij = serialIdx = M_2D_B0(ithRow, ithCol, EGv2_GLOBAL_1KM_EW_PIXELS);

            // if within tile boundaries, get data from the tile 
            // CRITICAL-NOTE: tileref.v|h images are now uint8 {0..255}, so we broadcast +1 to all pixels, so to 
            // recapture original semantics, we SHIFT All Pixels Values DOWN BY ONE, e.g. subtract 1 here
            thisTile_vert = (int16)tileref_p->v[ease_p->ij] -1; // decrement by 1 to retain orig -1..14 semantics
            thisTile_hori = (int16)tileref_p->h[ease_p->ij] -1; // decrement by 1 to retain orig -1..35 semantics

            // Check for match on both the vertical and horizontal tile indices
            if( thisTile_vert == tile_p->v && thisTile_hori == tile_p->h )
            {
                tile_p->row = tileref_p->r[ ease_p->ij ]; // .r member prior values left intact
                tile_p->col = tileref_p->c[ ease_p->ij ]; // .c member prior values left intact
                
                // for this row,col, determine the equivalent serialized 1D index
                tile_p->ij = M_2D_B0(tile_p->row, tile_p->col, 1200);
                // <JMG> 2021-07-14 DEPRECATED branch for NDVI, consolidate (MODIS|VIIRS) to share this logic
                ease_p->fpar[ease_p->ij] = tile_p->fpar[tile_p->ij];
                ease_p->qc[ease_p->ij]   = tile_p->qc[tile_p->ij];
                // <JMG> 2021-07-24, VERIFY if extra_qc plays a role here
                // earlier it appears that we DEPRECATED qc_extra, to allow for pass through
                // ease_p->qc_extra[ease_p->ij] = tile_p->qc_extra[tile_p->ij];
                
                nSumPixelsMosaiced++ ; // checksum
            } // end:: if thisTile_vert block
      
        } // end:: ithCol (j) column loop
    
  } // end:: ithRow, (i) row loop

  // return count of pixels mosaicked
  return(nSumPixelsMosaiced);
  
} // end:: pop_tile_to_1kmease2()


// ----------------------------------------------------------------------------
// purpose: Determine a single corner (integer row,col) index for tile passed in.
int find_corner(tile_struct *tile_p, int *row_p, int *col_p)
{

  float64   lat, lon;
  float64   row_dest_flt, col_dest_flt;

  // reference tile matrix
  tile_p->ij = M_2D_B0(tile_p->row, tile_p->col, MODIS_1KM_EW_PIXELS);

  lat = tile_p->lat[tile_p->ij];
  lon = tile_p->lon[tile_p->ij];

  // Check if lat/lon are within bounds
  lat = (lat <=  -90.0) ? -89.99999  : lat;
  lat = (lat >=   90.0) ? 89.99999   : lat;
  lon = (lon <= -180.0) ? -179.99999 : lon;
  lon = (lon >=  180.0) ? 179.99999  : lon;
  
  /* Get into the vicinity with the forward transform */
  
  // int32 l4c_easegrid2_wgs84_fwd(const int32 mapProjectionCode,const float64 lat,const float64 lon, float64 *ithRow, float64 *ithColumn)
  l4c_easegrid2_wgs84_fwd(IsEG_Global_1km, lat, lon, &row_dest_flt, &col_dest_flt);
  // ORIG: easegrid_v2_wgs84_forwardtransform("EASE2_M", "1km", lat, lon, &row_dest_flt, &col_dest_flt);
 
  *row_p = (int) FRM_API_ROUND(row_dest_flt);
  *col_p = (int) FRM_API_ROUND(col_dest_flt); 
  
  return(0);
  
} // end::find_corner()


// ----------------------------------------------------------------------------
int get_patch_dims(tile_struct *tile_p, int *istart, int *istop, int *jstart, int *jstop)
{
  int       ibuf, jbuf;
  int       corner_row[4];
  int       corner_col[4];

  // This provides a small edge buffer just in case the min distance lies in an adjacent tile
  ibuf = 1;
  jbuf = 1;

  // Locate the tiles (4) corner pixels

  /*lower left*/
  tile_p->row = 0;
  tile_p->col = 0;
  find_corner(tile_p, &corner_row[0], &corner_col[0]);

  /*lower right*/
  tile_p->row = 0;
  tile_p->col = 1200-1;
  find_corner(tile_p, &corner_row[1], &corner_col[1]);

  /*upper left*/
  tile_p->row = 1200-1;
  tile_p->col = 0;
  find_corner(tile_p, &corner_row[2], &corner_col[2]);

  /*upper right*/
  tile_p->row = 1200-1;
  tile_p->col = 1200-1;
  find_corner(tile_p, &corner_row[3], &corner_col[3]);

  // Find bounding coords to determine a rectangular ease grid patch entirely 
  //  covering the roughly trapezoidal-shaped tile region

  // furthest left col
  if (corner_col[0] < corner_col[2])
    *jstart = corner_col[0];
  else
    *jstart = corner_col[2];

  // furthest right col
  if (corner_col[1] > corner_col[3])
    *jstop = corner_col[1];
  else
    *jstop = corner_col[3];

  /*furthest upper row*/
  /*since ease grid origin in upper left, upper = smaller number*/    
  if (corner_row[2] < corner_row[3])
    *istart = corner_row[2];
  else
    *istart = corner_row[3];

  /*furthest lower  row*/
  if (corner_row[0] > corner_row[1])
    *istop = corner_row[0];
  else
    *istop = corner_row[1];

  // Apply buffer
  *istart -= ibuf;
  *istop += ibuf;

  *jstart -= jbuf;
  *jstop += jbuf;

  // Check boundary conditions (ease grid corners & edges)

  // lower left corner
  if (*istart < 0)
    *istart = 0;
  
  if (*jstart < 0)
    *jstart = 0;

  if (*istop > EGv2_GLOBAL_1KM_NS_PIXELS)
    *istop = EGv2_GLOBAL_1KM_NS_PIXELS;

  if (*jstop > EGv2_GLOBAL_1KM_EW_PIXELS)
    *jstop = EGv2_GLOBAL_1KM_EW_PIXELS;

  return(0);
} // end::get_patch_dims()


// ----------------------------------------------------------------------------
// function: parse_filename()
// purpose : parses off the 'VV' and 'HH' tile coords from MODIS (or VIIRs) filename
// inputs  : char *fname, char *substr, int *x, char str[]
// returns : int, 'h' {0..35} or 'v' {0..17} base0 tile index
// author  : Lucas Jones
// WARNING : Note this logic cannot handle fully qualified pathnames!! it must
//           be supplied with a basename.
int parse_filename(char *inputTilePath, char *substr, int *x, char str[])
{
  int resultCoord = -1 ;
  
  char *p = NULL;

  p = strstr(inputTilePath, substr) + 1; // increment base address

  strncpy(str,p,2);
  strcat(str,"\0");
  
  *x = atoi(str);

  resultCoord = *x ;
  // return the resulting (integer) tile-position index
  return( resultCoord);
} // end::parse_filename()



// ----------------------------------------------------------------------------
// function: get_yeardoy_from_filename
// purpose : Pulls YYYYDOY tile date format from MODIS filename
// inputs  : char *fname -- path
//           str
// author  : Lucas Jones
int get_yeardoy_from_filename(char *fname, char str[])
{
  char *p;

  p = strstr(fname, ".") + 2; /*increment base address*/

  strncpy(str,p,7);
  strcat(str,"\0");

  return(0);
} // end::get_yeardoy_from_filename()

 
// ----------------------------------------------------------------------------
// function: get_sinusoid_latlon()
// purpose : Assign correct lat,lon (decimal degree units) for each pixel
// inputs  : float64 upleft[], float64 lowright[], float64 lat[], float64 lon[], float64 *dx, float64 *dy
int get_sinusoid_latlon(float64 upleft[], float64 lowright[], float64 lat[], float64 lon[], float64 *dx, float64 *dy)
{
        int i=0, j=0, ij =0;
        float64 ll_upleft[2]   = {0},
                ll_lowright[2] = {0};
 
        // xLogMsg("@A get_sinusoid_latlon:: upLeft %.2f %2f lowright %.2f %.2f ~\n",upleft[0],upleft[1], lowright[0],lowright[0]);
 
        // Establish the actual cell size that splits up corner coords. evenly (close to 1 km but not exactly)
        *dx = abs(lowright[0] - upleft[0])/((float64) MODIS_1KM_NS_PIXELS);
        *dy = abs(upleft[1] - lowright[1])/((float64) MODIS_1KM_EW_PIXELS);
        
        // Retrieve the corner latitude,longitude coordinates, just to check 
        ll_upleft[0]=upleft[0] * 1.0e-3 / (RG*cos(upleft[1] * 1.0e-3/RG * PI/180.0));
        ll_upleft[1]=upleft[1]*  1.0e-3 / RG;
        
        ll_lowright[0]=lowright[0]*1.0e-3/(RG*cos(lowright[1]* 1.0e-3/RG * PI/180.0));
        ll_lowright[1]=lowright[1]*1.0e-3/RG;

        // assign lat/lon to each grid cell 
        for (i=0;i < MODIS_1KM_NS_PIXELS;i++)
        {
            for (j=0;j < MODIS_1KM_EW_PIXELS;j++)
            {          
                ij = M_2D_B0(i,j,MODIS_1KM_EW_PIXELS);

                lat[ij] = (i*(*dx) + lowright[1])*1.0e-3/ RG;
                lon[ij] = (j*(*dy) + upleft[0])*1.0e-3/ ( RG * cos(lat[ij] * PI /180.0));

                //  Warning: lat/lon can potentially be given out-of-range values
                //  TBD: in future, if above true, at least count them!!
            } // end:: inner j loop
            
        } // end:: outer i loop 

    // xLogMsg("@A get_sinusoid_latlon:: completed ~\n");       

    return(0);
    
} // end::get_sinusoid_latlon()


// ----------------------------------------------------------------------------
// function : dimswap()
// purpose  : perform vertical row flip
// revised  : 2017-02-09T17:42, modified to accomodate 1KM or 500m FPAR res tiles
//            so also added sesCtl_p import. Verify this is a 1km vs. 500m sensitive operation.
// revised  : 2013-11-09T10:10:00-07:00
// author   : Joe Glassy
// note     : Assumes, and works, only on uint8 datatype images. So, if
//            a non-uint8 image is passed in, in the future this code must
//            be adjusted for int16 or int8!
void dimswap(uint8 *A)
{
  SesCtl *sesCtl_p   = l4c_GetSesCtl() ;  // <JMG> 2017-02-09, added new import
  int   i;
  int   x, y;
  int   ix, iy;
  // <JMG> 2017-02-09T17:42 added nTotalPixels, nEW_Pixels,nNS_Pixels to
  // accomodate 1km vs. 500m FPAR res
  int32 nEW_Pixels =0 ;
  int32 nNS_Pixels =0 ;

  // ORIG CODE:  C[MODIS_1KM_NS_PIXELS] ..unsafe use of stack
  uint8  *pixTemp_p = NULL ; // alloc off the heap, not the stack

  if(sesCtl_p->fparResolutionMeters >= 1000)
  {
     nEW_Pixels = MODIS_1KM_EW_PIXELS ;
     nNS_Pixels = MODIS_1KM_NS_PIXELS ;
  } else {
     nEW_Pixels = MODIS_500M_EW_PIXELS ;
     nNS_Pixels = MODIS_500M_NS_PIXELS ;
  }

  x = 0; 
  y = nEW_Pixels - 1;  // this doesn't look right? Shouldn't it be nNS_Pixels for a "y" term?
  
  pixTemp_p = (uint8 *)memCalloc(nNS_Pixels, sizeof(uint8));

    while(x < y)
    {
        for (i=0; i < nNS_Pixels; i++)
        {
            // <JMG> verify if ONE of these below should be nNS_Pixels!
            // as this works only by the accident of the NS equaling the EW pixels
            ix = M_2D_B0(x,i, nEW_Pixels);
            iy = M_2D_B0(y,i, nEW_Pixels);
            // C[i] = A[ix];
            pixTemp_p[i] = A[ix];
            A[ix] = A[iy];
            // A[iy] = C[i];
            A[iy] = pixTemp_p[i] ;
        } // end:: i loop through MODIS pixels
    x++;
    y--;
  } // end::while x < y
  
  memFree(pixTemp_p);

} // end::dimswap()


// ----------------------------------------------------------------------------
// function : dimswap_int16()
// purpose  : perform vertical row flip
// revised  : 2013-11-09T10:10:00-07:00,ruby,TwoMed
// note     : here this is specifically only for int16
void dimswap_int16(int16 *A)
{

  int   i;
  int   x, y;
  int   ix, iy;

  // uint8  C[MODIS_1KM_NS_PIXELS];
  // uint8  *pixTemp_p = NULL ; // alloc off the heap, not the stack
  int16  *pixTemp_p = NULL ; // alloc off the heap, not the stack
 

  x = 0; y = MODIS_1KM_EW_PIXELS-1;
  
  pixTemp_p = (int16 *)memCalloc(MODIS_1KM_NS_PIXELS, sizeof(int16));

    while(x < y)
    {
        for (i=0; i < MODIS_1KM_NS_PIXELS; i++)
        {
            ix = M_2D_B0(x,i,MODIS_1KM_EW_PIXELS);
            iy = M_2D_B0(y,i,MODIS_1KM_EW_PIXELS);
            // C[i] = A[ix];
            pixTemp_p[i] = A[ix];
            A[ix] = A[iy];
            // A[iy] = C[i];
            A[iy] = pixTemp_p[i] ;
        } // end:: i loop through MODIS pixels
    x++;
    y--;
  } // end::while x < y
  
  memFree(pixTemp_p);

} // end::dimswap_int16()


// ----------------------------------------------------------------------------
// function : dimswap_int8()
// purpose  : perform vertical row flip?
// revised  : 2013-11-09T10:10:00-07:00,ruby,TwoMed
// note     : This specifically for int8 datatypes
void dimswap_int8(int8 *A)
{

  int   i;
  int   x, y;
  int   ix, iy;

  // uint8  C[MODIS_1KM_NS_PIXELS];
  // uint8  *pixTemp_p = NULL ; // alloc off the heap, not the stack
  int8  *pixTemp_p = NULL ; // alloc off the heap, not the stack
  

  x = 0; y = MODIS_1KM_EW_PIXELS-1;
  
  pixTemp_p = (int8 *)memCalloc(MODIS_1KM_NS_PIXELS, sizeof(int8));

    while(x < y)
    {
        for (i=0; i < MODIS_1KM_NS_PIXELS; i++)
        {
            ix = M_2D_B0(x,i,MODIS_1KM_EW_PIXELS);
            iy = M_2D_B0(y,i,MODIS_1KM_EW_PIXELS);
            // C[i] = A[ix];
            pixTemp_p[i] = A[ix];
            A[ix] = A[iy];
            // A[iy] = C[i];
            A[iy] = pixTemp_p[i] ;
        } // end:: i loop through MODIS pixels
    x++;
    y--;
  } // end::while x < y
  
  memFree(pixTemp_p);

} // end::dimswap_int8()


// ----------------------------------------------------------------------------
// function :set_land_mask_1km()
// purpose  : using 9km Plant Function Type (PFT) veg mask {0|1} image, 
//            OUTPUT raster's pixels to out-of-domain value for those that apply, e.g. 0
//            for both fpar and qc. PENDING: ndviImg_p WILL BE DEPRECATED SOON
// inputs   : uint8* mask
//            uint8 *fpar 
//            uint8 *ndvi 
//            uint8 *qc
// revised  : 2021-07-14, EvalDiag, deprecate ndviImg_p, eventually support fparImg_p from MODIS or VIIRs
// revised  : 2014-08-19T16:08, changed call interface, see above
// revised  : 2013-11-08, original.
// author   : Joe glassy
// research code : lucas jones
// notes    : <JMG> 2014-08-19, changed call interface to pass in ease_p->ndvi and
//            exclude the deprecated extra_qc
// notes    :  set_land_mask_1km(ease_p->mask, ease_p->fpar, ease_p->ndvi, ease_p->qc );
//
// notes    : FILL_VALUE 255, and, OUT_OF_DOMAIN_VALUE      254
//
void set_land_mask_1km(uint8 mask[], uint8 *fparImg_p, int16 *ndviImg_p, uint8 *fparQcImg_p,int8 *ndviQcImg_p)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  char fxName[256] ;
  int     i, j, in, jn;
  uint32  ij_9km, ij_1km;
  
    strcpy(fxName,__func__);
  

    xLogMsg("@A In(%s) starting for fparMethodCode %d ~ \n",fxName,sesCtl_p->fparMethodCode);

    // 9km grid loop
    for (i=0; i < NROW9KM; i++)
    {
        for (j=0; j < NCOL9KM; j++)
        {
            // derive the serial 1D index for 9km grid
            ij_9km = M_2D_B0(i,j,NCOL9KM);

            // iterate through pixels in this NESTED 1km grid loop
            // NCELL = 9, as the Number of nested 1km rows/cols within each 9km row/col
            // <JMG> 2021-07-14 TBD, rename "NCELL" to better describe what it is!
            for (in=0; in < NCELL; in++)
            {
                for (jn=0; jn < NCELL; jn++)
                {
                    ij_1km = M_2D_N9_B0(i,j,in,jn,NCOL1KM);
                    // NOTE: the FPAR 1km images out-of-domain value intentionally differs
                    // from its fill value and, is also intentionally different from the NDVI out-of-domain value
                    // <JMG> 2021-07-14, VERIFY how VIIRs handles this as well.
                    if(mask[ij_9km] == 0)
                    {
                       // <JMG> 2021-07-14 DEPRECATED branch on NDVI, share (MODIS|VIIRS) logic here
                       // FPAR and its QC fill values are both uint8 valued at 255, out-of-domain is 254
                       fparImg_p[ij_1km]    = OUT_OF_DOMAIN_VALUE; // Assign 254 out-of-domain cells
                       fparQcImg_p[ij_1km]  = OUT_OF_DOMAIN_VALUE; // (254) Assign value for out-of-domain cell
                        
                    } // end:: mask block
                } // end:: jn loop
            } // end: in loop, end nested 1km 2D grid loop
        
        } // end:: j loop
    } // end:: i loop ...9km 2D grid loop
  
  xLogMsg("@A In(%s) complete for fparMethodCode %d ~ \n",fxName,sesCtl_p->fparMethodCode);
  
} // end::set_land_mask_1km()


// ----------------------------------------------------------------
// function : reorg_VIIRS_Qc()
// signature: int32 nPix,uint8 *fparQc_p,uint8 *extraQc_p)
// inputs   : int32 nPix -- number of 500m pixels in this raw tile
//            fparQc_p -- ptr to 500m uint8 layer for fparQc standard from VIIRS
//            extraQc_p -- ptr to raw 500m uint8 layer for extraQc from VIIRS 
// Purpose : Reorganize QC pixel contents as input to L4CMOD, and passed thru.
// Notes   : uses L4CAPI's frm_utilGetBits() frm_utilSetBits() for bit manip.
// Notes   : We need to re-define sensorId bit as 0 signifying VIIRS by default
// revised : 2021-10-15T16:08 <JMG> replaced frm_utilSetBits() with frm_utilAssignBits()
//           in reorg_VIIRS_Qc() function.
// revision: 2021-10-06T09:44
// revision: 2021-10-05T15:00
// history : Replaced frm_utilSetBits() with frm_utilAssignBits() by <JMG>
//           in reorg_VIIRS_Qc() function.
// history : orig 2021-10-05T15:36
/*
 NOTE: If there is no good estimation in this composite, this value is set to 255!
 STANDARD uint8 VIIRS qc word layout: 
 FparLai_QC 8 BITFIELDS IN 8 BITWORD
0-2	Algorithm path (equivalent to SCF QC in MODIS, see below)
	000=0 Main (RT) algorithm used, best result possible (no saturation)
	001=1 Main (RT) algorithm used, saturation occured. Good, very usable.
	010=2 Main algorithm failed due to bad geometry, empirical algorithm used
	011=3 Main algorithm faild due to problems other than geometry, empirical algorithm used
	100=4 Pixel not produced at all, value coudn't be retrieved (possible reasons: bad L1B data, unusable VNP09GA data, water pixel)
3-3	Dead detector
	0 = Both red and NIR detectors are fine
	1 = At least one band has dead detector
4-7	Biome type
	0000=0 Water
	0001=1 Grasses/cereal crops
	0010=2 Shrubs
	0011=3 Broadleaf crops
	0100=4 Savanna
	0101=5 Evergreen broadleaf forest
	0110=6 Deciduous broadleaf forest
	0111=7 Evergreen needleleaf forest
	1000=8 Deciduous needleleaf forest
	1001=9 Unvegetated
	1010=10 Urban
	1011=11 Unclassified
	1100=12 Filled Value
	*
 FparExtra_QC 7 BITFIELDS IN 8 BITWORD (Extracted from VNP09GA)
0-1 Cloud detection and confience
	00 = 0 Confident clear
	01 = 1 Probably clear
	10 = 2 Probably cloudy
	11 = 3 Confident cloudy
2-2 Cloud shadow
	0 = 0 No cloud shadow
	1 = 1 Shadow
3-3 Thin cirrus flag
	0 = 0 No
	1 = 1 Yes
4-5 Aerosol quantity
	00 = 0 Climatology
	01 = 1 Low
	10 = 2 Average
	11 = 3 High
6-6 Snow/Ice
	0 = 0 No
	1 = 1 Yes 
 ------------------------------------------------------------------
 MODIS:  FparLai_QC 5 BITFIELDS IN 8 BITWORD

 MODLAND_QC START 0 END 0 VALIDS 2
 MODLAND_QC   0 = Good Quality (main algorithm with or without saturation)
 MODLAND_QC   1 = Other Quality (back-up algorithm or fill value)
 SENSOR START 1 END 1 VALIDS 2
 SENSOR       0  = Terra
 SENSOR       1  = Aqua
 DEADDETECTOR START 2 END 2 VALIDS 2
 DEADDETECTOR 0 = Detectors apparently fine for up to 50% of channels 1,2
 DEADDETECTOR 1 = Dead detectors caused >50% adjacent detector retrieval
 CLOUDSTATE START 3 END 4 VALIDS 4 (this inherited from Aggregate_QC bits {0,1} cloud state)
 CLOUDSTATE   00 = 0 Significant clouds NOT present (clear)
 CLOUDSTATE   01 = 1 Significant clouds WERE present
 CLOUDSTATE   10 = 2 Mixed cloud present on pixel
 CLOUDSTATE   11 = 3 Cloud state not defined,assumed clear
 SCF_QC START 5 END 7 VALIDS 5
 SCF_QC       000=0 Main (RT) algorithm used, best result possible (no saturation)
 SCF_QC       001=1 Main (RT) algorithm used, saturation occured. Good, very usable.
 SCF_QC       010=2 Main algorithm failed due to bad geometry, empirical algorithm used
 SCF_QC       011=3 Main algorithm faild due to problems other than geometry, empirical algorithm used
 SCF_QC       100=4 Pixel not produced at all, value coudn't be retrieved (possible reasons: bad L1B data, unusable MOD09GA data)

*/
int reorg_VIIRS_Qc(int32 nPix,uint8 *viirsFparQc_p,uint8 *viirsExtraQc_p)
{
 int32 nCorrected = 0 ;
 int32 ithPix = 0 ;
 uint8 destQcWordPix = 0 ; // destination (adjusted) QC result uint8 word to correct
 uint8 rawQcWord = 0 ;
 uint8 rawExtraQcWord = 0 ;
 uint8 modlandQc = 0 ; 
 uint8 sensorId  = 0 ; // <JMG> ASSERT: we need to redefine 0 as indicating VIIRS by default
 // contributed by VIIRS std Qc
 uint8 flagAlgorPath    = 0 ;    // from {0-2} VIIRS std Qc
 uint8 flagDeadDetector = 0 ;    // from {3-3} VIIRS std Qc
 uint8 flagBiomeType    = 0 ;    // from {4-7} VIIRS std Qc
 // contributed by VIIRS extraQc
 uint8 flagCloudConfidence = 0 ; // from {0-1} VIIRS extraQc 
 uint8 flagCloudShadow  = 0 ;    // from {2-2} VIIRS extraQc
 uint8 flagThinCirrus   = 0 ;    // from {3-3} VIIRS extraQc
 uint8 flagAerosolQuant = 0 ;    // from {4-5} VIIRS extraQc
 uint8 flagSnowIce      = 0 ;    // from {6-6} VIIRS extraQc, boolean
 
	// Usage template for 8 bit word:
	for(ithPix=0; ithPix < nPix; ithPix++)
	{
		rawQcWord      = viirsFparQc_p[ithPix]  ; // fetch VIIRS whole std Qc word
		rawExtraQcWord = viirsExtraQc_p[ithPix] ; // fetch VIIRS whole extraQc word
		
		// EXTRACTING PHASE
		// from VIIRS std qc word, extract each flag entity
		flagAlgorPath       = (uint8)frm_utilGetBits(0,2,rawQcWord);
		flagDeadDetector    = (uint8)frm_utilGetBits(3,3,rawQcWord);
		flagBiomeType       = (uint8)frm_utilGetBits(4,7,rawQcWord);
		// from VIIRS extraQc word, extrac each flag entity
		flagCloudConfidence = (uint8)frm_utilGetBits(0,1,rawExtraQcWord);
		flagCloudShadow     = (uint8)frm_utilGetBits(2,2,rawExtraQcWord);
		flagThinCirrus      = (uint8)frm_utilGetBits(3,3,rawExtraQcWord);
		flagAerosolQuant    = (uint8)frm_utilGetBits(4,5,rawExtraQcWord);
		flagSnowIce         = (uint8)frm_utilGetBits(6,6,rawExtraQcWord);
    
		// BIT-PACKING PHASE
		destQcWordPix = 0 ; 
		// <JMG> 2021-10-15T14:08, now using AssignBits in place of SetBits
		modlandQc = (flagAlgorPath <= 1)? 0 : 1 ;
		destQcWordPix =(uint8)frm_utilAssignBits(0,0,&destQcWordPix,sizeof(uint8),
		                        &modlandQc,sizeof(uint8)) ;
		// ASSERT: sensorId when "0" stands for VIIRS not MODIS...VERIFY HOW WE WANT THIS
		destQcWordPix =(uint8)frm_utilAssignBits(1,1,&destQcWordPix,sizeof(uint8),
		                        &sensorId, sizeof(uint8)) ;		                        
		destQcWordPix =(uint8)frm_utilAssignBits(2,2,&destQcWordPix,sizeof(uint8),
		                        &flagDeadDetector,sizeof(uint8)) ;
		// MODIS "cloud state" verify equivalence to VIIRS CloudConfidence semantics?		                        
		destQcWordPix =(uint8)frm_utilAssignBits(3,4,&destQcWordPix,sizeof(uint8),
		                        &flagCloudConfidence,sizeof(uint8)) ;
		// Substitute VIIRS AlgorPath as prior SCF_QC in bits 5,7		                        
		destQcWordPix =(uint8)frm_utilAssignBits(5,7,&destQcWordPix,sizeof(uint8),
		                        &flagAlgorPath,sizeof(uint8)) ;		                        		                        

		// ASSIGNMENT PHASE (over-writes prior contents of VIIRS uint8 Qc word
		viirsFparQc_p[ithPix] = destQcWordPix; // RELOAD the adjusted QC WORD back into 500m VIIRS tile
		
		nCorrected++ ;   // update tally of adjusted QC pixels
		
	} // end;: ithPix loop transferring QC bits from source input to dest tile

    // return count of corrected Qc pixels in case confirmation is needed
	return nCorrected ;
	
} // end::reorg_VIIRS_Qc()

// ----------------------------------------------------------------
// function: tile_loop()
// inputs  : tileref_struct *tileref_p, char datestr[], char yeardoy[],
//           char *fnames[], int ntiles, ease_struct *ease
// Purpose : Iterate through all 286 MODIS 10-deg tiles, mosaic and reprojecting each
// Notes   : Changed call interface to accept pointers, for wrapping efficiency,
//           and moved cont_p reference inside.
// revision: 2021-07-21, added 1km and 500m extra_qc in anticipation of QA/QC logic usage
// revision: 2021-07-14T18:03, eval-diagnostic code prep for VIIRs integration
// revision: 2017-01-26, note that when Collection 6 500m MOD15AH input tiles are specified
//           there are typically 274 tiles making up a fully populated set, not 286 as for 1KM default res.
// revision: 2013-11-07
// history : 2021-07-14, now use the HDFEOS5 HE5_fx() for VIIRs and HDFEOS2 for MODIS
// history : 2021-07-18 Added 'calloc' and 'free' of extra_qc to match alloc, accomodating potential use
// author  : joe glassy
int tile_loop(char *datestr, char *yeardoyString, char *inputTileList[], int ntiles) 
{
    // imports. Fetch larger container packages....
    ImgPkg *imagePkg_p = l4c_GetImagePackage(); // import in-memory INPUT image layers
    SesCtl *sesCtl_p   = l4c_GetSesCtl(); // import session controls
    
    int32 ithTile    = 0 ;
    int32 rsltStatus = 0 ;
    
    // char fpath[MXSTRLEN] = {0};
    // char *fpar_p = NULL ;
    // char memNamBufr[256] = {0};
    char fxName[256] = {0}; // fxName to register with __func__
    char workPath[FILENAME_MAX] = {0};
    char currTilePath[FILENAME_MAX] = {0};
    char inputLabel[32] = {0};

    // Establish equiv symbols, but as pointers, that resemble legacy instances
    ease_struct    *ease_p    = &imagePkg_p->ease ; // set convenience pointers
    tileref_struct *tileref_p = &imagePkg_p->tileref;

    // <LAJ> 2017-02-01 REMOVE:  tile_struct    *tile_p    = &imagePkg_p->tile ; 
    // ptr to large container object -- tile, for convenience 
    // <LAJ> 2017-02-01 ADD below. 
    // imagePkg_p now accomodates both simultaneous 1km and 500m tiles
    tile_struct *tile_1km_p  = &imagePkg_p->tile_1km ; 
    tile_struct *tile_500m_p = &imagePkg_p->tile_500m; 

    // REGISTER fx tile_loop
    strcpy(fxName,(char *)__func__); 

    // Set input label as FPAR or VIIRS and assign nElemQa for new Qa bit transfer as below
    if(sesCtl_p->fparMethodCode == UseMODISInputs)
    {
         strcpy(inputLabel,"MODIS FPAR");
    } else {
         strcpy(inputLabel,"VIIRs FPAR");
    }

    // <JMG> 2021-07-14, Note that for Collection 6, BOTH the MODIS and VIIRs Fpar are 500m
    // <JMG> 2017-01-26, adjusted to accomodate either 500m or 1KM MODIS inputs, to 
    //   allocate MODIS 500m|1km Sinusoidal (1200x1200)|(2400x2400) image layers, 
    //   just once for re-use.
    //  <JMG> 2021-07-14, TBD re-conform we will never need to compute lat/lon at 500m
    //  <LAJ> 2017-02-01 NOTE: These only need to be for 1km, we will never need to compute lat/lon for 500m
    //  <LAJ> 2017-02-01 ADD: Convert 'tile_p' to 'tile_1km_p' 
    xLogMsg("@A In(%s): INIT : allocating 1KM tile.lat for (%s) res of %d meters ~\n",fxName,inputLabel,sesCtl_p->fparResolutionMeters);
    tile_1km_p->lat      = memObjCalloc("tile_1km.lat",sizeof(float64),sesCtl_p->modis_Tile_1km_SumPixels,IsFloat64,fxName,__FILE__);
    
    xLogMsg("@A In(%s): INIT allocating 1KM tile.lon for (%s) res of %d meters ~\n",fxName,inputLabel,sesCtl_p->fparResolutionMeters);
    tile_1km_p->lon      = memObjCalloc("tile_1km.lon",sizeof(float64),sesCtl_p->modis_Tile_1km_SumPixels,IsFloat64,fxName,__FILE__);
         
    xLogMsg("@A In(%s): mode(%s) allocating 1KM tile.fpar of %d pixels~\n",fxName,inputLabel,sesCtl_p->modis_Tile_1km_SumPixels);
    tile_1km_p->fpar = memObjCalloc("tile_1km.fpar",sizeof(uint8)  ,sesCtl_p->modis_Tile_1km_SumPixels,IsUint8,fxName,__FILE__);
        
    xLogMsg("@A In(%s): mode(%s) allocating 1KM tile.qc as uint8 of %d pixels ~\n",fxName,inputLabel,sesCtl_p->modis_Tile_1km_SumPixels);
    tile_1km_p->qc   = memObjCalloc("tile_1km.qc",sizeof(uint8)  ,sesCtl_p->modis_Tile_1km_SumPixels,IsUint8,fxName,__FILE__);
    // <JMG> 2021-07-18, added extra_qc in case it is needed    
    xLogMsg("@A In(%s): mode(%s) allocating 1KM tile.extra_qc as uint8 of %d pixels ~\n",fxName,inputLabel,sesCtl_p->modis_Tile_1km_SumPixels);
    tile_1km_p->extra_qc = memObjCalloc("tile_1km.extra_qc",sizeof(uint8)  ,sesCtl_p->modis_Tile_1km_SumPixels,IsUint8,fxName,__FILE__);
    
    //  <LAJ> 2017-02-01 ADD: Memory allocation for tile_500m_p->fpar and tile_500m_p->fpar_qc arrays of size 2400x2400.  
    //       Should be done where-ever tile_1km_p->fpar, etc. is allocated.

     // <JMG> 2017-02-09 added memory allocation for 500 m processing when this option in force        
     // note these arrays are always allocated in addition to 1km arrays above.
     if( sesCtl_p->fparResolutionMeters <= 500)
     {
          xLogMsg("@A In(%s): mode(%s) allocating 500m tile.fpar of %d pixels~\n",fxName,inputLabel,sesCtl_p->modis_Tile_500m_SumPixels);
          tile_500m_p->fpar = memObjCalloc("tile_500m.fpar",sizeof(uint8),sesCtl_p->modis_Tile_500m_SumPixels,IsUint8,fxName,__FILE__);
          
          xLogMsg("@A In(%s): mode(%s) allocating 500m qc as uint8 of %d pixels ~\n",fxName,inputLabel,sesCtl_p->modis_Tile_500m_SumPixels);
          tile_500m_p->qc   = memObjCalloc("tile_500m.qc"  ,sizeof(uint8),sesCtl_p->modis_Tile_500m_SumPixels,IsUint8,fxName,__FILE__);
          
          // <JMG> 2021-07-18T16:41, allocate for future QC/QC use
          xLogMsg("@A In(%s): mode(%s) allocating 500m extra_qc as uint8 of %d pixels ~\n",fxName,inputLabel,sesCtl_p->modis_Tile_500m_SumPixels);
          tile_500m_p->extra_qc = memObjCalloc("tile_500m.extra_qc"  ,sizeof(uint8),sesCtl_p->modis_Tile_500m_SumPixels,IsUint8,fxName,__FILE__);
     }

    xLogMsg("@A In(%s): mode(%s) tile_loop: processing %d (%d meter resolution) input Sinusoidal tiles~\n",fxName, inputLabel,
                  ntiles, sesCtl_p->fparResolutionMeters);

    // Iterate:  through set of tiles for this one 8-day compositing period
    for (ithTile=0; ithTile < ntiles; ithTile++)
    {
        // retrieve tile coord info, store in tile_p->{h|v}str ?
        strcpy(workPath, l4c_strGetBaseNamePath(inputTileList[ithTile],workPath,'/'));
        strcpy(currTilePath,inputTileList[ithTile]) ;
        // <JMG> 2017-02-09, TBD --check with Lucas if we refer to "1km" or "500m" here..
        //  <LAJ> 2017-02-01 NOTE: Ensure that parse_filename() doesn't get tripped
        // up on '/pilv5/' or another other locations with 'v' or 'h' in the path name.
        parse_filename(workPath, "h", &tile_1km_p->h, tile_1km_p->hstr);
        parse_filename(workPath, "v", &tile_1km_p->v, tile_1km_p->vstr);
        // xLogMsg("@A In(%s) TilePath(%s) H(%s) V(%s) YYYYddd yeardoy initialized to [%s]~\n",fxName, workPath,tile_1km_p->hstr, tile_1km_p->vstr, yeardoyString);
        
        // retrieve date info from first tile (same for all tiles)
        if (ithTile == 0)
        {
            // Extract just the basename tiles path and store in workPath 
            strcpy(workPath, l4c_strGetBaseNamePath(inputTileList[ithTile],workPath,'/'));
            get_yeardoy_from_filename( workPath, yeardoyString);
            xLogMsg("@A In(%s) TilePath(%s) YYYYddd yeardoy initialized to [%s]~\n",fxName,workPath,yeardoyString);
        } // end:: block if FIRST TILE, used to seed date into yeardoyString
        
        // CLEAR Lat,Lon buffers (always at 1km for now)
        // <LAJ> 2017-02-01 Now referring to orig 'tile_p' as 'tile_1km_p' and clearing these
        memset(tile_1km_p->lat     , '\0', sizeof(float64) * sesCtl_p->modis_Tile_1km_SumPixels );
        memset(tile_1km_p->lon     , '\0', sizeof(float64) * sesCtl_p->modis_Tile_1km_SumPixels );
        
        // <LAJ> 2017-02-01 REMOVE: References to 'tile_p->fpar, fpar_qc'
        // <LAJ> 2017-02-01 ADD: References to 'tile_1km_p' and 'tile_500m_p'
        
        // <JMG> 2021-07-14, TBD we still need to verify logic here
        // DEPRECATED: if(sesCtl_p->gppMethodCode == IsFparSrcForGpp)
        
        memset(tile_1km_p->fpar       , '\0', sizeof(uint8) * sesCtl_p->modis_Tile_1km_SumPixels);
        memset(tile_1km_p->qc         , '\0', sizeof(uint8) * sesCtl_p->modis_Tile_1km_SumPixels );
        // <JMG> 2021-07-24, enabled CLEAR op on 1km extra_qc to match that of fpar,qc variables
        memset(tile_1km_p->extra_qc, '\0', sizeof(uint8) * sesCtl_p->modis_Tile_1km_SumPixels );
        
       // ASSERT: also handle special case of 500m support, clear 500m buffers
       if( sesCtl_p->fparResolutionMeters <= 500) 
       {
              memset(tile_500m_p->fpar    , '\0', sizeof(uint8) * sesCtl_p->modis_Tile_500m_SumPixels );
              memset(tile_500m_p->qc      , '\0', sizeof(uint8) * sesCtl_p->modis_Tile_500m_SumPixels );
              memset(tile_500m_p->extra_qc, '\0', sizeof(uint8) * sesCtl_p->modis_Tile_500m_SumPixels );
       }
            
        // <JMG> 2017-02-09 added "if" branch based on fparResolutionMeters for session
        // <JMG> 2017-01-26 verify that both 500m or 1KM tiles now supported in I/O
        // Read current MODIS Sinusoidal 1200x1200 OR (2400x2400) tile into buffer
        if(sesCtl_p->fparResolutionMeters >= 1000)
        {
            // xLogMsg("@D In(%s) tile_loop: READING 1km INPUT TILE %d of %d ~\n",fxName,ithTile,ntiles);
            
            l4c_read_hdfeos2(inputTileList[ithTile], tile_1km_p, ithTile, ntiles);
            
            // GET 1 KM LAT LON USING THE 500 M TILE CORNERS laj 01-26-2017
            // SciCode: get_sinusoid_latlon(tile_500m.upleft, tile_500m.lowright, tile_1km.lat, tile_1km.lon, &tile_1km.dx, &tile_1km.dy);

            // <JMG> 2017-02-13T18:03, place call to get_sinusoid_latlon() here to exactly
            // match Lucas science code.
            // Retrieve the current pixel's latitude and longitude coordinate
            // and assign the pixel wise delta for x and y
            get_sinusoid_latlon(tile_1km_p->upleft, tile_1km_p->lowright, 
                       tile_1km_p->lat   , tile_1km_p->lon, &tile_1km_p->dx   , &tile_1km_p->dy);
            
            // <JMG> 2017-02-09T17:51 verify which res "tileref_p" and "tile_p" here should be!
            // Insert (mosaic component) tile into single 1km EASEGRID v2 grid result image
            pop_tile_to_1kmease2(tileref_p, tile_1km_p, ease_p);

        } else {  // ASSERT: all other cases assume 500m inputs
            
            // xLogMsg("@D In(%s) tile_loop: READING 500M INPUT TILE %d of %d ~\n",fxName,ithTile,ntiles);
            if(sesCtl_p->fparMethodCode == UseMODISInputs)
               l4c_read_hdfeos2(inputTileList[ithTile], tile_500m_p, ithTile, ntiles); 
            else if(sesCtl_p->fparMethodCode == UseVIIRsInputs)
               l4c_read_hdfeos5(inputTileList[ithTile], tile_500m_p, ithTile, ntiles); 
            else
               xLogMsg("@F fparMethodCode (%d) NOT YET SUPPORTED.Only MODIS and VIIRS are.\n~",
                   sesCtl_p->fparMethodCode);

			      // <JMG> 2021-10-05, added for VIIRS integration Qc support
            // <KAE> 2021-10-12 This should only occur for VIIRS instrument mode
            if ( strcmp(sesCtl_p->fparMethodStr, "VIIRS") == 0 ) {
              rsltStatus = reorg_VIIRS_Qc(sesCtl_p->modis_Tile_500m_SumPixels,
                             tile_500m_p->qc,          // VIIRS 500m tile qc...
                             tile_500m_p->extra_qc) ;  // VIIRS 500m tile extraQc...
            }
            
            // <JMG> 2017-02-13T18:03, place call to get_sinusoid_latlon() here to exactly
            // match Lucas science code.
            get_sinusoid_latlon(tile_500m_p->upleft, 
                                tile_500m_p->lowright, 
                                tile_1km_p->lat   , tile_1km_p->lon,
                                &tile_1km_p->dx   , &tile_1km_p->dy);
            
            // <LAJ> 2017-02-01 ADD:  Function call to resample_tile_500m_to_1km() - (modify pointers at interface)
            // xLogMsg("@A In(%s) tile_loop: RESAMPLING 500m to 1km ~\n",fxName);
            // <JMG> 2021-07-23T17:53, within resample_tile xxx here we now call resample_fpar_window TWICE, once
            // for legacy qc, and now once for extra_qc.
            resample_tile_500m_to_1km(tile_500m_p, tile_1km_p);
            
            // <JMG> 2017-02-10T17:12-- Inserts (mosaic component) tile into 1km EASEGRID v2 grid
            pop_tile_to_1kmease2(tileref_p, tile_1km_p, ease_p);
        }
        
        // <JMG> 2021-07-13, temporarily changed from 50 to 10 here for granular diagnostics
        if(ithTile % 50 == 0)
        {
            xLogMsg("@A In(%s) mode(%s) tile [%s] %3d of %3d~\n",fxName, inputLabel, inputTileList[ithTile],ithTile, ntiles);
        }
        
    } // end:: tile loop


    // Free the shared (re-usable) MODIS Sinusoidal 1km 1200x1200 layers.
    // <LAJ> 2017-02-01  REMOVE:  References to 'tile_p' 
    // <LAJ> 2017-02-01  ADD:  References to 'tile_1km_p' and 'tile_500m_p' 

    if(sesCtl_p->fparResolutionMeters <= 500)
    {
        xLogMsg("@A In(%s) memObjFree mem for tile_500m.fpar and tile_500m.qc",fxName);
        memObjFree("tile_500m.fpar", NULL);
        memObjFree("tile_500m.qc"  , NULL);
        // <JMG> 2021-07-18 Added 'free' of extra_qc to match alloc, accomodating potential use
        memObjFree("tile_500m.extra_qc",NULL);
    }
    
    // <JMG> 2017-02-10 the 1km layers are used regardless, in both 1km and 500m cases
    xLogMsg("@A In(%s) memObjFree mem for tile_1km and tile_1km_qc",fxName);
    memObjFree("tile_1km.fpar"    , NULL);
    memObjFree("tile_1km.qc"      , NULL);
    // <JMG> 2021-07-18, added "free" of 1KM resolution extra_qc to match calloc above
    memObjFree("tile_1km.extra_qc", NULL);
  
  // lat,lon always at 1km as before...
  xLogMsg("@A In(%s) memObjFree mem for tile_1km.lat and tile_1km.lon",fxName);
  memObjFree("tile_1km.lat", NULL);
  memObjFree("tile_1km.lon", NULL);
  
  // return count of tiles processed
  return(ntiles);
  
} // end:: tile_loop() 
