// ---------------------------------------------------------------
// module       : l4c_session.c
// purpose      : Session management functions for l4c_MODIS.ex
// project      : NASA SMAP L4C science development.
// dependencies : HDF4, HdfEos , HDF5 and UM Framework (libfrm.a) libraries
// revision     : 2021-07-18T17:04, updated for EVAL/DIAG of VIIRs integration
// revision     : 2017-03-20T16:57, in l4c_CloseSession() added calls to
//                nsbReleaseInstance(nsbObj_p), and  pkgMgr_ReleaseMgr(pkgMgr_p)
//                to properly release all interior heap objects owned by the
//                pkgMgr and nsbOverseers
// revision     : 2017-02-13T18:41, verify final datestamped log file is being re-named
// revision     : 2014-10-14T12:49:00, implemented L4C_ISO_MetaElements object
//                and l4c_GetISO_MetaElements() accessor, to match L4CMDL and others
// history      : 2014-08-26T10:20:00, changed macro name AppVers to L4C_MODIS_APP_VERS to
//                harmonize with l4c_DAILY and l4c_MET
// history      : 2014-08-19T13:08, added call to l4c_readAndWriteLatLon_9kmCoordVars() and
//                changed imagePkg to match implementation in l4c_MET
// history      : 2014-07-29T14:05, integration with v5.0.1
// history      : 2013-11-18T09:36:00, corrected bug where OUTPUT root dir not prepended to output path
// history      : 2013-11-18T09:34:00, corrected bug where OUTPUT root dir not prepended to output path
//              : 2013-11-08T08:32:00, initial version delivered to NCCS
// author       : jglassy
// notes        : Reminder: session log facility is now in l4c_MODIS_log.c, AS xLogMsg()
//
// ---------------------------------------------------------------

#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"


// Image package is a container for various in-memory input , output images
// external callers gain visibility to it via l4c_GetImagePackage()
// Now also contains long-lived images and NsbMgrOverseer and PkgMgrOverseer
static ImgPkg imgPackageMaster;

// ------------- Session Control Structure (new) -----------------------
static SesCtl sesCtl ;
static L4C_ISO_MetaElements isoMetaMaster;

// ---------------------------------------------------------------------
// function : l4c_GetISO_MetaElements(void)
// purpose  : Retrieve public ptr to ISO metadata container, to support
//            more of the required ISO metadata, for pass-thru as needed
//            from L4CMOD to L4CMDL.
// inputs   : void
// returns  : char *, static (long-lived) string containing executables version id
// notes    : Implemented across all components as a macro {L4_C_|L4_C_MODIS_|L4C_MET_}
L4C_ISO_MetaElements *l4c_GetISO_MetaElements(void)
{
  // return pointer to master instance 
  return &isoMetaMaster ;
    
} // end:: l4c_GetISO_MetaElements()

// ---------------------------------------------------------------------
// function : l4c_GetExecutablVersionId()
// purpose  : Retrieve public ptr to executables datetime stamped version ID
// inputs   : void
// returns  : char *, static (long-lived) string containing executables version id
// notes    : Implemented across all components as a macro {L4_C_|L4_C_MODIS_|L4C_MET_}
char *l4c_GetExecutableVersionId(void)
{
 static char execSwVersion[64]= {0} ;
 
    // technically, this assignment only needs to be done once
    strcpy(execSwVersion,(char *)L4C_MODIS_VERS_TAG);

        // return pointer to long-lived version identifier
    return execSwVersion ;
    
} // end:: l4c_GetExecutableVersionId()

// ---------------------------------------------------------------------
// function : l4c_GetNsbOverseer()
// purpose  : Fetch public ptr to private master instance of NSB
// inputs   : void
// returns  : NsbObj*, ptr to master instance of NSB (overseer)
// notes    : NSB is the name value pair stringblock database
NsbObj *l4c_GetNsbOverseer(void)
{
  ImgPkg *imagePkg_p = &imgPackageMaster;
  
  return imagePkg_p->NsbMgrOverseer_p ;
    
} // end:: l4c_GetNsbOverseer()


// ---------------------------------------------------------------------
// function : l4c_GetPkgMgrOverseer()
// purpose  : Fetch public ptr to private master instance of PkgMgr
// inputs   : void
// returns  : PkgMgr *, ptr to master instance of PkgMgr (overseer)
PkgMgr *l4c_GetPkgMgrOverseer(void)
{
  ImgPkg *imagePkg_p = &imgPackageMaster;
  
  return imagePkg_p->PkgMgrOverseer_p ;
    
} // end::l4c_GetPkgMgrOverseer()


// ---------------------------------------------------------------
// function :l4c_GetSesCtl()
// purpose  : return a pointer to the master (private) session control structure
// revision : 2013-11-04T14:55
// history  : 2013-10-17T15:40:00, original refactoring
// note     : in progress...
SesCtl *l4c_GetSesCtl(void)
{
 return &sesCtl  ;
} // end:: l4c_GetSesCtl()


// ---------------------------------------------------------------
// function: l4c_GetImagePackage()
// inputs  : void
// returns : ImgPkg *, ptr to image package
// purpose : serve as a container for in-memory pointers to I/O images
//           and discrete data objects
ImgPkg *l4c_GetImagePackage(void)
{
 return &imgPackageMaster ;
} // end::l4c_GetImagePackage()



// ---------------------------------------------------------------
// function   : l4c_EventHighWaterMarkMsg()
// Purpose    : Announces the point at which current 
//              increment build (high-water-mark) testing has
//              been completed to.
// Inputs     : const char *thisFunc_p
// Returns    : double, cumulative sec since session began
// Notes      :  n/a
// Author     : joe glassy
double l4c_EventHighWaterMarkMsg(const char *thisFunction_p)
{
 SesCtl *sesCtl_p  = &sesCtl ;
 char timeStr[32] = {0};
 
 time_t secNow = time(0);
 // snapshot cumulative running time
 double cumulativeSec = difftime(secNow,sesCtl_p->sesStrtSec) ;
 
    xLogMsg("@A <%s> HIGH-WATER-MARK-COMPLETION %s (%.2f elap sec)~\n",
        thisFunction_p,  frm_utilGetCtime(timeStr,0), cumulativeSec );
    
    // PERFORM INTENTIONAL, ORDERLY EARLY SHUTDOWN OF SESSION   
    l4c_CloseSession(sesCtl_p);
    
    // logic will never get to here but this is shown for completeness
    return cumulativeSec ;
    
} // end:: l4c_EventHighWaterMarkMsg()


// ---------------------------------------------------------------
// function   : EventCompletionMsg()
// Purpose    : Announces that function 'thisFunc_p' has
//              completed, with a snapshot of current time,
// Inputs     : const char *thisFunc_p
// Returns    : double, cumulative sec since session began
// Notes      :  n/a
// Author     : joe glassy
double l4c_EventCompletionMsg(const char *thisFunction_p)
{
 SesCtl *sesCtl_p  = &sesCtl ;
 time_t secNow = time(0);
 char timeStr[32] = {0};
 
 // update a snapshot of cumulative running time
 double cumulativeSec = difftime(secNow,sesCtl_p->sesStrtSec) ;
 
    // <JMG> 4/28/2014, only display completion-message when
    // explicitly enabled, to reduce clutter...
    if(sesCtl_p->enableCompletionMsg == 1)
        xLogMsg("@A <%s> COMPLETED %s (%.2f elap sec)~\n",
            thisFunction_p,  frm_utilGetCtime(timeStr,0), cumulativeSec );
        
    return cumulativeSec;
    
} // end:: l4c_EventCompletionMsg()

// ---------------------------------------------------------------------
// function   : l4c_GetISO8601_datetime()
// inputs     : time_t usersSec -- optional, 0, or user provided seconds =time(0L)
//            : char *loc_dtStr_p -- resulting localtime string YYYY-MM-DDThh:mm:ss
//              char *utc_dtStr_p -- resulting UTC string YYYY-MM-DDThh:mm:000Z
// returns    : int, timezone difference in hours between UTC and local time (not adjusted for sign)
// purpose    : Retrieve an ISO8601 conformal datetime string, YYYY-MM-DDThh:mm:000Z for UTC
//              position 'startPos' and ending at bit position endPos.
// revised    : 2014-10-14T12:53, updated format to pass back ".000Z" ending
// history    : 2014-05-13T18:02:00,jmg
// notes      : Joe A. requested this form 2014-09-12T09:53:00.000Z not 2014-09-12T09:53:000Z
int l4c_GetISO8601_datetime(time_t usersSec,char *loc_dtStr_p,char *utc_dtStr_p)
{
 char fmtSpecStr[64]={0} ;
 char tzBuf1[16] = {0} ;
 char tzBuf2[16] = {0} ;
 
 time_t dtSecUsed ;
 struct tm *loc_tmPtr = NULL ; 
 struct tm  *gm_tmPtr = NULL ;
 int    tz_offset_hrdiff = 0;
  
    dtSecUsed = (usersSec <= 0)? time(0L): usersSec;
    
    // capture both the local time and UTC, GMT
    gm_tmPtr  = gmtime(&dtSecUsed);
    // <JMG> 2014-09-12T09:57, GMAP needs datetime stamp in 
    // exactly this format: 2014-09-12T09:57:00.000Z, presume
    // by rounding the 's' param to nearest even sec to result
    // in stamp ending with ".000Z" but we hard code in the .000Z for now
    strcpy(fmtSpecStr,"%Y-%2m-%2dT%H:%M:%S.000Z" );
    strftime(utc_dtStr_p, 64-1,fmtSpecStr,gm_tmPtr);
    strftime(tzBuf1,15,"%H", gm_tmPtr);
    // printf("\ngetISO8601datetime UTC [%s] tzBuf %s\n",utc_dtStr_p,tzBuf1);

    loc_tmPtr = localtime(&dtSecUsed);  
    strcpy(fmtSpecStr,"%Y-%2m-%2dT%H:%M:%S" );
    strftime(loc_dtStr_p,64-1,fmtSpecStr,loc_tmPtr);
    strftime(tzBuf2,15,"%H", loc_tmPtr);
    // printf("\ngetISO8601datetime MST [%s] tzBuf %s\n", loc_dtStr_p, tzBuf2);
    
    tz_offset_hrdiff = abs(atoi(tzBuf2) - atoi(tzBuf1))+1;
    // NOTE: need to correct time-zone hour "sign" to negative for West longitudes?
    // printf("\ntzBuf1 (%s) tzBuf2(%s) %d\n",tzBuf1,tzBuf2,tz_offset_hrdiff);
    
 return tz_offset_hrdiff ;
 
} // end:: l4c_getISO8601_datetime()


// ---------------------------------------------------------------
// function   : l4c_InitSession()
// Purpose    : Initializes session for a given task executable invocation. 
//            Actions include: opens the session logfile, logs initial
//             information, initializes the local memory manager/garbage collector,
//             snapshots starting time of session in seconds
// Inputs     : int argc,char *argv[],SesCtl *sesCtl_p
// Invocation : ./l4c_MODIS.ex <./l4c_MODIS.rip>
// Revision   : 2014-08-19, added call to l4c_readAndWriteLatLon_9kmCoordVars()
// Returns    : time_t --starting time in seconds.
// Notes      :  n/a
// Author     : joe glassy
time_t l4c_InitSession(int argc,char *argv[])
{
 SesCtl *sesCtl_p = &sesCtl ;
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // local pointer to support image package
 
 FILE *ripf = NULL; // handle for runtime input parameter (RIP) text file

 char timeBuffer[64]= {0} ;
 char fxName[256] = {0} ;
 char nvpBuf[FILENAME_MAX] = {0} ;
 char logPath[FILENAME_MAX];       // <JMG> 5/1/2014...
 
 int  t =0; // tile index
 int ithGrn =0;

 pid_t thisPID = 0 ;
 pid_t sesParentPID = 0 ;
 
    
    strcpy(fxName,__func__);
    
    if(argc <= 1)
    {
        fprintf(stderr,"@F l4c_session_init: missing required runtime input parameters (RIP) file!~\n");
        exit(-1);
    } else {
        // store path to runtime input parameters (RIP) file defining ALL configuration for this session
        strcpy(sesCtl_p->inputRipPath,argv[1]);
    }

    // set hard defaults
    sesCtl_p->sesStrtSec  = time(0L);
    sesCtl_p->sesEndiSec  = time(0L);
    sesCtl_p->compressLev = 3;     // fallback default
    sesCtl_p->enableCompletionMsg = 0 ; 

    // COMMON VERSION ASSIGNMENTS -- Retrieve hard-wired L4C versions
    strcpy(sesCtl_p->swVersion    ,l4c_GetExecutableVersionId());
    
    // swRelease is written as attribute L4_C_SW_DATETIME
    strcpy(sesCtl_p->swRelease, (char *)L4C_MODIS_VERS_TAG);
    
    strcpy(sesCtl_p->frmApiVersion, frm_GetApiVersion(nvpBuf));
    strcpy(sesCtl_p->pkgApiVersion, pkgMgr_GetVersion(nvpBuf));
    strcpy(sesCtl_p->nsbApiVersion, nsbGetVersion(nvpBuf));
    
    // (3) themes now
    imagePkg_p->nPkgThemes = L4CMOD_PKGMGR_THEME_COUNT ; 
    // set 9km dimensions, to match l4c_MET style ops
    imagePkg_p->nElem9km  = (NROW9KM * NCOL9KM);
    imagePkg_p->nElem1km  = (NROW1KM * NCOL1KM);

    sesCtl_p->nFparGranules = 0 ;
    // DEPCRECATE sesCtl_p->nVIIRsNDVI_granules 
    
    // register function to call atexit..
    atexit(memObjGarbageCollect);
    
    // Initialize the local memory manager and garbage collector
    memObjInit(stdout);
    // Toggle OFF the verbose memObj messaging for now
    memObjAdjustVerbose(0);

    // capture name of host we're running on
    memset(sesCtl_p->thisHost,'\0',256);
    gethostname(sesCtl_p->thisHost,255) ;
    // Capture name of user controlling terminal session. This will
    // fail if process is daemonized, but l4c_DAILY.ex is not normally
    // run as a daemon
    getlogin_r(sesCtl_p->thisUsername,128);
    
    // capture PID and any other diagnostic info
    thisPID = l4c_getPid(&sesParentPID);    

    // Open session log with fixed name, will likely close,rename as end of session
    // in order to datetime-stamp, etc.  
    strcpy(sesCtl_p->logInitialPath, "l4c_MODIS.log");
    xOpenLogfile( sesCtl_p->logInitialPath, sesCtl_p->logf);

    // Fetch and store name of runtime input parameter (RIP) file
    if(argc < 2)
        xLogMsg("@F l4c_MODIS invocation: <l4c_MODIS.rip> missing required RIP file!~ \n");

    // -------------------- SESSION STARTUP MESSAGE -------------
    xLogMsg("\n~");
    xLogMsg("=====================================================================~\n");
    xLogMsg("@A L4_C task[%s] started %s on host [%s] at pid %zu parents-pid %zu ~\n",
            argv[0],  frm_utilGetCtime(timeBuffer, sesCtl_p->sesStrtSec),
            sesCtl_p->thisHost, (size_t)thisPID,(size_t) sesParentPID);

    xLogMsg("@A Executable %s version %s ~ \n",argv[0], l4c_GetExecutableVersionId());      
    xLogMsg("@A In(%s) API: UM Framework version (%s)~ \n" , fxName,sesCtl_p->frmApiVersion);
    xLogMsg("@A In(%s) API: PkgMgr       version (%s)~ \n" , fxName,sesCtl_p->pkgApiVersion);
    xLogMsg("@A In(%s) API: NSB          version (%s)~ \n" , fxName,sesCtl_p->nsbApiVersion);

    xLogMsg("@A size of SesCtl is: %d ~\n", (int)sizeof(SesCtl));   

    xLogMsg("@A In(%s) host (%s) user: (%s) ~\n",
            fxName,sesCtl_p->thisHost, sesCtl_p->thisUsername);


    // Create MASTER NSB, noting other clients access this via
    // a call to *l4c_GetNsbOverseer(). Here we refer to the 
    // private, local instance of it.
    // ----------------------------------------------------------------
    // OPEN and PARSE the main RIP file, retrieve ALL runtime inputs
    if((ripf=fopen(sesCtl_p->inputRipPath,"rt"))==NULL)
    {
        xLogMsg("@F In(%s) missing required RUNTIME INPUTS (RIP) FILE (%s)~\n",
            fxName, sesCtl_p->inputRipPath);
        
    } else {
        xLogMsg("@A In(%s) Runtime input parameters (RIP) from %s ~\n",
                        fxName,sesCtl_p->inputRipPath);
    }
    
    // CREATE: Name-Value-Pair String Block (NSB) in-memory database instance
    imagePkg_p->NsbMgrOverseer_p = nsbCreateInstance("L4_C_MODIS_NSB_Manager");
   
    // NSB available, use it to parse and store all RIP name-value-pairs, 
    //   where individual key pair values are retrieved using access functions like:
    //      nsbGetValueByKey(nsbObj_p, "KEY",nsbBuffer) or GetValueByIndex()
    sesCtl_p->nTotalNvp = nsbParseIntoStringblock(l4c_GetNsbOverseer(),"",ripf);
        
    // CLOSE the runtime input parameters text file
    fclose(ripf);

    // CREATE: PKG MGR master instance as an object scoping tool for session. 
    //  Initially we declare 0 themes, since we'll manually create the actual
    //   themes using pkgMgr_CreateThemes() below
    imagePkg_p->PkgMgrOverseer_p = pkgMgr_CreateMgr("L4_C_MODIS_PkgMgr", 0, "L4_C_MODIS_Data");
    
    // Initialize all PkgMgr Themes and known properties, as many as possible
    // from external RIP name-value-pair definitions
    l4c_PkgMgr_Init();
    
    // COLLECT: first organize CORE name-value-pairs (Nvps).
    // This also populates the version ID's for all major FRM API mini-apis
    l4c_SesCollectCoreNvps(sesCtl_p);
    
    // COLLECT: reminder of NVPs into local variables, using expected keyword set
    //  note that all FPAR, VIIRs and NDVI input granules list qualification and
    //  sorting occurs here
    l4c_SesCollectNvps(sesCtl_p);

    // <JMG> 2021-07-14, Legacy deprecated here is left intact until cutover to new method
    // xLogMsg("@A In(%s) DATASOURCE for Fpar  (%s) Code(%d)~ \n" , fxName, 
    //            sesCtl_p->fparMethodStr,sesCtl_p->fparMethodCode);
    
    // <JMG> 2021-07-14 (add) new   
    xLogMsg("@A In(%s) DATASOURCE FPAR Method     (%s) Code(%d)~ \n" , fxName, 
                sesCtl_p->fparMethodStr,sesCtl_p->fparMethodCode);                
    
    // <JMG> 2017-01-26T13:48, added output of MODIS FPAR input tile resolution to log 
    xLogMsg("@A In(%s) INPUT FPAR Tile PIXEL RESOLUTION: %d meters~\n", fxName,sesCtl_p->fparResolutionMeters);
    if(sesCtl_p->fparResolutionMeters <= 500)
    {
        xLogMsg("@A In(%s) INPUT FPAR tile dimensions: rows %d columns %d~\n", fxName,
            fxName, sesCtl_p->modis_ns_500m_Pixels,sesCtl_p->modis_ew_500m_Pixels);
    } else {
        xLogMsg("@A In(%s) INPUT FPAR tile dimensions: rows %d columns %d~\n", fxName,
            fxName, sesCtl_p->modis_ns_1km_Pixels,sesCtl_p->modis_ew_1km_Pixels);
    }
    
    xLogMsg("@A In(%s) Total runtime name-value-pairs (NVPs) parsed (%d)~ \n" , fxName, sesCtl_p->nTotalNvp);

    xLogMsg("@A In(%s) FPAR Global 1km EASEGRID OUTPUT HDF5 file [%s]~\n",fxName, sesCtl_p->outputFparPath);
    xLogMsg("@A In(%s) Ancillary file used                       [%s]~\n",fxName, sesCtl_p->inputAncPath);
    
    xLogMsg("@A In(%s) FPAR Output root directory      (%s)~\n",fxName, sesCtl_p->outputFparRootDir);
    xLogMsg("@A In(%s) FPAR Session Log root directory (%s)~\n",fxName, sesCtl_p->logRootDir);
    xLogMsg("@A In(%s) FPAR Session Log                (%s)~\n",fxName, sesCtl_p->logInitialPath);
    
    // CREATE: New OUTPUT HDF5 FILE,  (parent container) and store new file_id for output HDF5 file
    xLogMsg("@A In(%s) CREATING new output HDF5 (%s)~ \n",fxName,sesCtl_p->outputFparPath);
    
    if((sesCtl_p->fparOutput_file_id = H5Fcreate(sesCtl_p->outputFparPath, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT)) < 0)
    {
        xLogMsg("@F In(%s) FATAL-ERROR failed to create FPAR OUTPUT (%s)w\n~",fxName,sesCtl_p->outputFparPath);
    }
    xLogMsg("@A In(%s) DONE Creating new HDF5 (%s) with id %d ~ \n",fxName,sesCtl_p->outputFparPath,sesCtl_p->fparOutput_file_id);

    // OPEN and READ ANCILLARY file to retrieve (4) tileref layers, 
    //  and the (1) EG 9km mask from SMAP_L4_C_Ancillary.h5 file
    l4c_readSmapAncillary() ;
    
    // Open and read CF latitude,longitude images,
    // CREATE new OUTPUT FILE, and immediately WRITE lat,lon into NEW DAILY output file
    //  as CF style "cell_lat,cell_lon" reference images
    l4c_readAndWriteLatLon_1kmCoordVars();
    
 // return start time in unix seconds
 return sesCtl_p->sesStrtSec;

} // end :: l4c_InitSession()


// ---------------------------------------------------------------
// function :l4c_Ses_LogLabelTime()
// inputs   : const time_t finalSec, char *finalLogPath
// returns  : char *, ptr to final log path
// revised  : 2014-02-13T09:41:00, patched in frm_utilGetCtime().
// Purpose  : Using initial log path, construct a FINAL datetime stamped
//            log file path that includes the datetime stamp, and return it.
// notes    : since initial log path (l4c_DAILY.log) is fixed, we don't pass
//            it in here explicitly.
char *l4c_Ses_LogLabelTime(const time_t finalSec,char *finalLogPath)
{
 // char timeBuffer[64] = {0} ;
 // buffer for copying log file to final destination
 char logWorkingPath[FILENAME_MAX] = {0};
 char tzQualifier[16] = {0} ; // time zone qualifier
 struct tm workingTime  ;
 int curYear,curMonth,curDayOfMonth, curHour,curMin, curSec;

    // ONLY AFTER working log is closed, datestamp it and copy it to final log location
    //  Function: struct tm * gmtime_r (const time_t *time, struct tm *resultp)
    // using thread-safe GMT time fx, for final labelling of log file currently
    // NOTE: check if this SHOULD be 'UTC' instead of GMT
#ifdef L4_C_USE_GMT
    gmtime_r(&finalSec, &workingTime);
    strcpy(tzQualifier,"UTC");
#else   
    localtime_r(&finalSec, &workingTime);
    strcpy(tzQualifier,"LCL"); // local time zone used...
#endif
    // de-reference the struct tm, noting subtle ways the
    // curYear and curMonth are actually defined as units SINCE a ref mark
    curYear       = workingTime.tm_year + 1900 ;
    curMonth      = workingTime.tm_mon  + 1;
    curDayOfMonth = workingTime.tm_mday;
    curHour       = workingTime.tm_hour ;
    curMin        = workingTime.tm_min  ;
    curSec        = workingTime.tm_sec  ;
    
    // NOTE: as an optional strategy, we could use the BASE NAME of the
    // current days L4CMOD output file, and append a .log extension to 
    // it: sesCtl_p->outputFparPath  (substitute extension ".h5" with ".log"
    // TBD: could change this to L4CMOD if requested by SMAP team or Joe A etc.
    sprintf(logWorkingPath,"l4c_MODIS_%4.4d-%2.2d-%2.2dT%-2.2d%-2.2d%-2.2d_%s.log",
               curYear, curMonth, curDayOfMonth, curHour,curMin,curSec, tzQualifier);
               
    strcpy(finalLogPath, logWorkingPath);

 return finalLogPath;   
} // end:: l4c_Ses_LogLabelTime()


// ---------------------------------------------------------------
// function :l4c_Ses_DatestampLog()
// inputs   : const time_t finalSec, char *finalLogPath
// returns  : char *, ptr to final log path
// revised  : 2017-02-13T18:41, verify final log is being renamed
// revised  : 2014-02-13T09:41:00, patched in frm_utilGetCtime().
// Purpose  : Using initial log path, produce a datetime stamped log
//            file, using l4c_Ses_LogLabelTime(), by copying the current
//            fixed name log to this new datetime stamped log filed 
//            as located in the proper user-defined final log destination dir.
// notes    : The sesCtl_p->logf MUST be flushed and closed prior to calling this.
int l4c_Ses_DatestampLog(SesCtl *sesCtl_p)
{
 char logWorkingPath[FILENAME_MAX] = {0};
 char sysCmdLogCopy[FILENAME_MAX*2] = {0} ;


    // note: could perform a 'mv' here instead of 'cp' but for now, to be safe, do a cp
    sprintf(sysCmdLogCopy,"cp -p %s %s\n" ,sesCtl_p->logInitialPath, sesCtl_p->logFinalPath);

    // Perform system OS (shell) call to copy the final file
    // Note: could migrate this to use a silent pipe instead of system() 
    system(sysCmdLogCopy);

    fprintf(stdout,"\n@A FINAL-log-path (%s)~\n",sesCtl_p->logFinalPath);
        
    return 0;   
    
} // end:: l4c_Ses_DatestampLog()


// ---------------------------------------------------------------
// function : l4c_CloseSession()
// inputs   : SesCtl *sesCtl_p
// Purpose  : Closes session for a given task executable invocation. Stops the runtime clock, logs timing, and 
//          program exit information and closes out the logfile. 
// Revision: 2017-03-20T16:57, investigating PkgMgr unfree memory
// Notes  : if a @F fatal error trapped, l4c_CloseSession() is preemptively called..
//
void l4c_CloseSession(SesCtl *sesCtl_p)
{
 char timeBuffer[64] = {0} ;
 char fxName[256] = {0};
 char logWorkingPath[FILENAME_MAX]={0};
 
 size_t highestSlot= 0;
 size_t totalBytesTracked=0;
 size_t hwmMem =0;
 herr_t ret = 0 ;
 
 // fetch pointer to container of pointers to legacy in-memory images
 ImgPkg *imagePkg_p = l4c_GetImagePackage();
 NsbObj *nsbObj_p = l4c_GetNsbOverseer();
 PkgMgr *pkgMgr_p = l4c_GetPkgMgrOverseer();

    strcpy(fxName,__func__);
    
    // close ancillary file
    ret =H5Fclose(sesCtl_p->anc_file_id);
    xLogMsg("@A In(%s) CLOSED ANCILLARY (%s) with status %d~\n",fxName, sesCtl_p->inputAncPath,ret);

    // close the primary output file itself
    ret = H5Fclose(sesCtl_p->fparOutput_file_id);
    xLogMsg("@A In(%s) CLOSED PRIMARY OUTPUT HDF5 file %d with ret %d~\n",fxName, sesCtl_p->fparOutput_file_id,ret);

    // Depending on if FPAR or NDVI collection used, free the appropriate granule list
    if(sesCtl_p->nFparGranules > 0)
        memFreeStringTbl(sesCtl_p->input_Fpar_pathList, sesCtl_p->nFparGranules); 
 
    // Free tile support layers, created via memObjCalloc() so use memObjFree() to release
    memObjFree("ease.mask" , imagePkg_p->ease.mask);
    memObjFree("tileref.v" , imagePkg_p->tileref.v);
    memObjFree("tileref.h" , imagePkg_p->tileref.h);
    memObjFree("tileref.r" , imagePkg_p->tileref.r);
    memObjFree("tileref.c" , imagePkg_p->tileref.c);
    
    // <JMG> 2021-07-14 CHANGE consolidated to eliminate NDVI branch here
    // Free primary l4c_MODIS pre-processor output image buffers.
    // Depending on if NDVI or FPAR used, free up primary buffer
    memObjFree("ease.fpar"    , imagePkg_p->ease.fpar);
    memObjFree("ease.qc"      , imagePkg_p->ease.qc);       
    
     // <JMG> 2017-03-20T16:57 Investigating PkgMgr related un-freed memory
    // Free the NSB overseer, legacy method:  memFree(nsbObj_p);
    nsbReleaseInstance(nsbObj_p);
    
    // <JMG> 2017-03-20T16:57 Investigating PkgMgr related un-freed memory
    // Legacy method: Free the PkgMgr overseer, uses:   memFree(pkgMgr_p)
    pkgMgr_ReleaseMgr(pkgMgr_p);
 
    xLogMsg("@A \n");
    memObjGetMetrics(&highestSlot,&totalBytesTracked,&hwmMem);
    xLogMsg("@A In(%s) sesssion memory resource: high-slot %zu total-bytes %zu high-water-mark %zu (%.1f Gb)~\n",
            fxName, highestSlot, totalBytesTracked, hwmMem, ((float)hwmMem/1024000000.0) );

    sesCtl_p->sesEndiSec = time(0L);
    sesCtl_p->sesElapSec = difftime(sesCtl_p->sesEndiSec,sesCtl_p->sesStrtSec);

    // Trigger final final garbage collection, to release all memory not otherwise
    // explicitly free'd
    memObjGarbageCollect();

    // Finally, copy the (flushed,closed) log file to a datetime
    // stamped instance of it in the final log location
    // ONLY AFTER working log is closed, datestamp it and copy it to final log location
    //  Function: struct tm * gmtime_r (const time_t *time, struct tm *resultp)
    // using thread-safe GMT time fx, for final labelling of log file currently
    strcpy(logWorkingPath, sesCtl_p->logInitialPath);
    l4c_Ses_LogLabelTime(sesCtl_p->sesEndiSec, logWorkingPath);
    strcpy(sesCtl_p->logFinalPath,logWorkingPath);
    
    // only uses a log root dir if one is defined
    // Step 1a: Do one-time-only path qualifications
    frm_ioQualifyPathWithRoot(sesCtl_p->logRootDir, sesCtl_p->logFinalPath);
    
    xLogMsg("@A In(%s) final log file will be (%s)~\n",fxName,sesCtl_p->logFinalPath);

    xLogMsg("@A In(%s) task [%s] version (%s) completed at [%s] requiring %.2lf sec~\n",
        fxName, (char *)L4C_MODIS_APP_TAG, l4c_GetExecutableVersionId(),
         l4c_getCtime(timeBuffer,sesCtl_p->sesEndiSec), 
        sesCtl_p->sesElapSec);

    // finally, close and flush the log file
    xCloseLog(sesCtl_p->logf);

    // now that log is closed, copy log to final destination
    l4c_Ses_DatestampLog(sesCtl_p) ;

    
} // end :: l4c_CloseSession()


// --------------------------------------------------------------------
// function :l4c_TIVDatatypeToDatatypeString()
// Purpose  : Convert a TIV type to a HDF4 style datatype string
//            identifier ("int32")
// Inputs   : const int TIV_datatypeId
char *l4c_TIVDatatypeToDatatypeString(const int TIV_datatypeId,char *typBuf)
{

        switch(TIV_datatypeId)
        {
        case IsInt8    : strcpy(typBuf,"int8"   ); break;
        case IsUint8   : strcpy(typBuf,"uint8"  ); break;
        case IsInt16   : strcpy(typBuf,"int16"  ); break;
        case IsUint16  : strcpy(typBuf,"uint16" ); break;
        case IsInt32   : strcpy(typBuf,"int32"  ); break;
        case IsUint32  : strcpy(typBuf,"uint32" ); break;
        case IsInt64   : strcpy(typBuf,"int64"  ); break;
        case IsUint64  : strcpy(typBuf,"uint64" ); break;
        case IsFloat32 : strcpy(typBuf,"float32"); break;
        case IsFloat64 : strcpy(typBuf,"float64"); break;
        default :
            xLogMsg("@E ERROR: l4c_TIVDatatypeToDatatypeString unsupported TIV type(%d)\n", 
              TIV_datatypeId);
        } // end:: TIV_datatypeId
        
        return typBuf ;
} // end::      l4c_TIVDatatypeToDatatypeString()


// ---------------------------------------------------------------------
// function: l4c_DiagImage()
// purpose : accept a raw binary (2D) numeric array of known TIV type and 
//           generate a simple diagnostic image (pgm if uint8, or
//           jpeg if float32) using a python support utility that
//           calls numpy,scipy to produce the actual image)
//
// inputs  : char *imgPrefix, 
//           char *outputPath,
//           int tivDatatype,
//           size_t nPixels,
//           int *dims,
//           void *imgData_p
// returns      : status 0
// dependencies : genDiagImg.py python script in current directory
//
int l4c_DiagImage(char *imgPrefix, char *outputPath,int tivDatatype,size_t nPixels, size_t *dims,void *imgData_p)
{
 char osCmd[FILENAME_MAX] = {0};    
 char inpAryPath[FILENAME_MAX] = {0} ;
 char tivTypeStr[32]      = {0} ;
 char typeBuffer[32] = {0};
 char fxName[256] = {0} ;
 char pyDiagPath[FILENAME_MAX] = {0} ; // full unambiguous path to genDiagImg.py script
 char utilDirPath[FILENAME_MAX] = {0} ; // prefix directory where genDiagImg.py located 
 
 size_t nImgBytes   = DFKNTsize(tivDatatype) * nPixels;
 uint8 *imgBuf_p = (uint8 *)imgData_p;
 int status = -1 ;
 
 strcpy(fxName,__func__);
 
 // first, verify that genDiagImg.py exists
 strcpy(utilDirPath,"."); // default for now...
 sprintf(pyDiagPath,"%s/%s",utilDirPath,"genDiagImg.py");
 
 if(!frm_ioDoesFileExist(pyDiagPath))
 {
    xLogMsg("@E In(%s) utilDirPath(%s). DiagnosticImage capability requires python support utility [%s] ~\n",
        fxName,  utilDirPath,pyDiagPath);
    return -1 ;
 }
 
 // convert numeric TIV (HDF4) datatype identifier to its string equivalent
 strcpy(tivTypeStr, l4c_TIVDatatypeToDatatypeString(tivDatatype,typeBuffer));
 
 // write out the raw numeric array as a file, "{imgLabel}.{datatype}"
 sprintf(inpAryPath,"%s.%s",imgPrefix,tivTypeStr);
 frm_ioWritebinary(inpAryPath,tivDatatype, nPixels, (uint8*)imgBuf_p);
 
 // the genDiagImg.py utility generates an image as well as 
 // displaying it immediately.  
 // python ../../l4c_DAILY/src/genDiagImg.py float32 1624 3856 smrz_mx.bin smrz_TEST.png 
 
 // convert the numeric TIV datatype to its string representation
 // like "float32" etc
 sprintf(osCmd,"python %s %s %zu %zu %s %s \n",
        pyDiagPath, tivTypeStr, dims[0], dims[1], inpAryPath, outputPath);
 xLogMsg("@A In(%s) osCmd[%s] ~\n",fxName,osCmd);

 system(osCmd); 

 return status ;
} // end:: l4c_DiagImage()


// ---------------------------------------------------------------------
// function: l4c_GenFreqDistrib()
// purpose : Accepts a raw binary (2D) numeric array of known TIV type and 
//           generate a simple diagnostic frequency distribution and 
//           (when available) produces an image of it. Uses numpy and
//           scipy.
// inputs       : char *imgPrefix, char *outputPath,int tivDatatype,
//              size_t nPixels, void *imgArray_p
// returns      : int, status
// dependencies : genFreqDistrib.py script, and Python with numpy, 
//                matplotlib, later scipy as well
int l4c_GenFreqDistrib(char *imgPrefix, char *outputPath,int tivDatatype,
            size_t nPixels, void *imgArray_p)
{
 char osCmd[FILENAME_MAX] = {0};    
 char inpAryPath[FILENAME_MAX] = {0} ;
 char tivTypeStr[32]      = {0} ;
 char typeBuffer[32] = {0};
 char fxName[256] = {0} ;
 size_t nImgBytes   = DFKNTsize(tivDatatype) * nPixels;
 uint8 *imgBuf_p = (uint8 *)imgArray_p;
 int status = -1 ;
 FILE *outf = NULL;
 
 strcpy(fxName,__func__);
 
 // first, verify that genDiagImg.py exists
 if(!frm_ioDoesFileExist("genDiagImg.py"))
 {
    xLogMsg("@E In(%s) requires python utility [%s] ~\n",
        fxName,"genDiagImg.py" );    
    return -1 ;
 }
 
 // Convert numeric TIV (HDF4) datatype identifier to its string equivalent
 strcpy(tivTypeStr,l4c_TIVDatatypeToDatatypeString(tivDatatype,typeBuffer));
 
 // Write out the raw numeric array as a file, "{imgLabel}.{datatype}"
 sprintf(inpAryPath,"%s.%s",imgPrefix,tivTypeStr);
 outf=fopen(inpAryPath,"wb");
 status = fwrite(imgBuf_p,sizeof(uint8),nImgBytes,outf);
 fclose(outf);
 
 // the genFreqDistrib.py utility uses numpy to generate a 
 // frequency distribution.
 // python ../../l4c_DAILY/src/genFreqDistrib.py float32 smrz_mx_1624x3856.flt32 smrz_TEST.png 
 
 // convert the numeric TIV datatype to its string representation
 // like "float32" etc
 sprintf(osCmd,"python genFreqDistrib.py %s %s %s\n",
        tivTypeStr, inpAryPath, outputPath);
 xLogMsg("@A In(%s) osCmd[%s] ~\n",fxName,osCmd);

 system(osCmd); 

 return 0 ;
} // end:: l4c_GenFreqDistrib()
